/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.io;

import iso.mpeg.streamezzo.laser.*;
import iso.mpeg.streamezzo.resource.Resource;
import iso.mpeg.streamezzo.saf.*;
import iso.mpeg.streamezzo.sdl.BitReaderD;
import iso.mpeg.streamezzo.sdl.SDLClasses;
import iso.mpeg.streamezzo.util.ImageInfo;
import iso.mpeg.streamezzo.util.MPEG4Constants;
import iso.mpeg.streamezzo.util.Scene;

import java.io.IOException;
import java.io.PrintWriter;

public abstract class PreDecoder implements MPEG4Constants {
    public final static boolean debug = false;

    public abstract Object updates(BitReaderD br, String msg) throws IOException;

    public abstract Object updatable_elements(BitReaderD br, String msg) throws IOException;

    public abstract Object elements(BitReaderD br, String msg) throws IOException;

    public abstract void decode(XprivateElementContainer pec, BitReaderD br) throws IOException;

    private ImageInfo info = new ImageInfo();

    public int readLength(BitReaderD br, String s) throws IOException {
        return br.readVlcInt(5, s);
    }

    public void decodeRare(X b, BitReaderD br) throws IOException {
        int nb = br.readBits(6, "nbOfRares");
        for (int i = 0; i < nb; i++) {
            int rareId = br.readBits(6, "rareId");
            switch (rareId) {
                case /* RARE_CLASS */ 0:
                    ((XpropElem) b).classe = SDLClasses.class_byteAlignedString(br, "class");
                    break;
                case /* RARE_AUDIO_LEVEL */ 1:
                    ((XpropElem) b).audio_level = SDLClasses.class_0to1float(br, "audio-level");
                    break;
                case /* RARE_COLOR */ 2:
                    ((XpropElem) b).color = SDLClasses.class_paint(br, "color");
                    break;
                case /* RARE_COLOR_RENDERING */ 3:
                    ((XpropElem) b).color_rendering = br.readBits(2, "colorRendering");
                    break;
                case /* RARE_DISPLAY */ 4:
                    ((Idpv) b).setDisplay(br.readBits(5, "display"));
                    break;
                case /* RARE_DISPLAY_ALIGN */ 5:
                    ((XpropElem) b).display_align = br.readBits(3, "display-align");
                    break;
                case /* RARE_FILL_OPACITY */ 6:
                    ((XpropElem) b).fill_opacity = SDLClasses.class_0to1float(br, "fill-opacity");
                    break;
                case /* RARE_FILL_RULE */ 7:
                    ((XpropElem) b).fill_rule = br.readBits(2, "fill-rule");
                    break;
                case /* RARE_IMAGE_RENDERING */ 8:
                    ((XpropElem) b).image_rendering = br.readBits(2, "imageRendering");
                    break;
                case /* RARE_LINE_INCREMENT */ 9:
                    ((XpropElem) b).line_increment = SDLClasses.lineIncrementType(br, "line-increment");
                    break;
                case /* RARE_POINTER_EVENTS */ 10:
                    ((Idpv) b).setPointer_events(br.readBits(4, "pointerEvents"));
                    break;
                case /* RARE_SHAPE_RENDERING */ 11:
                    ((XpropElem) b).shape_rendering = br.readBits(3, "shapeRendering");
                    break;
                case /* RARE_SOLID_COLOR */ 12:
                    ((XpropElem) b).solid_color = SDLClasses.class_paint(br, "solid-color");
                    break;
                case /* RARE_SOLID_OPACITY */ 13:
                    ((XpropElem) b).solid_opacity = SDLClasses.class_0to1float(br, "solid-opacity");
                    break;
                case /* RARE_STOP_COLOR */ 14:
                    ((XpropElem) b).stop_color = SDLClasses.class_paint(br, "stop-color");
                    break;
                case /* RARE_STOP_OPACITY */ 15:
                    ((XpropElem) b).stop_opacity = SDLClasses.class_0to1float(br, "stop-opacity");
                    break;
                case /* RARE_STROKE_DASHARRAY */ 16:
                    if (! br.readBit("isInherit")) {
                        int max = readLength(br, "nb of numbers");
                        ((XpropElem) b).stroke_dasharray = new int[max];
                        for (int j = 0; j < max; j++) {
                            ((XpropElem) b).stroke_dasharray[j] = SDLClasses.class_fixed_16_8(br, "dash");
                        }
                    }
                    break;
                case /* RARE_STROKE_DASHOFFSET */ 17:
                    ((XpropElem) b).stroke_dashoffset = SDLClasses.class_fixed_16_8i(br, "strokeDashoffset");
                    break;
                case /* RARE_STROKE_LINECAP */ 18:
                    ((XpropElem) b).stroke_linecap = br.readBits(2, "strokeLinecap");
                    break;
                case /* RARE_STROKE_LINEJOIN */ 19:
                    ((XpropElem) b).stroke_linejoin = br.readBits(2, "strokeLinejoin");
                    break;
                case /* RARE_STROKE_MITERLIMIT */ 20:
                    ((XpropElem) b).stroke_miterlimit = SDLClasses.class_fixed_16_8i(br, "strokeMiterlimit");
                    break;
                case /* RARE_STROKE_OPACITY */ 21:
                    ((XpropElem) b).stroke_opacity = SDLClasses.class_0to1float(br, "stroke-opacity");
                    break;
                case /* RARE_STROKE_WIDTH */ 22:
                    ((XpropElem) b).stroke_width = SDLClasses.class_fixed_16_8i(br, "stroke-width");
                    break;
                case /* RARE_TEXT_ANCHOR */ 23:
                    ((XpropElem) b).text_anchor = br.readBits(2, "text-align");
                    break;
                case /* RARE_TEXT_RENDERING */ 24:
                    ((XpropElem) b).text_rendering = br.readBits(3, "textRendering");
                    break;
                case /* RARE_VIEWPORT_FILL */ 25:
                    ((XpropElem) b).viewport_fill = SDLClasses.class_paint(br, "viewport-fill");
                    break;
                case /* RARE_VIEWPORT_FILL_OPACITY */ 26:
                    ((XpropElem) b).viewport_fill_opacity = SDLClasses.class_0to1float(br, "viewport-fill-opacity");
                    break;
                case /* RARE_VECTOR_EFFECT */ 27:
                    ((XpropElem) b).vector_effect = br.readBits(4, "vector-effect");
                    break;
                case /* RARE_VISIBILITY */ 28:
                    ((Idpv) b).setVisibility(br.readBits(2, "visibility"));
                    break;
                case /* RARE_REQUIREDEXTENSIONS */ 29:
                    ((Ibasic) b).setRequiredExtensions(SDLClasses.class_byteAlignedString(br, "requiredExtensions"));
                    break;
                case /* RARE_REQUIREDFEATURES */ 30:
                    int len = br.readVlcInt(5, "nb Features");
                    byte by[] = new byte[len];
                    for (int j = 0; j < len; j++) {
                        by[j] = (byte) br.readBits(6, "requiredFeatures");
                    }
                    ((Ibasic) b).setRequiredFeatures(by);
                    break;
                case /* RARE_REQUIREDFORMATS */ 31:
                    ((Ibasic) b).setRequiredFormats(SDLClasses.class_byteAlignedString(br, "requiredFormats"));
                    break;
                case /* RARE_SYSTEMLANGUAGE */ 32:
                    ((Ibasic) b).setSystemLanguage(SDLClasses.class_byteAlignedString(br, "systemLanguage"));
                    break;
                case /* RARE_XML_BASE */ 33:
                    b.base = SDLClasses.class_byteAlignedString(br, "xmlBase");
                    break;
                case /* RARE_XML_LANG */ 34:
                    b.lang = SDLClasses.class_byteAlignedString(br, "xmlLang");
                    break;
                case /* RARE_XML_SPACE */ 35:
                    b.space = br.readBits(1, "xml:space");
                    break;
                case /* RARE_FOCUSNEXT */ 36:
                    b.navNext = SDLClasses.focusType(br, "navNext");
                    break;
                case /* RARE_FOCUSNORTH */ 37:
                    b.navUp = SDLClasses.focusType(br, "navUp");
                    break;
                case /* RARE_FOCUSNORTHEAST */ 38:
                    b.navUpRight = SDLClasses.focusType(br, "navUpRight");
                    break;
                case /* RARE_FOCUSNORTHWEST */ 39:
                    b.navUpLeft = SDLClasses.focusType(br, "navUpLeft");
                    break;
                case /* RARE_FOCUSPREV */ 40:
                    b.navPrev = SDLClasses.focusType(br, "navPrev");
                    break;
                case /* RARE_FOCUSSOUTH */ 41:
                    b.navDown = SDLClasses.focusType(br, "navDown");
                    break;
                case /* RARE_FOCUSSOUTHEAST */ 42:
                    b.navDownRight = SDLClasses.focusType(br, "navDownRight");
                    break;
                case /* RARE_FOCUSSOUTHWEST */ 43:
                    b.navDownLeft = SDLClasses.focusType(br, "navDownLeft");
                    break;
                case /* RARE_FOCUSWEST */ 44:
                    b.navLeft = SDLClasses.focusType(br, "navLeft");
                    break;
                case /* RARE_FOCUSABLE */ 45:
                    b.focusable = br.readBits(2, "focusable");
                    break;
                case /* RARE_FOCUSEAST */ 46:
                    b.navRight = SDLClasses.focusType(br, "navRight");
                    break;
                case /* RARE_TRANSFORM */ 47:
                    ((Ilocatable) b).setTransform(SDLClasses.class_transform(br, "matrix"));
                    break;
                    /*case 48:
                    case 49:*/
                case /* RARE_FONT_VARIANT */ 50:
                    ((XpropElem) b).font_variant = br.readBits(2, "font index");
                    break;
                case /* RARE_FONT_FAMILY */ 51:
                    ((XpropElem) b).font_family = SDLClasses.class_fontFamily(br, "font index");
                    break;
                case /* RARE_FONT_SIZE */ 52:
                    ((XpropElem) b).font_size = SDLClasses.class_fixed_16_8i(br, "font-size");
                    break;
                case /* RARE_FONT_STYLE */ 53:
                    ((XpropElem) b).font_style = br.readBits(3, "font style");
                    break;
                case /* RARE_TEXT_DECORATION */ 48:
                    ((XpropElem) b).text_decoration = br.readBits(4, "font style");
                    break;
                case /* RARE_FONT_WEIGHT */ 54:
                    ((XpropElem) b).font_weight = br.readBits(4, "fontWeight");
                    break;
                case /* RARE_HREF_TITLE */ 55:
                    ((Ixlink) b).setTitle(SDLClasses.class_byteAlignedString(br, "title"));
                    break;
                case /* RARE_HREF_TYPE */ 56:
                    ((Ixlink) b).setXlinkType(br.readBits(3, "type"));
                    break;
                case /* RARE_HREF_ROLE */ 57:
                    ((Ixlink) b).setRole(SDLClasses.class_anyURI(br, "role"));
                    break;
                case /* RARE_HREF_ARCROLE */ 58:
                    ((Ixlink) b).setArcrole(SDLClasses.class_anyURI(br, "arcrole"));
                    break;
                case /* RARE_HREF_ACTUATE */ 59:
                    ((Ixlink) b).setActuate(br.readBits(2, "actuate"));
                    break;
                case /* RARE_HREF_SHOW */ 60:
                    ((Ixlink) b).setShow(br.readBits(3, "show"));
                    break;
                case /* RARE_END */ 61:
                    if(b instanceof Xset) {
                        ((Xset)b).end = SDLClasses.timesType(br, "end");
                    } else {
                        ((Xaudio)b).end = SDLClasses.timesType(br, "end");
                    }
                    break;
                case /* RARE_MAX */ 62:
                    if (b instanceof Xset) ((Xset) b).max = SDLClasses.class_time(br, "max", true);
                    if (b instanceof Xaudio) ((Xaudio) b).max = SDLClasses.class_time(br, "max", true);
                    break;
                case /* RARE_MIN */ 63:
                    if (b instanceof Xset) ((Xset) b).min = SDLClasses.class_time(br, "min", true);
                    if (b instanceof Xaudio) ((Xaudio) b).min = SDLClasses.class_time(br, "min", true);
                    break;
                case 49: // extension (rare2)
                    do {
                        int extID = br.readBits(SDLClasses.extensionIDBits, "extID");
                        int skipLen = br.readVlcInt(5, "skipLen");
                        if (extID == 2) {
                            int max = br.readBits(2, "nbOfAttrs");
                            for (int j = 0; j < max; j++) {
                                int ar = br.readBits(3, "attributeRARE");
                                switch(ar) {
                                    case 0:
                                        if (b instanceof Xaudio) {
                                            ((Xaudio) b).syncMaster = (br.readBit("syncMaster") ? 1 : 0);
                                        }
                                        break;
                                    case 1:
                                        ((XpropElem)b).focusHighlight = br.readBits(2, "focusHighlight");
                                        break;
                                    case 2:
                                        if (b instanceof Xvideo) {
                                            ((Xvideo)b).initialVisibility = br.readBits(2, "initialVisibility");
                                        }
                                        if (b instanceof Xanimation) {
                                            ((Xanimation)b).initialVisibility = br.readBits(2, "initialVisibility");
                                        }
                                        break;
                                    case 3:
                                        // fullscreen
                                        if (b instanceof Xvideo) {
                                            ((Xvideo)b).fullscreen = br.readBits(2, "fullscreen");
                                        }
                                        if (b instanceof Xsvg) {
                                            ((Xsvg)b).fullscreen = br.readBits(2, "fullscreen");
                                        }
                                        break;
                                    case 4:
                                        // requiredFonts
                                        b.requiredFonts = SDLClasses.class_byteAlignedString(br, "requiredFonts");
                                        break;
                                    case 5:
                                        // active
                                        b.active = br.readBits(1, "active");
                                        break;
                                }
                            }
                        } else {
                            while (skipLen > 32) {
                                br.readInt("skipping");
                                skipLen -= 32;
                            }
                            br.readBits(skipLen, "skipping");
                        }
                    } while (br.readBit("moreRareExt"));
                    break;
            }
        }
    }

    /*public void decodeAnimate(Xanimate self, BitReaderD br) throws IOException {
        self.accumulate = br.readBit("accumulate");
        self.additive = br.readBit("additive");
        if (br.readBit("by")) { self.by = SDLClasses.class_AnimatedValue(br, "by");}
        if (br.readBit("calcMode")) { self.calcMode = br.readBits(2, "calcMode");}
        if (br.readBit("from")) { self.from = SDLClasses.class_AnimatedValue(br, "from");}
        if (br.readBit("keySplines")) { self.keySplines = SDLClasses.class_floatList(br, "keySplines");}
        if (br.readBit("keyTimes")) { self.keyTimes = SDLClasses.class_timeList(br, "keyTimes");}
        if (br.readBit("values")) { self.values = SDLClasses.class_AnimatedValues(br, "values");}
    }*/

    public void decodeContent(X self, BitReaderD br) throws IOException {
        SDLClasses.class_anyAttribute(br, "anyAttribute", self);
        self.privateAttributes = SDLClasses.privateAttributeContainer(br, "privateAttributeContainer");
        if (br.readBit("loop")) {
            int occ0 = br.readVlcInt(5, "occ0");
            for (int i = 0; i < occ0; i++) {
                self.addChild(elements(br, "child"));
            }
        }
    }

    public void decodeSameContent(X self, BitReaderD br) throws IOException {
        if (br.readBit("loop")) {
            int occ0 = br.readVlcInt(5, "occ0");
            for (int i = 0; i < occ0; i++) {
                self.addChild(elements(br, "child"));
            }
        }
    }

    public void decodeSimpleUpdateContent(U self, BitReaderD br) throws IOException {
        SDLClasses.class_anyAttribute(br, "anyAttribute", null);
    }

    public void decodeInsertContent(XInsert self, BitReaderD br) throws IOException {
        SDLClasses.class_anyAttribute(br, "anyAttribute", null);
        if (br.readBit("loop")) {
            self.addChild(updatable_elements(br, "child"));
        }
    }

    public void decodeReplaceContent(XReplace self, BitReaderD br) throws IOException {
        SDLClasses.class_anyAttribute(br, "anyAttribute", null);
        if (br.readBit("loop")) {
            int occ0 = br.readVlcInt(5, "occ0");
            for (int i = 0; i < occ0; i++) {
                self.addChild(updatable_elements(br, "child"));
            }
        }
    }

    public void decodeOneChildUpdateContent(U self, BitReaderD br) throws IOException {
        SDLClasses.class_anyAttribute(br, "anyAttribute", null);
        self.addChild(updatable_elements(br, "child"));
    }

    public void decodeLContent(X self, BitReaderD br) throws IOException {
        SDLClasses.class_anyAttribute(br, "anyAttribute", null);
        self.privateAttributes = SDLClasses.privateAttributeContainer(br, "privateAttributeContainer");
    }

    public static class SAF_SL_PacketHeader {
        public int packetSequenceNumber = 0;
        public boolean randomAccessFlag = false;
        public int time = 0; // compositionTimeStamp or validity
        public int accessUnitLength;

        public void read(BitReaderD br) throws Exception {
            randomAccessFlag = br.readBit("randomAccessFlag");
            packetSequenceNumber = br.readBits(15, "packetSequenceNumber");
            if (br.readBits(2, "flags") != 1) {
                throw new RuntimeException("wrong flags in SAF header");
            }
            time = br.readBits(30, "time");
            SDLClasses.currentTimeInTics = time;
            accessUnitLength = br.readBits(16, "accessUnitLength");
        }
    }

    public static class SAF_AU_Header {
        public int accessUnitType = -1;
        public int streamID = -1;
        public int payloadLength = -1;
        public boolean isFragment = false;

        public void read(BitReaderD br, SAF_SL_PacketHeader ssph) throws Exception {
            accessUnitType = br.readBits(4, "accessUnitType");
            streamID = br.readBits(12, "href");
            if (streamID == 0xFFF) {
                streamID = -1;
            }
            if (ssph.accessUnitLength == 0) {
                isFragment = true;
                payloadLength = br.readBits(16, "payloadLength");
            } else if (ssph.accessUnitLength == 1) {
                throw new RuntimeException("length=1 is forbidden in the SAF SL Header");
            }
        }
    }

    public static class SimpleDecoderConfigDescriptor {
        public byte objectTypeIndication = -1;
        public byte streamType = -1;
        public int timeStampResolution = -1;
        public int bufferSize = -1;

        public void read(BitReaderD br) throws Exception {
            objectTypeIndication = (byte) br.readBits(8, "objectTypeIndication");
            streamType = (byte) br.readBits(8, "streamType");
            timeStampResolution = br.readBits(24, "timStampResolution");
            bufferSize = br.readBits(16, "bufferSize");
        }
    }

    public void decode(Scene scene, BitReaderD br, PrintWriter pw) throws Exception {
        SDLClasses.currentScene = scene;
        SDLClasses.man = scene.man;
        boolean goOn = true;
        SAF_SL_PacketHeader ssph = new SAF_SL_PacketHeader();
        SAF_AU_Header sah = new SAF_AU_Header();
        try {
            while (goOn) {
                pw.println();
                pw.println("     <@> <@> <@>      start packet");
                pw.println();
                ssph.read(br);
                sah.read(br, ssph);
                if (sah.isFragment) {
                    // here, deal with fragments
                    // in this version, skip
                    br.readByteArray(sah.payloadLength);
                } else {
                    boolean permanent = false;
                    switch (sah.accessUnitType) {
                        case 0x02:
                            // header, permanent
                            permanent = true;
                            if (debug) System.out.print("permanent ");
                        case 0x01:
                            // header, not permanent
                            SimpleDecoderConfigDescriptor sdcd = new SimpleDecoderConfigDescriptor();
                            sdcd.read(br);
                            if (sdcd.streamType == SceneDescriptionStreamType) {
                                if (sdcd.objectTypeIndication == FontDataOTI) {
                                    // this is a fontStreamHeader
                                    parseFontHeader(scene, new FontHeader(), ssph, sah, sdcd, br, permanent);
                                    if (debug) System.out.println("font header");
                                } else if (sdcd.objectTypeIndication == LASeROTI) {
                                    // this is a sceneHeader
                                    parseSceneHeader(scene, new SceneHeader(), ssph, sah, sdcd, br, permanent);
                                    if (debug) System.out.println("scene header");
                                }
                            } else if (sdcd.streamType == AudioStreamType) {
                                // this is a MediaHeader
                                parseMediaHeader(scene, new MediaHeader(), ssph, sah, sdcd, br, permanent);
                                if (debug) System.out.println("media header");
                            } else if (sdcd.streamType == VisualStreamType) {
                                if (sdcd.objectTypeIndication == JPEGOTI || sdcd.objectTypeIndication == PNGOTI) {
                                    // image header
                                    parseImageHeader(scene, new ImageHeader(), ssph, sah, sdcd, br, permanent);
                                    if (debug) System.out.println("image header");
                                } else {
                                    // video header
                                    parseMediaHeader(scene, new MediaHeader(), ssph, sah, sdcd, br, permanent);
                                    if (debug) System.out.println("media header");
                                }
                            }
                            break;
                        case 0x03:
                            // end of stream
                            EndOfStream au = new EndOfStream();
                            parseEndOfStream(scene, au, ssph, sah, br);
                            if (debug) System.out.println("end of stream");
                            break;
                        case 0x04:
                            // access unit
                            SAFHeader h = scene.lastHeaderWithThisStreamID("Stream" + sah.streamID);
                            if (h == null) {
                                throw new RuntimeException("Missing header for stream " + sah.streamID);
                            } else if (h.streamType == SceneDescriptionStreamType) {
                                if (h.objectTypeIndication == FontDataOTI) {
                                    // this is a fontStreamHeader
                                    parseFontUnit(scene, new FontUnit(), ssph, sah, br);
                                    if (debug) System.out.println("font unit");
                                } else if (h.objectTypeIndication == LASeROTI) {
                                    // this is a sceneHeader
                                    parseSceneUnit(scene, new SceneUnit(), ssph, sah, br);
                                    if (debug) System.out.println("scene unit");
                                }
                            } else if (h.streamType == AudioStreamType) {
                                // this is a MediaHeader
                                parseMediaUnit(scene, new MediaUnit(), ssph, sah, br);
                                if (debug) System.out.println("media unit");
                            } else if (h.streamType == VisualStreamType) {
                                if (h.objectTypeIndication == JPEGOTI || h.objectTypeIndication == PNGOTI) {
                                    // image header
                                    parseImageUnit(scene, new ImageUnit(), ssph, sah, br);
                                    if (debug) System.out.println("image unit");
                                } else {
                                    // video header
                                    parseMediaUnit(scene, new MediaUnit(), ssph, sah, br);
                                    if (debug) System.out.println("media unit");
                                }
                            }
                            break;
                        case 0x05:
                            // end of saf session
                            EndOfSAFSession au1 = new EndOfSAFSession();
                            parseEndOfSAFSession(scene, au1, ssph, sah, br);
                            goOn = false;
                            if (debug) System.out.println("end of saf session");
                            break;
                        case 0x06:
                            // cache unit
                            CacheObject co = new CacheObject();
                            parseCacheObject(scene, co, ssph, sah, br);
                            if (debug) System.out.println("cache object");
                            break;
                        case 0x07:
                            // remote header
                            RemoteStreamHeader rmh = new RemoteStreamHeader();
                            SimpleDecoderConfigDescriptor sdcd1 = new SimpleDecoderConfigDescriptor();
                            sdcd1.read(br);
                            parseRemoteStreamHeader(scene, rmh, ssph, sah, sdcd1, br, false);
                            if (debug) System.out.println("remote header");
                            break;
                            /*case 0x08:
                            // saf config
                            SAFConfiguration sc = new SAFConfiguration();
                            parseSAFConfiguration(scene, sc, ssph, sah, br);
                            break;*/
                    }
                    br.flush();
                }
            }
        } catch (Exception e) {
            pw.flush();
            pw.close();
            throw e;
        }
    }

    public void decodeAny(Scene scene, BitReaderD br, PrintWriter pw) throws Exception {
        SDLClasses.currentScene = scene;
        SDLClasses.man = scene.man;
        SDLClasses.class_initialisation(br, "init", null);
        try {
            XprivateElementContainer pec = new XprivateElementContainer();
            decode(pec, br);
            scene.addChild(pec);
        } catch (Exception e) {
            pw.flush();
            pw.close();
            throw e;
        }
    }

    protected void parseImageUnit(Scene scene, ImageUnit au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah, BitReaderD br)
            throws IOException {
        scene.addChild(au);
        au.streamID = "Stream" + sah.streamID;
        //scene.idMan.setStreamID(au.streamID, sah.streamID);
        byte data[] = br.readByteArray(ssph.accessUnitLength - 2, "image data");
        // end of parsing, the rest is package-specific
        info.setInput(new java.io.ByteArrayInputStream(data));
        info.check();
        String name = scene.name + "_Image" + sah.streamID;
        switch (info.getFormat()) {
            case ImageInfo.FORMAT_GIF:
                name += ".gif";
                break;
            case ImageInfo.FORMAT_JPEG:
                name += ".jpg";
                break;
            case ImageInfo.FORMAT_PNG:
                name += ".png";
                break;
            default:
                name += ".unknown";
        }
        au.resource = Resource.create(name, data);
    }

    protected void parseSceneHeader(Scene scene, SceneHeader au3, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                    SimpleDecoderConfigDescriptor sdcd, BitReaderD br, boolean permanent) throws Exception, IOException {
        scene.addChild(au3);
        au3.bufferSize = sdcd.bufferSize;
        au3.objectTypeIndication = sdcd.objectTypeIndication;
        au3.streamType = sdcd.streamType;
        au3.timeStampResolution = sdcd.timeStampResolution;
        String streamName = "Stream" + sah.streamID;
        scene.idMan.setStreamID(streamName, sah.streamID);
        au3.setStreamID(streamName);
        au3.lh = new XLASeRHeader();
        decode(au3.lh, br);
        SDLClasses.colorComponentBits = au3.lh.colorComponentBits_minus_1 + 1;
        SDLClasses.scaleBits = au3.lh.scaleBits_minus_coordBits;
        SDLClasses.resolution = au3.lh.resolution;
        SDLClasses.extensionIDBits = au3.lh.extensionIDBits;
    }

    public abstract void decode(XLASeRHeader self, BitReaderD br) throws IOException;

    private static int reframeComponent(int comp) throws IOException {
        if (SDLClasses.colorComponentBits > 8) {
            return 0xFF & (comp >> SDLClasses.colorComponentBits - 8);
        } else {
            return 0xFF & (comp << (8 - SDLClasses.colorComponentBits));
        }
    }

    protected void parseSceneUnit(Scene scene, SceneUnit au1, SAF_SL_PacketHeader ssph, SAF_AU_Header sah, BitReaderD br)
            throws Exception {
        au1.isRAP = ssph.randomAccessFlag;
        au1.streamID = "Stream"+sah.streamID;
        au1.time = ssph.time;
        scene.addChild(au1);
        br.traceFile.println("---------------------AU");
        // LASeRUnitHeader
        au1.resetEncodingContext = br.readBit("resetEncodingContext");
        if (br.readBit("extension")) {
            SDLClasses.class_extension(br, "extension");
        }
        SDLClasses.class_initialisation(br, "init", au1);
        int occ1 = br.readVlcInt(5, "occ1");
        for (int t = 0; t <= occ1; t++) {
            au1.addChild(updates(br, "update"));
        }
        if (br.readBit("extension")) {
            SDLClasses.class_extension(br, "extension");
        }
    }

    protected void parseEndOfStream(Scene scene, EndOfStream au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                    BitReaderD br) {
        scene.addChild(au);
        au.streamID = "Stream" + sah.streamID;
    }

    protected void parseEndOfSAFSession(Scene scene, EndOfSAFSession au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                        BitReaderD br) {
        scene.addChild(au);
    }

    protected void parseMediaUnit(Scene scene, MediaUnit au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                  BitReaderD br) throws IOException {
        au.setRAP(ssph.randomAccessFlag);
        au.setStreamID("Stream" + sah.streamID);
        scene.addChild(au);
        au.setTime(ssph.time);
        byte data[] = br.readByteArray(ssph.accessUnitLength - 2, "media data");
        /*ByteArrayOutputStream stream = (ByteArrayOutputStream)tmpResources.get(au.getStreamID());
        int offset = stream.delta();
        stream.write(data);
        au.setStartOffset(offset);
        au.setLength(data.length);*/
    }

    /*protected void parseSAFConfiguration(Scene scene, SAFConfiguration au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                  BitReaderD br) throws IOException {
        au.setStreamID("Stream" + sah.streamID);
        scene.addChild(au);
        au.timeStampResolution = br.readInt("timeStampResolution");
    }*/

    protected void parseFontUnit(Scene scene, FontUnit au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah, BitReaderD br)
            throws IOException {
        au.setRAP(ssph.randomAccessFlag);
        au.setStreamID("Stream" + sah.streamID);
        scene.addChild(au);
        au.setTime(ssph.time);
        byte data[] = br.readByteArray(ssph.accessUnitLength - 2, "font data");
        /*ByteArrayOutputStream stream = (ByteArrayOutputStream)tmpResources.get(au.getStreamID());
        int offset = stream.delta();
        stream.write(data);
        au.setStartOffset(offset);
        au.setLength(data.length);*/
    }

    protected void parseMediaHeader(Scene scene, MediaHeader au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                    SimpleDecoderConfigDescriptor sdcd, BitReaderD br, boolean permanent) throws Exception {
        int streamID = sah.streamID;
        String streamName = "Stream" + streamID;
        scene.idMan.setStreamID(streamName, streamID);
        au.setStreamID(streamName);
        scene.addChild(au);
        au.bufferSize = sdcd.bufferSize;
        au.objectTypeIndication = sdcd.objectTypeIndication;
        au.streamType = sdcd.streamType;
        au.timeStampResolution = sdcd.timeStampResolution;
        int len = 0;
        if (au.objectTypeIndication == 0xFF && au.streamType == 0xFF) {
            len = br.readBits(16, "mimeTypeLength");
            au.mimeType = new String(br.readByteArray(len, "mimeType"), "UTF-8");
            len += 2; // to get the number of bytes read in this branch
        }
        if (ssph.accessUnitLength > 9 + len) {
            byte data[] = br.readByteArray(ssph.accessUnitLength - 9, "media header");
            String name = scene.name + "_mediaheader_" + streamID;
            au.resource = Resource.create(name, data);
            /*Output out = oset.get(name, true);
            out.getOutputStream().write(data);
            // premaring output stream for mediaAu...
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            tmpResources.put(streamName, stream);*/
        }
    }

    protected void parseRemoteStreamHeader(Scene scene, RemoteStreamHeader au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                           SimpleDecoderConfigDescriptor sdcd, BitReaderD br, boolean permanent) throws Exception {
        int streamID = sah.streamID;
        String streamName = "Stream" + streamID;
        scene.idMan.setStreamID(streamName, streamID);
        au.setStreamID(streamName);
        scene.addChild(au);
        au.bufferSize = sdcd.bufferSize;
        au.objectTypeIndication = sdcd.objectTypeIndication;
        au.streamType = sdcd.streamType;
        au.timeStampResolution = sdcd.timeStampResolution;
        int len = 0;
        if (au.objectTypeIndication == 0xFF && au.streamType == 0xFF) {
            len = br.readBits(16, "mimeTypeLength");
            au.mimeType = new String(br.readByteArray(len, "mimeType"), "UTF-8");
            len += 2; // to get the number of bytes read in this branch
        }
        int len2 = br.readBits(16, "urlLength");
        au.source = new String(br.readByteArray(len2, "url"), "UTF-8");
        len2 += 2; // to get the number of bytes read in this branch
        if (ssph.accessUnitLength > 9 + len + len2) {
            byte data[] = br.readByteArray(ssph.accessUnitLength - 9, "media header");
            String name = scene.name + "_mediaheader_" + streamID;
            au.resource = Resource.create(name, data);
            /*Output out = oset.get(name, true);
            out.getOutputStream().write(data);
            // premaring output stream for mediaAu...
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            tmpResources.put(streamName, stream);*/
        }
    }

    protected void parseImageHeader(Scene scene, ImageHeader au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                    SimpleDecoderConfigDescriptor sdcd, BitReaderD br, boolean permanent) throws Exception {
        int streamID = sah.streamID;
        String streamName = "Stream" + streamID;
        scene.idMan.setStreamID(streamName, streamID);
        au.setStreamID(streamName);
        scene.addChild(au);
        au.bufferSize = sdcd.bufferSize;
        au.objectTypeIndication = sdcd.objectTypeIndication;
        au.streamType = sdcd.streamType;
        au.timeStampResolution = sdcd.timeStampResolution;
        if (ssph.accessUnitLength > 9) {
            byte data[] = br.readByteArray(ssph.accessUnitLength - 9, "image header");
            String name = scene.name + "_imageheader_" + streamID;
            au.resource = Resource.create(name, data);
            /*Output out = oset.get(name, true);
            out.getOutputStream().write(data);
            // premaring output stream for mediaAu...
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            tmpResources.put(streamName, stream);*/
        }
    }

    protected void parseFontHeader(Scene scene, FontHeader au, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                   SimpleDecoderConfigDescriptor sdcd, BitReaderD br, boolean permanent) throws Exception {
        au.bufferSize = sdcd.bufferSize;
        au.objectTypeIndication = sdcd.objectTypeIndication;
        au.streamType = sdcd.streamType;
        au.timeStampResolution = sdcd.timeStampResolution;
        int streamID = sah.streamID;
        String streamName = "Stream" + streamID;
        scene.idMan.setStreamID(streamName, streamID);
        au.setStreamID(streamName);
        scene.addChild(au);
        if (ssph.accessUnitLength > 9) {
            byte data[] = br.readByteArray(ssph.accessUnitLength - 9);
            String name = scene.name + "_fontheader_" + streamID;
            au.resource = Resource.create(name, data);
            /*Output out = oset.get(name, true);
            out.getOutputStream().write(data);
            // premaring output stream for mediaAu...
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            tmpResources.put(streamName, stream);*/
        }
    }

    private static int cacheCount = 1;

    protected void parseCacheObject(Scene scene, CacheObject au2, SAF_SL_PacketHeader ssph, SAF_AU_Header sah,
                                    BitReaderD br) throws Exception {
        au2.isReplace = br.readBit("replace");
        au2.isPermanent = br.readBit("permanent");
        br.readBits(6);
        au2.endTime = ssph.time;
        scene.addChild(au2);
        String tmpURL [] = br.readString1("url");
        au2.url = tmpURL[0];
        br.flush();
        //read data
        byte data[] = br.readByteArray(ssph.accessUnitLength - 3 - Integer.parseInt(tmpURL[1]), "cache unit payload");
        // end of parsing, the rest is package-specific
        String name = scene.name + "_cache_" + (++cacheCount) + ".lsr";
        au2.resource = Resource.create(name, data);
        /*Output out = oset.get(name, true);
        out.getOutputStream().write(data);*/
    }

}
