/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.sdl;

import java.io.DataInputStream;
import java.io.InputStream;

public class BitReader {
    private DataInputStream dataIn;
    private int buffer;
    protected int valid = 0;
    protected int byteRead = 0;

    public BitReader(InputStream in) {
        dataIn = new java.io.DataInputStream(in);
    }

    void read() throws java.io.IOException {
        if (valid == 0) {
            buffer = 0xFF & dataIn.readByte();
            valid = 8;
            byteRead++;
        }
    }

    private final static int maskBit[] = {0x00, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};

    public boolean readBit() throws java.io.IOException {
        read();
        return ((buffer & maskBit[valid--]) != 0);
    }

    private final static int maskRight[] = {0x00000000, 0x00000001, 0x00000003, 0x00000007, 0x0000000F, 0x0000001F,
                                            0x0000003F, 0x0000007F, 0x000000FF};

    public int readByte() throws java.io.IOException {
        read();
        if (valid == 8) {
            valid = 0;
            return buffer;
        } else {
            int tmp = (buffer & maskRight[valid]) << (8 - valid);
            buffer = 0xFF & dataIn.readByte();
            byteRead++;
            tmp |= (buffer >>> valid);
            return tmp;
        }
    }

    public int readInt() throws java.io.IOException {
        byteRead += 4;
        if (valid == 0) {
            return dataIn.readInt();
        } else {
            int tmp = (buffer & maskRight[valid]) << (32 - valid);
            int t = dataIn.readInt();
            buffer = 0xFF & t;
            tmp |= (t >>> valid);
            return tmp;
        }
    }

    public int readBits(int n) throws java.io.IOException {
        if (n > 32) {
            throw new RuntimeException("Cannot read more than 32 bits as an int: " + n);
        }
        if (n <= 0) {
            return 0;
        }
        read();
        if (n <= valid) {
            // either we take everything from buffer
            return (buffer >>> (valid -= n)) & maskRight[n];
        } else {
            // or we will have to do more
            n -= valid;
            int tmp = (buffer & maskRight[valid]) << n;
            valid = 0;
            while (n >= 8) {
                int byt = 0xFF & dataIn.readByte();
                byteRead++;
                n -= 8;
                tmp |= byt << n;
            }
            if (n > 0) {
                buffer = 0xFF & dataIn.readByte();
                byteRead++;
                valid = 8 - n;
                tmp |= buffer >>> valid;
            }
            return tmp;
        }
    }

    public byte[] readByteArray(int l) throws java.io.IOException {
        byte a[] = new byte[l];
        for (int i = 0; i < l; i++) {
            a[i] = (byte)(0xFF & readByte());
        }
        return a;
    }

    public String readString(int l) throws java.io.IOException {
        byte a[] = new byte[l + 2];
        a[0] = (byte)((l >>> 8) & 0xFF);
        a[1] = (byte)(l & 0xFF);
        for (int i = 2; i < l + 2; i++) {
            a[i] = (byte)(0xFF & readByte());
        }
        java.io.ByteArrayInputStream bais = new java.io.ByteArrayInputStream(a);
        return new java.io.DataInputStream(bais).readUTF();
    }

    /*public String readString() throws java.io.IOException {
        return readString(readBits(readBits(5)));
    }*/

    public String[] readString1() throws java.io.IOException {
        int bitsUrlLen = readBits(5);
        int l = readBits(bitsUrlLen);
        byte a[] = new byte[l + 2];
        a[0] = (byte)((l >>> 8) & 0xFF);
        a[1] = (byte)(l & 0xFF);
        for (int i = 2; i < l + 2; i++) {
            a[i] = (byte)(0xFF & readByte());
        }
        java.io.ByteArrayInputStream bais = new java.io.ByteArrayInputStream(a);
        String tmp[] = new String[2];
        tmp[0] = new java.io.DataInputStream(bais).readUTF();
        //this is the number of bytes taken by the encoded string with 5 bits + X (url length in bytes) bits
        tmp[1] = Integer.toString(l + (5 + bitsUrlLen + 8 - (5 + bitsUrlLen) % 8) / 8);
        return tmp;
    }

    public void flush() {
        if (valid == 8) {
            return;
        }
        valid = 0;
        buffer = 0;
    }
}
