/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.sdl;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.OutputStream;

public final class BitWriter extends java.io.OutputStream {
    private static final boolean debug = false;
    DataOutputStream dataOut;
    int buffer;
    int valid = 0;
    public java.io.PrintWriter trace;

    public BitWriter(OutputStream os) {
        dataOut = new DataOutputStream(os);
    }

    public void close() throws java.io.IOException {
        flush();
    }

    public void flush() throws java.io.IOException {
        if (valid > 0) {
            // System.out.println("flushing "+(8 - valid));
            valid = 8;
            write();
        }
    }

    public void write(int i) throws java.io.IOException {
        putByte(i);
    }

    public int length() {
        if (dataOut != null) {
            return dataOut.size();
        }
        return 0;
    }

    public int bitlength() {
        return valid + (length() << 3);
    }

    public boolean notEmpty() {
        return ((dataOut != null && dataOut.size() > 0) || (valid > 0));
    }

    void write() throws java.io.IOException {
        if (valid == 8) {
            dataOut.writeByte(buffer);
            if (debug && trace != null) {
                trace.println(" 8 " + Integer.toHexString(0xFF & buffer));
            }
            buffer = 0;
            valid = 0;
        }
    }

    private final static int maskBit[] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

    public void putBit(boolean b) throws java.io.IOException {
        if (b) {
            buffer |= maskBit[valid++];
        } else {
            valid++;
        }
        write();
    }

    private final static int maskRight[] = {0x00000000, 0x00000001, 0x00000003, 0x00000007, 0x0000000F, 0x0000001F,
                                            0x0000003F, 0x0000007F, 0x000000FF, 0x000001FF, 0x000003FF, 0x000007FF,
                                            0x00000FFF, 0x00001FFF, 0x00003FFF, 0x00007FFF, 0x0000FFFF, 0x0001FFFF,
                                            0x0003FFFF, 0x0007FFFF, 0x000FFFFF, 0x001FFFFF, 0x003FFFFF, 0x007FFFFF,
                                            0x00FFFFFF, 0x01FFFFFF, 0x03FFFFFF, 0x07FFFFFF, 0x0FFFFFFF, 0x1FFFFFFF,
                                            0x3FFFFFFF, 0x7FFFFFFF, 0xFFFFFFFF};

    public void putInt(int i) throws java.io.IOException {
        if (valid == 0) {
            dataOut.writeInt(i);
            //if (trace != null) trace.println(" > "+Integer.toHexString(i));
            if (debug && trace != null) {
                trace.println(" 32 " + Integer.toHexString(i));
            }
        } else {
            int last = i & maskRight[valid]; // save left-most "valid" bits
            int tmp = buffer;                // copy buffer in int
            tmp = (tmp << 24) | (i >>> valid);     // shift buffer on top and add remaining bits from input
            dataOut.writeInt(tmp);           // write out value
            //if (trace != null) trace.println(" > "+Integer.toHexString(tmp));
            if (debug && trace != null) {
                trace.println(" 32 " + Integer.toHexString(tmp));
            }
            buffer = last << (8 - valid);        // normalize left-over bits into buffer
        }
    }

    private int work[] = new int[4];

    public void putByte(int i) throws java.io.IOException {
        if (valid == 0) {
            dataOut.writeByte(i);
            //if (trace != null) trace.println(" > "+Integer.toHexString(i&0xFF));
            if (debug && trace != null) {
                trace.println(" 8 " + Integer.toHexString(0xFF & i));
            }
        } else {
            int last = i & maskRight[valid]; // save left-most "valid" bits
            int tmp = buffer;                // copy buffer in int
            tmp = tmp | ((0xFF & i) >>> valid);         // add remaining bits from input
            dataOut.writeByte(tmp);          // write out value
            //if (trace != null) trace.println(" > "+Integer.toHexString(tmp&0xFF));
            if (debug && trace != null) {
                trace.println(" 8 " + Integer.toHexString(0xFF & tmp));
            }
            buffer = last << (8 - valid);        // normalize left-over bits into buffer
        }
    }

    public void putBits(int n, int l) throws java.io.IOException {
        if (n > 32 || n < 0) {
            throw new RuntimeException("Cannot write more than 32 bits or less than 0: " + n);
        }
        //if (l<0 && n != 32)
        //  throw new RuntimeException("Cannot write negative number as an int: "+l);
        if (n == 0) {
            return;
        }
        l = l & maskRight[n];
        if (valid + n <= 8) {
            // either we can pack everything in buffer
            buffer |= l << (8 - valid - n);
            valid += n;
            write();
        } else {
            // or we will have to do more
            int lastBits = (valid + n) % 8;
            int last = l & maskRight[lastBits]; // these are remaining bits at the end;
            int i = 0;
            work[i++] = l >>> lastBits; // buffer+work = one or more complete bytes
            n -= lastBits;
            for (; n > 8; n -= 8, i++) {
                work[i] = work[i - 1] >>> 8;
            }
            i--;
            dataOut.writeByte(buffer = buffer | work[i--]);
            //if (trace != null) trace.println(" > "+Integer.toHexString(buffer));
            if (debug && trace != null) {
                trace.println(" 8 " + Integer.toHexString(buffer));
            }
            for (; i >= 0; i--) {
                int j = work[i] & 0xFF;
                dataOut.writeByte(j);
                if (debug && trace != null) {
                    trace.println(" 8 " + Integer.toHexString(0xFF & j));
                }
                //if (trace != null) trace.println(" > "+Integer.toHexString(j));
            }
            buffer = last << (8 - lastBits);
            valid = lastBits;
        }
    }

    /*public void putString(String s) throws java.io.IOException {
        java.io.ByteArrayOutputStream baos = new java.io.ByteArrayOutputStream();
        new java.io.DataOutputStream(baos).writeUTF(s);
        byte g[] = baos.toByteArray();
        int i = g.length - 2, j = 0;
        for (; i > 0; j++) {
            i = i >>> 1;
        }
        putBits(5, j);
        putBits(j, g.length - 2);
        for (int k = 2; k < g.length; k++) {
            putBits(8, 0xFF & g[k]);
            //System.out.print(ensureTwoChars(Integer.toHexString(0xFF & g[k])));
        }
        //System.out.println();
    }*/

    public void putVlcString(String s) throws java.io.IOException {
        //System.out.println(s);
        java.io.ByteArrayOutputStream baos = new java.io.ByteArrayOutputStream();
        new java.io.DataOutputStream(baos).writeUTF(s);
        byte g[] = baos.toByteArray();
        putVlcInt(8, g.length - 2);
        for (int k = 2; k < g.length; k++) {
            putBits(8, 0xFF & g[k]);
            //System.out.print(ensureTwoChars(Integer.toHexString(0xFF & g[k])));
        }
        //System.out.println();
    }

    public int getUTFStringLength(String s) throws java.io.IOException {
        RandomAccessMemoryStream tempDataOut = new RandomAccessMemoryStream();
        java.io.DataOutputStream tempDDO = new java.io.DataOutputStream(tempDataOut);
        tempDDO.writeUTF(s);
        return tempDDO.size();
    }

    public void putUTFString(String s) throws java.io.IOException {
        dataOut.writeUTF(s);
    }

    public void putStringNoLength(String s) throws java.io.IOException {
        java.io.ByteArrayOutputStream baos = new java.io.ByteArrayOutputStream();
        new java.io.DataOutputStream(baos).writeUTF(s);
        byte g[] = baos.toByteArray();
        for (int k = 2; k < g.length; k++) {
            putBits(8, 0xFF & g[k]);
            if (trace != null) {
                charPrint.setCharAt(0, (char)(0xFF & g[k]));
                trace.println(" char '" + charPrint + "' " + (0xFF & g[k]));
            }
        }
        putBits(8, 0); // null terminator
        if (trace != null) {
            trace.println(" char '\\0' ");
        }
    }

    StringBuffer charPrint = new StringBuffer("a");

    public void putByteArray(byte a[]) throws java.io.IOException {
        for (int i = 0; i < a.length; i++) {
            putByte(a[i]);
        }
    }

    public void putBitArray(byte a[], int len, int bitOffset, int forceVlcLen) throws java.io.IOException {
        //System.out.println("bitOffset "+bitOffset+" len "+len);
        putVlcInt(5, len, forceVlcLen);
        BitReader br = new BitReader(new ByteArrayInputStream(a));
        br.readBits(bitOffset);
        while (len > 8) {
            putByte(br.readBits(8));
            len -= 8;
        }
        putBits(len, br.readBits(len));
    }

    public void putByteArrayWithLength(byte a[]) throws java.io.IOException {
        putVlcInt(5, a.length);
        putByteArray(a);
    }

    public void insertFile(String fn) throws java.io.IOException {
        File f = new File(fn);
        int len = (int)f.length();
        putInt(len);
        flush();
        FileInputStream fis = new FileInputStream(f);
        for (int i = 0; i < len; i++) {
            putByte(fis.read());
        }
        fis.close();
    }

    public void putVlcInt(int n, int i) throws java.io.IOException {
        if (i < 0) {
            throw new RuntimeException("vlcint needs to be positive:"+i);
        }
        int j = i, k = 1;
        while (j > maskRight[n-1]) {
            putBit(true);
            j = j >> (n-1);
            k++;
        }
        putBit(false);
        putBits(k*(n-1), i);
    }

    public void putVlcInt(int n, int i, int forceVlcLen) throws java.io.IOException {
        if (i < 0) {
            throw new RuntimeException("vlcint needs to be positive:"+i);
        }
        int j = i, k = 1;
        while (j > maskRight[n-1]) {
            putBit(true);
            j = j >> (n-1);
            k++;
        }
        if (k > forceVlcLen) {
            throw new RuntimeException("forced vlc length less than actual length:"+forceVlcLen+"<"+k+" ("+i);
        } else {
           while (k < forceVlcLen) {
               putBit(true);
               k++;
           }
        }
        putBit(false);
        putBits(k*(n-1), i);
    }
}
