/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.sdl;

public class Matrix {
    public boolean isRefSVG = false;
    public int xx = 1<<12, xy = 0, xz = 0, yx = 0, yy = 1<<12, yz = 0;
    public boolean xxValid = false, yyValid = false, xyValid = false, yxValid = false, xzValid = false, yzValid = false;
    public Matrix() {}
    public Matrix(int tab[]) {
        xx = tab[0];
        yx = tab[1];
        xy = tab[2];
        yy = tab[3];
        xz = tab[4];
        yz = tab[5];
        xxValid = xyValid = xzValid = yxValid = yyValid = yzValid = true;
    }
    /**
     * Translation is the only one that can be easily added (left multiplication) to an existing
     * matrix (without matrix multiplication)
     *
     * @param tx horizontal component of the translation
     * @param ty vertical component of the translation
     */
    public void concatTranslation(int tx, int ty) {
        if (xzValid) xz += tx;
        else {
            xzValid = true;
            xz = tx;
        }
        if (yzValid) yz += ty;
        else {
            yzValid = true;
            yz = ty;
        }
    }

    private int mul(int x, int y) {
        long v = x;
        v *= y;
        v >>= 12;
        return (int)v;
    }

    public void concatScale(int sx, int sy) {
        if (xxValid) xx = mul(sx, xx);
        else {
            xx = sx;
            xxValid = true;
        }
        if (yxValid) yx = mul(sy, yx);
        if (xyValid) xy = mul(sx, xy);
        if (yyValid) yy = mul(sy, yy);
        else {
            yy = sy;
            yyValid = true;
        }
        if (xzValid) xz = mul(sx, xz);
        if (yzValid) yz = mul(sy, yz);
    }

    public void concatRotation(int angle) {
        float a = ((float)angle)/4096f;
        float cosa = (float)Math.cos(a);
        float sina = (float)Math.sin(a);
        if (!xxValid) {
            xxValid = true;
            xx = 4096;
        }
        if (!yyValid) {
            yyValid = true;
            yy = 4096;
        }
        if (!xyValid) {
            xyValid = true;
            xy = 0;
        }
        if (!yxValid) {
            yxValid = true;
            yx = 0;
        }
        float xx1 = (cosa * xx) - (sina * yx);
        float yx1 = (sina * xx) + (cosa * yx);
        float xy1 = (cosa * xy) - (sina * yy);
        float yy1 = (sina * xy) + (cosa * yy);
        xx = (int)xx1;
        yx = (int)yx1;
        xy = (int)xy1;
        yy = (int)yy1;
        if (xzValid || yzValid) {
            if (!xzValid) {
                xzValid = true;
                xz = 0;
            }
            if (!yzValid) {
                yzValid = true;
                yz = 0;
            }
            float xz1 = (cosa * xz) - (sina * yz);
            float yz1 = (sina * xz) + (cosa * yz);
            xz = (int)xz1;
            yz = (int)yz1;
        }
    }

    /**
     * detects is a matrix is a simple matrix == scale (sx > 0; sy != 0) then rotation then translation
     *
     * @return null if not a simple matrix, and the angle, scale and transl. factors if a simple matrix
     */
    public Matrix.SimpleMatrix isSimpleMatrix() {
        //System.out.println(this);
        if (!xxValid || !yxValid || !xyValid || !yyValid || !xzValid || !yzValid) {
            if (xzValid && yzValid) {
                // translation OK
                SimpleMatrix sm = new SimpleMatrix();
                if (xxValid && yyValid) {
                    // scale OK, then skew cannot be OK
                    sm.sx = xx;
                    sm.sy = yy;
                } // if scale not OK, no use computing rotation
                sm.tx = xz;
                sm.ty = yz;
                //System.out.println(sm);
                return sm;
            } else {
                // no translation
                if (xxValid && yyValid) {
                    if (yxValid && xyValid) {
                        // scale and skew OK
                        SimpleMatrix sm = computeScaleAndRotation();
                        //System.out.println(sm);
                        return sm;
                    }
                    // only scale OK
                    SimpleMatrix sm = new SimpleMatrix();
                    sm.sx = xx;
                    sm.sy = yy;
                    //System.out.println(sm);
                    return sm;
                }
                //System.out.println("null");
                return null;
            }
        }
        // else all components are OK, so try for full interpretation
        SimpleMatrix sm = computeScaleAndRotation();
        if (sm != null) {
            sm.tx = xz;
            sm.ty = yz;
        }
        //System.out.println(sm);
        return sm;
    }

    private final static int pi = (int)(Math.PI * 4096);

    private Matrix.SimpleMatrix computeScaleAndRotation() {
        int determinant = mul(xx, yy) - mul(xy, yx);
        float fyx = ((float)yx)/4096f;
        float fyy = ((float)yy)/4096f;
        float fxx = ((float)xx)/4096f;
        float fxy = ((float)xy)/4096f;
        if (Math.abs(fxx) > 1E-3f && Math.abs(fyy) > 1E-3f) {
            float v1 = (float)Math.atan(fyx / fxx), v2 = -(float)Math.atan(fxy / fyy);
            if (Math.abs(v1 - v2) < 0.02f) { // 0.02 rad ~= 1 deg
                SimpleMatrix sm = new SimpleMatrix();
                float fangle = (v1 + v2) / 2f;
                sm.angle = (int)(4096f*fangle);
                float cos = (float)Math.cos(fangle);
                float sin = (float)Math.sin(fangle);
                if (Math.abs(cos) > Math.abs(sin)) {
                    sm.sx = (int)(4096f*(fxx / cos));
                    sm.sy = (int)(4096f*(fyy / cos));
                } else {
                    sm.sy = (int)(4096f*(-fxy / sin));
                    sm.sx = (int)(4096f*(fyx / sin));
                }
                if (sm.sx < 0F) {
                    sm.angle = (int)(4096f*((float)Math.PI - fangle));
                    sm.sx = -sm.sx;
                }
                if (determinant > 0 && sm.sy < 0f) {
                    sm.angle = -sm.angle;
                    sm.sy = -sm.sy;
                } else if (determinant < 0 && sm.sy > 0f) {
                    sm.angle = -sm.angle;
                    sm.sy = -sm.sy;
                }
                if (sm.angle < -pi) {
                    sm.angle += 2 * pi;
                } else if (sm.angle > pi) {
                    sm.angle -= 2 * pi;
                }
                return sm;
            }
        } else if (Math.abs(fxx) < 1E-3f && Math.abs(fyy) < 1E-3f) {
            SimpleMatrix sm = new SimpleMatrix();
            sm.angle = pi / 2;
            sm.sx = (int)(4096f*(-fxy));
            sm.sy = (int)(4096f*fyx);
            if (sm.sx < 0F) {
                sm.angle = (int)(4096f*(float)Math.PI) - sm.angle;
                sm.sx = -sm.sx;
            }
            if (determinant > 0 && sm.sy < 0) {
                sm.angle = -sm.angle;
                sm.sy = -sm.sy;
            }
            return sm;
        }
        return null;
    }

    /**
     * A simple matrix is the matrix resulting of a scale (sx > 0; sy != 0) then rotation then translation
     */
    public static class SimpleMatrix {
        public int angle = Integer.MIN_VALUE;
        public int sx = Integer.MIN_VALUE, sy = Integer.MIN_VALUE,
            tx = Integer.MIN_VALUE, ty = Integer.MIN_VALUE;

        public String toString() {
            StringBuffer sb = new StringBuffer("rot:");
            sb.append(angle);
            sb.append(" sc:");
            sb.append(sx);
            sb.append(",");
            sb.append(sy);
            sb.append(" tr:");
            sb.append(tx);
            sb.append(",");
            sb.append(ty);
            return sb.toString();
        }
    }
}
