/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.util;

public final class FloatVector {
    /**
     * The array buffer into which the components of the vector are
     * stored. The capacity of the vector is the length of this array buffer.
     */
    protected float elementData[];
    /**
     * The number of valid components in the vector.
     */
    protected int elementCount;
    /**
     * The amount by which the capacity of the vector is automatically
     * incremented when its delta becomes greater than its capacity. If
     * the capacity is <code>0</code>, the capacity of the vector is
     * doubled each time it needs to grow.
     */
    protected int capacityIncrement;

    /**
     * Constructs an empty vector with the specified initial capacity and
     * capacity increment.
     *
     * @param initialCapacity   the initial capacity of the vector.
     * @param capacityIncrement the amount by which the capacity is
     *                          increased when the vector overflows.
     */
    public FloatVector(int initialCapacity, int capacityIncrement) {
        this.elementData = new float[initialCapacity];
        this.capacityIncrement = capacityIncrement;
    }

    /**
     * Constructs an empty vector with the specified initial capacity.
     *
     * @param initialCapacity the initial capacity of the vector.
     */
    public FloatVector(int initialCapacity) {
        this(initialCapacity, 0);
    }

    /**
     * Constructs an empty vector.
     */
    public FloatVector() {
        this(10);
    }

    /**
     * Increases the capacity of this vector, if necessary, to ensure
     * that it can hold at least the number of components specified by
     * the minimum capacity argument.
     *
     * @param minCapacity the desired minimum capacity.
     */
    public final synchronized void ensureCapacity(int minCapacity) {
        int oldCapacity = elementData.length;
        if (minCapacity > oldCapacity) {
            float oldData[] = elementData;
            int newCapacity = (capacityIncrement > 0) ? (oldCapacity + capacityIncrement) : (oldCapacity * 2);
            if (newCapacity < minCapacity) {
                newCapacity = minCapacity;
            }
            elementData = new float[newCapacity];
            System.arraycopy(oldData, 0, elementData, 0, elementCount);
        }
    }

    /**
     * Returns the number of components in this vector.
     *
     * @return the number of components in this vector.
     */
    public final int size() {
        return elementCount;
    }

    /**
     * Returns the component at the specified index.
     *
     * @param index an index into this vector.
     * @return the component at the specified index.
     * @throws ArrayIndexOutOfBoundsException if an invalid index was
     *                                        given.
     */
    public final synchronized float elementAt(int index) {
        if (index >= elementCount) {
            throw new ArrayIndexOutOfBoundsException(index + " >= " + elementCount);
        }
        try {
            return elementData[index];
        } catch (ArrayIndexOutOfBoundsException e) {
            throw new ArrayIndexOutOfBoundsException(index + " < 0");
        }
    }

    /**
     * Sets the component at the specified <code>index</code> of this
     * vector to be the specified object. The previous component at that
     * position is discarded.
     * <p/>
     * The index must be a value greater than or equal to <code>0</code>
     * and less than the current delta of the vector.
     *
     * @param obj   what the component is to be set to.
     * @param index the specified index.
     * @throws ArrayIndexOutOfBoundsException if the index was invalid.
     * @see java.util.Vector#size()
     */
    public final synchronized void setElementAt(float obj, int index) {
        if (index >= elementCount) {
            throw new ArrayIndexOutOfBoundsException(index + " >= " + elementCount);
        }
        elementData[index] = obj;
    }

    /**
     * Adds the specified component to the end of this vector,
     * increasing its delta by one. The capacity of this vector is
     * increased if its delta becomes greater than its capacity.
     *
     * @param obj the component to be added.
     * @since JDK1.0
     */
    public final synchronized void addElement(float obj) {
        ensureCapacity(elementCount + 1);
        elementData[elementCount++] = obj;
    }

    /**
     * Returns a clone of this vector.
     *
     * @return a clone of this vector.
     */
    public synchronized Object clone() {
        FloatVector v = new FloatVector();
        v.elementData = new float[elementCount];
        System.arraycopy(elementData, 0, v.elementData, 0, elementCount);
        v.elementCount = elementCount;
        return v;
    }

    public void clear() {
        elementCount = 0;
    }
}

final class FloatVectorEnumerator implements FloatEnum {
    FloatVector vector;
    int count;

    FloatVectorEnumerator(FloatVector v) {
        vector = v;
        count = 0;
    }

    public boolean hasMoreElements() {
        return count < vector.elementCount;
    }

    public float nextElement() {
        synchronized (vector) {
            if (count < vector.elementCount) {
                return vector.elementData[count++];
            }
        }
        throw new java.util.NoSuchElementException("IntVectorEnumerator");
    }
}
