/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.util;

import iso.mpeg.streamezzo.sdl.*;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public final class IDManager implements Serializable {
    /**
     * the vector containing the node defs
     */
    private Vector elementIDs = new Vector();
    /**
     * whether there are unused slots in the elementIDs vector
     */
    private boolean elementIDHole = false;
    /**
     * the vector containing the stream defs
     */
    private Vector streamDefs = new Vector();
    /**
     * whether there are unused slots in the streamDefs vector
     */
    private boolean streamHole = false;

    private Vector paintURI = new Vector();

    /**
     * to get the idMan corresponding to an existing def or to create a new idMan for a new def
     *
     * @param s the def name
     * @return the idMan, either existing, or chosen in a hole if any, or next available number
     */
    public int getOrCreateNodeID(String s) {
        if (s == null || s.equals("")) {
            throw new SceneIDParsingException(s);
        }
        int i = elementIDs.indexOf(s);
        if (i < 0) {
            int u = -1;
            if (s.startsWith("ID_")) {
                try {
                    u = Integer.parseInt(s.substring(3));
                } catch (Exception e) {
                }
            }
            if (u != -1 && u > elementIDs.size()) {
                for (int k = 0; k < u - elementIDs.size(); k++) {
                    elementIDs.addElement(null);
                }
                elementIDs.addElement(s);
                elementIDHole = true;
            } else if (u != -1 && elementIDs.elementAt(u - 1) != null && !s.equals(elementIDs.elementAt(u - 1))) {
                throw new SceneIDParsingException(s);
            } else if (u != -1) {
                elementIDs.setElementAt(s, u - 1);
            } else if (elementIDHole) {
                i = elementIDs.indexOf(null);
                if (i > -1) {
                    elementIDs.setElementAt(s, i);
                    return i + 1;
                }
                elementIDHole = false;
            }
            i = elementIDs.size();
            elementIDs.addElement(s);
        }
        return i + 1;
    }

    public boolean isNodeID(String s) {
        if (s == null || s.equals("")) {
            return false;
        }
        return (elementIDs.indexOf(s) >= 0);
    }

    /**
     * to get the idMan corresponding to an existing def or to create a new idMan for a new def
     *
     * @param s the def name
     * @return the idMan, either existing, or chosen in a hole if any, or next available number
     */
    public int getOrCreateStreamID(String s) {
        if (s == null || s.equals("")) {
            throw new SceneIDParsingException(s);
        }
        int i = streamDefs.indexOf(s);
        if (i < 0) {
            int u = -1;
            if (s.startsWith("ID_")) {
                try {
                    u = Integer.parseInt(s.substring(3));
                } catch (Exception e) {
                }
            }
            if (u != -1 && u > streamDefs.size()) {
                for (int k = 0; k < u - streamDefs.size(); k++) {
                    streamDefs.addElement(null);
                }
                streamDefs.addElement(s);
                streamHole = true;
            } else if (u != -1 && streamDefs.elementAt(u) != null && !s.equals(streamDefs.elementAt(u))) {
                throw new SceneIDParsingException(s);
            } else if (u != -1) {
                streamDefs.setElementAt(s, u);
            } else if (streamHole) {
                i = streamDefs.indexOf(null);
                if (i > -1) {
                    streamDefs.setElementAt(s, i);
                    return i;
                }
                streamHole = false;
            }
            i = streamDefs.size();
            streamDefs.addElement(s);
        }
        return i;
    }

    public String getStreamID(int i) {
        if (i < 0 || i > streamDefs.size()) return null;
        return (String)streamDefs.elementAt(i);
    }

    public String getID(int i) {
        if (i <= 0 || i > elementIDs.size()) return null;
        return (String)elementIDs.elementAt(i-1);
    }

    /**
     * to set a def-idMan correspondance, maybe leaving holes
     *
     * @param s the def
     * @param j the idMan
     */
    public void setID(String s, int j) {
        if (j <= 0) {
            throw new SceneIDParsingException(Integer.toString(j));
        }
        j--; // compensate the +1 on the return of getOrCreateNodeID
        if (elementIDs.size() <= j) {
            int max = j - elementIDs.size();
            for (int i = 0; i < max; i++) {
                elementIDs.addElement(null);
                elementIDHole = true;
            }
            elementIDs.addElement(s);
        } else {
            String ss = (String)elementIDs.elementAt(j);
            if (ss != null && !ss.equals(s)) {
                throw new RuntimeException("Duplicate ID: "+ ss + " <> " + s + " j:" + j);
            }
            elementIDs.setElementAt(s, j);
        }
    }

    /**
     * to set a def-id correspondance, maybe leaving holes
     *
     * @param s the def
     * @param j the id
     */
    public void setStreamID(String s, int j) {
        if (streamDefs.size() <= j) {
            int max = j - streamDefs.size();
            for (int i = 0; i < max; i++) {
                streamDefs.addElement(null);
                streamHole = true;
            }
            streamDefs.addElement(s);
        } else {
            String ss = (String)streamDefs.elementAt(j);
            if (ss != null && !ss.equals(s)) {
                throw new RuntimeException("DuplicateStreamID : "+ ss + " <> " + s + " j:" + j);
            }
            streamDefs.setElementAt(s, j);
        }
    }

    private static class Tag {
        QName name = null;
        Vector attrNames = new Vector();
    }

    public int registerPaintURI(URI u) {
        //System.out.println("register "+u+" "+this);
        for (int i = 0; i < paintURI.size(); i++) {
            URI v = (URI)paintURI.elementAt(i);
            if (v.equals(u)) return i;
        }
        paintURI.addElement(u);
        return paintURI.size()-1;
    }

    public URI getPaintURI(int i) {
        //System.out.println("get "+i+" "+this);
        return (URI)paintURI.elementAt(i);
    }
}