/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.SensitiveZone;

public class Bitmap extends RenderedGraphObject
{
	public Xlink m_xlink = new Xlink();
    public int opacity = Integer.MIN_VALUE;
    public int transformBehavior = Integer.MIN_VALUE;
    
    public Object m_cacheImage = null;

    /** the sensitive zone */
    public SensitiveZone m_senstiveZone = new SensitiveZone();

    // x coordinate
    private int m_rActualX;
    private int m_rPrevActualX = 0;
    public int m_rX = 0;

    // y coordinate
    private int m_rActualY;
    private int m_rPrevActualY = 0;
    public int m_rY = 0;

    public int m_rWidth = -1;
    public int m_rHeight = -1;
    private int m_xScaleFactor = 0;
    private int m_yScaleFactor = 0;

    public String contentType = null;
    
    public Bitmap(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
    }

    public int getImageStreamId() {
    	if (m_xlink.getHref() != null)
    		return m_xlink.getHref().streamID;
    	else 
    		throw new RuntimeException("Null href, could not get image stream id");
    }
    
    /**
     * Set DOM X
     * @param rX the new x value
     */
    public void setDOMX(int rX)
    {
        m_rX = rX;
        wasUpdated();
    }

    /**
     * Update actual X
     * @param rNewX the new x value
     */
    public void updateActualX(int rNewX)
    {
        // Check changes
        if (rNewX != m_rPrevActualX)
        {
            m_rActualX = rNewX;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualX = m_rPrevActualX;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual X
     */
    public int getActualX()
    {
        return m_rActualX;
    }

    /**
     * Set DOM CY
     * @param rY the new y value
     */
    public void setDOMY(int rY)
    {
        m_rX = rY;
        wasUpdated();
    }

    /**
     * Update actual Y
     * @param rNewY the new y value
     */
    public void updateActualY(int rNewY)
    {
        // Check changes
        if (rNewY != m_rPrevActualY)
        {
            m_rActualY = rNewY;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualY = m_rPrevActualY;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual Y
     */
    public int getActualY()
    {
        return m_rActualY;
    }

    public void addDOMX(int r_x)
    {
        m_rX += r_x;
    }
    public void addDOMY(int r_y)
    {
        m_rY += r_y;
    }

    public void addDOMWidth(int r_width)
    {
        m_rWidth += r_width;
    }
    public void addDOMHeight(int r_height)
    {
        m_rHeight += r_height;
    }


    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Changed ?
        if (m_rX != m_rActualX || m_rY != m_rActualY)
        {
            // Store previous values
            m_rPrevActualX = m_rActualX;
            m_rPrevActualY = m_rActualY;

            // Set DOM values
            m_rActualX = m_rX;
            m_rActualY = m_rY;

            m_wasUpdated = true;
        }
    }


    public void specificRender(int sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Check visibility
        if ( (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_HIDDEN) ||
             (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_COLLAPSE) )
        {
            return;
        }

        // Apply matrix
        if (parentHasChanged)
        {
            m_wasUpdated = true;
        }

        // apply x/y translation and normalize
        Point2D bitmapXY = m_actualTransform.apply(new Point2D(m_rActualX, m_rActualY));
        bitmapXY.x_r = m_graphObjectManager.m_renderer.normalizeX(bitmapXY.x_r);
        bitmapXY.y_r = m_graphObjectManager.m_renderer.normalizeX(bitmapXY.y_r);

        // Check for resizing
        if ((m_xScaleFactor != m_graphObjectManager.m_renderer.getXScaleFactor())
         || (m_yScaleFactor != m_graphObjectManager.m_renderer.getYScaleFactor()))
        {
            m_xScaleFactor = m_graphObjectManager.m_renderer.getXScaleFactor();
            m_yScaleFactor = m_graphObjectManager.m_renderer.getYScaleFactor();

            int id = getImageStreamId();
            byte array[]= (byte[])super.m_graphObjectManager.getSingleObject(id);
            if (array != null)
            {
                m_cacheImage = m_graphObjectManager.m_renderer.resizeImage(array, m_xScaleFactor, m_yScaleFactor);
            }
        }

        if (m_cacheImage != null)
        {
            // Draw the image
            m_graphObjectManager.m_renderer.drawImageAtCenter(m_cacheImage, bitmapXY.x_r, bitmapXY.y_r);
        }
        else if (getImageStreamId() >= 0)
        {
            byte array[]= (byte[])super.m_graphObjectManager.getSingleObject(getImageStreamId());
            if (array != null)
            {
                m_cacheImage = super.m_graphObjectManager.m_renderer.createImage(array);
            }
        }

        // Compute bounding box and check sensitive zone
        if ((null == useContext) && (m_cacheImage != null))
        {
            m_senstiveZone.setZoneRect(bitmapXY.x_r, bitmapXY.y_r,
                                       m_graphObjectManager.m_renderer.getImageSize(m_cacheImage).x_r,
                                       m_graphObjectManager.m_renderer.getImageSize(m_cacheImage).y_r);
            if (Debug.DEBUG_BBOX) System.out.println(this + " : bbox= (" + m_senstiveZone.m_rect.x_r + "," + m_senstiveZone.m_rect.y_r + "," + m_senstiveZone.m_rect.width_r + "," + m_senstiveZone.m_rect.height_r + ")");
            m_senstiveZone.checkZone(m_graphObjectManager);
        }


        m_wasUpdated = false;
    }

    public void setDOMWidth(int r_width)
    {
        m_rWidth = r_width;
        wasUpdated();
    }

    public void setDOMHeight(int r_height)
    {
        m_rHeight = r_height;
        wasUpdated();
    }
}