/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.BasicRendering;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.tool.SceneLoader;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.TimedElement;

public class Conditional extends RenderedGraphObject implements TimedElement
{
    // Begin time
    public int m_begin = BasicRendering.TIME_INDEFINITE;

    // Buffer that contains encoded LASeR commands
    public byte[] m_buffer = null;

    // The decoder of encoded data
	public SceneLoader m_elementParser = null;

    // the relative decoding parameters
    public int m_relativeIdBits;
    public int m_relativeLenBits;
    public int m_relativeScaleBits;
    public int m_relativeColorIndexBits;
    public int m_relativeResolution;
    public int m_relativeTimeOffset;
    public int m_fontIndexBits;

    public String contentType = null;
    
    // The saved scene time
    private int m_savedSceneTime = BasicRendering.TIME_INDEFINITE;
    
    public byte m_EventPost = EventListener.TIME_BEGIN;

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////// Setter/Getter, construction ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * The constructor
     * @param graphObjectManager
     */
    public Conditional(GraphObjectManager graphObjectManager)
	{
        super(graphObjectManager);
        m_isTimeDependent = true;
	}

    /**
     * Saves parser current properties
     * @param elmtParser
     */
    public void setElementParser(SceneLoader elmtParser)
	{
        // read the scene loader properties
        m_relativeScaleBits = elmtParser.m_decodingContext.m_scaleBits_minus_coordBits;
        m_relativeColorIndexBits = elmtParser.m_decodingContext.m_colorIndexBits;
        m_relativeResolution = elmtParser.m_decodingContext.m_resolution;
        m_relativeTimeOffset = elmtParser.m_decodingContext.m_offsetAppendSceneTime;
        m_fontIndexBits = elmtParser.m_decodingContext.m_fontIndexBits;
		m_elementParser = elmtParser;
	}

////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////// TimedNode implementation ///////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Sets the new begin time of the element
     * @param sceneTimeInMillis
     * @param delay the delay for the begin time to set
     */
    public void setBeginTime(int sceneTimeInMillis, int delay)
    {
        m_begin = sceneTimeInMillis + delay;
    }

    /**
     * Sets the new end time of the element
     * @param sceneTimeInMillis
     * @param delay the delay for end time to set
     */
    public void setEndTime(int sceneTimeInMillis, int delay)
    {
        // no end attribute for Script element
    }

    /**
     * Gets the begin time of an element
     * @return the begin time
     */
    public int getBeginTime()
    {
        return m_begin;
    }


////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// Pre/Post Render ///////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Check if the script needs to be executed
     * @param sceneTimeInMillis
     */
    public void preRender(int sceneTimeInMillis)
    {
        // Check now timestamp
        if (m_begin == BasicRendering.TIME_NOW)
        {
            m_begin = sceneTimeInMillis;
        }

//        System.out.println(m_begin+"    |||   "+sceneTimeInMillis);
        if ((m_begin != BasicRendering.TIME_INDEFINITE) && (m_begin <= sceneTimeInMillis))
        {
            // Register for PostRender
            m_graphObjectManager.registerForPostRender(this);
            m_savedSceneTime = m_begin;
            m_begin = BasicRendering.TIME_INDEFINITE;
            notifyEvent(m_EventPost, new Integer(sceneTimeInMillis));
//            m_EventPost = EventListener.TIME_BEGIN;
//            System.out.println("id : "+this.m_id+" >> "+m_savedSceneTime+"    |||   "+sceneTimeInMillis);
        }
    }


    /**
     * Execute Laser commands
     */
    public void postRender()
    {
        if (Debug.DEBUG_EXECCOND) System.out.println("Script: PostRender()");

        // save the scene loader properties
        int scaleBits = m_elementParser.m_decodingContext.m_scaleBits_minus_coordBits;
        int colorIndexBits = m_elementParser.m_decodingContext.m_colorIndexBits;
        int resolution = m_elementParser.m_decodingContext.m_resolution;
        int timeOffset = m_elementParser.m_decodingContext.m_offsetAppendSceneTime;
        int fontIndexBits = m_elementParser.m_decodingContext.m_fontIndexBits;

        // set the relative ones
        m_elementParser.m_decodingContext.m_scaleBits_minus_coordBits = m_relativeScaleBits;
        m_elementParser.m_decodingContext.m_colorIndexBits = m_relativeColorIndexBits;
        m_elementParser.m_decodingContext.m_resolution = m_relativeResolution;
        m_elementParser.m_decodingContext.m_offsetAppendSceneTime = m_relativeTimeOffset;
        m_elementParser.m_decodingContext.m_fontIndexBits = m_fontIndexBits;

        try
        {
            // have the buffer parsed
            m_elementParser.readUpdateList(m_buffer, m_savedSceneTime);
        }
        catch(Exception e) { 
        	System.out.println(e);
        	e.printStackTrace();
        }

        // set again the saved properties
        // set the relative ones
        m_elementParser.m_decodingContext.m_scaleBits_minus_coordBits = scaleBits;
        m_elementParser.m_decodingContext.m_colorIndexBits = colorIndexBits;
        m_elementParser.m_decodingContext.m_resolution = resolution;
        m_elementParser.m_decodingContext.m_offsetAppendSceneTime = timeOffset;
        m_elementParser.m_decodingContext.m_fontIndexBits = fontIndexBits;
    }

	public void specificRender(int sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext) {
		// TODO I temporarily made this a RenderedGraphObject, for being capable of calling readObject
		// in future we should move the m_properties down in the hierarchy
	}
}