/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.SensitiveZone;

import java.util.Vector;

public class CursorManager extends GraphObject
{
    /** cursor point, relative to the CTM */
    public Point2D m_pointer = new Point2D(0, 0);

    /** actual normalized cursor point */
    public Point2D m_actualPointer = new Point2D(0, 0);

    /** the id of the graph part that represents the cursor */
    public Xlink m_xlink = new Xlink();

    /** a flag telling whether the cursor have to perform an activation (simulates click) */
    public boolean m_bActivated = false;

    /** cursor states : enter, exit, stay inside, click */
    public cursorState m_enterState = null;
    public cursorState m_exitState = null;
    public cursorState m_stayInsideState = null;
    public cursorState m_clickState = null;

    /**
     * Private internal class that defines a cursor state
     */
    private class cursorState
    {
        /** the current cursor event : mouseover, mouseout, click */
        public byte m_event = EventListener.INTERNAL_UNDEFINED;

        /** the sensitive zone in which the event occures */
        public SensitiveZone m_sensitiveZone = null;

        /** the listeners for the current event */
        public Vector m_listeners = new Vector();

        /** constructor */
        public cursorState(byte eventType, SensitiveZone zone, Vector listeners)
        {
            // Set the vent
            m_event = eventType;

            // Copy the listener list
            m_listeners.setSize(0);
            for (int i = 0; i < listeners.size(); i++)
            {
                m_listeners.addElement(listeners.elementAt(i));
            }
        }
        
        /** notify listeners */
        public void notifyListeners()
        {
            GraphObject listener = null;
            for (int i = 0; i < m_listeners.size(); i++)
            {
                listener = (GraphObject)m_listeners.elementAt(i);
                if (listener instanceof EventListener)
                {
                    ((EventListener)listener).update(null, m_event, null);
                }
                else if (listener instanceof A)
                {
                    if (EventListener.MOUSE_CLICK == m_event)
                    {
                        ((A)listener).onEvent(EventListener._UI_ACTIVATE, null);
                    }
                }
            }
        }
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////// Setter/Getter, construction ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * The constructor
     * @param graphObjectManager
     */
    public CursorManager(GraphObjectManager graphObjectManager)
	{
        super(graphObjectManager);
	}

    /**
     * Returns a reference to the normalized cursor position
     */
    public Point2D getPosition()
    {
        return m_actualPointer;
    }
    
    public int getTargetID()
    {
        if (m_xlink.getHref() != null)
        	return m_xlink.getHref().id;
        else
        	return -1;
    }

    /**
     * Set a cursor event and the listeners for this event.
     * It ony store the event for processing it at post rendering step.
     * @param event
     * @param listeners
     * @param zone
     */
    public void setCursorEvent(byte event, SensitiveZone zone, Vector listeners)
    {
        switch (event)
        {
            // Entering
            case EventListener.MOUSE_OVER:
                // store the event
                m_enterState = new cursorState(event, zone, listeners);

                // perhaps it means that a zone lost the focus
                if (null != m_stayInsideState)
                {
                    m_exitState = m_stayInsideState;
                    m_exitState.m_event = EventListener.MOUSE_OUT;
                    m_stayInsideState = null;
                }
                break;

            // Stay inside
            case EventListener.MOUSE_MOVE:
                // store the event
                m_stayInsideState = new cursorState(event, zone, listeners);

                // Any other event is invalid
                m_exitState = null;
                m_clickState = null;
                m_enterState = null;
                break;

            // Exit
            case EventListener.MOUSE_OUT:
                // store the event
                m_exitState = new cursorState(event, zone, listeners);

                // Perhaps it means that it enters into another zone
                if (null != m_stayInsideState)
                {
                    m_enterState = m_stayInsideState;
                    m_enterState.m_event = EventListener.MOUSE_OVER;
                    m_stayInsideState = null;
                }
                break;

            // Click
            case EventListener.MOUSE_CLICK:
                // store the event
                m_clickState = new cursorState(event, zone, listeners);
                break;
        }
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Composition ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Called by the scene composer to compose this node
     * @param sceneTimeInMillis the current qscene time
     * @param parentHasChanged ignored
     * @param parentTransform ignored
     * @param parentProperties ignored
     * @param useContext
     */
    public void render(int sceneTimeInMillis, boolean parentHasChanged,
                       Matrix2D parentTransform, NodeProperties parentProperties, GraphObject useContext)
    {
        // Always post render this element
        m_graphObjectManager.registerForPostRender(this);

        // Get the graphical item that represents the cursor
        GraphObject cursorShape = null;
        if (-1 != getTargetID())
        {
            cursorShape = m_graphObjectManager.getNode(getTargetID());
        }
        if ( (null == cursorShape) || ( ! (cursorShape instanceof RenderedGraphObject)) )
        {
            return;
        }

        // Register itself as active cursor
        m_graphObjectManager.setActiveCursor(this);

        // Compute the active point
        Point2D shapeOrigin = ((RenderedGraphObject)cursorShape).m_actualTransform.getTranslation();
        m_actualPointer.x_r = m_pointer.x_r + shapeOrigin.x_r;
        m_actualPointer.y_r = m_pointer.y_r + shapeOrigin.y_r;
        m_actualPointer.x_r = m_graphObjectManager.m_renderer.normalizeX(m_actualPointer.x_r);
        m_actualPointer.y_r = m_graphObjectManager.m_renderer.normalizeX(m_actualPointer.y_r);
        if (Debug.DEBUG_CURSOR_FULL) 
        	System.out.println("CURSOR : active point = (" +  m_actualPointer.x_r + "," + m_actualPointer.y_r + ")");
//        	System.out.println("SHAPE : active point = (" +  (shapeOrigin.x_r >> 12) + "," + (shapeOrigin.y_r >> 12) + ")");
        	

        m_wasUpdated = false;
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Post rendering /////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Called at post-rendering time
     *
     * It resets the flag indicating if the cursor is activated
     */
    public void postRender()
    {
        // Process events
        if (null != m_exitState)
        {
            m_exitState.notifyListeners();
            if (Debug.DEBUG_CURSOR) System.out.println("CURSOR : notify exit");
        }
        if (null != m_enterState)
        {
            m_enterState.notifyListeners();
            if (Debug.DEBUG_CURSOR) System.out.println("CURSOR : notify enter");
        }
        if (null != m_clickState)
        {
            m_clickState.notifyListeners();
            if (Debug.DEBUG_CURSOR) System.out.println("CURSOR : notify click");
        }

        // Reset internal state
        if (m_bActivated)
        {
            m_bActivated = false;
            if (Debug.DEBUG_CURSOR) System.out.println("CURSOR : deactivation");
        }
        m_enterState = null;
        m_exitState = null;
        m_stayInsideState = null;
        m_clickState = null;
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// OnEvent //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Is called when receiving an event
     * @param eventType the event type
     * @param eventData the associated event data
     */
    public void onEvent(byte eventType, Object eventData)
    {
        if (EventListener.UI_ACTIVATE == eventType)
        {
            m_bActivated = true;
            if (Debug.DEBUG_CURSOR) System.out.println("CURSOR : activation");
        }
    }

}
