/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Polygonizer;
import iso.mpeg.streamezzo.player.library.Real;
import iso.mpeg.streamezzo.player.library.Rect2D;
import iso.mpeg.streamezzo.player.library.SetOfPolygon2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.SensitiveZone;

import java.util.Vector;

/**
 * This class provides a higher level implementation of a 2D polygon shape.
 * Its direct parent shall be a Group node.
 * @see Group
 */
public class Ellipse extends RenderedGraphObject
{
    public final static int POLYGON_INTERPOLATION_PRECISION = 8;

    /** the sensitive zone */
    public SensitiveZone m_senstiveZone = new SensitiveZone();

    // ellipse radius
    public int m_rActualXRadius;
    private int m_rPrevActualXRadius;
    public int m_rXRadius;

    public int m_rActualYRadius;
    private int m_rPrevActualYRadius;
    public int m_rYRadius;

    // center x coordinate
    public int m_rActualCX;
    private int m_rPrevActualCX = 0;
    public int m_rCX = 0;

    // center y coordinate
    public int m_rActualCY;
    private int m_rPrevActualCY = 0;
    public int m_rCY = 0;

    // circle of radius of 1.0 defined as path with quadratic bezier
    private static Vector m_refCirclePoints = null;
    private static byte[] m_refCircleTypes = null;

    // circle of radius m_r defined as path with cubic bezier
    private Vector m_circlePoints = null;

    // The interpolated path
    private SetOfPolygon2D m_interpolatedPath = null;

    // The bounding box
    public Rect2D m_bbox = new Rect2D(-1, -1, -1, -1);

    {
        // Initialize the reference circle of radius 1.0
        // d="M 0,1 C 0.552,1 1,0.552 1,0 C 1,-0.552 0.552,-1 0,-1 C -0.552,-1 -1,-0.552 -1,0 C -1,0.552 -0.552,1 0,1 z"
        // 'magicValue' = (sqrt(2)-1)*4/3 = 0.552 
        // Circle points
        m_refCirclePoints = new Vector(13);
        int r_magicValue = Real.applyMul(Real.applySub(Real.sqrt(Real.convertIntToRealInt(2)), Real.VALUE1), Real.applyDiv(Real.convertIntToRealInt(4), Real.convertIntToRealInt(3)));
        m_refCirclePoints.addElement(new Point2D(0, Real.VALUE1));
        m_refCirclePoints.addElement(new Point2D(r_magicValue, Real.VALUE1));
        m_refCirclePoints.addElement(new Point2D(Real.VALUE1, r_magicValue));
        m_refCirclePoints.addElement(new Point2D(Real.VALUE1, 0));
        m_refCirclePoints.addElement(new Point2D(Real.VALUE1, -r_magicValue));
        m_refCirclePoints.addElement(new Point2D(r_magicValue, -Real.VALUE1));
        m_refCirclePoints.addElement(new Point2D(0, -Real.VALUE1));
        m_refCirclePoints.addElement(new Point2D(-r_magicValue, -Real.VALUE1));
        m_refCirclePoints.addElement(new Point2D(-Real.VALUE1, -r_magicValue));
        m_refCirclePoints.addElement(new Point2D(-Real.VALUE1, 0));
        m_refCirclePoints.addElement(new Point2D(-Real.VALUE1, r_magicValue));
        m_refCirclePoints.addElement(new Point2D(-r_magicValue, Real.VALUE1));
        m_refCirclePoints.addElement(new Point2D(0, Real.VALUE1));

        // Circle types
        m_refCircleTypes = new byte[5];
        // implicit m_refCircleTypes[0] = 0; // MoveTo
        m_refCircleTypes[0] = 3; // Quadratic CurveTo
        m_refCircleTypes[1] = 3; // Quadratic CurveTo
        m_refCircleTypes[2] = 3; // Quadratic CurveTo
        m_refCircleTypes[3] = 3; // Quadratic CurveTo
        m_refCircleTypes[4] = 2; // Close

        m_wasUpdated = true;
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////// Properties setters/getters /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    public Ellipse(GraphObjectManager graphObjectManager)
    {
        super( graphObjectManager);
        m_actualTransform = Matrix2D.createIdentityMatrix();

        m_circlePoints = new Vector(13);
        for(int i=0;i<m_refCirclePoints.size();i++)
        {
            m_circlePoints.addElement(new Point2D(0, 0));
        }
    }

    public void setDOMRadiusX(int r_radius)
    {
        m_rXRadius = r_radius;
        wasUpdated();
    }
    public void setDOMRadiusY(int r_radius)
    {
        m_rYRadius = r_radius;
        wasUpdated();
    }

    /**
     * Update actual x radius
     * @param rNewXRadius New x radius (real 20.12)
     */
    public void updateActualXRadius(int rNewXRadius)
    {
        // Check changes
        if (rNewXRadius != m_rPrevActualXRadius)
        {
            m_rActualXRadius = rNewXRadius;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualXRadius = m_rPrevActualXRadius;
            m_wasUpdated = false;
        }
    }

    /**
     * Update actual y radius
     * @param rNewYRadius New y radius (real 20.12)
     */
    public void updateActualYRadius(int rNewYRadius)
    {
        // Check changes
        if (rNewYRadius != m_rPrevActualYRadius)
        {
            m_rActualYRadius = rNewYRadius;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualYRadius = m_rPrevActualYRadius;
            m_wasUpdated = false;
        }
    }

        /**
     * Set DOM CX
     * @param rX the new center x value
     */
    public void setDOMX(int rX)
    {
        m_rCX = rX;
        wasUpdated();
    }

    /**
     * Update actual CX
     * @param rNewX the new center x value
     */
    public void updateActualX(int rNewX)
    {
        // Check changes
        if (rNewX != m_rPrevActualCX)
        {
            m_rActualCX = rNewX;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualCX = m_rPrevActualCX;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual CX
     */
    public int getActualX()
    {
        return m_rActualCX;
    }

    /**
     * Set DOM CY
     * @param rY the new center y value
     */
    public void setDOMY(int rY)
    {
        m_rCY = rY;
        wasUpdated();
    }

    /**
     * Update actual CY
     * @param rNewY the new center x value
     */
    public void updateActualY(int rNewY)
    {
        // Check changes
        if (rNewY != m_rPrevActualCY)
        {
            m_rActualCY = rNewY;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualCY = m_rPrevActualCY;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual CY
     */
    public int getActualY()
    {
        return m_rActualCY;
    }

    public void addDOMX(int r_x)
    {
        m_rCX += r_x;
    }
    public void addDOMY(int r_y)
    {
        m_rCY += r_y;
    }

    public void addDOMRadiusX(int r_radius)
    {
        m_rXRadius += r_radius;
    }
    public void addDOMRadiusY(int r_radius)
    {
        m_rYRadius += r_radius;
    }

    /**
     * Generates an ellipse with the desired radius
     */
    private void generateEllipseAsPath()
    {
        Point2D refPoint = null, point = null;
        // Go through each coordinates of the reference circle of radius 1.0 and
        // multiply by the desired radius to generate the corresponding path
        for(int i=0;i<m_refCirclePoints.size();i++)
        {
            refPoint = (Point2D)m_refCirclePoints.elementAt(i);
            point = (Point2D)m_circlePoints.elementAt(i);
            point.x_r = Real.applyAdd(Real.applyMul(refPoint.x_r, m_rActualXRadius), m_rActualCX);
            point.y_r = Real.applyAdd(Real.applyMul(refPoint.y_r, m_rActualYRadius), m_rActualCY);
        }
    }

    public void specificRender(int a_sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Check if it must be drawn
        if ( (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_HIDDEN) ||
             (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_COLLAPSE) )
        {
            return;
        }

        // need to reprocess the interpolation if the path was modified
        if (m_wasUpdated)
        {
            generateEllipseAsPath();
            m_interpolatedPath = Polygonizer.generateInterpolatedPath(m_circlePoints, m_refCircleTypes,
                    POLYGON_INTERPOLATION_PRECISION);
        }

        // render the path
        if (m_interpolatedPath != null)
        {
            // Fill the whole path
            if (m_actualProperties.m_fill != Value.COLOR_NONE)
            {
                // set the color
                m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_fill, m_actualProperties.m_fillOpacity&0xFF);
                // set the opacity
                m_interpolatedPath.applyMatrixAndfillPolygonSet(m_actualTransform, m_graphObjectManager.m_renderer);
            }

            // Draw polylines
            if (m_actualProperties.m_stroke != Value.STROKE_NONE)
            {
                m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_stroke, m_actualProperties.m_strokeOpacity&0xFF);
                m_interpolatedPath.applyMatrixAndDrawPolyLineSet(m_actualTransform, m_graphObjectManager.m_renderer,
                                                                     m_actualProperties.m_strokeWidth);
            }

            // Get the bounding box and check sensitive zone
            if (null == useContext)
            {
                m_interpolatedPath.getBoundingBox(m_senstiveZone.getZone());
                if (Debug.DEBUG_BBOX) System.out.println(this + " : bbox= (" + m_senstiveZone.m_rect.x_r + "," + m_senstiveZone.m_rect.y_r + "," + m_senstiveZone.m_rect.width_r + "," + m_senstiveZone.m_rect.height_r + ")");
                m_senstiveZone.checkZone(m_graphObjectManager);
            }
        }

        m_wasUpdated = false;
    }

    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Changed ?
        if ((m_rXRadius != m_rActualXRadius) || (m_rYRadius != m_rActualYRadius)
         || (m_rCX != m_rActualCX) || (m_rCY != m_rActualCY))
        {
            // Store previous values
            m_rPrevActualXRadius = m_rActualXRadius;
            m_rPrevActualYRadius = m_rActualYRadius;
            m_rPrevActualCX = m_rActualCX;
            m_rPrevActualCY = m_rActualCY;

            // Set DOM values
            m_rActualXRadius = m_rXRadius;
            m_rActualYRadius = m_rYRadius;
            m_rActualCX = m_rCX;
            m_rActualCY = m_rCY;

            m_wasUpdated = true;
        }
    }
}
