/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Polygon2D;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.tool.Rendering;
import iso.mpeg.streamezzo.player.util.AnimationManager;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.EventManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.QName;

import java.util.Vector;


/**
 * A scene is defined by a graph of GraphObjects.Each scene is handled by its
 * Root node that shall be a Group Node. The method render() of the root
 * node launchs the rendering process of the graph. The rendering process is done
 * by the BasicRenderer.
 */

public abstract class GraphObject
{
    //newly added attributes
	public String stringId = null;
	public QName[] privateAttributes = null;
    public static int autoID = Integer.MAX_VALUE;

    public Vector m_children;
    protected int m_id = -1;

    /** The parent node */
    protected GraphObject m_parent = null;

    /** The graphObject manager */
    protected GraphObjectManager m_graphObjectManager;

    /** If a change occures */
    public boolean m_wasUpdated = true;

    /** a flag which tells if pre rendering is necessary */
    protected boolean m_needPreRender = true;

    /** EventManager */
    protected EventManager m_eventManager = null;

    /** Used to know if the node has begin/end properties */
    protected boolean m_isTimeDependent = false;

    /** AnimationManager */
    protected AnimationManager m_animationManager = null;


////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////// Construction, Getters/Setters ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Default constructor, the DefName is not set.
     */
    public GraphObject(GraphObjectManager graphObjectManager)
    {
        m_graphObjectManager = graphObjectManager;
        m_children = new Vector();
    }

    /**
     * Default constructor with set the DefName.
     * When the DefName is set, the object is referenced in the DefName table
     * and it can be reach by calling the method getID(). In order to release
     * the memory of this object, one must call deleteNode(String)
     * If no DefName is set, then the object cannot be retreived but by using
     * its parent.
     * @param   id the DefName of this object.
     * @see     #getID
     * @see     #setID
     */
    public GraphObject(int id)
    {
        setID(id);
    }

     /**
     * When the DefName is set, the object is referenced in the DefName table
     * and it can be reach by calling the method getID(). In order to release
     * the memory of this object, one must call deleteNode(String)
     * @param   id the id of this object.
     * @see     #getID
     */
    public void setID(int id)
    {
        m_id = id;

        m_graphObjectManager.registerNode(id, this);
    }

    public int getID()
    {
        return m_id;
    }

    public void setAutoID() {
        m_id = autoID--;
        m_graphObjectManager.registerNode(m_id, this);
    }

    public void setParent(GraphObject a_parent)
    {
        m_wasUpdated = true;
        m_parent = a_parent;
    }

    /**
     * Give the parent node. If the parent is null, then it indicates that the
     * current object is the root node of the scene.
     * @return     GraphOject the parent object
     */
    public GraphObject getParent()
    {
        return m_parent;
    }

    public void setRenderer(Rendering a_renderer)
    {
        m_graphObjectManager.setRenderer(a_renderer);
    }

    /**
    * Notify that some properties influencing rendering have changed.
    * This oprations set flags to ask for a pre-render.
    */
    public void wasUpdated()
    {
        m_wasUpdated = true;
        m_needPreRender = true;
    }

    /**
     * Tell whether the node is time dependant or not i.e. it has begin/end properties
     * @return the time dependancy value
     */
    public boolean isTimeDependent()
    {
        return m_isTimeDependent;
    }

    /**
     * In the case of a node targeting another node (i.e. Trigger, Animation, Use),
     * it returns the targeted node ID
     * @return the ID of the targeted node
     */
    public int getTargetID()
    {
        return -1;
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// Children management ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * This operation add a new child to this node.
     *
     * @param child The sibling of the current node
     */
    public void appendChild(Object child)
    {
        if (child != null)
        {
            if (m_children == null) m_children = new Vector();
            m_children.addElement(child);
            if (child instanceof GraphObject) ((GraphObject)child).setParent(this);
            wasUpdated();
        }
    }

    /**
     * This operation insert a child at the specified position.
     *
     * @param child The new Child replacing the previous.
     * @param childPos The position of the replaced Child.
     */
    public void insertChild(int childPos, GraphObject child)
    {
        if ( (childPos > m_children.size()) || (childPos < 0) )
        {
            // At the end
            m_children.addElement(child);
        }
        else
        {
            m_children.insertElementAt(child, childPos);
        }

        child.setParent(this);
        wasUpdated();
    }

    /**
     * Remove the child identified by its position.
     * @param childPos The position of the child to remove
     */
    public void removeChild(int childPos)
    {
        if (m_children.size() > childPos)
        {
            // Get the child and remove it
            GraphObject child = (GraphObject) m_children.elementAt(childPos);
            m_children.removeElementAt(childPos);

            // Unregister this node
            m_graphObjectManager.unRegisterNode(child);

            // Remove all children of this child
            child.removeAllChildren();

            wasUpdated();
        }
    }

    /**
     * Remove the passed-in child
     * @param child The child object to remove
     */
    public void removeChild(GraphObject child)
    {
        if (m_children.contains(child))
        {
            // Remove it
            m_children.removeElement(child);

            // Unregister this node
            m_graphObjectManager.unRegisterNode(child);

            // Remove all children of this child
            child.removeAllChildren();

            wasUpdated();
        }
    }

    /**
     * Remove all children
     */
    public void removeAllChildren()
    {
        wasUpdated();

        // Iterate through children
        Object child = null;
        for (int childIndex = 0; childIndex < m_children.size(); childIndex++)
        {
            // Get a child
            child = m_children.elementAt(childIndex);

            // Unregister this node
            if (child instanceof GraphObject)
            {
                m_graphObjectManager.unRegisterNode(((GraphObject) child));
            }

            // Bitmap is a special case
            if (child instanceof Bitmap)
            {
                ((Bitmap) child).m_cacheImage = null;
            }
            else
                if (child instanceof GraphObject)
                {
                    ((GraphObject) child).removeAllChildren();
                }
        }

        //clean the vector by removing all elements
        m_children.removeAllElements();
    }

    /**
     * The children count
     * @return the number of children
     */
    public int getChildrenCount()
    {
        return m_children.size();
    }

    /**
     * Return the child at the requested index
     * @param index the index of the child
     * @return the child at the requested index
     */
    public Object getChild(int index)
    {
        if (index >= 0 && index < m_children.size())
        {
            return (GraphObject)m_children.elementAt(index);
        }
        else
        {
            return null;
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Rendering operations ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    public void preRender(int sceneTimeInMillis)
    {
        // The implementation of this function is located in the appropriate node that need to use this function.
    }

    public void render(int sceneTimeInMillis, boolean parentHasChanged,
                       Matrix2D parentTransform, NodeProperties parentProperties, GraphObject useContext)
    {
        m_wasUpdated = false;
    }

    public void postRender()
    {
       // The implementation of this function is located in the appropriate node that need to use this function.
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Scene management ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /*public final void deleteScene()
    {
        m_graphObjectManager.deleteScene();
    }*/

    public final void cleanScene()
    {
        m_graphObjectManager.cleanScene();
    }

    /*public void setSingleObject(int streamID, Object data, boolean isPermanent)
    {
        m_graphObjectManager.setSingleObject(streamID, data, isPermanent);
    }*/

    /*public Object getSingleObject(int streamID)
    {
        return m_graphObjectManager.getSingleObject(streamID);
    }*/


////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// Event Listeners Management ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Registers a listener
     * @param listener
     */
    public void attachEventListener(EventListener listener)
    {
        if (m_eventManager == null)
        {
            m_eventManager = new EventManager();
        }
        m_eventManager.attach(listener);
    }

    /**
     * Removes a listener from the list
     * @param listener
     */
    public void detachEventListener(EventListener listener)
    {
        if (m_eventManager != null)
        {
            m_eventManager.detach(listener);
        }
    }

    /**
     * Notifies all the regitered listeners
     * @param eventType the event type
     * @param eventData the associated event data
     */
    public void notifyEvent(byte eventType, Object eventData)
    {
        if (m_eventManager != null)
        {
            m_eventManager.notify(this, eventType, eventData);
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// OnEvent //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Is called when receiving an event
     * @param event the event type
     * @param eventData the associated event data
     */
    public void onEvent(Event event, Object eventData) {}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// Animation Management /////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Registers an animation
     * @param anim
     */
    public void attachAnimation(SVGset anim)
    {
        if (m_animationManager == null)
        {
            m_animationManager = new AnimationManager();
        }
        m_animationManager.attach(anim);
    }

    /**
     * Removes an animation from the list
     * @param anim
     */
    public void detachAnimation(SVGset anim)
    {
        if (m_animationManager != null)
        {
            m_animationManager.detach(anim);
        }
    }

    /**
     * Process all animations
     * @param sceneTimeInMillis
     */
    public void animateAll(int sceneTimeInMillis)
    {
        if (m_animationManager != null)
        {
            m_animationManager.animateAll(sceneTimeInMillis);
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// DOM Updates management ///////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
    public void setDOMEditable(boolean isTrue)
    {
        // implemented by derived class
    }
    public void setDOMFocusable(boolean isTrue)
    {
        // implemented by derived class
    }
    public void setDOMZoomAndPan(boolean isTrue)
    {
        // implemented by derived class
    }
    public void setDOMDisplay(byte display)
    {
        // implemented by derived class
    }
    public void setDOMDisplayAlign(byte displayAlign)
    {
        // implemented by derived class
    }
    public void setDOMVisibility(byte visibility)
    {
        // implemented by derived class
    }
    public void setDOMTextAlign(byte textAlign)
    {
        // implemented by derived class
    }
    public void setDOMChoice(int choice)
    {
        // implemented by derived class
    }
    public void setDOMFontStyle(byte fontstyle)
    {
        // implemented by derived class
    }
    public void setDOMHRef(int href)
    {
        // implemented by derived class
    }
    public void setDOMFillColor(int color)
    {
        // implemented by derived class
    }
    public void addDOMFillColor(int color)
    {
        // implemented by derived class
    }
    public void setDOMStroke(int color)
    {
        // implemented by derived class
    }
    public void addDOMStroke(int color)
    {
        // implemented by derived class
    }
    public void setDOMBackgroundFill(int color)
    {
        // implemented by derived class
    }
    public void addDOMBackgroundFill(int color)
    {
        // implemented by derived class
    }
    public void setDOMFillOpacity(byte opacity)
    {
        // implemented by derived class
    }
    public void addDOMFillOpacity(byte opacity)
    {
        // implemented by derived class
    }
    public void setDOMStrokeOpacity(byte opacity)
    {
        // implemented by derived class
    }
    public void addDOMStrokeOpacity(byte opacity)
    {
        // implemented by derived class
    }
    public void setDOMStrokeWidth(int r_width)
    {
        // implemented by derived class
    }
    public void addDOMStrokeWidth(int r_width)
    {
        // implemented by derived class
    }
    public void setDOMFontSize(int r_fontsize)
    {
        // implemented by derived class
    }
    public void addDOMFontSize(int r_fontsize)
    {
        // implemented by derived class
    }
    public void setDOMRotation(int r_rotation)
    {
        // implemented by derived class
    }
    public void addDOMRotation(int r_rotation)
    {
        // implemented by derived class
    }
    public void setDOMWidth(int r_width)
    {
        // implemented by derived class
    }
    public void addDOMWidth(int r_width)
    {
        // implemented by derived class
    }
    public void setDOMHeight(int r_height)
    {
        // implemented by derived class
    }
    public void addDOMHeight(int r_height)
    {
        // implemented by derived class
    }
    public void setDOMRadius(int r_radius)
    {
        // implemented by derived class
    }
    public void addDOMRadius(int r_radius)
    {
        // implemented by derived class
    }
    public void setDOMRadiusX(int r_radius)
    {
        // implemented by derived class
    }
    public void addDOMRadiusX(int r_radius)
    {
        // implemented by derived class
    }
    public void setDOMRadiusY(int r_radius)
    {
        // implemented by derived class
    }
    public void addDOMRadiusY(int r_radius)
    {
        // implemented by derived class
    }
    public void setDOMX2(int r_x)
    {
        // implemented by derived class
    }
    public void addDOMX2(int r_x)
    {
        // implemented by derived class
    }
    public void setDOMY2(int r_y)
    {
        // implemented by derived class
    }
    public void addDOMY2(int r_y)
    {
        // implemented by derived class
    }
    public void setDOMX(int r_x)
    {
        // implemented by derived class
    }
    public void addDOMX(int r_x)
    {
        // implemented by derived class
    }
    public void setDOMY(int r_y)
    {
        // implemented by derived class
    }
    public void addDOMY(int r_y)
    {
        // implemented by derived class
    }
    public void setDOMScale(Point2D scale)
    {
        // implemented by derived class
    }
    public void addDOMScale(Point2D scale)
    {
        // implemented by derived class
    }
    public void setDOMSize(Point2D size)
    {
        // implemented by derived class
    }
    public void addDOMSize(Point2D size)
    {
        // implemented by derived class
    }
    public void setDOMTranslation(Point2D translation)
    {
        // implemented by derived class
    }
    public void addDOMTranslation(Point2D translation)
    {
        // implemented by derived class
    }
    public void setDOMTransform(Matrix2D matrix)
    {
        // implemented by derived class
    }
    public void addDOMTransform(Matrix2D matrix)
    {
        // implemented by derived class
    }
    public void setDOMUrl(String url)
    {
        // implemented by derived class
    }
    public void addDOMUrl(String url)
    {
        // implemented by derived class
    }
    public void setDOMPoints(Polygon2D polygon, int index)
    {
        // implemented by derived class
    }
    public void addDOMPoints(Polygon2D polygon)
    {
        // implemented by derived class
    }
    public void setDOMD(Polygon2D polygon, byte[] types)
    {
        // implemented by derived class
    }
    public void setDOMViewBox(int[] viewbox)
    {
        // implemented by derived class
    }
    public void addDOMViewBox(int[] viewbox)
    {
        // implemented by derived class
    }
    public void setDOMTextContent(Vector textContent)
    {
        // implemented by derived class
    }
    public void addDOMTextContent(Vector textContent)
    {
        // implemented by derived class
    }
    public void setDOMFontFamily(String fontfamily)
    {
        // implemented by derived class
    }


    /**
     * Set a new attribute value into the dom tree
     * @param attributeID attribute id that identify the target attibute
     * @param value the new value for the target attribute
     * @param index index into target attribute value
     */
    // TODO when passed to the new constants, there were a few without a match
    // TODO all the values that are not implemented should not be treated, and the default case should show a WARNING message
    public void setDOMAttribute(int attributeID, Value value, int index)
    {
    	wasUpdated();
    	
    	switch (attributeID)
    	{
    	case Value.UPDATE_ACCUMULATE :
    		//todo
    		break;
    	case Value.UPDATE_ADDITIVE :
    		//todo
    		break;
//    	case Value.UPDATE_APPEND :
    		//todo
//    		break;
    	case Value.UPDATE_EDITABLE :
    		setDOMEditable(value.m_boolean);
    		break;
    	case Value.UPDATE_FOCUSABLE :
    		setDOMFocusable(value.m_boolean);
    		break;
//    	case Value.UPDATE_ANIMFILL :
    		//todo
//    		break;
    	case Value.UPDATE_EXTERNALRESOURCESREQUIRED :
    		//todo
    		break;
    	case Value.UPDATE_ZOOMANDPAN :
    		setDOMZoomAndPan(value.m_boolean);
    		break;
    	case Value.UPDATE_CALCMODE :
    		//todo
    		break;
    	case Value.UPDATE_DISPLAY :
    		setDOMDisplay((byte)value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_FILLRULE :
    		//todo
    		break;
    	case Value.UPDATE_RESTART :
    		//todo
    		break;
    	case Value.UPDATE_DISPLAYALIGN :
    		setDOMDisplayAlign((byte)value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_VISIBILITY :
    		setDOMVisibility((byte)value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_TEXTANCHOR :
    		setDOMTextAlign((byte)value.m_enumOrColorOrRealOrID);
    		break;
//    	case Value.UPDATE_TYPE :
    		//todo
//    		break;
    	case Value.UPDATE_STROKELINECAP :
    		//todo
    		break;
    	case Value.UPDATE_STROKELINEJOIN :
    		//todo
    		break;
    	case Value.UPDATE_OVERFLOW :
    		//todo
    		break;
    	case Value.UPDATE_FONTVARIANT :
    		//todo
    		break;
    	case Value.UPDATE_COLORRENDERING :
    		//todo
    		break;
    	case Value.UPDATE_IMAGERENDERING :
    		//todo
    		break;
    	case Value.UPDATE_SHAPERENDERING :
    		//todo
    		break;
    	case Value.UPDATE_TEXTRENDERING :
    		//todo
    		break;
//    	case Value.UPDATE_TIMEATTRIBUTE :
    		//todo
//    		break;
//    	case Value.UPDATE_ATTRIBUTENAME :
    		//todo
//    		break;
    	case Value.UPDATE_EVENT :
    		//todo
    		break;
    	case Value.UPDATE_CHOICE :
    		setDOMChoice(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_FONTSTYLE :
    		setDOMFontStyle((byte)value.m_enumOrColorOrRealOrID);
    		break;
//  		case Value.UPDATE_KEYCODE :
//  		//todo
//  		break;
    	case Value.UPDATE_PRESERVEASPECTRATIO :
    		//todo
    		break;
    	case Value.UPDATE_REPEATCOUNT :
    		//todo
    		break;
    	case Value.UPDATE_POINTEREVENTS :
    		//todo
    		break;
    	case Value.UPDATE_FONTWEIGHT :
    		//todo
    		break;
    	case Value.UPDATE_FONTFAMILY :
    		setDOMFontFamily(value.m_string);
    		break;
//  		case Value.UPDATE_ID :
//  		//todo
//  		break;
//  		case Value.UPDATE_MPATH :
//  		//todo
//  		break;
//  		case Value.UPDATE_NAVINDEX :
//  		//todo
//  		break;
//  		case Value.UPDATE_NAVUP :
//  		//todo
//  		break;
//  		case Value.UPDATE_NAVDOWN :
//  		//todo
//  		break;
//  		case Value.UPDATE_NAVLEFT :
//  		//todo
//  		break;
//  		case Value.UPDATE_NAVRIGHT :
//  		//todo
//  		break;
    		
    	case Value.UPDATE_XLINK_HREF :
    		setDOMHRef(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_OBSERVER :
    		//todo
    		break;
    	case Value.UPDATE_COLOR_FILL :
    		setDOMFillColor(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_STROKE :
    		setDOMStroke(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_STOPCOLOR :
    		//todo
    		break;
    	case Value.UPDATE_COLOR :
    		//todo
    		break;
    	case Value.UPDATE_VIEWPORTFILL :
    		setDOMBackgroundFill(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_FILLOPACITY :
    		setDOMFillOpacity((byte)value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_STROKEOPACITY :
    		setDOMStrokeOpacity((byte)value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_STOPOPACITY :
    		//todo
    		break;
    	case Value.UPDATE_VIEWPORTFILLOPACITY :
    		//todo
    		break;
    	case Value.UPDATE_STROKEWIDTH :
    		setDOMStrokeWidth(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_STROKEDASHOFFSET :
    		//todo
    		break;
    	case Value.UPDATE_STROKEMITERLIMIT :
    		//todo
    		break;
    	case Value.UPDATE_FONTSIZE :
    		setDOMFontSize(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_ROTATION :
    		setDOMRotation(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_HEIGHT :
    		setDOMHeight(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_WIDTH :
    		setDOMWidth(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_R :
    		setDOMRadius(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_RX :
    		setDOMRadiusX(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_RY :
    		setDOMRadiusY(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_X2 :
    		setDOMX2(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_Y2 :
    		setDOMY2(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_X :
    		setDOMX(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_Y :
    		setDOMY(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_X1 :
    		setDOMX(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_Y1 :
    		setDOMY(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_CX :
    		setDOMX(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_CY :
    		setDOMY(value.m_enumOrColorOrRealOrID);
    		break;
    	case Value.UPDATE_BEGIN :
    		//todo
    		break;
    	case Value.UPDATE_END :
    		//todo
    		break;
    	case Value.UPDATE_DUR :
    		//todo
    		break;
    	case Value.UPDATE_DELAY :
    		//todo
    		break;
    	case Value.UPDATE_REPEATDUR :
    		//todo
    		break;
    	case Value.UPDATE_SIZE :
    		setDOMSize(value.m_point);
    		break;
    	case Value.UPDATE_TRANSFORM :
    		setDOMTransform(value.m_matrix);
    		break;
    	case Value.UPDATE_REQUIREDFORMATS :
    		//todo
    		break;
    	case Value.UPDATE_REQUIREDEXTENSIONS :
    		//todo
    		break;
    	case Value.UPDATE_SYSTEMLANGUAGE :
    		//todo
    		break;
    	case Value.UPDATE_XML_LANG :
    		//todo
    		break;
    	case Value.UPDATE_XML_BASE :
    		//todo
    		break;
//  		TODO url doesn't exist any more
//  		case Value.UPDATE_URL :
//  		setDOMUrl(value.m_string);
//  		break;
    	case Value.UPDATE_POINTS :
    		setDOMPoints(value.m_polygon, index);
    		break;
    	case Value.UPDATE_D :
    		setDOMD(value.m_polygon, value.m_types);
    		break;
    	case Value.UPDATE_VIEWBOX :
    		setDOMViewBox(value.m_floatsOrInts);
    		break;
    	case Value.UPDATE_REQUIREDFEATURES :
    		//todo
    		break;
//    	case Value.UPDATE_KEYCODES :
    		//todo
//    		break;
    	case Value.UPDATE_STROKEDASHARRAY :
    		//todo
    		break;
    	case Value.UPDATE_KEYPOINTS :
    		//todo
    		break;
    	case Value.UPDATE_KEYSPLINES :
    		//todo
    		break;
    	case Value.UPDATE_KEYTIMES :
    		//todo
    		break;
    	case Value.UPDATE_TEXTCONTENT :
    		setDOMTextContent(value.m_textContent);
    		break;
    	case Value.UPDATE_TRANSLATION :
    		setDOMTranslation(value.m_point);
    		break;
    	case Value.UPDATE_SCALE :
    		setDOMScale(value.m_point);
    		break;
    	case Value.UPDATE_ROTATE :
    		setDOMRotation(value.m_enumOrColorOrRealOrID);
    		break;
//  		case Value.UPDATE_SKEWX :
//  		//todo
//  		break;
//  		case Value.UPDATE_TRANSFORMSKEWY :
//  		//todo
//  		break;
    	}
    }



    /**
     * Set a new attribute value into the dom tree. This new value is the result of addition
     * between passed-in value and current dom value.
     * @param attributeTag attribute id that identify the target attibute
     * @param valueToAdd the value to add to the current dom value
     */
    // TODO when passed to the new constants, there were a few without a match
    // TODO all the values that are not implemented should not be treated, and the default case should show a WARNING message
    public void addDOMAttribute(int attributeTag, Value valueToAdd)
    {
        wasUpdated();

        switch (attributeTag)
        {
//            case Value.UPDATE_TIMEATTRIBUTE :
                //todo
//                break;
            case Value.UPDATE_REPEATCOUNT :
                //todo
                break;
            case Value.UPDATE_FONTWEIGHT :
                //todo
                break;
//            case Value.UPDATE_NAVINDEX :
//                //todo
//                break;
//            case Value.UPDATE_NAVUP :
//                //todo
//                break;
//            case Value.UPDATE_NAVDOWN :
//                //todo
//                break;
//            case Value.UPDATE_NAVLEFT :
//                //todo
//                break;
//            case Value.UPDATE_NAVRIGHT :
//                //todo
//                break;
            case Value.UPDATE_COLOR_FILL :
                addDOMFillColor(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_STROKE :
                addDOMStroke(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_STOPCOLOR :
                //todo
                break;
            case Value.UPDATE_COLOR :
                //todo
                break;
            case Value.UPDATE_VIEWPORTFILL :
                addDOMBackgroundFill(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_FILLOPACITY :
                addDOMFillOpacity((byte)valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_STROKEOPACITY :
                addDOMStrokeOpacity((byte)valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_STOPOPACITY :
                //todo
                break;
            case Value.UPDATE_VIEWPORTFILLOPACITY :
                //todo
                break;
            case Value.UPDATE_STROKEWIDTH :
                addDOMStrokeWidth(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_FONTSIZE :
                addDOMFontSize(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_ROTATION :
                addDOMRotation(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_HEIGHT :
                addDOMHeight(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_WIDTH :
                addDOMWidth(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_R :
                addDOMRadius(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_RX :
                addDOMRadiusX(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_RY :
                addDOMRadiusY(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_X2 :
                addDOMX2(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_Y2 :
                addDOMY2(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_X :
                addDOMX(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_Y :
                addDOMY(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_X1 :
                addDOMX(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_Y1 :
                addDOMY(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_CX :
                addDOMX(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_CY :
                addDOMY(valueToAdd.m_enumOrColorOrRealOrID);
                break;
            case Value.UPDATE_BEGIN :
                //todo
                break;
            case Value.UPDATE_END :
                //todo
                break;
            case Value.UPDATE_DUR :
                //todo
                break;
            case Value.UPDATE_DELAY :
                //todo
                break;
            case Value.UPDATE_REPEATDUR :
                //todo
                break;
            case Value.UPDATE_SIZE :
                addDOMSize(valueToAdd.m_point);
                break;
            case Value.UPDATE_TRANSFORM :
                addDOMTransform(valueToAdd.m_matrix);
                break;
            case Value.UPDATE_XML_BASE :
                //todo
                break;
//            case Value.UPDATE_URL :
//                addDOMUrl(valueToAdd.m_string);
//                break;
            case Value.UPDATE_POINTS :
                addDOMPoints(valueToAdd.m_polygon);
                break;
            case Value.UPDATE_VIEWBOX :
                addDOMViewBox(valueToAdd.m_floatsOrInts);
                break;
            case Value.UPDATE_KEYTIMES :
                //todo
                break;
            case Value.UPDATE_TEXTCONTENT :
                addDOMTextContent(valueToAdd.m_textContent);
                break;
            case Value.UPDATE_TRANSLATION :
                addDOMTranslation(valueToAdd.m_point);
                break;
            case Value.UPDATE_SCALE :
                addDOMScale(valueToAdd.m_point);
                break;
            case Value.UPDATE_ROTATE :
                addDOMRotation(valueToAdd.m_enumOrColorOrRealOrID);
                break;
//            case Value.UPDATE_TRANSFORMSKEWX :
//                //todo
//                break;
//            case Value.UPDATE_TRANSFORMSKEWY :
//                //todo
//                break;
        }
    }


}
