/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;

/**
 * This class handles a graphic object that can group children and applies affine
 * transformation to its siblings (i.e., scaling, rotation, translation).
 * It is possible to specify the order of the rendering process (see setOrder)
 * or to specify which unique child will be rendered (see setChoice).
 * A scene is defined by a graph of GraphObjects. The Group objects can be
 * considered as the grouping nodes of the graph. Each scene is handled by its
 * Root node that shall be a Group Node. The scene graph is built using the
 * method appendChild() ( or insertChild() ) that adds a child to the current node.
 * The method render() of the root node launchs the rendering process of
 * the graph.
 *
 * @see iso.mpeg.streamezzo.player.library.Real
 * @see iso.mpeg.streamezzo.player.library.Matrix2D
 * @see iso.mpeg.streamezzo.player.library.Point2D
 */
public class Group extends GenericGroup {
    public int m_choicePosition = -3;

    public Point2D m_size;

    /** Presentation value of size */
    public Point2D m_actualSize;

    /** Previous presentation value of size */
    public Point2D m_prevActualSize;


////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////// Properties management /////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Default constructor, it sets the transform matrix to the identity matrix.
     *
     * @see iso.mpeg.streamezzo.player.node.RenderedGraphObject
     */
    public Group(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
        m_actualTransform = Matrix2D.createIdentityMatrix();
        m_definedTransform = Matrix2D.createIdentityMatrix();
    }

    /**
     * Update actual size
     * @param size New size
     */
    public void updateActualSize(Point2D size)
    {
        // Check changes
        if ( (size.x_r != m_prevActualSize.x_r) || (size.y_r != m_prevActualSize.y_r) )
        {
            m_actualSize.x_r = size.x_r;
            m_actualSize.y_r = size.y_r;
            wasUpdated();
        }
        else
        {
            m_actualSize.x_r = m_prevActualSize.x_r;
            m_actualSize.y_r = m_prevActualSize.y_r;
            m_wasUpdated = false;
        }
    }

    public void setDOMSize(Point2D size)
    {
        // size doesn't exist before
        if (m_size == null)
        {
            m_size = new Point2D(size.x_r, size.y_r);
            m_actualSize = new Point2D(0, 0);
            m_prevActualSize = new Point2D(0, 0);
        }

        m_size.x_r = size.x_r;
        m_size.y_r = size.y_r;
        wasUpdated();
    }

    public void addDOMSize(Point2D size)
    {
        if (m_size == null)
        {
            setDOMSize(size);
        }
        else
        {
            m_size.x_r += size.x_r;
            m_size.y_r += size.y_r;    
        }
    }

    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Is there a size attribute ?
        if (m_size != null)
        {
            if (m_actualSize == null)
            {
                m_actualSize = new Point2D(0, 0);
            }
            if (m_prevActualSize == null)
            {
                m_prevActualSize = new Point2D(0, 0);
            }
        }

        if ( (m_size == null) || (m_actualSize == null) || (m_prevActualSize == null) )
        {
            return;
        }

        // Changed ?
        if ( (m_size.x_r != m_actualSize.x_r) || (m_size.y_r != m_actualSize.y_r) )
        {
            // Store previous values
            m_prevActualSize.x_r = m_actualSize.x_r;
            m_prevActualSize.y_r = m_actualSize.y_r;

            // Set DOM values
            m_actualSize.x_r = m_size.x_r;
            m_actualSize.y_r = m_size.y_r;

            wasUpdated();
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// Rendering operations ////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * This method computes all the tranform in the graph according to the
     * stored transform matrix. If the node did not change, only propagation is needed
     */
    public void specificRender(int a_sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Copy the flag that indicates if something was changed and reset the original flag
        // We have to use this copy because a child node can modify the original flag
        boolean wasUpdatedCurrent = m_wasUpdated;
        m_wasUpdated = false;

        // Ignore visibility because even if this transform is hidden, one of its children can be visible
        // and have to be rendered.
        // Furthermore inheritable properties and matrix must be spread
        if (parentHasChanged)
        {
            wasUpdatedCurrent = true;
        }

        // Process on children accroding to choice attribute
        int i, N = m_children.size();
        GraphObject a_child = null;
        if (m_choicePosition > -1) // TML : selector node
        {
            // Render the choosen child
            a_child = (GraphObject) m_children.elementAt(m_choicePosition);
            a_child.render(a_sceneTimeInMillis, wasUpdatedCurrent, m_actualTransform, m_actualProperties, useContext);
        }
        else
            if (m_choicePosition == -2) // Choice = clip, TML : rectClip node
            {
                // Clipping
                // save previous clip rect
                Point2D previousOrigin = m_graphObjectManager.m_renderer.getClipRectOrigin();
                Point2D previousDimension = m_graphObjectManager.m_renderer.getClipRectDimension();
                Point2D clipOrigin = new Point2D(m_graphObjectManager.m_renderer.normalizeX(m_actualTransform.getTranslation().x_r), m_graphObjectManager.m_renderer.normalizeY(m_actualTransform.getTranslation().y_r));
                Point2D clipDim = new Point2D(m_graphObjectManager.m_renderer.normalizeW(m_actualSize.x_r), m_graphObjectManager.m_renderer.normalizeH(m_actualSize.y_r));
                clipOrigin.x_r = clipOrigin.x_r - clipDim.x_r / 2;
                clipOrigin.y_r = clipOrigin.y_r - clipDim.y_r / 2;
                m_graphObjectManager.m_renderer.setClipRect(clipOrigin, clipDim);

                // render children with the provided clipping rect
                for (i = 0; i < N; i++)
                {
                    a_child = (GraphObject) m_children.elementAt(i);
                    a_child.render(a_sceneTimeInMillis, wasUpdatedCurrent, m_actualTransform, m_actualProperties, useContext);
                }

                // set previous clip rect
                m_graphObjectManager.m_renderer.setClipRect(null, null);
                m_graphObjectManager.m_renderer.setClipRect(previousOrigin, previousDimension);
            }
            else
                if (m_choicePosition == -3) // Choice = delta
                {
                    if (m_actualSize == null) // check if we are in delta Mode
                    { // TML : group node
                        for (i = 0; i < N; i++)
                        {
                            a_child = (GraphObject) m_children.elementAt(i);
                            a_child.render(a_sceneTimeInMillis, wasUpdatedCurrent, m_actualTransform, m_actualProperties, useContext);
                        }
                    }
                    else // TML : simpleLayout node
                    {
                        Matrix2D the_deltaMat = new Matrix2D(m_actualTransform);
                        Matrix2D d = Matrix2D.createIdentityMatrix();
                        d.setTranslation(m_actualSize);
                        for (i = 0; i < N; i++)
                        {
                            a_child = (GraphObject) m_children.elementAt(i);
                            a_child.render(a_sceneTimeInMillis, true, the_deltaMat, m_actualProperties, useContext);
                            Matrix2D.Mul(d, the_deltaMat, the_deltaMat);
                        }
                    }
                }
        // -1 : no child rendered
    }
}
