/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Polygon2D;
import iso.mpeg.streamezzo.player.library.Real;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.SensitiveZone;

import java.util.Vector;

/**
 * This class provides a higher level implementation of a 2D polygon shape.
 * Its direct parent shall be a Group node.
 * @see Group
 */
public class Line extends RenderedGraphObject
{
    /** begin point of the line */
    public Point2D m_X1Y1 = null;

    /** Presentation value of begin point of the line. It reflects applied matrix */
    public Point2D m_actualX1Y1 = null;

    /** Previous presentation values */
    public Point2D m_prevActualX1Y1= null;

    /** end point of the line */
    public Point2D m_X2Y2 = null;

    /** Presentation value of end point of the line. It reflects applied matrix */
    public Point2D m_actualX2Y2 = null;

    /** Previous presentation values */
    public Point2D m_prevActualX2Y2= null;

    /** Line representation as polygon */
    public Polygon2D m_polygon = null;

    /** the sensitive zone */
    public SensitiveZone m_senstiveZone = new SensitiveZone();


////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////// Properties setters/getters /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    public Line(GraphObjectManager graphObjectManager)
    {
        super( graphObjectManager);
        m_actualTransform = Matrix2D.createIdentityMatrix();
        m_X1Y1 = new Point2D(0, 0);
        m_X2Y2 = new Point2D(0, 0);

        // Create internal representation as polygon
        m_polygon = new Polygon2D();
        m_polygon.m_points = new Vector(2);
        for (int i=0; i<2; i++)
        {
            Point2D point = new Point2D(0, 0);
            m_polygon.m_points.addElement(point);
        }

        // Initialize actual values. They will reflect applied matrix
        m_actualX2Y2 = new Point2D(0, 0);
        m_actualX1Y1 = new Point2D(0, 0);
        m_prevActualX1Y1 = new Point2D(0, 0);
        m_prevActualX2Y2 = new Point2D(0, 0);
    }

    /**
     * Create a polygon from rectangle properties
     */
    protected void convertToPolygon()
    {
        Point2D p = (Point2D)m_polygon.m_points.elementAt(0);
        p.x_r = m_actualX1Y1.x_r;
        p.y_r = m_actualX1Y1.y_r;
        p = (Point2D)m_polygon.m_points.elementAt(1);
        p.x_r = m_actualX2Y2.x_r;
        p.y_r = m_actualX2Y2.y_r;
    }

    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Changed ?
        if ( (m_X2Y2.x_r != m_actualX2Y2.x_r) || (m_X2Y2.y_r != m_actualX2Y2.y_r)
          || (m_X1Y1.x_r != m_actualX1Y1.x_r) || (m_X1Y1.y_r != m_actualX1Y1.y_r))
        {
            // Store previous values
            m_prevActualX2Y2.x_r = m_actualX2Y2.x_r;
            m_prevActualX2Y2.y_r = m_actualX2Y2.y_r;
            m_prevActualX1Y1.x_r = m_actualX1Y1.x_r;
            m_prevActualX1Y1.y_r = m_actualX1Y1.y_r;

            // Set DOM values
            m_actualX2Y2.x_r = m_X2Y2.x_r;
            m_actualX2Y2.y_r = m_X2Y2.y_r;
            m_actualX1Y1.x_r = m_X1Y1.x_r;
            m_actualX1Y1.y_r = m_X1Y1.y_r;

            wasUpdated();
        }
    }

    public void setDOMX2(int r_x)
    {
        m_X2Y2.x_r = r_x;
    }
    public void setDOMY2(int r_y)
    {
        m_X2Y2.y_r = r_y;
    }

    public void addDOMX2(int r_x)
    {
        m_X2Y2.x_r += r_x;
    }
    public void addDOMY2(int r_y)
    {
        m_X2Y2.y_r += r_y;
    }

    public void setDOMX(int r_x)
    {
        m_X1Y1.x_r = r_x;
    }
    public void setDOMY(int r_y)
    {
        m_X1Y1.y_r = r_y;
    }

    public void addDOMX(int r_x)
    {
        m_X1Y1.x_r += r_x;
    }
    public void addDOMY(int r_y)
    {
        m_X1Y1.y_r += r_y;
    }

    /**
     * Update actual x
     * @param rX New end x (real 20.12)
     */
    public void updateActualX2(int rX)
    {
        // Check changes
        if (rX != m_prevActualX2Y2.x_r)
        {
            m_actualX2Y2.x_r = rX;
            wasUpdated();
        }
        else
        {
            m_actualX2Y2.x_r = m_prevActualX2Y2.x_r;
            m_wasUpdated = false;
        }
    }

    /**
     * Update actual y
     * @param rY New y (real 20.12)
     */
    public void updateActualY2(int rY)
    {
        // Check changes
        if (rY != m_prevActualX2Y2.y_r)
        {
            m_actualX2Y2.y_r = rY;
            wasUpdated();
        }
        else
        {
            m_actualX2Y2.y_r = m_prevActualX2Y2.y_r;
            m_wasUpdated = false;
        }
    }

    /**
     * Update actual x
     * @param rX New end x (real 20.12)
     */
    public void updateActualX(int rX)
    {
        // Check changes
        if (rX != m_prevActualX1Y1.x_r)
        {
            m_actualX1Y1.x_r = rX;
            wasUpdated();
        }
        else
        {
            m_actualX1Y1.x_r = m_prevActualX1Y1.x_r;
            m_wasUpdated = false;
        }
    }

    /**
     * Update actual y
     * @param rY New y (real 20.12)
     */
    public void updateActualY(int rY)
    {
        // Check changes
        if (rY != m_prevActualX1Y1.y_r)
        {
            m_actualX1Y1.y_r = rY;
            wasUpdated();
        }
        else
        {
            m_actualX1Y1.y_r = m_prevActualX1Y1.y_r;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual X
     */
    public int getActualX()
    {
        return m_actualX1Y1.x_r;
    }

    /**
     * Get the actual X
     */
    public int getActualY()
    {
        return m_actualX1Y1.y_r;
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// Rendering operations //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    public void specificRender(int a_sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Compute polygon that represents the rect
        if ( (m_wasUpdated) || (m_polygon == null) )
        {
            convertToPolygon();
        }

        // Apply matrix
        if (parentHasChanged)
        {
            m_wasUpdated = true;
        }
        
        // Check if it must be drawn
        if ( (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_HIDDEN) ||
             (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_COLLAPSE) )
        {
            return;
        }

        // Fill if needed
        if (m_actualProperties.m_fill != Value.COLOR_NONE)
        {
            // set the color
            m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_fill, m_actualProperties.m_fillOpacity&0xFF);
            // set the opacity
            m_polygon.applyMatrixAndfillPolygon(m_actualTransform, m_graphObjectManager.m_renderer);
        }

        // Draw the stroke if needed
        if (m_actualProperties.m_stroke != Value.STROKE_NONE)
        {
            m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_stroke, m_actualProperties.m_strokeOpacity&0xFF);
            if (m_actualProperties.m_strokeWidth != Real.VALUE1)
            {
                m_polygon.applyMatrixAndDrawPolyLine(m_actualTransform, m_graphObjectManager.m_renderer,
                                                     m_actualProperties.m_strokeWidth);
            }
            else
            {
               m_polygon.applyMatrixAndDrawPolyLine(m_actualTransform, m_graphObjectManager.m_renderer);
            }
        }

//        System.out.println("line: " + 
//        		(((Point2D)(m_polygon.m_points.elementAt(0))).x_r >> 12) + ", " +
//        		(((Point2D)(m_polygon.m_points.elementAt(0))).y_r >> 12) + ", " +
//        		(((Point2D)(m_polygon.m_points.elementAt(1))).x_r >> 12) + ", " +
//        		(((Point2D)(m_polygon.m_points.elementAt(1))).y_r >> 12) + ", "); 
        		

        // Get the bounding box and check sensitive zone
        if (null == useContext)
        {
            m_polygon.getBoundingBox(m_senstiveZone.getZone());
            if (Debug.DEBUG_BBOX) System.out.println(this + " : bbox= (" + m_senstiveZone.m_rect.x_r + "," + m_senstiveZone.m_rect.y_r + "," + m_senstiveZone.m_rect.width_r + "," + m_senstiveZone.m_rect.height_r + ")");
            m_senstiveZone.checkZone(m_graphObjectManager);
        }

        m_wasUpdated = false;
    }
}
