/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.BasicRendering;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.tool.SceneLoader;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.TimedElement;

public class Listener extends RenderedGraphObject implements EventListener {
    //newly added attrs
    public Event event = null;
    public URI handler = null;
    public int keyCode = Integer.MIN_VALUE;
    public int m_delay = 0;
    public int m_observer = -1;
    // TODO delete m_url
    public String m_url = null;
    public boolean m_beginOrEnd = true; // target attribute: begin(true) or end(false)
    public boolean m_enabled = true;
    private int m_savedSceneTime;
    private int m_savedKeyPressed;
    public int propagate = Integer.MIN_VALUE;
    public int defaultAction = Integer.MIN_VALUE;
    public URI target = null;

    /**
     * Constructor
     *
     * @param graphObjectManager
     * @see RenderedGraphObject
     */
    public Listener(GraphObjectManager graphObjectManager) {
        super(graphObjectManager);
    }

    /**
     * Check if the event is global or has an initiator.
     *
     * @return true if the it's a global event, without any source element
     */
    public boolean hasGlobalEvent() {
        switch (event.event) {
            case UI_FOCUS_IN:
            case UI_FOCUS_OUT:
            case UI_ACTIVATE:
            case MOUSE_OVER:
            case MOUSE_DOWN:
            case MOUSE_UP:
            case MOUSE_CLICK:
            case MOUSE_OUT:
            case TIME_BEGIN:
            case TIME_END:
            case TIME_REPEAT:
                return false;
            default:
                return true;
        }
    }

    /**
     * Check whether this trigger is listening for a mouse event
     *
     * @return true if the trigger is listening for a mouse event
     */
    public boolean isListeningForMouseEvent() {
        switch (event.event) {
            case MOUSE_OVER:
            case MOUSE_DOWN:
            case MOUSE_UP:
            case MOUSE_CLICK:
            case MOUSE_OUT:
                return true;
            default:
                return false;
        }
    }

    /**
     * Overriden method to manage trigger without specified event source
     *
     * @param parent The parent node
     */
    public void setParent(GraphObject parent) {
        // Base behavior
        super.setParent(parent);

        // Check if the parent is the event source or target
        if ((-1 == m_observer) && (! hasGlobalEvent())) {
            m_graphObjectManager.attachListener(this, parent);
        }
    }

    /**
     * In the case of a node targeting another node (i.e. Trigger, Animation, Use),
     * it returns the targeted node ID
     *
     * @return the ID of the targeted node
     */
    public int getTargetID() {
        if (handler != null) 
        	return handler.id;
        else {
        	if (m_parent.getID() == -1)
        		m_parent.setAutoID();
        	return m_parent.getID();
        }
    }

    /**
     * Set a new href for this trigger
     *
     * @param href the ID of the targeted node
     */
    public void setDOMHRef(int href) {
        if (handler == null)
        	handler = new URI();
    	handler.id = href;
    }

    public void setDOMUrl(String url) {
        m_url = url;
    }

    public void addDOMUrl(String url) {
        m_url = m_url + url;
    }

    /**
     * Do the post render
     */
    public void postRender() {
        // Hyperlinking
        if (m_url != null) {
            // Load a new scene
            SceneLoader sceneLoader = super.m_graphObjectManager.m_renderer.getSceneLoader();
            sceneLoader.reLoadScene(m_url);
        }
    }

    /**
     * Do the processing related to the event
     */
    private void processEvent() {
        // Hyperlinking
        if (m_url != null) {
            // Load a new scene in the Post Render
            super.m_graphObjectManager.registerForPostRender(this);
        }
        // Activation
        else {
            if (getTargetID() != -1) {
                // get the node
                GraphObject node = super.m_graphObjectManager.getNode(getTargetID());
                if (node != null) {
                    if (node.isTimeDependent()) {
                        // begin or end ?
                    	if (m_beginOrEnd) {
                            // begin
                            ((TimedElement)node).setBeginTime(m_savedSceneTime, m_delay);
                        } else {
                            // end
                            ((TimedElement)node).setEndTime(m_savedSceneTime, m_delay);
                        }
                    } else {
                        // send an activate event
                        node.onEvent(EventListener._UI_ACTIVATE, null);
                    }
                }
            }
        }
    }

    /**
     * Check if the trigger reacts to a specified key code
     *
     * @param currentKeyCode the key code to test
     * @return true if the key code is registered in the trigger
     */
    protected boolean isRegiteredKeyPressed(int currentKeyCode) {
        if (currentKeyCode != BasicRendering.NO_KEY &&
                (event.keyCode == BasicRendering.ANY_KEY || ConvertionKeyTab[event.keyCode] == currentKeyCode)) {
            return true;
        } else {
            return false;
        }
    }
    
    // TML TODO : Remove this !!!!!!!
    private static int[] ConvertionKeyTab = {-1, -1, -1, -1, -1,
									    	-1, -1, -1, -1, -1,
									    	-1, -1, 1, 4, 2,
									    	-1, 3, -1, -1, -1,
									    	0, -1, -1, -1, -1};

    //////////////////////////////////
    // EventListener implementation //
    //////////////////////////////////

    /**
     * Method called back to inform the listener when a new event is generated
     *
     * @param initiator the object which initiated the event
     * @param eventType the event type
     * @param eventData the associated event data
     * @return true if the event has been handled
     */
    public boolean update(Object initiator, byte eventType, Object eventData) {
        if (!m_enabled) {
            return false;
        }
        if (Debug.DEBUG_EVENT) {
            System.out
                    .println("Trigger " + this + " receives an event " + eventType + "(compare to event.event "+event.event+") from " + initiator);
        }
        if (eventType == event.event) {
            // depending on the event type, the event data is different
            switch (eventType) {
                case EventListener.TIME_BEGIN:
                case EventListener.TIME_END:
                case EventListener.TIME_REPEAT:
                    m_savedSceneTime = ((Integer)eventData).intValue();
                    break;
                case EventListener.KEY_PRESS:
                    int[] params = (int[])eventData;
                    m_savedSceneTime = params[0];
                    m_savedKeyPressed = params[1];
                    // check if the key is registered
                    if (!isRegiteredKeyPressed(m_savedKeyPressed)) {
                        return false;
                    }
                    break;
                case EventListener.UI_FOCUS_IN:
                case EventListener.UI_FOCUS_OUT:
                case EventListener.UI_ACTIVATE:
                case EventListener.UI_ACTIVATE_EVENT:
                case EventListener.UI_DESACTIVATE_EVENT:
                    m_savedSceneTime = ((Integer)eventData).intValue();
                    break;
                default:
                    break;
            }

            // event has been decoded, now process it
            processEvent();
            return true;
        }
        if (Debug.DEBUG_EVENT) {
            System.out
                    .println("Trigger " + this + " ignores an event " + eventType + " from " + initiator);
        }
        return false;
    }

    public void specificRender(int sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext) {
        // TODO I temporarily made this a RenderedGraphObject, for being capable of calling readObject
        // in future we should move the m_properties down in the hierarchy
    }

}
