/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.TimedElement;

import java.util.Vector;

public class Multimedia extends RenderedGraphObject implements TimedElement
{
    /** Timing constants */
    protected static final byte TIME_INDEFINITE = -1;
    protected static final byte TIME_UNDEFINED = -2;

    /** Repeat constants */
    protected static final byte REPEAT_INFINITE = -1;

    /** Machine state constants */
    protected static final byte STATE_IDLE = 0;                   //! Idle, no interval has been searched
    protected static final byte STATE_WAITFORSTART = 1;           //! Current interval available, wait for begin
    protected static final byte STATE_STARTPLAYING = 2;           //! Start playing the current interval
    protected static final byte STATE_PLAYING = 3;                //! Playing the current interval
    protected static final byte STATE_ENDPLAYING = 4;             //! End playing the current interval
    protected static final byte STATE_WAITFORINTERVAL = 7;        //! Idle, no suitable interval exist. Times lists have already been
                                                                  //! checked, wait for times lists update (by events for example).

    /** Specifies the total duration for repeat. */
    public int m_repeatDur = TIME_UNDEFINED;
    /** Specifies the number of iterations of the animation function. */
    public int m_repeatCount = 1;
    /** Defines when the element sould begin */
    public Vector m_beginTimes = new Vector();
    /** Defines when the element should stop playing */
    public Vector m_endTimes = new Vector();

    /** Simple duration */
    private int m_duration = TIME_INDEFINITE;
    /** Current simple duration end time */
    private int m_simpleDurEnd = TIME_INDEFINITE;
    /** Beginning of the current interval */
    private int m_intervalBegin = TIME_INDEFINITE;
    /** Beginning of the next interval */
    private int m_nextintervalBegin = TIME_INDEFINITE;
    /** End time of the current interval */
    private int m_intervalEnd = TIME_INDEFINITE;
    /** End time of the active duration */
    private int m_activeEnd = TIME_INDEFINITE;
    /** End time of the previously played active duration */
    private int m_previousActiveEnd = TIME_INDEFINITE;
    /** The current repeat iteration count */
    private int m_iterationCount = 0;
    /** The actual begin time */
    private int m_actualBegin = TIME_INDEFINITE;
    /** The actual duration */
    private int m_actualDuration = TIME_INDEFINITE;
    /** Internal state */
    protected byte m_state = STATE_IDLE;
    public String contentType = null;

////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////// Construction, Getters and Setters  ////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    public Multimedia(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
        m_isTimeDependent = true;
        m_beginTimes.addElement(new Integer(0));
        m_endTimes.addElement(new Integer(TIME_INDEFINITE));
    }

    public void setDOMBegin(int beginTime)
    {
        m_beginTimes.setSize(0);
        addBeginTime(beginTime);
    }

    public void setDOMEnd(int endTime)
    {
        m_endTimes.setSize(0);
        addEndTime(endTime);
    }

    public void addBeginTime(int beginTime)
    {
        int i = 0;
        for (i = 0; i < m_beginTimes.size(); i++)
        {
            if ( (((Integer)m_beginTimes.elementAt(i)).intValue() > beginTime) || (TIME_INDEFINITE == ((Integer)m_beginTimes.elementAt(i)).intValue()) )
            {
                break;
            }
        }
        m_beginTimes.insertElementAt(new Integer(beginTime), i);

        // According to the automaton current state
        switch (m_state)
        {
            case STATE_WAITFORINTERVAL:
            case STATE_WAITFORSTART:
                // A new interval can be found, force the automaton to check if it must start playing a new interval.
                m_state = STATE_IDLE;
                m_actualDuration = m_duration;
                m_actualBegin = TIME_INDEFINITE;
                break;

            case STATE_PLAYING:
            case STATE_STARTPLAYING:
                // Adjust the next interval begin time
                if ( (beginTime < m_nextintervalBegin) || (m_nextintervalBegin == TIME_INDEFINITE) )
                {
                    m_nextintervalBegin = beginTime;
                    if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : adjusted next begin time=" + m_nextintervalBegin);
                }
                // Adjust the active duration end time
                if ( ( (m_intervalEnd != TIME_INDEFINITE) && (m_nextintervalBegin < m_intervalEnd) ) ||
                       (m_intervalEnd == TIME_INDEFINITE) )
                {
                    // Set the end to the next begin time and compute the active duration once again
                    m_intervalEnd = m_nextintervalBegin;
                    if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : adjusted currentInterval=[" + m_intervalBegin + ";" + m_intervalEnd +"]");
                    // Compute active duration
                    computeActiveDuration();
                }
                break;
        }
    }

    public void addEndTime(int endTime)
    {
        int i = 0;
        for (i = 0; i < m_endTimes.size(); i++)
        {
            if ( (((Integer)m_endTimes.elementAt(i)).intValue() > endTime) || (TIME_INDEFINITE == ((Integer)m_endTimes.elementAt(i)).intValue()) )
            {
                break;
            }
        }
        m_endTimes.insertElementAt(new Integer(endTime), i);

        // Adjust active duration and current interval end time
        if ( (endTime < m_intervalEnd) || (m_intervalEnd == TIME_INDEFINITE) )
        {
            m_intervalEnd = endTime;
            if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : adjusted currentInterval=[" + m_intervalBegin + ";" + m_intervalEnd +"]");
            // Compute active duration
            computeActiveDuration();
        }
    }

    public void setBeginTime(int sceneTimeMillis, int delay)
    {
        // Add a new time into the list
        addBeginTime(sceneTimeMillis + delay);
    }

    public void setEndTime(int sceneTimeMillis, int delay)
    {
        // Add new end time into the list
        addEndTime(sceneTimeMillis + delay);
    }

    public int getBeginTime()
    {
        return m_intervalBegin;
    }

    protected boolean isDurationNeeded()
    {
        return ( (m_actualDuration == TIME_INDEFINITE) && (m_actualBegin != TIME_INDEFINITE) );
    }

    protected boolean isStarted()
    {
        return (m_actualBegin != TIME_INDEFINITE);
    }

    protected void onDurationAvailable(int sceneTimeMillis, int duration)
    {
        if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeMillis+ " - onDurationAvailable : "+duration);
        // Set simple duration
        m_actualDuration = duration;
        // Compute active end
        computeActiveDuration();
    }

    protected void onStart(int sceneTimeMillis, int offset)
    {
        if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeMillis+ " - onStart (offset="+offset+")");
        // Adjust begin time
        if (m_iterationCount == 1)
        {
            m_actualBegin = sceneTimeMillis - (offset != -1 ? offset : 0);
            m_intervalBegin = m_actualBegin;
            // Compute active end
            computeActiveDuration();
        }
    }

    protected void onEnd(int sceneTimeMillis)
    {
        if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeMillis+ " - onEnd");
        // Adjust simple duration
        if (m_iterationCount == 1)
        {
            m_actualDuration = sceneTimeMillis - m_actualBegin;
            // Compute active end
            computeActiveDuration();
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// OnEvent //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Is called when receiving an event
     * @param eventType the event type
     * @param eventData the associated event data
     */
    public void onEvent(byte eventType, Object eventData)
    {
        switch(eventType)
        {
            case EventListener.UI_ACTIVATE:
                setBeginTime(((Integer)eventData).intValue(), 0);
                break;
        }
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////// Pre-Rendering - Timing  /////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    public void preRender(int sceneTimeInMillis)
    {
        // Events to post : no event as default
        byte beginOrRepeatEvent = EventListener.INTERNAL_UNDEFINED;
        byte endEvent = EventListener.INTERNAL_UNDEFINED;

        // Automaton
        switch(m_state)
        {
            // WAITING : Interval has been revolved. Check if it must begin
            case STATE_WAITFORSTART:
                if ( (m_intervalBegin != TIME_INDEFINITE) && (m_intervalBegin <= sceneTimeInMillis) )
                {
                    // It's time to begin
                    m_iterationCount++;
                    m_actualBegin = TIME_INDEFINITE;
                    m_state = STATE_STARTPLAYING;
                    if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - state = START PLAYING");
                    beginOrRepeatEvent = EventListener.TIME_BEGIN;
                }
                break;

            // PLAYING THE CURRENT INTERVAL: Check if it must end or repeat
            case STATE_STARTPLAYING:
                m_state = STATE_PLAYING;
                if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - state = PLAYING");
            case STATE_PLAYING:
                // Check if active duration end time has come
                if ( (m_activeEnd != TIME_INDEFINITE) && (m_activeEnd <= sceneTimeInMillis) )
                {
                    // It's time to end
                    m_iterationCount = 0;
                    m_previousActiveEnd = m_activeEnd;
                    endEvent = EventListener.TIME_END;
                    m_state = STATE_ENDPLAYING;
                    m_actualDuration = m_duration;
                    m_actualBegin = TIME_INDEFINITE;
                    if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - state = END PLAYING");
                }
                // Check if simple duration time has come
                else if ( (m_simpleDurEnd != TIME_INDEFINITE) && (m_simpleDurEnd <= sceneTimeInMillis) )
                {
                    // Repeat
                    m_iterationCount++;
                    if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - repeat iteration #"+m_iterationCount);
                    m_simpleDurEnd = m_intervalBegin + (m_actualDuration * m_iterationCount);
                    beginOrRepeatEvent = EventListener.TIME_REPEAT;
                    if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - current simple duration begin/end times = "+sceneTimeInMillis+"/"+m_simpleDurEnd);
                }
                break;

            // END OF PLAYING THE CURRENT INTERVAL
            case STATE_ENDPLAYING :
                m_state = STATE_IDLE;
                m_actualDuration = m_duration;
                m_actualBegin = TIME_INDEFINITE;
                if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - state = IDLE");
                break;
        }

        // Outsource IDLE state check to be able to restart before the next frame
        if (STATE_IDLE == m_state)
        {
            if (resolveCurrentInterval(sceneTimeInMillis))
            {
                if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - currentInterval=["+m_intervalBegin+";"+m_intervalEnd+"]");
                // Set new state
                m_state = STATE_WAITFORSTART;
                if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - state = WAIT FOR START");
                // Is it time to start ?
                if (sceneTimeInMillis >= m_intervalBegin)
                {
                    // It's the time to begin
                     m_iterationCount++;
                    m_state = STATE_STARTPLAYING;
                    if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - state = START PLAYING");
                    beginOrRepeatEvent = EventListener.TIME_BEGIN;
                }
            }
            else
            {
                m_state = STATE_WAITFORINTERVAL;
                if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - state = WAIT FOR INTERVAL");
            }
        }

        // Possibly post event
        // CAUTION : this code line must be the last line of this method to avoid re-entrant call
        if (beginOrRepeatEvent != EventListener.INTERNAL_UNDEFINED)
        {
            notifyEvent(beginOrRepeatEvent, new Integer(sceneTimeInMillis));
        }
        if (endEvent != EventListener.INTERNAL_UNDEFINED)
        {
            notifyEvent(endEvent, new Integer(sceneTimeInMillis));
        }
    }

    public boolean resolveCurrentInterval(int sceneTimeInMillis)
    {
        // Clean the lists by removing times before the previous active duration end time
        if (m_previousActiveEnd != TIME_INDEFINITE)
        {
            while (m_beginTimes.size() > 0)
            {
                if ( ((Integer)m_beginTimes.firstElement()).intValue() < m_previousActiveEnd )
                {
                    m_beginTimes.removeElementAt(0);
                }
                else
                {
                    break;
                }
            }
            while (m_endTimes.size() > 0)
            {
                if ( ((Integer)m_endTimes.firstElement()).intValue() <= m_previousActiveEnd )
                {
                    m_endTimes.removeElementAt(0);
                }
                else
                {
                    break;
                }
            }
        }

        // Find the begin time
        if (m_beginTimes.size() > 0)
        {
            m_intervalBegin = ((Integer)m_beginTimes.firstElement()).intValue();
        }
        else
        {
            m_intervalBegin = TIME_INDEFINITE;
        }

        // Find the next begin time
        if (m_beginTimes.size() > 1)
        {
            m_nextintervalBegin = ((Integer)m_beginTimes.elementAt(1)).intValue();
        }
        else
        {
            m_nextintervalBegin = TIME_INDEFINITE;
        }

        // Get end time
        if (m_endTimes.size() > 0)
        {
            m_intervalEnd = ((Integer)m_endTimes.firstElement()).intValue();
        }
        else
        {
            m_intervalEnd = TIME_INDEFINITE;
        }

        // Check begin
        if (m_intervalBegin == TIME_INDEFINITE)
        {
            if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : sceneTime="+sceneTimeInMillis+ " - No begin defined");
            return false;
        }

        // Compute active duration
        computeActiveDuration();
        return true;
    }

    public void computeActiveDuration()
    {
        int activeDuration = TIME_INDEFINITE;

        if ( (REPEAT_INFINITE == m_repeatCount) && (TIME_INDEFINITE == m_repeatDur) )
        {
            if (m_intervalEnd != TIME_INDEFINITE)
            {
                activeDuration = m_intervalEnd - m_intervalBegin;
            }
            else
            {
                activeDuration = TIME_INDEFINITE;
            }
        }
        else if (REPEAT_INFINITE == m_repeatCount)
        {
            if (TIME_INDEFINITE != m_intervalEnd)
            {
                activeDuration = m_intervalEnd - m_intervalBegin;
            }
            else
            {
                activeDuration = TIME_INDEFINITE;
            }
        }
        else if (TIME_INDEFINITE == m_repeatDur)
        {
            if (TIME_INDEFINITE != m_intervalEnd)
            {
                activeDuration = m_intervalEnd - m_intervalBegin;
            }
            else
            {
                activeDuration = TIME_INDEFINITE;
            }
        }
        else if ( (TIME_INDEFINITE == m_actualDuration) && ( (TIME_INDEFINITE != m_repeatDur) && (TIME_UNDEFINED != m_repeatDur)) )
        {
            if (TIME_INDEFINITE != m_intervalEnd)
            {
                int endMinusBegin = m_intervalEnd - m_intervalBegin;
                if (m_repeatDur < endMinusBegin)
                {
                    activeDuration = m_repeatDur;
                }
                else
                {
                    activeDuration = endMinusBegin;
                }
            }
            else
            {
                activeDuration = m_repeatDur;
            }
        }
        else if ( (TIME_INDEFINITE == m_actualDuration) && (TIME_INDEFINITE != m_intervalEnd) )
        {
            activeDuration = m_intervalEnd - m_intervalBegin;
        }
        else if (TIME_INDEFINITE == m_actualDuration)
        {
            activeDuration = TIME_INDEFINITE;
        }
        else if ( (TIME_INDEFINITE != m_actualDuration) && (REPEAT_INFINITE == m_repeatCount) && ((TIME_INDEFINITE != m_repeatDur) && (TIME_UNDEFINED != m_repeatDur)) )
        {
            int repeatCountAsTime = m_repeatCount * m_actualDuration;
            if (TIME_INDEFINITE != m_intervalEnd)
            {
                int endMinusBegin = m_intervalEnd - m_intervalBegin;
                if (repeatCountAsTime < m_repeatDur)
                {
                    if (repeatCountAsTime < endMinusBegin)
                    {
                        activeDuration = repeatCountAsTime;
                    }
                    else
                    {
                        activeDuration = endMinusBegin;
                    }
                }
                else
                {
                    if (m_repeatDur < endMinusBegin)
                    {
                        activeDuration = m_repeatDur;
                    }
                    else
                    {
                        activeDuration = endMinusBegin;
                    }
                }
            }
            else
            {
                if (repeatCountAsTime < m_repeatDur)
                {
                    activeDuration = repeatCountAsTime;
                }
                else
                {
                    activeDuration = m_repeatDur;
                }
            }
        }
        else if ( (TIME_INDEFINITE != m_actualDuration) && (REPEAT_INFINITE != m_repeatCount) )
        {
            int repeatCountAsTime = m_repeatCount * m_actualDuration;
            if (TIME_INDEFINITE != m_intervalEnd)
            {
                int endMinusBegin = m_intervalEnd - m_intervalBegin;
                if (repeatCountAsTime < endMinusBegin)
                {
                    activeDuration = repeatCountAsTime;
                }
                else
                {
                    activeDuration = endMinusBegin;
                }
            }
            else
            {
                activeDuration = repeatCountAsTime;
            }
        }
        else if ( (TIME_INDEFINITE != m_actualDuration) && ((TIME_INDEFINITE != m_repeatDur) && (TIME_UNDEFINED != m_repeatDur)) )
        {
            if (TIME_INDEFINITE != m_intervalEnd)
            {
                int endMinusBegin = m_intervalEnd - m_intervalBegin;
                if (m_repeatDur < endMinusBegin)
                {
                    activeDuration = m_repeatDur;
                }
                else
                {
                    activeDuration = endMinusBegin;
                }
            }
            else
            {
                activeDuration = m_repeatDur;
            }
        }
        else if (TIME_INDEFINITE != m_actualDuration)
        {
            if (TIME_INDEFINITE != m_intervalEnd)
            {
                int endMinusBegin = m_intervalEnd - m_intervalBegin;
                if (m_actualDuration < endMinusBegin)
                {
                    activeDuration = m_actualDuration;
                }
                else
                {
                    activeDuration = endMinusBegin;
                }
            }
            else
            {
                activeDuration = m_actualDuration;
            }
        }
        else
        {
            activeDuration = TIME_INDEFINITE;
        }

        if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : active duration = " + activeDuration);

        // Compute simple duration end
        if (TIME_INDEFINITE != m_actualDuration)
        {
            m_simpleDurEnd = m_intervalBegin + m_actualDuration;
            if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : current simple duration end time = " + m_simpleDurEnd);
        }
        else
        {
            m_simpleDurEnd = TIME_INDEFINITE;
            if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : current simple duration end time = infinite");
        }

        // Compute active duration end time
        if (TIME_INDEFINITE != activeDuration)
        {
            m_activeEnd = m_intervalBegin + activeDuration;
            if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : active duration end time = " + m_activeEnd);
        }
        else
        {
            m_activeEnd = TIME_INDEFINITE;
            if (Debug.DEBUG_MULTIMEDIA) System.out.println("MULTIMEDIA : active duration end time = infinite");
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// Rendering  //////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    public void specificRender(int sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Nothing generic to do
    }
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////Accessor ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////
	public byte getMultimediaPlayingState()
	{
		return this.m_state;
	}
	
	public void setMultimediaPlayingState(int NewState)
	{
		this.m_state = (byte)NewState;
	}
}
