/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Polygon2D;
import iso.mpeg.streamezzo.player.library.Polygonizer;
import iso.mpeg.streamezzo.player.library.SetOfPolygon2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.SensitiveZone;


/**
 * This class provides a higher level implementation of a 2D polygon shape.
 * Its direct parent shall be a Group node.
 * @see iso.mpeg.streamezzo.player.node.Group
 */
public class Path extends RenderedGraphObject
 {
    public final static int POLYGON_INTERPOLATION_PRECISION = 8;

    // the path in LASeR specs, i.e. moveto, lineto, ...
    public Polygon2D m_polygon = null; // the coordinates
    public byte[] m_types = null; // the types

    // a list of polygons reprensting the path
    private SetOfPolygon2D m_interpolatedPath = null;

    // The presentation value of the point sequence
    public Polygon2D m_actualPolygon = null;
    public Polygon2D m_previousActualPolygon = null;

    /** the sensitive zone */
    public SensitiveZone m_senstiveZone = new SensitiveZone();

    /**
     * The Path constructor
     * @param graphObjectManager
     */
    public Path(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
    }

    public void specificRender(int a_sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        if (m_actualPolygon == null) return;

        // Apply matrix
        if (parentHasChanged)
        {
            m_wasUpdated = true;
        }

        if ( (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_HIDDEN) ||
             (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_COLLAPSE) )
        {
            return;
        }

        // need to reprocess the interpolation if the path was modified
        if (m_wasUpdated)
        {
/*            for (int i = 0; i < m_actualPolygon.m_points.size(); i++) {
                Point2D p = (Point2D)m_actualPolygon.m_points.elementAt(i);
                System.out.print(p.x_r>>12);
                System.out.print(" ");
                System.out.print(p.y_r>>12);
                System.out.print(" ");
            System.out.println();
            }*/

            m_interpolatedPath = Polygonizer.generateInterpolatedPath(m_actualPolygon.m_points, m_types,
                    POLYGON_INTERPOLATION_PRECISION);
        }

        // render the path
        if (m_interpolatedPath != null)
        {
            // Fill the whole path
            if (m_actualProperties.m_fill != Value.COLOR_NONE)
            {
                // set the color
                m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_fill, m_actualProperties.m_fillOpacity&0xFF);
                // set the opacity
                m_interpolatedPath.applyMatrixAndfillPolygonSet(m_actualTransform, m_graphObjectManager.m_renderer);
            }

            // Draw polylines
            if (m_actualProperties.m_stroke != Value.STROKE_NONE)
            {
                m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_stroke, m_actualProperties.m_strokeOpacity&0xFF);
                m_interpolatedPath.applyMatrixAndDrawPolyLineSet(m_actualTransform, m_graphObjectManager.m_renderer,
                                                                     m_actualProperties.m_strokeWidth);
            }
            // Get the bounding box and check sensitive zone
            if (null == useContext)
            {
                m_interpolatedPath.getBoundingBox(m_senstiveZone.getZone());
                if (Debug.DEBUG_BBOX) System.out.println(this + " : bbox= (" + m_senstiveZone.m_rect.x_r + "," + m_senstiveZone.m_rect.y_r + "," + m_senstiveZone.m_rect.width_r + "," + m_senstiveZone.m_rect.height_r + ")");
                m_senstiveZone.checkZone(m_graphObjectManager);
            }
        }

        m_wasUpdated = false;
    }

    /**
     * Updates path attributes to define a new path
     * @param newPolygon New points sequence
     * @param newTypes New types sequence
     */
    public void updateActualPathDataAndTypes(Polygon2D newPolygon, byte[] newTypes)
    {
        // Sanity check
        if (newPolygon == null)
        {
            return;
        }

        // Check changes : if no changes do not polygonize
        if (! newPolygon.equals(m_previousActualPolygon))
        {
            m_actualPolygon = newPolygon;
            m_wasUpdated = true;
        }
        else
        {
            m_actualPolygon = m_previousActualPolygon;
            m_wasUpdated = false;
        }
    }

    public void setDOMD(Polygon2D polygon, byte[] types)
    {
        m_polygon = polygon;
        m_types = types;
    }

    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Changed ?
        if (! m_polygon.equals(m_actualPolygon))
        {
            // Store previous
            m_previousActualPolygon = m_actualPolygon;

            // Reset polygon
            m_actualPolygon = m_polygon;

            m_wasUpdated = true;
        }
    }
}
