/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Polygon2D;
import iso.mpeg.streamezzo.player.library.Real;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.SensitiveZone;

/**
 * The Polygon class defines a closed shape consisting of
 * a set of connected straight line segments.
 */
public class Polygon extends RenderedGraphObject
{
    // the coordinates
    public Polygon2D m_polygon;
    // The presentation value of the point sequence
    public Polygon2D m_actualPolygon = null;

    /** the sensitive zone */
    public SensitiveZone m_senstiveZone = new SensitiveZone();

    /**
     * Constructor
     * @param graphObjectManager
     */
    public Polygon(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
    }

    public void setDOMPoints(Polygon2D polygon, int index)
    {
        // Replace all points
        if (index == Value.NO_INDEX)
        {
            m_polygon = polygon;
        }
        // Repalce one point
        else
        {
            if ( (polygon.m_points != null) && (polygon.m_points.size() > 0) )
            {
                m_polygon.m_points.setElementAt(polygon.m_points.elementAt(0), index);
            }
        }

        m_actualPolygon = m_polygon.getCopy();
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// Rendering operations //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Renders the polygon
     * @param a_sceneTimeInMillis
     * @param parentHasChanged
     * @param useContext
     */
    public void specificRender(int a_sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Check if it must be drawn
        if ( (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_HIDDEN) ||
             (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_COLLAPSE) )
        {
            return;
        }

        // Fill if needed
        if (m_actualProperties.m_fill != Value.COLOR_NONE)
        {
            // set the color
            m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_fill, m_actualProperties.m_fillOpacity&0xFF);
            // set the opacity
            m_actualPolygon.applyMatrixAndfillPolygon(m_actualTransform, m_graphObjectManager.m_renderer);
        }

        // Draw the stroke if needed
        if (m_actualProperties.m_stroke != Value.STROKE_NONE)
        {
            m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_stroke, m_actualProperties.m_strokeOpacity&0xFF);
            if (m_actualProperties.m_strokeWidth != Real.VALUE1)
            {
                m_actualPolygon.applyMatrixAndDrawPolyLine(m_actualTransform, m_graphObjectManager.m_renderer,
                                                     m_actualProperties.m_strokeWidth);
            }
            else
            {
                m_actualPolygon.applyMatrixAndDrawPolyLine(m_actualTransform, m_graphObjectManager.m_renderer);
            }
        }

        // Get the bounding box and check sensitive zone
        if (null == useContext)
        {
            m_actualPolygon.getBoundingBox(m_senstiveZone.getZone());
            if (Debug.DEBUG_BBOX) 
            	System.out.println(this + " : bbox= (" + m_senstiveZone.m_rect.x_r + "," + m_senstiveZone.m_rect.y_r + "," + m_senstiveZone.m_rect.width_r + "," + m_senstiveZone.m_rect.height_r + ")");
            m_senstiveZone.checkZone(m_graphObjectManager);
        }
    }

    /**
     * Updates polygon points
     * @param newPolygon New points sequence
     */
    public void updateActualPolygonPoints(Polygon2D newPolygon)
    {
        // Sanity check
        if (newPolygon == null)
        {
            return;
        }

        // todo find another way, differentiate polygon and polyline rendering instead of closing the shape
        // close the polygon
        Point2D initialPoint = (Point2D)newPolygon.m_points.elementAt(0);
        newPolygon.m_points.addElement(new Point2D(initialPoint.x_r, initialPoint.y_r));

        m_actualPolygon = newPolygon;
    }


    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Reset polygon
        m_actualPolygon = m_polygon;
    }
}
