/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Polygon2D;
import iso.mpeg.streamezzo.player.library.Real;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.SensitiveZone;

import java.util.Vector;

/**
 * This class provides a higher level implementation of a 2D polygon shape.
 * Its direct parent shall be a Group node.
 * @see Group
 */
public class Rect extends RenderedGraphObject
{
    /** Width and height of the rectangle */
    public int m_rWidth, m_rHeight;

    /** Top/Left coordinate for DOM, presentation and previous presentation values*/
    public int m_x = 0, m_y = 0, m_rActualX = 0, m_rActualY = 0, m_rPrevActualX = 0, m_rPrevActualY = 0;

    /** rounded corner radius */
    public int m_rx, m_ry;

    /** Rectangle representation as polygon */
    public Polygon2D m_polygon = null;

    /** Actual top left point of the rectangle. It reflects applied matrix */
    public Point2D m_actualTopLeft = null;

    /** Actual bottom right point of the rectangle. It reflects applied matrix */
    public Point2D m_actualBottomRight = null;

    /** Acutal Width and Height of the rectangle. It reflects applied matrix */
    public int m_rActualWidth = -1, m_rActualHeight = -1;

    /** Previous presentation values */
    public int m_rPrevActualWidth = -1, m_rPrevActualHeight = -1;

    /** the sensitive zone */
    public SensitiveZone m_senstiveZone = new SensitiveZone();



////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////// Properties setters/getters /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    public Rect(GraphObjectManager graphObjectManager)
    {
        super( graphObjectManager);
        m_actualTransform = Matrix2D.createIdentityMatrix();

        // Create internal representation as polygon
        m_polygon = new Polygon2D();
        m_polygon.m_points = new Vector(5);
        for (int i=0; i<5; i++)
        {
            Point2D point = new Point2D(0, 0);
            m_polygon.m_points.addElement(point);
        }

        // Initialize actual values. They will reflect applied matrix
        m_actualTopLeft = new Point2D(0, 0);
        m_actualBottomRight = new Point2D(0, 0);
        m_rActualWidth = 0;
        m_rActualHeight = 0;
    }

    /**
     * Create a polygon from rectangle properties
     */
    protected void convertToPolygon()
    {
        Point2D point;
        for (int i=0; i<5; i++)
        {
            point = (Point2D)m_polygon.m_points.elementAt(i);
            point.x_r = m_rActualX;
            point.y_r = m_rActualY;
        }
        ((Point2D)m_polygon.m_points.elementAt(1)).x_r += m_rActualWidth;
        ((Point2D)m_polygon.m_points.elementAt(2)).x_r += m_rActualWidth;
        ((Point2D)m_polygon.m_points.elementAt(2)).y_r += m_rActualHeight;
        ((Point2D)m_polygon.m_points.elementAt(3)).y_r += m_rActualHeight;

    }

    /**
     * Applies a matrix and compute actual topleft and size.
     * @param transform Group matrix to apply.
     */
    public void applyMatrixToGetActualRect(Matrix2D transform)
    {
        // Apply on top-left point
        m_actualTopLeft.x_r = m_rActualX;
        m_actualTopLeft.y_r = m_rActualY;
        Point2D tmpPoint = transform.apply(m_actualTopLeft);
        m_actualTopLeft.x_r = tmpPoint.x_r;
        m_actualTopLeft.y_r = tmpPoint.y_r;

        // Apply to bottom-right point
        m_actualBottomRight.x_r = m_rActualX + m_rActualWidth;
        m_actualBottomRight.y_r = m_rActualY + m_rActualHeight;
        tmpPoint = transform.apply(m_actualBottomRight);
        m_actualBottomRight.x_r = tmpPoint.x_r;
        m_actualBottomRight.y_r = tmpPoint.y_r;

        // Compute actuel width and height
        m_rActualWidth = m_actualBottomRight.x_r - m_actualTopLeft.x_r;
        m_rActualHeight = m_actualBottomRight.y_r - m_actualTopLeft.y_r;
    }

    public void setDOMWidth(int r_width)
    {
        m_rWidth = r_width;
        wasUpdated();
    }
    public void setDOMHeight(int r_height)
    {
        m_rHeight = r_height;
        wasUpdated();
    }

    /**
     * Update actual with
     * @param rNewWidth New width (real 20.12)
     */
    public void updateActualWidth(int rNewWidth)
    {
        // Check changes
        if (rNewWidth != m_rPrevActualWidth)
        {
            m_rActualWidth = rNewWidth;
            wasUpdated();
        }
        else
        {
            m_rActualWidth = m_rPrevActualWidth;
            m_wasUpdated = false;
        }
    }

    /**
     * Update actual height
     * @param rNewHeight New width (real 20.12)
     */
    public void updateActualHeight(int rNewHeight)
    {
        // Check changes
        if (rNewHeight != m_rPrevActualHeight)
        {
            m_rActualHeight = rNewHeight;
            wasUpdated();
        }
        else
        {
            m_rActualHeight = m_rPrevActualHeight;
            m_wasUpdated = false;
        }
    }

    /**
     * Set DOM X
     * @param rX the new top/left x value
     */
    public void setDOMX(int rX)
    {
        m_x = rX;
        wasUpdated();
    }

    /**
     * Update actual X
     * @param rNewX the new top/left x value
     */
    public void updateActualX(int rNewX)
    {
        // Check changes
        if (rNewX != m_rPrevActualX)
        {
            m_rActualX = rNewX;
            wasUpdated();
        }
        else
        {
            m_rActualX = m_rPrevActualX;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual X
     */
    public int getActualX()
    {
        return m_rActualX;
    }

    /**
     * Set DOM Y
     * @param rY the new top/left y value
     */
    public void setDOMY(int rY)
    {
        m_x = rY;
        wasUpdated();
    }

    /**
     * Update actual Y
     * @param rNewY the new top/left y value
     */
    public void updateActualY(int rNewY)
    {
        // Check changes
        if (rNewY != m_rPrevActualY)
        {
            m_rActualY = rNewY;
            wasUpdated();
        }
        else
        {
            m_rActualY = m_rPrevActualY;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual CY
     */
    public int getActualY()
    {
        return m_rActualY;
    }

    public void addDOMX(int r_x)
    {
        m_x += r_x;
    }
    public void addDOMY(int r_y)
    {
        m_y += r_y;
    }

    public void addDOMWidth(int r_width)
    {
        m_rWidth += r_width;
    }
    public void addDOMHeight(int r_height)
    {
        m_rHeight += r_height;
    }

    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Changed ?
        if ( (m_rWidth != m_rActualWidth) || (m_rHeight != m_rActualHeight)
          || (m_x != m_rActualX) || (m_y != m_rActualY) )
        {
            // Store previous values
            m_rPrevActualWidth = m_rActualWidth;
            m_rPrevActualHeight = m_rActualHeight;
            m_rPrevActualX = m_rActualX;
            m_rPrevActualY = m_rActualY;

            // Set DOM values
            m_rActualWidth = m_rWidth;
            m_rActualHeight = m_rHeight;
            m_rActualX = m_x;
            m_rActualY = m_y;

            wasUpdated();
        }
    }

    /**
     * RenderedGraphObject overriding
     * Notify that some properties influencing rendering have changed.
     * This oprations set flags to ask for a pre-render.
     * This operation is useful because parent can be a text element (flowroot)
     */
    public void wasUpdated()
    {
        m_wasUpdated = true;
        m_needPreRender = true;

        // Nofity parent if this is a textbase object
        if ( (m_parent != null) && (m_parent instanceof TextBase) )
        {
            ((TextBase)m_parent).wasUpdated();
        }
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// Rendering operations //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    public void specificRender(int a_sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Compute polygon that represents the rect
        if ( (m_wasUpdated) || (m_polygon == null) )
        {
            convertToPolygon();
        }

        // Apply matrix
        if (parentHasChanged)
        {
            m_wasUpdated = true;
        }
        
        // Apply matrix to compute actual rectangle
        applyMatrixToGetActualRect(m_actualTransform);

        // Check if it must be drawn
        if ( (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_HIDDEN) ||
             (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_COLLAPSE) )
        {
            return;
        }

        // Fill if needed
        if (m_actualProperties.m_fill != Value.COLOR_NONE)
        {
            // set the color
            m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_fill, m_actualProperties.m_fillOpacity&0xFF);
            // set the opacity
            m_polygon.applyMatrixAndfillPolygon(m_actualTransform, m_graphObjectManager.m_renderer);
        }

        // Draw the stroke if needed
        if (m_actualProperties.m_stroke != Value.STROKE_NONE)
        {
            m_graphObjectManager.m_renderer.setColor(m_actualProperties.m_stroke, m_actualProperties.m_strokeOpacity&0xFF);
            if (m_actualProperties.m_strokeWidth != Real.VALUE1)
            {
                m_polygon.applyMatrixAndDrawPolyLine(m_actualTransform, m_graphObjectManager.m_renderer,
                                                     m_actualProperties.m_strokeWidth);
            }
            else
            {
               m_polygon.applyMatrixAndDrawPolyLine(m_actualTransform, m_graphObjectManager.m_renderer);
            }
        }

        // Get the bounding box and check sensitive zone
        if (null == useContext)
        {
            m_polygon.getBoundingBox(m_senstiveZone.getZone());
            if (Debug.DEBUG_BBOX) System.out.println(this + " : bbox= (" + m_senstiveZone.m_rect.x_r + "," + m_senstiveZone.m_rect.y_r + "," + m_senstiveZone.m_rect.width_r + "," + m_senstiveZone.m_rect.height_r + ")");
            m_senstiveZone.checkZone(m_graphObjectManager);
        }

        m_wasUpdated = false;
    }
}
