/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Rect2D;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;



public abstract class RenderedGraphObject extends GraphObject
{
    /** DOM properties */
    public NodeProperties m_properties = new NodeProperties();

    /** Properties that handles presentation values */
    public NodeProperties m_actualProperties = new NodeProperties();

    /** Properties that handles previous presentation values */
    public NodeProperties m_previousActualProperties = new NodeProperties();

    /** DOM matrix */
    public Matrix2D m_definedTransform = Matrix2D.createIdentityMatrix();

    /** Presentation value of the matrix */
    public Matrix2D m_actualTransform = Matrix2D.createIdentityMatrix();

    /** Previous presentation value of the matix */
    public Matrix2D m_prevActualTransform = Matrix2D.createIdentityMatrix();


    /**
     * Default constructor, the DefName is not set.
     */

    public RenderedGraphObject(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
    }

    /**
     * Default constructor with set the DefName.
     * When the DefName is set, the object is referenced in the DefName table
     * and it can be reach by calling the method getID(). In order to release
     * the memory of this object, one must call deleteNode(String)
     * If no DefName is set, then the object cannot be retreived but by using
     * its parent.
     * @param   id the DefName of this object.
     * @see     #getID
     * @see     #setID
     */
    public RenderedGraphObject(int id)
    {
        super(id);
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////// Setters/getters and parsing operations ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * SVG recommendation : Evaluates to "true" if one of the languages indicated by user preferences
     * exactly equals one of the languages given in the value of this parameter, or if one of the
     * languages indicated by user preferences exactly equals a prefix of one of the languages given
     * in the value of this parameter such that the first tag character following the prefix is "-".
     * If the attribute is not present, then its implicit return value is "true". If a null string or
     * empty string value is given to attribute systemLanguage, the attribute returns "false".
     * @param languageToCheck
     * @return boolean
     */
    public boolean isSystemLanguageSuitable(String languageToCheck)
    {
        // Check simple true case : no spcified
        if (m_properties.isInherited(Value.UPDATE_SYSTEMLANGUAGE))
        {
            return true;
        }

        // Check simple false case : null or empty
        if (m_properties.m_systemLanguage == null)
        {
            return false;
        }
        else if (m_properties.m_systemLanguage.length() == 0)
        {
            return false;
        }

        // check language list
        int i = 0;
        boolean found = false;
        String language = null;
        while( (i < m_properties.m_systemLanguage.length()) && (! found) )
        {
            // search ',' char
            int separatorPos = m_properties.m_systemLanguage.indexOf(',', i);
            if (separatorPos == -1)
            {
                language = m_properties.m_systemLanguage.substring(i);
                i = m_properties.m_systemLanguage.length();
            }
            else
            {
                language = m_properties.m_systemLanguage.substring(i, separatorPos);
                i = separatorPos + 1;
            }

            // check language
            // SVG recommendation : equals one of the languages given in the value of this parameter,
            // or if exactly equals a prefix of one of the languages such that the first tag character
            // following the prefix is "-".
            if ( (language != null) && (language.length() != 0) )
            {
                language = language.trim();
                if (language.equals(languageToCheck))
                {
                    // exactly equals
                    found = true;
                }
                else
                {
                    // check prefix
                    int prefixPos = language.indexOf('-');
                    if (prefixPos != -1)
                    {
                        language = language.substring(0 ,prefixPos);
                    }
                    if (language.equals(languageToCheck))
                    {
                        found = true;
                    }
                }
            }
        }

        return found;
    }

    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Overrriden when needed
    }

    public void updateActualFill(int newFill)
    {
        m_actualProperties.m_fill = newFill;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualStroke(int newStroke)
    {
        m_actualProperties.m_stroke = newStroke;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualFillOpacity(byte newFillOpacity)
    {
        m_actualProperties.m_fillOpacity = newFillOpacity;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualStrokeOpacity(byte newStrokeOpacity)
    {
        m_actualProperties.m_strokeOpacity = newStrokeOpacity;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualStrokeWidth(int newStrokeWidth)
    {
        m_actualProperties.m_strokeWidth = newStrokeWidth;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualFontFamily(String newFontFamily)
    {
        m_actualProperties.m_fontFamily = newFontFamily;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualFontSize(int newFontSize)
    {
        m_actualProperties.m_fontSize = newFontSize;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualFontStyle(byte newFontStyle)
    {
        m_actualProperties.m_fontStyle = newFontStyle;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualVisibility(byte newVisibility)
    {
        m_actualProperties.m_visibility = newVisibility;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualDisplay(byte newDisplay)
    {
        m_actualProperties.m_display = newDisplay;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualFillRule(byte newFillRule)
    {
        m_actualProperties.m_fillRule = newFillRule;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualFocusable(boolean newFocusable)
    {
        m_actualProperties.m_focusable = (byte)(newFocusable ? 1 : 0);

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualTextAlign(byte newAlign)
    {
        m_actualProperties.m_text_anchor = newAlign;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }
    public void updateActualDisplayAlign(byte newAlign)
    {
        m_actualProperties.m_displayAlign = newAlign;

        // Notify if properties change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }

    public void setDOMTransform(Matrix2D matrix)
    {
        Matrix2D.copyMatrixVariables(m_definedTransform, matrix);
    }

    public void setDOMTranslation(Point2D translation)
    {
        updateTranslation(m_definedTransform, translation, false);
    }

    public void addDOMTranslation(Point2D translation)
    {
        Point2D domTranslation = m_definedTransform.getTranslation();
        m_definedTransform.setTranslation(new Point2D(domTranslation.x_r + translation.x_r, domTranslation.y_r + translation.y_r));
    }

    /**
     * Updates the presentation value of the translation
     * @param newTranslation New translation
     * @param additive True is newTranslation must be added to the current actual matrix
     */
    public void updateActualTranslation(Point2D newTranslation, boolean additive)
    {
        updateTranslation(m_actualTransform, newTranslation, additive);
    }

    /**
     * Updates the value of the translation
     * @param targetTransform The transform matrix that will be modified
     * @param newTranslation New translation
     * @param additive True is newTranslation must be added to the current actual matrix
     */
    public void updateTranslation(Matrix2D targetTransform, Point2D newTranslation, boolean additive)
    {
        // Sanity check
        if (newTranslation == null)
        {
            return;
        }

        // Create translation matrix
        Matrix2D translation = Matrix2D.createIdentityMatrix();
        translation.setTranslation(newTranslation);

        // additive ?
        if (! additive)
        {
            // Delete translation in actual transform
            targetTransform.setTranslation(new Point2D(0, 0));
        }

        // Compute new actual transform
        Matrix2D result = Matrix2D.createIdentityMatrix();
        Matrix2D.Mul(targetTransform, translation, result);
        Matrix2D.copyMatrixVariables(targetTransform, result);

        // Notify if presentation value has been changed
        if (! m_actualTransform.equals(m_prevActualTransform))
        {
            wasUpdated();
        }
    }

    /**
    * Return a copy of actual translation
    * @return Point2D
    */
    public Point2D getActualTranslation()
    {
        Point2D tmpPoint = m_actualTransform.getTranslation();
        Point2D toReturn = new Point2D(tmpPoint.x_r, tmpPoint.y_r);
        return toReturn;
    }

    /**
    * Return a copy of actual scale
    * @return Point2D
    */
    public Point2D getActualScale()
    {
        Point2D tmpPoint = m_actualTransform.getScale();
        Point2D toReturn = new Point2D(tmpPoint.x_r, tmpPoint.y_r);
        return toReturn;
    }

    /**
     * Updates the presentation value of the scale
     * @param newScale New scale
     * @param additive True is newScale must be added to the current actual matrix
     */
    public void updateActualScale(Point2D newScale, boolean additive)
    {
        updateScale(m_actualTransform, newScale, additive);
    }

    public void setDOMScale(Point2D scale)
    {
        updateScale(m_definedTransform, scale, false);
    }

    /**
     * Updates the value of the scale
     * @param targetTransform The transform matrix that will be modified
     * @param newScale New scale
     * @param additive True is newScale must be added to the current actual matrix
     */
    public void updateScale(Matrix2D targetTransform, Point2D newScale, boolean additive)
    {
        // Sanity check
        if (newScale == null)
        {
            return;
        }

        // additive ?
        if (! additive)
        {
            targetTransform.setScale(newScale);
        }
        else
        {
            Matrix2D translation = Matrix2D.createIdentityMatrix();
            translation.setScale(newScale);
            Matrix2D result = Matrix2D.createIdentityMatrix();
            Matrix2D.Mul(targetTransform, translation, result);
            Matrix2D.copyMatrixVariables(targetTransform, result);
        }

        // Notify if presentation value has been changed
        if (! m_actualTransform.equals(m_prevActualTransform))
        {
            wasUpdated();
        }
    }

    public void setDOMRotation(int r_rotation)
    {
        updateRotation(m_definedTransform, r_rotation, false);
    }

    /**
     * Updates the presentation value of the rotation
     * @param rAngle New angle in degrees (20.12 fiwed point)
     * @param additive True is rAngle must be added to the current actual matrix
     */
    public void updateActualRotation(int rAngle, boolean additive)
    {
        updateRotation(m_actualTransform, rAngle, additive);
    }

    /**
     * Updates the presentation value of the rotation
     * @param targetTransform The transform matrix that will be modified
     * @param rAngle New angle in degrees (20.12 fiwed point)
     * @param additive True is rAngle must be added to the current actual matrix
     */
    public void updateRotation(Matrix2D targetTransform, int rAngle, boolean additive)
    {
        // additive ?
        if (! additive)
        {
            targetTransform.setRotation(rAngle);
        }
        else
        {
            Matrix2D translation = Matrix2D.createIdentityMatrix();
            translation.setRotation(rAngle);
            Matrix2D result = Matrix2D.createIdentityMatrix();
            Matrix2D.Mul(targetTransform, translation, result);
            Matrix2D.copyMatrixVariables(targetTransform, result);
        }

        // Notify if presentation value has been changed
        if (! m_actualTransform.equals(m_prevActualTransform))
        {
            wasUpdated();
        }
    }
    
    /**
     * Updates the presentation value of the rotation
     * @param rAngle New angle in degrees (20.12 fiwed point)
     * @param additive True is rAngle must be added to the current actual matrix
     */
    public void updateActualRotationWithCenter(int rAngle, int rCenterX, int rCenterY, boolean additive)
    {
        // additive ?
        if (! additive)
        {
            m_actualTransform.setRotationWithCenter(rAngle, rCenterX, rCenterY);
        }
        else
        {
            Matrix2D translation = Matrix2D.createIdentityMatrix();
            translation.setRotationWithCenter(rAngle, rCenterX, rCenterY);
            Matrix2D result = Matrix2D.createIdentityMatrix();
            Matrix2D.Mul(m_actualTransform, translation, result);
            m_actualTransform = result;
        }

        // Notify if something changes
        if (! m_actualTransform.equals(m_prevActualTransform))
        {
            wasUpdated();
        }
    }

    /**
     * Return the rotation value of the presentation matrix.
     * @return Angle in degrees (20.12 fiwed point)
     */
    public int getActualRotation()
    {
        return m_actualTransform.getRotation();
    }

    public void setDOMX(int rX)
    {
        // see implementation in the related node
    }

    public void updateActualX(int rX)
    {
        // see implementation in the related node
    }

    public int getActualX()
    {
        // see implementation in the related node
        return 0;
    }

    public int getActualY()
    {
        // see implementation in the related node
        return 0;
    }

    public void setDOMY(int rY)
    {
        // see implementation in the related node
    }
    public void updateActualY(int rY)
    {
        // see implementation in the related node
    }

    public void updateActualWidth(int rWidth)
    {
        // see implementation in the related node
    }

    public void updateActualHeight(int rHeight)
    {
        // see implementation in the related node
    }

    public int getActualWidth()
    {
        // see implementation in the related node
        return 0;
    }

    public int getActualHeight()
    {
        // see implementation in the related node
        return 0;
    }

    public void setDOMFocusable(boolean isTrue)
    {
        m_properties.m_focusable = (byte)(isTrue?1:0);
        m_properties.setInheritance(Value.UPDATE_FOCUSABLE, false);
    }
    public void setDOMTextAlign(byte textAlign)
    {
        m_properties.m_text_anchor = textAlign;
        m_properties.setInheritance(Value.UPDATE_TEXTANCHOR, false);
    }
    public void setDOMDisplayAlign(byte displayAlign)
    {
        m_properties.m_displayAlign = displayAlign;
        m_properties.setInheritance(Value.UPDATE_DISPLAYALIGN, false);
    }
    public void setDOMFontStyle(byte fontstyle)
    {
        m_properties.m_fontStyle = fontstyle;
        m_properties.setInheritance(Value.UPDATE_FONTSTYLE, false);
    }
    public void setDOMFontSize(int r_fontsize)
    {
        m_properties.m_fontSize = r_fontsize;
        m_properties.setInheritance(Value.UPDATE_FONTSIZE, false);
    }
    public void setDOMStrokeWidth(int r_width)
    {
        m_properties.m_strokeWidth = r_width;
        m_properties.setInheritance(Value.UPDATE_STROKEWIDTH, false);
    }
    public void setDOMFillOpacity(byte opacity)
    {
        m_properties.m_fillOpacity = opacity;
        m_properties.setInheritance(Value.UPDATE_FILLOPACITY, false);
    }
    public void setDOMStrokeOpacity(byte opacity)
    {
        m_properties.m_strokeOpacity = opacity;
        m_properties.setInheritance(Value.UPDATE_STROKEOPACITY, false);
    }
    public void setDOMFillColor(int color)
    {
        m_properties.m_fill = color;
        m_properties.setInheritance(Value.UPDATE_COLOR_FILL, false);
    }
    public void setDOMStroke(int color)
    {
        m_properties.m_stroke = color;
        m_properties.setInheritance(Value.UPDATE_STROKE, false);
    }
    public void setDOMDisplay(byte display)
    {
        m_properties.m_display = display;
        m_properties.setInheritance(Value.UPDATE_DISPLAY, false);
    }
    public void setDOMVisibility(byte visibility)
    {
        m_properties.m_visibility = visibility;
        m_properties.setInheritance(Value.UPDATE_VISIBILITY, false);
    }
    public void setDOMFontFamily(String fontfamily)
    {
        m_properties.m_fontFamily = fontfamily;
        m_properties.setInheritance(Value.UPDATE_FONTFAMILY, false);
    }

    public void addDOMFontSize(int r_fontsize)
    {
        m_properties.m_fontSize += r_fontsize;
        m_properties.setInheritance(Value.UPDATE_FONTSIZE, false);
    }
    public void addDOMStrokeWidth(int r_width)
    {
        m_properties.m_strokeWidth += r_width;
        m_properties.setInheritance(Value.UPDATE_STROKEWIDTH, false);
    }
    public void addDOMFillOpacity(byte opacity)
    {
        m_properties.m_fillOpacity = (byte)(((byte)m_properties.m_fillOpacity) + ((byte)opacity));
        m_properties.setInheritance(Value.UPDATE_FILLOPACITY, false);
    }
    public void addDOMStrokeOpacity(byte opacity)
    {
        m_properties.m_strokeOpacity = (byte)(((byte)m_properties.m_strokeOpacity) + ((byte)opacity));
        m_properties.setInheritance(Value.UPDATE_STROKEOPACITY, false);
    }
    public void addDOMFillColor(int color)
    {
        Value value1 = new Value();
        value1.m_enumOrColorOrRealOrID = color;
        Value value2 = new Value();
        if (m_properties.m_stroke == Value.COLOR_NONE)
        {
            value2.m_enumOrColorOrRealOrID = 0x00000000;
        }
        else
        {
            value2.m_enumOrColorOrRealOrID = m_properties.m_fill;
        }
        Value result = value2.Add(value1, Value.COLOR);
        m_properties.m_fill = result.m_enumOrColorOrRealOrID;
        m_properties.setInheritance(Value.UPDATE_COLOR_FILL, false);
    }
    public void addDOMStroke(int color)
    {
        Value value1 = new Value();
        value1.m_enumOrColorOrRealOrID = color;
        Value value2 = new Value();
        if (m_properties.m_stroke == Value.COLOR_NONE)
        {
            value2.m_enumOrColorOrRealOrID = 0x00000000;
        }
        else
        {
            value2.m_enumOrColorOrRealOrID = m_properties.m_stroke;
        }
        Value result = value2.Add(value1, Value.COLOR);
        m_properties.m_stroke = result.m_enumOrColorOrRealOrID;
        m_properties.setInheritance(Value.UPDATE_STROKE, false);
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////// Properties inheritance management /////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    protected void processTranformationMatrices(Matrix2D parentTransform)
    {
        Matrix2D tmpMatrix = Matrix2D.createIdentityMatrix();

        if (parentTransform != null)
        {
            Matrix2D.Mul(parentTransform, m_actualTransform, tmpMatrix);
        }

        Matrix2D.copyMatrixVariables(m_actualTransform, tmpMatrix);
    }
    
    // TODO not all the attributes from m_properties are processed (copied) to m_actualproperties
    /**
     * This operation process properties inheritance and then provides the actual properties to use.
     * It also reset DOM values and matrix.
     * @param parentProperties Actual parent properties (inherited from ancestors)
     */
    protected void resetDOMValuesAndProcessInheritance(/*Matrix2D parentTransform,*/ NodeProperties parentProperties)
    {
        // Reset DOM values
        resetDOMValues();

        // Previous property
        m_previousActualProperties = m_actualProperties.getCopy();

        // Matrix
//        if (m_definedTransform == null && parentTransform != null)
//        	Matrix2D.copyMatrixVariables(m_actualTransform, parentTransform);
//        else
        	Matrix2D.copyMatrixVariables(m_actualTransform, m_definedTransform);

        // Non heritable properties
        m_actualProperties.m_focusable = m_properties.m_focusable;
        /*m_actualProperties.m_navIndex = m_properties.m_navIndex;
        m_actualProperties.m_navLeft = m_properties.m_navLeft;
        m_actualProperties.m_navRight = m_properties.m_navRight;
        m_actualProperties.m_navUp = m_properties.m_navUp;
        m_actualProperties.m_navDown = m_properties.m_navDown;*/

        // Fill
        if (m_properties.isInherited(Value.UPDATE_COLOR_FILL))
        {
            m_actualProperties.m_fill = parentProperties.m_fill;
        }
        else
        {
             m_actualProperties.m_fill = m_properties.m_fill;
        }

        // Stroke
        if (m_properties.isInherited(Value.UPDATE_STROKE))
        {
            m_actualProperties.m_stroke = parentProperties.m_stroke;
        }
        else
        {
             m_actualProperties.m_stroke = m_properties.m_stroke;
        }

        // Fill opacity
        if (m_properties.isInherited(Value.UPDATE_FILLOPACITY))
        {
            m_actualProperties.m_fillOpacity = parentProperties.m_fillOpacity;
        }
        else
        {
             m_actualProperties.m_fillOpacity = m_properties.m_fillOpacity;
        }

        // Stroke opacity
        if (m_properties.isInherited(Value.UPDATE_STROKEOPACITY))
        {
            m_actualProperties.m_strokeOpacity = parentProperties.m_strokeOpacity;
        }
        else
        {
             m_actualProperties.m_strokeOpacity = m_properties.m_strokeOpacity;
        }

        // Stroke width
        if (m_properties.isInherited(Value.UPDATE_STROKEWIDTH))
        {
            m_actualProperties.m_strokeWidth = parentProperties.m_strokeWidth;
        }
        else
        {
             m_actualProperties.m_strokeWidth = m_properties.m_strokeWidth;
        }

        // Font family
        if (m_properties.isInherited(Value.UPDATE_FONTFAMILY))
        {
            m_actualProperties.m_fontFamily = parentProperties.m_fontFamily;
        }
        else
        {
             m_actualProperties.m_fontFamily = m_properties.m_fontFamily;
        }

        // Font size
        if (m_properties.isInherited(Value.UPDATE_FONTSIZE))
        {
            m_actualProperties.m_fontSize = parentProperties.m_fontSize;
        }
        else
        {
             m_actualProperties.m_fontSize = m_properties.m_fontSize;
        }
        
        if (this instanceof Text) {
	        m_actualProperties.m_fontFamily = parentProperties.m_fontFamily;
	        m_actualProperties.m_fontSize = parentProperties.m_fontSize;
        }
        
        if (this instanceof TextArea) {
	        m_actualProperties.m_fontFamily = parentProperties.m_fontFamily;
	        m_actualProperties.m_fontSize = parentProperties.m_fontSize;
        }

        // Font style
        if (m_properties.isInherited(Value.UPDATE_FONTSTYLE))
        {
            m_actualProperties.m_fontStyle = parentProperties.m_fontStyle;
        }
        else
        {
             m_actualProperties.m_fontStyle = m_properties.m_fontStyle;
        }

        // Display
        if ( (m_properties.isInherited(Value.UPDATE_DISPLAY)) || (m_properties.m_display == NodeProperties.DISPLAY_INHERIT) )
        {
            m_actualProperties.m_display = parentProperties.m_display;
        }
        else
        {
            m_actualProperties.m_display = m_properties.m_display;
        }

        // Visibility
        if ( (m_properties.isInherited(Value.UPDATE_VISIBILITY)) || (m_properties.m_visibility == NodeProperties.VISIBILITY_INHERIT) )
        {
            m_actualProperties.m_visibility = parentProperties.m_visibility;
        }
        else
        {
            m_actualProperties.m_visibility = m_properties.m_visibility;
        }

        // Fill rule
        if (m_properties.isInherited(Value.UPDATE_FILLRULE))
        {
            m_actualProperties.m_fillRule = parentProperties.m_fillRule;
        }
        else
        {
            m_actualProperties.m_fillRule = m_properties.m_fillRule;
        }

        // Text align
        if (m_properties.isInherited(Value.UPDATE_TEXTANCHOR))
        {
            m_actualProperties.m_text_anchor = parentProperties.m_text_anchor;
        }
        else
        {
            m_actualProperties.m_text_anchor = m_properties.m_text_anchor;
        }

        // Display align
        if (m_properties.isInherited(Value.UPDATE_DISPLAYALIGN))
        {
            m_actualProperties.m_displayAlign = parentProperties.m_displayAlign;
        }
        else
        {
            m_actualProperties.m_displayAlign = m_properties.m_displayAlign;
        }

        // Notify if something change
        if (! m_actualProperties.equals(m_previousActualProperties))
        {
            wasUpdated();
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Rendering operations ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    public abstract void specificRender(int sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext);

    public void render(int sceneTimeInMillis, boolean parentHasChanged,
                       Matrix2D parentTransform, NodeProperties parentProperties, GraphObject useContext)
    {
        // Disable the cursor if this element is the cursor representation
        boolean bHasDisabledTheCursor = false;
        if (this == (GraphObject)m_graphObjectManager.getActiveCursorView())
        {
            bHasDisabledTheCursor = true;
            m_graphObjectManager.disableCursor(true);
        }

        // Reset DOM values and Process inheritance
        resetDOMValuesAndProcessInheritance(parentProperties);

        // Animation
        animateAll(sceneTimeInMillis);

        processTranformationMatrices(parentTransform);

        // Check display because animation can modify it
        if (m_actualProperties.m_display != NodeProperties.DISPLAY_INLINE)
        {
            return;
        }

        // Register itself as navigatable
        registerAsNavigatable();

        // Push its picking listeners
        int pushedCount = pushPickingListeners();

        // Render strictly speaking
        // TML : TODO :  test must be in each node like visibility test
        if(!(this.m_properties.m_Deactivated || this.m_properties.m_Deactivating || this.m_properties.m_DeactivatedOwner))
        	specificRender(sceneTimeInMillis, parentHasChanged, useContext);

        // Pop the pushed picking listeners
        m_graphObjectManager.popPickingListeners(pushedCount);

        // Enable cursor if it was disabled here
        if (bHasDisabledTheCursor)
        {
            m_graphObjectManager.disableCursor(false);
        }
    }

    /**
     * Push the picking listeners listening for this element
     * @return the number of listeners pushed
     */
    protected int pushPickingListeners()
    {
        int numberOfPushedListeners = 0;
        GraphObject listener = null;
        if (null != m_eventManager)
        {
            for (int i = 0; i < m_eventManager.listeners.size(); i++)
            {
                listener = (GraphObject)m_eventManager.listeners.elementAt(i);
                if (listener instanceof Listener)
                {
                    if (((Listener)listener).isListeningForMouseEvent())
                    {
                        m_graphObjectManager.pushPickingListener(listener);
                        numberOfPushedListeners++;
                    }
                }
            }
        }
        if (this instanceof A)
        {
            m_graphObjectManager.pushPickingListener(this);
            numberOfPushedListeners++;
        }
        return numberOfPushedListeners;
    }

    /**
     * Register this node as navigatable if it is focusable
     * This operation can be overriden. In particular for text node.
     */
    protected void registerAsNavigatable()
    {
        if (m_actualProperties.m_focusable != 0)
        {
            m_graphObjectManager.registerNodeAsNavigatable(this);
        }
    }

    /**
     * Sets the passed Rect2D object with the bounding box of all the underlying rendered object
     * @param boundingBox the Rect2D to update
     */
    public void updateBoundingBox(Rect2D boundingBox)
    {
        // each rendered objects has to override the method
        if (boundingBox != null)
        {
            // set the bounding box with undefined dimensions
            boundingBox.width_r = -1;
            boundingBox.height_r = -1;
        }
    }
}
