/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.BasicRendering;
import iso.mpeg.streamezzo.player.library.Real;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;

public class SVGanimate extends SVGset
{

    // Calc modes
    public static final byte CALCMODE_DISCRETE = 0;
    public static final byte CALCMODE_LINEAR = 1;
    public static final byte CALCMODE_PACED = 2;
    public static final byte CALCMODE_SPLINE = 3;

    // Animation classes
    public static final byte CLASS_UNDEFINED = -1;
    public static final byte CLASS_FROMTO = 0;
    public static final byte CLASS_FROMBY = 1;
    public static final byte CLASS_BY = 2;
    public static final byte CLASS_TO = 3;
    public static final byte CLASS_VALUES = 4;

    /** The animation class */
    protected byte m_animClass = CLASS_UNDEFINED;

    /** Specifies the interpolation mode for the animation */
    public byte m_calcMode = CALCMODE_LINEAR;

    /** Specifies the starting value of the animation. */
    public Value m_from = null;

    /** Specifies a relative offset value for the animation. */
    public Value m_by = null;

    /** Each time in the list corresponds to a value in the m_values list, and defines when the value
     *  is used in the animation function. */
    public int m_keyTimes[] = null;
        public Value m_values[] = null;

    /** A set of Bzier control points associated with the keyTimes list, defining a cubic Bzier function
     *  that controls interval pacing. The keyTimes values that define the associated segment are the
     *  Bzier "anchor points", and the keySplines values are the control points.
     *  This attribute is ignored unless the calcMode is set to "spline" */
    public int m_keySplines[] = null;

    /** Controls whether or not the animation is additive. */
    public boolean m_additive = false;

    /** Controls whether or not the animation is cumulative. */
    public boolean m_accumulate = false;

    /** Values for interpolation */
    protected ValuesForInterpolation m_valuesInterpol = new ValuesForInterpolation();

    public SVGanimate(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////// Properties management /////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * This operation generates keyTimes for values according to svg specification
     */
    private void generatesKeyTimes()
    {
        // In a list of keyTimes, there must be exactly as many values in the keyTimes list as in the values list.
        // It depends on calcMode and attribute type (discrete or not)
        if ( (Value.isDiscreteAttribute(m_attributeType)) || (m_calcMode == CALCMODE_DISCRETE) )
        {
            // The duration is divided into n even periods
            m_keyTimes = new int[m_values.length];
            int rPeriod = Real.VALUE1 / m_values.length;
            for (int i = 0; i < m_values.length - 1; i++)
            {
                m_keyTimes[i] = i * rPeriod;
            }
            m_keyTimes[m_values.length - 1] = Real.VALUE1;
        }
        else 
        	/*if ( (m_calcMode == CALCMODE_LINEAR) || (m_calcMode == CALCMODE_SPLINE) )*/
        	//TODO is there any special processing for CALCMODE_PACED, or we treat it here?
        {
            // The duration is divided into n-1 even periods
            m_keyTimes = new int[m_values.length];
            int rPeriod = Real.VALUE1 / (m_values.length - 1);
            for (int i = 0; i < m_values.length - 1; i++)
            {
                m_keyTimes[i] = i * rPeriod;
            }
            m_keyTimes[m_values.length - 1] = Real.VALUE1;
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////// Specific animation control operations //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Starts animation. Overrides SVGSet operation.
     * @param targetNode The targetted node
     * @param sceneTimeInMillis The current scene time
     */
    protected void startAnimation(RenderedGraphObject targetNode, int sceneTimeInMillis)
    {
        // Super
        super.startAnimation(targetNode, sceneTimeInMillis);

        // Get the animation class
        if (m_values != null)
        {
            // Values class
            m_animClass = CLASS_VALUES;
            // initialize m_to for freezeAnimation
            m_to = m_values[m_values.length - 1];

            // Check KeyTimes
            if ( (m_keyTimes == null) || (m_keyTimes.length != m_values.length) )
            {
                // No key times or key times error so creates key times
                generatesKeyTimes();
            }
        }
        else if (m_from != null)
        {
            if (m_to != null)
            {
                m_animClass = CLASS_FROMTO;
            }
            else if (m_by != null)
            {
                m_animClass = CLASS_FROMBY;
            }
        }
        else
        {
            if (m_to != null)
            {
                m_animClass = CLASS_TO;
            }
            else if (m_by != null)
            {
                m_animClass = CLASS_BY;
            }
        }
        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "  animClass = " + m_animClass);

        // Simple duration must be defined for interpolation
        if (m_dur == BasicRendering.TIME_INDEFINITE)
        {
            // Not valid animation
            m_animClass = CLASS_UNDEFINED;
        }

        // Ceck animation class to adapt possibly keyTimes and values
        if ( (m_animClass != CLASS_VALUES) && (m_animClass != CLASS_UNDEFINED) )
        {
            // Allocate 2 keys to store initial and final state
            m_values = new Value[2];
            Value valueOfTarget;
            switch (m_animClass)
            {
            	
                case CLASS_FROMTO :
                    m_values[0] = m_from;
                    m_values[1] = m_to;
                    break;
                case CLASS_TO :
                    m_values[0] = getValueOfTarget(targetNode);
                    m_values[1] = m_to;
                    break;
                case CLASS_FROMBY :
                    m_values[0] = m_from;
                    m_values[1] = m_by.Add(m_from, m_attributeType);
                    break;
                case CLASS_BY :
                    valueOfTarget = getValueOfTarget(targetNode);
                    m_values[0] = valueOfTarget;
                    m_values[1] = m_by.Add(valueOfTarget, m_attributeType);
                    break;
            }
            generatesKeyTimes();

            if ( m_values[1] == null)
            {
                // addition not supproted for this value so animation is not valid
                m_animClass = CLASS_UNDEFINED;
            }
        }
    }

    /**
     * Process current simple animation (interpolation oparation)
     * @param targetNode
     * @param sceneTimeInMillis
     */
    protected void processSimpleAnimation(RenderedGraphObject targetNode, int sceneTimeInMillis)
    {
        // Nothing to do if animation class or target arre invalid
        if ((m_animClass == CLASS_UNDEFINED) || (targetNode == null))
        {
            return;
        }

        // Get 'from' value if not provided at decoding time
        switch (m_animClass)
        {
            case CLASS_TO :
                m_values[0] = getValueOfTarget(targetNode);
                break;
            case CLASS_BY :
                Value valueOfTarget = getValueOfTarget(targetNode);
                m_values[0] = valueOfTarget;
                m_values[1] = m_by.Add(valueOfTarget, m_attributeType);
                break;
        }

        // Retrieve the 2 values between which interpolation applies
        // Somtimes no interpolation is needed because it comes upon a keyTime
        getBoundValuesForInterpolation(sceneTimeInMillis);
        Value downValue = m_valuesInterpol.downValue;
        Value upValue = m_valuesInterpol.upValue;
        int rCoeff = m_valuesInterpol.rCoeff;
        
        // Check result for interpolate
        Value valueToSet = null;
        if ( (rCoeff == 0) && (downValue != null) )
        {
            // No interpolation
            valueToSet = downValue;
        }
        else if ( (downValue != null) && (upValue != null) )
        {
            // Interpolate
            valueToSet = Value.linearInterpolation(downValue, upValue, rCoeff, m_attributeType);
        }

        // Set computed value
        if (valueToSet != null)
        {
            // Check additive/accumulate modes
            if (m_accumulate)
            {
                // Add with the value of the last keytime
                /*!!*/if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "  valueToSet = " + Integer.toHexString(valueToSet.m_enumOrColorOrRealOrID));
                Value lastValue = m_values[m_values.length - 1];
                lastValue = lastValue.Mul(m_simpleAnimCount, m_attributeType);
                /*!!*/if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "  lastValue = " + Integer.toHexString(lastValue.m_enumOrColorOrRealOrID));
                valueToSet = valueToSet.Add(lastValue, m_attributeType);
            }
            if (m_additive)
            {
                // Add to current value
                /*!!*/if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "  valueToSet = " + Integer.toHexString(valueToSet.m_enumOrColorOrRealOrID));
                Value currentValue = getValueOfTarget(targetNode);
                /*!!*/if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "  currentValue = " + Integer.toHexString(currentValue.m_enumOrColorOrRealOrID));
                valueToSet = valueToSet.Add(currentValue, m_attributeType);
            }

            /*!!*/if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "  valueToSet = " + Integer.toHexString(valueToSet.m_enumOrColorOrRealOrID) + " , " + valueToSet.m_string);
          
            setValueToTarget(targetNode, valueToSet);
        }
    }


    /**
     * Provides usable values for interpolation.
     * If it comes upon a keyTime only downValue is filled and rCoeff is 0.
     * @param sceneTimeInMillis
     */
    protected void getBoundValuesForInterpolation(int sceneTimeInMillis)
    {
        // Get time offset from simple animation start time
        int elapsedTime = sceneTimeInMillis - m_simpleAnimBegin;

        // Intialize
        m_valuesInterpol.rCoeff = 0;
        m_valuesInterpol.downValue = null;
        m_valuesInterpol.upValue = null;

        // Search keys
        int keysCount = m_keyTimes.length;
        int indexInterpol = keysCount;
        int index = 0;
        boolean found = false;
        int timeDown = 0;
        int timeUp = 0;
        int currKeyTime= 0;
        for (index = 0; ((index < keysCount) && (!found)); index++)
        {
            currKeyTime = Real.applyMul(m_dur, m_keyTimes[index]);
            if (elapsedTime < currKeyTime)
            {
                found = true;
                indexInterpol = index - 1;
                timeDown = elapsedTime - Real.applyMul(m_dur, m_keyTimes[indexInterpol]);
                timeUp = currKeyTime - Real.applyMul(m_dur, m_keyTimes[indexInterpol]);
                m_valuesInterpol.rCoeff = Real.applyDiv(Real.convertIntToRealInt(timeDown), Real.convertIntToRealInt(timeUp));
            }
            else if (elapsedTime == currKeyTime)
            {
                found = true;
                m_valuesInterpol.downValue = m_values[index];
            }
        }

        if (m_valuesInterpol.downValue == null)
        {
            if (indexInterpol == -1)
            {
                m_valuesInterpol.downValue = m_values[0];
            }
            else if (indexInterpol == keysCount)
            {
                m_valuesInterpol.downValue = m_values[keysCount - 1];
            }
        }

        if (m_valuesInterpol.downValue == null)
        {
            m_valuesInterpol.downValue = m_values[indexInterpol];
            m_valuesInterpol.upValue = m_values[indexInterpol + 1];
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////// Structure which handles values for interpolation ////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
protected class ValuesForInterpolation
{
    /** lower value */
    public Value downValue = null;
    /** upper value */
    public Value upValue = null;
    /** Coeff to aplly to the upper value */
    public int rCoeff = 0;
}

}
