/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;

public class SVGanimateTransform extends SVGanimate
{

////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////// Properties management and parsing operations //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    public SVGanimateTransform(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// Target getter/setter operations ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
     * Provides a COPY of the current value of an attribute of the target
     * @param targetNode The target node whose value is animated
     */
    protected Value getValueOfTarget(RenderedGraphObject targetNode)
    {
        // Sanity check
        if (targetNode == null)
        {
            return null;
        }

        // Get the right value
        Value valueToReturn = new Value();
        switch (m_attribute)
        {
            case NodeProperties.TRANSFORMTRANSLATE :
                valueToReturn.m_floatsOrInts = new int[2];
                Point2D trans = targetNode.getActualTranslation();
                valueToReturn.m_floatsOrInts[0] = trans.x_r;
                valueToReturn.m_floatsOrInts[1] = trans.y_r;
                break;

            case NodeProperties.TRANSFORMSCALE :
                valueToReturn.m_floatsOrInts = new int[2];
                Point2D scale = targetNode.getActualScale();
                valueToReturn.m_floatsOrInts[0] = scale.x_r;
                valueToReturn.m_floatsOrInts[1] = scale.y_r;
                break;

            case NodeProperties.TRANSFORMROTATE :
                valueToReturn.m_floatsOrInts = new int[1];
                valueToReturn.m_floatsOrInts[0] = targetNode.getActualRotation();
                break;

            // todo : other cases
        }
        return valueToReturn;
    }

    /**
     * Process animation itself.
     * @param targetNode The target node whose value is animated
     */
    protected void setValueToTarget(RenderedGraphObject targetNode, Value valueToSet)
    {
        // Sanity check
        if (targetNode == null)
        {
            return;
        }

        if (valueToSet == null)
        {
            return;
        }

        // Set the right value
        switch (m_attribute)
        {
            case NodeProperties.TRANSFORMTRANSLATE :
                // Get translation
                Point2D translation = null;
                if (valueToSet.m_floatsOrInts != null)
                {
                    translation = new Point2D(valueToSet.m_floatsOrInts[0], valueToSet.m_floatsOrInts[1]);
                }
                // Set translation
                if (translation != null)
                {
                    targetNode.updateActualTranslation(translation, m_additive);
                }
                break;

            case NodeProperties.TRANSFORMSCALE :
                // Get scale factors
                Point2D scale = null;
                if (valueToSet.m_floatsOrInts != null)
                {
                    if (valueToSet.m_floatsOrInts.length == 2)
                    {
                        scale = new Point2D(valueToSet.m_floatsOrInts[0], valueToSet.m_floatsOrInts[1]);
                    }
                    else if (valueToSet.m_floatsOrInts.length == 1)
                    {
                        scale = new Point2D(valueToSet.m_floatsOrInts[0], valueToSet.m_floatsOrInts[0]);
                    }
                }
                // Set scale
                if (scale != null)
                {
                    targetNode.updateActualScale(scale, m_additive);
                }
                break;

            case NodeProperties.TRANSFORMROTATE :
                // Set rotation
                if (valueToSet != null && valueToSet.m_floatsOrInts != null) {
                    switch (valueToSet.m_floatsOrInts.length) {
                    case 1:
                        targetNode.updateActualRotation(valueToSet.m_floatsOrInts[0], m_additive);
                        break;
                    case 2:
                        targetNode.updateActualRotationWithCenter(valueToSet.m_floatsOrInts[0], valueToSet.m_floatsOrInts[1], valueToSet.m_floatsOrInts[1], m_additive);
                        break;
                    case 3:
                        targetNode.updateActualRotationWithCenter(valueToSet.m_floatsOrInts[0], valueToSet.m_floatsOrInts[1], valueToSet.m_floatsOrInts[2], m_additive);
                        break;
                }
                break;

            // todo : other cases
            }
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// Specific animation operations /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

/**
     * Process current simple animation (interpolation oparation)
     * @param targetNode
     * @param sceneTimeInMillis
     */
    protected void processSimpleAnimation(RenderedGraphObject targetNode, int sceneTimeInMillis)
    {
        // Nothing to do if animation class or target arre invalid
        if ((m_animClass == CLASS_UNDEFINED) || (targetNode == null))
        {
            return;
        }

        // Retrieve the 2 values between which interpolation applies
        // Somtimes no interpolation is needed because it comes upon a keyTime
        getBoundValuesForInterpolation(sceneTimeInMillis);
        Value downValue = m_valuesInterpol.downValue;
        Value upValue = m_valuesInterpol.upValue;
        int rCoeff = m_valuesInterpol.rCoeff;

        // Check result for interpolate
        Value valueToSet = null;
        if ( (rCoeff == 0) && (downValue != null) )
        {
            // No interpolation
            valueToSet = downValue;
        }
        else if ( (downValue != null) && (upValue != null) )
        {
            // Interpolate
            valueToSet = Value.linearInterpolation(downValue, upValue, rCoeff, m_attributeType);
        }

        // Set computed value
        if (valueToSet != null)
        {
            if (Debug.DEBUG_ANIMATION)
            {
                if (m_attribute == NodeProperties.TRANSFORMTRANSLATE)
                    System.out.println(this + "  " + sceneTimeInMillis + "  valueToSet = " + valueToSet.m_floatsOrInts[0] + ", " + valueToSet.m_floatsOrInts[1]);
                else
                    System.out.println(this + "  " + sceneTimeInMillis + "  valueToSet = " + valueToSet.m_floatsOrInts[0]);
            }

            // Check additive/accumulate modes
            if (m_accumulate)
            {
                // Get last value
                Value lastValue = m_values[m_values.length - 1];
                lastValue = lastValue.Mul(m_simpleAnimCount, m_attributeType);
                if (Debug.DEBUG_ANIMATION)
                {
                    if (m_attribute == NodeProperties.TRANSFORMTRANSLATE)
                        System.out.println(this + "  " + sceneTimeInMillis + "  lastValue = " + lastValue.m_floatsOrInts[0] + ", " + lastValue.m_floatsOrInts[1]);
                    else
                        System.out.println(this + "  " + sceneTimeInMillis + "  lastValue = " + lastValue.m_floatsOrInts[0]);
                }

                // Add to last value
                valueToSet = valueToSet.Add(lastValue, m_attributeType);
                if (Debug.DEBUG_ANIMATION)
                {
                    if (m_attribute == NodeProperties.TRANSFORMTRANSLATE)
                        System.out.println(this + "  " + sceneTimeInMillis + "  valueToSet = " + valueToSet.m_floatsOrInts[0] + ", " + valueToSet.m_floatsOrInts[1]);
                    else
                        System.out.println(this + "  " + sceneTimeInMillis + "  valueToSet = " + valueToSet.m_floatsOrInts[0]);
                }
            }
            if (m_additive)
            {
                // Cannot perform addition here because it must be matrix multiplication...
            }

            setValueToTarget(targetNode, valueToSet);
        }
        //todo
    }
}