/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.BasicRendering;
import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.TimedElement;

public class SVGset extends GraphObject implements TimedElement {

    //newly added attrs
    /** Specifies the target node to animate */
	public Xlink xlink = new Xlink();
	public int min = Integer.MIN_VALUE;
	public int max = Integer.MIN_VALUE;

	public static final byte RESTART_ALWAYS = 0; //can be restarted at any time.
    public static final byte RESTART_WHENNOTACTIVE = 1; //can only be restarted when it is not active (i.e. after the active end).
    public static final byte RESTART_NEVER = 2; //cannot be restarted for the remainder of the current simple duration of the parent time container.

    public static final byte REPEAT_INDEFINITLY = -1; // repeat indefinitly
    public static final byte REPEAT_NOTSPECIFIED = -2; // not specified

    /** Defines when the element should begin (i.e. become active). LASeR default value: 0*/
	public int m_begin = 0;
    /** Specifies the simple duration. If is indefinite (REPEAT_INDEFINITLY) interpolation will not work */
    public int m_dur = BasicRendering.TIME_INDEFINITE;
    /** Defines an end value for the animation that can constrain the active duration */
    public int m_end = BasicRendering.TIME_INDEFINITE;
    /** Restart rule. see above. */
    public byte m_restart = RESTART_ALWAYS;
    /** Specifies the number of iterations of the animation function. */
    public int m_repeatCount = REPEAT_NOTSPECIFIED;
    /** Specifies the total duration for repeat. */
    public int m_repeatDur = REPEAT_NOTSPECIFIED;
    /** to freeze the effect value at the last value of the active duration */
    public boolean m_freeze = false;

    /** Specifies the value for the attribute during the duration of the animation */
    public Value m_to = null;
    /** Specifies the attribute to animate */
    public byte m_attribute = -1;
    /** The attribute type */
    public byte m_attributeType = Value.UNKNOWN;

    public static final byte HREF_UNDEFINED     = -1; // HREF is not defined
    public static final byte HREF_DIRECT_PARENT = -2; // HREF is the node's direct parent

    /** Internal state */
    public static final byte ANIMSTATE_IDLE = 0; // Idle, animation not yet started
    public static final byte ANIMSTATE_IDLEAFTERSTOP = 1; //Idle, animation aleady done
    public static final byte ANIMSTATE_ANIMATING = 2;
    public static final byte ANIMSTATE_IDLEANDWAITFOREND = 3; //Simple animation and its repeats ended
    public static final byte ANIMSTATE_FREEZEDAFTEREND = 4; //Animation is freezed after end time
    public static final byte ANIMSTATE_FREEZEDAFTERDUR = 5; //Animation is freezed after duration time
    public static final byte ANIMSTATE_SIMPLEENDNOREPEAT = 6; // End of a simple animation, no repeat
    private byte m_state = ANIMSTATE_IDLE;

    /** Beginning of the current simple animation */
    protected int m_simpleAnimBegin = BasicRendering.TIME_INDEFINITE;
    /** Current repeat count */
    protected int m_simpleAnimCount = 0;
    /** Idle time of animation acording to possibly repeatDur. Anim could be still rendered if end time hasn't come */
    protected int m_animIdleTime = BasicRendering.TIME_INDEFINITE;


    public SVGset(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
        m_isTimeDependent = true;
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////// Properties management /////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * In the case of a node targeting another node (i.e. Trigger, Animation, Use),
     * it returns the targeted node ID
     * @return the ID of the targeted node
     */
    public int getTargetID()
    {
        return getHref();
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////// Rendering operations /////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    public void preRender(int sceneTimeInMillis)
    {
        // Check if animation is a direct child and needs to be attached
        if (getHref() == HREF_UNDEFINED)
        {
            m_parent.attachAnimation(this);
            if (xlink.getHref() == null) xlink.setHref(new URI());
            xlink.getHref().id = HREF_DIRECT_PARENT;
        }

        // Check now timestamp
        if (m_begin == BasicRendering.TIME_NOW)
        {
            m_begin = sceneTimeInMillis;
        }

        // Event to post : no event as default
        byte eventToPost = EventListener.INTERNAL_UNDEFINED;

        // Perform automaton transitions by time checks
        switch (m_state)
        {
            // ANIMSTATE_IDLE case : possibly start animation
            case ANIMSTATE_IDLE :
                // Check if begin time has come and end time hasn't come
                if ( (m_begin != BasicRendering.TIME_INDEFINITE) && (sceneTimeInMillis >= m_begin) )
                {
                    // Check end !
                    if ( ( (sceneTimeInMillis <= m_end) && (m_end != BasicRendering.TIME_INDEFINITE) ) ||
                         (m_end == BasicRendering.TIME_INDEFINITE)  )
                    {
                        // START animation
                    	
                        // Init members
                        m_simpleAnimBegin = sceneTimeInMillis;
                        m_simpleAnimCount++;

                        // start
                        genericStartAnimation(sceneTimeInMillis);
                        m_state = ANIMSTATE_ANIMATING;

                        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "starting - ANIMSTATE_ANIMATING");

                        // notify event listeners with begin event
                        eventToPost = EventListener.TIME_BEGIN;
                        
//                        System.out.println("REPEAT " + m_simpleAnimCount + " >> " + m_simpleAnimBegin); // TRACE
                    }
                }
                break;


            // ANIMSTATE_IDLEANDWAITFOREND and ANIMSTATE_FREEZEDAFTERDUR cases : simple animation and its repeats
            case ANIMSTATE_FREEZEDAFTERDUR :
            case ANIMSTATE_IDLEANDWAITFOREND :
                // Check end
                if ( (sceneTimeInMillis > m_end) && (m_end != BasicRendering.TIME_INDEFINITE) )
                {
                    if (m_freeze)
                    {
                        m_state = ANIMSTATE_FREEZEDAFTEREND;
                        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "End - ANIMSTATE_FREEZEDAFTEREND");
                    }
                    else
                    {
                        m_state = ANIMSTATE_IDLEAFTERSTOP;
                        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "End - ANIMSTATE_IDLEAFTERSTOP");
                    }

                    // notify event listeners with end event
                    eventToPost = EventListener.TIME_END;
                }
                break;


            case ANIMSTATE_ANIMATING :
            	//System.out.println("ANIMSTATE_ANIMATING >>" + m_simpleAnimBegin + m_dur); // TRACE 
            	
                if ( (sceneTimeInMillis > m_end) && (m_end != BasicRendering.TIME_INDEFINITE) )
                {
                    // STOP BEFORE DURATION HAS COME, BECAUSE OF END !

                    // Reinit members
                    m_simpleAnimBegin = BasicRendering.TIME_INDEFINITE;

                    // Freeze ?
                    if (m_freeze)
                    {
                        m_state = ANIMSTATE_FREEZEDAFTEREND;
                        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "StopBeforeDur - ANIMSTATE_FREEZEDAFTEREND");
                    }
                    else
                    {
                        m_state = ANIMSTATE_IDLEAFTERSTOP;
                        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "StopBeforeDur - ANIMSTATE_IDLEAFTERSTOP");
                    }

                    // notify event listeners with end event
                    eventToPost = EventListener.TIME_END;
                }
                else
                {
                    // Check if simple duration time has come and if repeat must occure
                    if ( (m_simpleAnimBegin != BasicRendering.TIME_INDEFINITE) &&
                         (m_dur != BasicRendering.TIME_INDEFINITE) &&
                         ( sceneTimeInMillis >= (m_simpleAnimBegin + m_dur) )   )
                    {
                        // This is the end of simple duration
                        m_state = ANIMSTATE_SIMPLEENDNOREPEAT;
                        
                        //System.out.println("ANIMSTATE_SIMPLEENDNOREPEAT >>" + sceneTimeInMillis); // TRACE

                        // Check repeats
                        // Total duration = 1 * m_dur + m_repeatDur
                        // Total duration = 1 * m_dur + m_repeatCount * m_dur
                        // Apply the min between the two repeat mode
                        // If no repeat specified at all (REPEAT_NOTSPECIFIED and REPEAT_NOTSPECIFIED) then do not repeat
                        if ((m_repeatCount != REPEAT_NOTSPECIFIED) || (m_repeatDur != REPEAT_NOTSPECIFIED))
                        {
                            boolean countDoesNotAllowRepeat = false;
                            boolean durDesoNotAllowRepeat = false;
                            if ( ((m_repeatCount != REPEAT_NOTSPECIFIED) && (m_repeatCount != REPEAT_INDEFINITLY)) &&
                                 (m_simpleAnimCount > m_repeatCount) )
                            {
                                countDoesNotAllowRepeat = true;
                            }
                            if ( ((m_repeatDur != REPEAT_NOTSPECIFIED) && (m_repeatDur != REPEAT_INDEFINITLY)) &&
                                 ((m_simpleAnimCount * m_dur) >= (m_dur + m_repeatDur)) )
                            {
                                durDesoNotAllowRepeat = true;
                            }
                            if ( (! countDoesNotAllowRepeat) && (! durDesoNotAllowRepeat) )
                            {
                                // REPEAT !

                                // Init members
                                m_simpleAnimBegin = sceneTimeInMillis;
//                                System.out.println("REPEAT " + m_simpleAnimCount + " >> " + m_simpleAnimBegin); // TRACE
                                m_simpleAnimCount++;

                                m_state = ANIMSTATE_ANIMATING;

                                if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "Repeat - ANIMSTATE_ANIMATING");

                                // notify event listeners with repeat event
                                eventToPost = EventListener.TIME_REPEAT;
                            }
                        }
                    }

                    // Check if repeat duration time has come
                    if ( ( (m_animIdleTime != BasicRendering.TIME_INDEFINITE) && (m_animIdleTime != REPEAT_NOTSPECIFIED) ) &&
                         (sceneTimeInMillis > m_animIdleTime) )
                    {
                        m_state = ANIMSTATE_SIMPLEENDNOREPEAT;
                        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "  ANIMSTATE_SIMPLEENDNOREPEAT");
                    }

                    // Check simple end with no repeat to notify event listeners with end event
                    if (m_state == ANIMSTATE_SIMPLEENDNOREPEAT)
                    {
                        // Reinit members
                        m_simpleAnimBegin = BasicRendering.TIME_INDEFINITE;

                        // Freeze ?
                        if (m_freeze)
                        {
                            m_state = ANIMSTATE_FREEZEDAFTERDUR;
                            if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "EndNoRepeat - ANIMSTATE_FREEZEDAFTERDUR");
                        }
                        else
                        {
                            m_state = ANIMSTATE_IDLEANDWAITFOREND;
                            if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "EndNoRepeat - ANIMSTATE_IDLEANDWAITFOREND");
                        }
                        
                        eventToPost = EventListener.TIME_END;
                    }

                }
                break;
        }

        // Possibly post event
        // CAUTION : this code line must be the last line of this method to avoid re-entrant call
        if (eventToPost != EventListener.INTERNAL_UNDEFINED)
        {
            notifyEvent(eventToPost, new Integer(sceneTimeInMillis));
        }
    }

    /**
     * Rendering operation : do nothing
     * @param sceneTimeInMillis The current scene time
     * @param parentHasChanged Flag indicating wether parent has changed
     * @param parentTransform Parent matrix
     */
    protected void render(int sceneTimeInMillis, boolean parentHasChanged, Matrix2D parentTransform)
    {
    }

    /**
     * Rendering operation : process animation
     * @param sceneTimeInMillis The current scene time
     */
    public void animate(int sceneTimeInMillis)
    {
//    	System.out.println(m_state); // TRACE
        switch (m_state)
        {
            case ANIMSTATE_ANIMATING :
                genericProcessSimpleAnimation(sceneTimeInMillis);
                break;

            case ANIMSTATE_FREEZEDAFTERDUR :
            case ANIMSTATE_FREEZEDAFTEREND :
                genericFreezeAnimation();
                break;
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// Animation control operations ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Starts save current value of the target and starts animation
     * @param sceneTimeInMillis The current scene time
     */
    private final void genericStartAnimation(int sceneTimeInMillis)
    {
        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + "  genericStartAnimation");

        // Get target
        GraphObject targetBaseNode = null;
        RenderedGraphObject targetNode = null;
        if (getHref() != HREF_UNDEFINED)
        {
            // get direct parent
            if (getHref() == HREF_DIRECT_PARENT)
            {
                targetBaseNode = m_parent;
            }
            else
            {
                targetBaseNode = super.m_graphObjectManager.getNode(getHref());
            }
            if ( (targetBaseNode != null) && (targetBaseNode instanceof RenderedGraphObject) )
            {
                targetNode = (RenderedGraphObject)targetBaseNode;
            }
            else
            {
                return;
            }
        }

        // Init members
        m_simpleAnimCount = 0;
        m_animIdleTime = REPEAT_NOTSPECIFIED;
        if ( (m_begin != BasicRendering.TIME_INDEFINITE) &&
             (m_dur != BasicRendering.TIME_INDEFINITE) )
        {
            if (m_repeatDur == BasicRendering.TIME_INDEFINITE)
            {
                m_animIdleTime = BasicRendering.TIME_INDEFINITE;
            }
            else if (m_repeatDur != REPEAT_NOTSPECIFIED)
            {
                 m_animIdleTime = m_begin + m_dur + m_repeatDur;
            }
        }

        // Specific start animation
        startAnimation(targetNode, sceneTimeInMillis);
    }

    /**
     * Process current simple animation (interpolation oparation)
     * @param sceneTimeInMillis
     */
    private final void genericProcessSimpleAnimation(int sceneTimeInMillis)
    {
        // This is the simple animation rendering operation.
        // Interpolation can occures.
        // m_simpleAnimBegin provides the time when this simple animation starts
        // m_simpleAnimCount provides the current count of simple animation (1 for the first)
        
        // Get target
        GraphObject targetBaseNode = null;
        RenderedGraphObject targetNode = null;
        if (getHref() != HREF_UNDEFINED)
        {
            if (getHref() == HREF_DIRECT_PARENT)
            {
                targetBaseNode = m_parent;
            }
            else
            {
                targetBaseNode = super.m_graphObjectManager.getNode(getHref());
            }
            if ( (targetBaseNode != null) && (targetBaseNode instanceof RenderedGraphObject) )
            {
                targetNode = (RenderedGraphObject)targetBaseNode;
            }
            else
            {
                return;
            }
        }

        processSimpleAnimation(targetNode, sceneTimeInMillis);
    }

    /**
     * Freeze animation at last value
     */
    private final void genericFreezeAnimation()
    {
        // Get target
        GraphObject targetBaseNode = null;
        RenderedGraphObject targetNode = null;
        if (getHref() != HREF_UNDEFINED)
        {
            if (getHref() == HREF_DIRECT_PARENT)
            {
                targetBaseNode = m_parent;
            }
            else
            {
                targetBaseNode = super.m_graphObjectManager.getNode(getHref());
            }
            if ( (targetBaseNode != null) && (targetBaseNode instanceof RenderedGraphObject) )
            {
                targetNode = (RenderedGraphObject)targetBaseNode;
            }
            else
            {
                return;
            }
        }

        freezeAnimation(targetNode);
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////// Specific animation control operations //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Starts animation
     * @param targetNode The targetted node
     * @param sceneTimeInMillis The current scene time
     */
    protected void startAnimation(RenderedGraphObject targetNode, int sceneTimeInMillis)
    {
        // Nothing to do
    }

    /**
     * Process current simple animation (interpolation oparation)
     * @param targetNode
     * @param sceneTimeInMillis
     */
    protected void processSimpleAnimation(RenderedGraphObject targetNode, int sceneTimeInMillis)
    {
        // This is the simple animation rendering operation.
        // Interpolation can occures.
        // m_simpleAnimBegin provides the time when this simple animation starts
        // m_simpleAnimCount provides the current count of simple animation (1 for the first)

        setValueToTarget(targetNode, m_to);
    }

    /**
     * Freeze animation at last value
     * @param targetNode
     */
    protected final void freezeAnimation(RenderedGraphObject targetNode)
    {
        if (m_to != null)
        	setValueToTarget(targetNode, m_to);
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// Attribute save/restore and set /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Process animation itself.
     * @param targetNode The target node whose value is animated
     */
    protected void setValueToTarget(RenderedGraphObject targetNode, Value valueToSet)
    {
        // Sanity check
        if (targetNode == null)
        {
            return;
        }

        switch (m_attribute) {
        case Value.ANIM_FILL:
        	targetNode.updateActualFill(valueToSet.m_enumOrColorOrRealOrID);
        	break;
        case Value.ANIM_R:
                ((Circle)targetNode).updateActualRadius(valueToSet.m_enumOrColorOrRealOrID);
            break;
        // Ellipse
        case Value.ANIM_RX:
            if (targetNode instanceof Ellipse)
            {
                ((Ellipse)targetNode).updateActualXRadius(valueToSet.m_enumOrColorOrRealOrID);
            }
            break;
        // Ellipse
        case Value.ANIM_RY :
            if (targetNode instanceof Ellipse)
            {
                ((Ellipse)targetNode).updateActualYRadius(valueToSet.m_enumOrColorOrRealOrID);
//                System.out.println("valueToSet.m_enumOrColorOrRealOrID >> " + valueToSet.m_enumOrColorOrRealOrID); // TRACE
            }
            break;
        case Value.ANIM_DISPLAY :
            targetNode.updateActualDisplay((byte)valueToSet.m_enumOrColorOrRealOrID);
            break;

        case Value.ANIM_VISIBILITY :
            targetNode.updateActualVisibility((byte)valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_FONTSTYLE :
            targetNode.updateActualFontStyle((byte)valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_FONTSIZE :
            targetNode.updateActualFontSize(valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_STROKEWIDTH :
            targetNode.updateActualStrokeWidth(valueToSet.m_enumOrColorOrRealOrID);
            break;

        case Value.ANIM_STROKE :
            targetNode.updateActualStroke(valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_FILLOPACITY :
            targetNode.updateActualFillOpacity((byte)valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_STROKEOPACITY :
            targetNode.updateActualStrokeOpacity((byte)valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_FILLRULE :
            targetNode.updateActualFillRule((byte)valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_FOCUSABLE :
            targetNode.updateActualFocusable((valueToSet.m_enumOrColorOrRealOrID == 1) ? true : false);
            break;
        case Value.ANIM_TEXTALIGN :
            targetNode.updateActualTextAlign((byte)valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_DISPLAYALIGN :
            targetNode.updateActualDisplayAlign((byte)valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_WIDTH :
            targetNode.updateActualWidth(valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_HEIGHT :
            targetNode.updateActualHeight(valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_X :
        case Value.ANIM_CX :
        case Value.ANIM_X1 :
            targetNode.updateActualX(valueToSet.m_enumOrColorOrRealOrID);
            break;
        case Value.ANIM_Y :
        case Value.ANIM_CY :
        case Value.ANIM_Y1 :
            targetNode.updateActualY(valueToSet.m_enumOrColorOrRealOrID);
            break;

        // STRING
        case Value.ANIM_FONTFAMILY :
            targetNode.updateActualFontFamily(valueToSet.m_string);
            break;

        // PATH
        case Value.ANIM_D :
            if (targetNode instanceof Path)
            {
                ((Path)targetNode).updateActualPathDataAndTypes(valueToSet.m_polygon, valueToSet.m_types);
            }
            break;

        // Polygon/Polyline
        case Value.ANIM_POINTS :
            if (targetNode instanceof Polyline)
            {
                ((Polyline)targetNode).updateActualPolylinePoints(valueToSet.m_polygon);
            }
            else if (targetNode instanceof Polygon)
            {
                ((Polygon)targetNode).updateActualPolygonPoints(valueToSet.m_polygon);
            }
            break;

        // VIEWBOX
        case Value.ANIM_VIEWBOX :
            if (targetNode instanceof Svg)
            {
                ((Svg)targetNode).updateActualViewbox(valueToSet.m_floatsOrInts);
            }
            break;

        case Value.ANIM_X2 :
            if (targetNode instanceof Line)
            {
                ((Line)targetNode).updateActualX2(valueToSet.m_enumOrColorOrRealOrID);
            }
            break;
        case Value.ANIM_Y2 :
            if (targetNode instanceof Line)
            {
                ((Line)targetNode).updateActualY2(valueToSet.m_enumOrColorOrRealOrID);
            }
            break;

        // G size
        case Value.ANIM_SIZE :
            if (targetNode instanceof Group)
            {
                ((Group)targetNode).updateActualSize(valueToSet.m_point);
            }
            break;
            
        case NodeProperties.TRANSFORMTRANSLATE :
            // Get translation
            Point2D translation = null;
            if (valueToSet.m_floatsOrInts != null)
            {
                translation = new Point2D(valueToSet.m_floatsOrInts[0], valueToSet.m_floatsOrInts[1]);
/*
                System.out.println(valueToSet.m_floatsOrInts[0]);
                System.out.println(valueToSet.m_floatsOrInts[1]);
                System.out.println("-----------");
*/                
            }
            // Set translation
            if (translation != null)
            {
            	targetNode.updateActualTranslation(translation, false);
            }
            break;

        // TODO other cases
            
        default:
        	System.out.println("WARNING: setValueToTarget undefined for : " + m_attribute);
        }
    }

    /**
     * Provides the current value of an attribute of the target
     * @param targetNode The target node whose value is animated
     */
    protected Value getValueOfTarget(RenderedGraphObject targetNode)
    {
    	// Sanity check
    	if (targetNode == null)
    	{
    		return null;
    	}
    	
    	// Get the right value
    	Value valueToReturn = new Value();
    	switch (m_attribute) {
    	case Value.ANIM_FILL:
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_fill;
    		break;
    		
    		// Circle
    	case Value.ANIM_R:
    		if (targetNode instanceof Circle)
    		{
    			valueToReturn.m_enumOrColorOrRealOrID = ((Circle)targetNode).m_rActualRadius;
    		}
    		break;
    		
    		// Ellipse
    	case Value.ANIM_RX:
    		if (targetNode instanceof Ellipse)
    		{
    			valueToReturn.m_enumOrColorOrRealOrID = ((Ellipse)targetNode).m_rActualXRadius;
    		}
    		break;
    		
    		// Ellipse
    	case Value.ANIM_RY:
    		if (targetNode instanceof Ellipse)
    		{
    			valueToReturn.m_enumOrColorOrRealOrID = ((Ellipse)targetNode).m_rActualYRadius;
    		}
    		break;
    	case Value.ANIM_DISPLAY :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_display;
    		break;
    		
    		// ENUM, INT, F01, COLOR
    	case Value.ANIM_VISIBILITY :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_visibility;
    		break;
    	case Value.ANIM_FONTSTYLE :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_fontStyle;
    		break;
    	case Value.ANIM_FONTSIZE :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_fontSize;
    		break;
    	case Value.ANIM_STROKEWIDTH :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_strokeWidth;
    		break;
    	case Value.ANIM_STROKE :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_stroke;
    		break;
    	case Value.ANIM_FILLOPACITY :
    		valueToReturn.m_enumOrColorOrRealOrID = (targetNode.m_actualProperties.m_fillOpacity & 0xFF);
    		break;
    	case Value.ANIM_STROKEOPACITY :
    		valueToReturn.m_enumOrColorOrRealOrID = (targetNode.m_actualProperties.m_strokeOpacity & 0xFF);
    		break;
    	case Value.ANIM_FILLRULE :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_fillRule;
    		break;
    	case Value.ANIM_FOCUSABLE :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_focusable != 0 ? 1 : 0;
    		break;
    	case Value.ANIM_TEXTALIGN :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_text_anchor;
    		break;
    	case Value.ANIM_DISPLAYALIGN :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.m_actualProperties.m_displayAlign;
    		break;
    	case Value.ANIM_WIDTH :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.getActualWidth();
    		break;
    	case Value.ANIM_HEIGHT :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.getActualHeight();
    		break;
    	case Value.ANIM_X :
    	case Value.ANIM_CX :
    	case Value.ANIM_X1 :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.getActualX();
    		break;
    	case Value.ANIM_Y :
    	case Value.ANIM_CY :
    	case Value.ANIM_Y1 :
    		valueToReturn.m_enumOrColorOrRealOrID = targetNode.getActualY();
    		break;
    		
    		// STRING
    	case Value.ANIM_FONTFAMILY :
    		valueToReturn.m_string = targetNode.m_actualProperties.m_fontFamily;
    		break;
    		
    		// PATH
    	case Value.ANIM_D :
    		if (targetNode instanceof Path)
    		{
    			valueToReturn.m_polygon = ((Path)targetNode).m_polygon;
    			valueToReturn.m_types = ((Path)targetNode).m_types;
    		}
    		break;
    		
    		// Polygon/Polyline
    	case Value.ANIM_POINTS :
    		if (targetNode instanceof Polyline)
    		{
    			valueToReturn.m_polygon = ((Polyline)targetNode).m_polygon;
    		}
    		else if (targetNode instanceof Polygon)
    		{
    			valueToReturn.m_polygon = ((Polygon)targetNode).m_polygon;
    		}
    		break;
    		
    		// VIEWBOX
    	case Value.ANIM_VIEWBOX :
    		if (targetNode instanceof Svg)
    		{
    			valueToReturn.m_floatsOrInts = ((Svg)targetNode).m_rViewbox;
    		}
    		break;
    		
    		// Line end point
    	case Value.ANIM_X2 :
    		if (targetNode instanceof Line)
    		{
    			valueToReturn.m_enumOrColorOrRealOrID = ((Line)targetNode).m_actualX2Y2.x_r;
    		}
    		break;
    	case Value.ANIM_Y2 :
    		if (targetNode instanceof Line)
    		{
    			valueToReturn.m_enumOrColorOrRealOrID = ((Line)targetNode).m_actualX2Y2.y_r;
    		}
    		break;
    		
    		// G size
    	case Value.ANIM_SIZE :
    		if (targetNode instanceof Group)
    		{
    			valueToReturn.m_point = ((Group)targetNode).m_actualSize;
    		}
    		break;
    	default:
    		System.out.println("WARNING: getValueOfTarget undefined for : " + m_attribute);
    	}
    	// TODO other cases
    	return valueToReturn;
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// OnEvent //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Is called when receiving an event
     * @param eventType the event type
     * @param eventData the associated event data
     */
    public void onEvent(byte eventType, Object eventData)
    {
        switch(eventType)
        {
            case EventListener.UI_ACTIVATE:
                setBeginTime(((Integer)eventData).intValue(), 0);
                break;
        }
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////// TimedElement implementation ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Sets the new begin time of the element
     * @param sceneTimeInMillis
     * @param delay the delay for the begin time to set
     */
    public void setBeginTime(int sceneTimeInMillis, int delay)
    {
        // Check restartable mode
        boolean restart = false;
        switch (m_restart)
        {
            case RESTART_ALWAYS :
                //can be restarted at any time.
                restart = true;
                break;
            case RESTART_WHENNOTACTIVE :
                //can only be restarted when it is not active (i.e. after the active end).
                if ( (m_state == ANIMSTATE_IDLE) || (m_state == ANIMSTATE_IDLEAFTERSTOP) ||
                     (m_state == ANIMSTATE_FREEZEDAFTEREND) )
                {
                    restart = true;
                }
                break;
            case RESTART_NEVER :
                //cannot be restarted for the remainder of the current simple duration.
                if ( (m_state == ANIMSTATE_IDLE) || (m_state == ANIMSTATE_IDLEAFTERSTOP) ||
                     (m_state == ANIMSTATE_IDLEANDWAITFOREND) || (m_state == ANIMSTATE_FREEZEDAFTERDUR) ||
                     (m_state == ANIMSTATE_FREEZEDAFTEREND) )
                {
                    restart = true;
                }
                break;
        }
        if (! restart)
        {
            return;
        }

        // Set begin time and reset automaton state
        m_begin = sceneTimeInMillis + delay;
        m_state = ANIMSTATE_IDLE;
        m_simpleAnimBegin = BasicRendering.TIME_INDEFINITE;
        m_simpleAnimCount = 0;
        m_animIdleTime = BasicRendering.TIME_INDEFINITE;
        if (Debug.DEBUG_ANIMATION) System.out.println(this + "  " + sceneTimeInMillis + " - ANIMSTATE_IDLE");

        // Check if it's time to start the anim, don't wait for the next iteration.
        preRender(sceneTimeInMillis);
    }

    /**
     * Sets the new end time of the element
     * @param sceneTimeInMillis
     * @param delay the delay for end time to set
     */
    public void setEndTime(int sceneTimeInMillis, int delay)
    {
        // Set end time
        m_end = sceneTimeInMillis + delay;

        // Check if it's time to end the anim
        preRender(sceneTimeInMillis);
    }

    /**
     * Gets the begin time of an element
     * @return the begin time
     */
    public int getBeginTime()
    {
        if (m_simpleAnimBegin == BasicRendering.TIME_INDEFINITE)
        {
            return m_begin;
        }
        else
        {
            return m_simpleAnimBegin;
        }
    }
    
    public int getHref() {
    	if (xlink.getHref() != null)
    		return xlink.getHref().id;
    	else 
    		return HREF_UNDEFINED;
    }
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////Accessor ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////
    public byte getAnimationState()
    {
    	return this.m_state;
    }

    public void setAnimationState(int NewState)
    {
    	this.m_state = (byte)NewState;
    }
}
