/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Rect2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.SensitiveZone;
import iso.mpeg.streamezzo.player.util.SensitiveZoneText;

import java.util.Vector;

public class TextArea extends TextBase {
    public boolean m_editable = false;
    public boolean m_actualEditable = false;

    public boolean m_isEditMode = false;

    private int m_intWrapWidth;
    private int m_intWrapHeight;

    // x coordinate
    public int m_rActualX;
    private int m_rPrevActualX = 0;
    public int m_rX = 0;

    // y coordinate
    public int m_rActualY;
    private int m_rPrevActualY = 0;
    public int m_rY = 0;
    
    public int m_Width = 0;
    public int m_Height = 0;

    /** The list of text lines **/
    protected Vector m_textLines = null;
    /** The intial text line. At the beginning all strings are set into a single line **/
    protected TextLine m_initTextLine = null;
    /** The list of sensitive zone for the whole text */
    protected Vector m_sensitivesZones = null;

    public TextArea(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);

        // Allocate memory
        m_children = new Vector();
        m_initTextLine = new TextLine();
        m_textLines = new Vector();
        m_sensitivesZones = new Vector();
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////// Properties setters/getters /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    public void setString(String a_string)
    {
        m_wasUpdated = true;
        m_children.removeAllElements();
        m_children.addElement(a_string);
        m_needPreRender = true;
    }

    public void addString(String a_string)
    {
        m_wasUpdated = true;
        m_needPreRender = true;
        m_children.addElement(a_string);
    }

    /**
     * Set DOM X
     * @param rX the new x value
     */
    public void setDOMX(int rX)
    {
        m_rX = rX;
        wasUpdated();
    }

    /**
     * Update actual X
     * @param rNewX the new x value
     */
    public void updateActualX(int rNewX)
    {
        // Check changes
        if (rNewX != m_rPrevActualX)
        {
            m_rActualX = rNewX;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualX = m_rPrevActualX;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual X
     */
    public int getActualX()
    {
        return m_rActualX;
    }

    /**
     * Set DOM Y
     * @param rY the new y value
     */
    public void setDOMY(int rY)
    {
        m_rX = rY;
        wasUpdated();
    }

    /**
     * Update actual Y
     * @param rNewY the new y value
     */
    public void updateActualY(int rNewY)
    {
        // Check changes
        if (rNewY != m_rPrevActualY)
        {
            m_rActualY = rNewY;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualY = m_rPrevActualY;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual CY
     */
    public int getActualY()
    {
        return m_rActualY;
    }

    public void setDOMEditable(boolean isTrue)
    {
        m_editable = isTrue;
    }

    public void addDOMX(int r_x)
    {
        m_rX += r_x;
    }
    public void addDOMY(int r_y)
    {
        m_rY += r_y;
    }

    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Changed ?
        if ( (m_rX != m_rActualX) || (m_rY != m_rActualY) || (m_editable != m_actualEditable))
        {
            // Store previous values
            m_rPrevActualX = m_rActualX;
            m_rPrevActualY = m_rActualY;

            // Set DOM values
            m_rActualX = m_rX;
            m_rActualY = m_rY;
            m_actualEditable = m_editable;

            m_wasUpdated = true;
        }
    }

    /**
     * Register this node as navigatable. Text node is peculiar :
     * if focusable=auto and editable=true then it is navigatable.
     */
    protected void registerAsNavigatable()
    {
        if (m_actualProperties.m_focusable != 0)
        {
            m_graphObjectManager.registerNodeAsNavigatable(this);
        }
        else if ( (m_actualEditable) && (m_properties.isInherited(Value.UPDATE_FOCUSABLE)) )
        {
            // Note : isInherited tells us if this property is set to 'auto' or 'true/false'
            m_graphObjectManager.registerNodeAsNavigatable(this);
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// OnEvent //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Is called when receiving an event
     * @param eventType the event type
     * @param eventData the associated event data
     */
    public void onEvent(byte eventType, Object eventData)
    {
        switch(eventType)
        {
            case EventListener.UI_ACTIVATE:
                m_isEditMode = true;
                break;
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// Rendering operations //////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     *
     * @param sceneTimeInMillis The scene time in ms
     * @param parentHasChanged If parent has been updated
     * @param useContext
     */
    public void specificRender(int sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        //  Was updated ?
        if (parentHasChanged)
        {
            m_wasUpdated = true;
        }

        if (m_isEditMode)
        {
            setFirstString(m_graphObjectManager.m_renderer.textInput("Edit", getFirstString()));
            m_isEditMode = false;
            wasUpdated();
        }

        // Set the font spec
        setFont(m_actualProperties.m_fontFamily, m_actualProperties.m_fontStyle, m_actualProperties.m_fontSize);

        // Build intial text management structure and process inheritance and animation
        // It also lists the involved sensitives zones
        m_initTextLine.clean();
        m_sensitivesZones.setSize(0);
        preprocessText(sceneTimeInMillis, this, useContext);

        // Format if needed
        if (m_needPreRender)
        {
            formatText();
        }

        // Render
        if (m_children.size() > 0)
        {
            drawText();
        }

        // Check sensitive zones. Because of unconventional text rendering there is an
        // unconventional listeners management.
        for (int i = 0; i < m_sensitivesZones.size(); i++)
        {
            SensitiveZoneText zone = (SensitiveZoneText)m_sensitivesZones.elementAt(i);
            Vector savedListeners = m_graphObjectManager.swapPickingListerners(zone.getListeners());
            zone.checkZone(m_graphObjectManager);
            m_graphObjectManager.swapPickingListerners(savedListeners);
        }

        m_wasUpdated = false;
    }

    private void drawText()
    {
        // Declare local var
        int usedX = 0; // top-left position for drawString operation
        int usedY = 0; // top-left position for drawString operation
        int upLimit = 0; // top limit of the bounding rect
        int downLimit = 0;  // down limit of the bounding rect
        int boundingX = 0; // left limit of the bounding rect

        // Get position and apply matrix
        Point2D topleft =  m_actualTransform.apply(new Point2D(m_rActualX, m_rActualY));
        int posX = m_graphObjectManager.m_renderer.normalizeX(topleft.x_r);
        int posY = m_graphObjectManager.m_renderer.normalizeY(topleft.y_r);

        // Get number of lines
        int nbLines = m_textLines.size();

        // Compute the text total height
        int totalTextHeight = 0;
        for (int i = 0; i < nbLines; i++)
        {
            totalTextHeight += ((TextLine) m_textLines.elementAt(i)).m_lineHeight;
        }

        // displayRect height forced with total height
        // displayRect width has been forced at prerender time
        m_intWrapHeight = totalTextHeight;

        // Compute limits and start point of drawing
        if (m_actualProperties.m_displayAlign == NodeProperties.DISPLAYALIGN_AFTER)
        {
            upLimit = posY;
            downLimit = posY + m_intWrapHeight;
            usedY = posY;
        }
        else if (m_actualProperties.m_displayAlign == NodeProperties.DISPLAYALIGN_BEFORE)
        {
            upLimit = posY - m_intWrapHeight;
            downLimit = posY;
            usedY = posY - totalTextHeight;
        }
        else // m_actualProperties.m_displayAlign == CENTER
        {
            upLimit = posY - m_intWrapHeight / 2;
            downLimit = posY + m_intWrapHeight / 2;
            usedY = posY - totalTextHeight / 2;
        }

        // Compute the top-left corner of the bounding box
        if (m_actualProperties.m_text_anchor == NodeProperties.TEXTALIGN_START)
        {
            boundingX = posX;
        }
        else if (m_actualProperties.m_text_anchor == NodeProperties.TEXTALIGN_END)
        {
            boundingX = posX - m_intWrapWidth;
        }
        else // m_actualProperties.m_text_anchor == CENTER
        {
            boundingX = posX - m_intWrapWidth / 2;
        }

        // Set the clipping for text node
        Point2D previousOrigin = null, previousDimension = null;
        if (m_intWrapWidth != -1)
        {
            // save previous clip rect (intersection of all parent clip rects)
            previousOrigin = m_graphObjectManager.m_renderer.getClipRectOrigin();
            previousDimension = m_graphObjectManager.m_renderer.getClipRectDimension();
            m_graphObjectManager.m_renderer.setClipRect(new Point2D(boundingX, upLimit),
                new Point2D(m_intWrapWidth, m_intWrapHeight));
        }

        // Draw all lines
        for (int ii = 0; ii < nbLines; ii++)
        {
            // Get a line
            TextLine line = (TextLine) m_textLines.elementAt(ii);

            // Compute X
            if (m_actualProperties.m_text_anchor == NodeProperties.TEXTALIGN_START)
            {
                usedX = posX;
            }
            else if (m_actualProperties.m_text_anchor == NodeProperties.TEXTALIGN_END)
            {
                usedX = posX - line.m_lineWidth;
            }
            else // m_actualProperties.m_text_anchor == NodeProperties.ALIGN_CENTER
            {
                usedX = posX - line.m_lineWidth / 2;
            }

            // Check if line can be drawn and draw it
            if (((usedY) <= (downLimit)) && ((usedY + line.m_lineHeight) >= upLimit))
            {
                line.drawLine(usedX, usedY);
            }

            // Compute next Y
            usedY = usedY + line.m_lineHeight;
        }

        // Reset the clipping to previous
        if (m_intWrapWidth != -1)
        {
            // set previous clip rect (intersection of all parent clip rects)
            m_graphObjectManager.m_renderer.setClip(previousOrigin.x_r, previousOrigin.y_r,
                previousDimension.x_r, previousDimension.y_r);
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////// Pre-Rendering and formatting operations ////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * This operation builds the text structure as a TextLine object to keep a link between strings and their properties.
     * Inheritance and animation are processed here.
     * A string is always owned by a Text object or a tspan object.
     * All strings are set into the same line.
     * @param sceneTimeInMillis
     * @param current The current node in whitch to find strings
     * @param useContext The current use context
     */
    protected void preprocessText(int sceneTimeInMillis, TextBase current, GraphObject useContext)
    {
        // check diplay
        if (current.m_actualProperties.m_display == NodeProperties.DISPLAY_NONE)
        {
            return;
        }

        // Configure and store the sensitive zone
        if (null == useContext)
        {
            current.m_senstiveZone.setNullZone();
            current.m_senstiveZone.setListeners(m_graphObjectManager.getPickingListeners());
            m_sensitivesZones.addElement(current.m_senstiveZone);
        }

        // Get the content
        Vector content = current.m_children;

        // Iterate through content
        for (int i = 0; i < content.size(); i++)
        {
            // Get child
            Object child = content.elementAt(i);

            // Check string as direct child
            if (child instanceof TextContent)
            {
                // Get actual properties of the string
                NodeProperties properties = current.m_actualProperties.getCopy();

                // Add element
                if (null == useContext)
                {
                    m_initTextLine.addString(((TextContent) child).getString(), properties, current.m_senstiveZone);
                }
                else
                {
                    m_initTextLine.addString(((TextContent) child).getString(), properties, null);
                }
            }
            // Check tspan child
            else if (child instanceof Tspan)
            {
                // Render to process inheritance and animation
                ((Tspan)child).render(sceneTimeInMillis, current.m_wasUpdated, current.m_actualTransform, current.m_actualProperties, useContext);

                // Push again the listeners because of unconventional text rendering
                int pushedCount = ((Tspan)child).pushPickingListeners();

                preprocessText(sceneTimeInMillis, (TextBase) child, useContext);

                // Pop the pused listeners
                m_graphObjectManager.popPickingListeners(pushedCount);
            }
            // Check use
            else if (child instanceof Use)
            {
                Use useAsChild = (Use)child;

                // Render to process inheritance and animation
                useAsChild.render(sceneTimeInMillis, current.m_wasUpdated, current.m_actualTransform, current.m_actualProperties, useContext);

                // Get use target
                GraphObject useTarget = ((Use)child).getTarget();

                if ((useTarget != null) && (useTarget instanceof Tspan) )
                {
                    // Render to process inheritance and animation
                    ((Tspan)useTarget).render(sceneTimeInMillis, useAsChild.m_wasUpdated, useAsChild.m_actualTransform, useAsChild.m_actualProperties, useContext);

                    preprocessText(sceneTimeInMillis, (TextBase) useTarget, (Use)child);
                }
            }
        }
    }

    /**
     * Is responsible for formatting text
     * Inheritance and animations are processed here
     */
    protected void formatText()
    {
        // Clean text lines structure
        m_textLines.removeAllElements();

        // Bounding rect not specified
        m_intWrapWidth = -1;
        m_intWrapHeight = -1;

        // Check end of lines chars
        m_intWrapWidth = formatWithEndOfLines();

        // Pre-render done
        m_needPreRender = false;
    }

    /**
     * Split the text into several lines according to '\n' chars.
     * Internal member handling text format into several lines is filled.
     * @return The width of the longest line
     */
    private int formatWithEndOfLines()
    {
        // Local declaration
        int maxWidth = 0;                   // the max width
        String segmentString = null;        // a segment string
        NodeProperties segmentOwner = null; // a segemnt string properties
        SensitiveZone sensitiveZone = null; // a segment string sensitive zone
        int segmentLength = 0;              // count of char into a segment string
        int lineWidth = 0;                  // width in pixel of a line
        int currPos = 0;                    // current position into a string
        int startPos = 0;                   // beginning of a substring
        String tempString = null;           // a temproray var for string

        Vector TMPtextLines = new Vector(); // textline for check the lenght with the area width
        
        // Create one line and add it to the text lines structure
        TextLine currentTextLine = new TextLine();
//      m_textLines.addElement(currentTextLine);
        TMPtextLines.addElement(currentTextLine);

        // Iterate through the text segments
        for (int seg = 0; (seg < m_initTextLine.m_strings.size() && seg < m_initTextLine.m_stringProperties.size()); seg++)
        {
        	// Get a segment string and its oproperties
            segmentString = (String) m_initTextLine.m_strings.elementAt(seg);
            segmentOwner = (NodeProperties) m_initTextLine.m_stringProperties.elementAt(seg);
            sensitiveZone = (SensitiveZone) m_initTextLine.m_stringSensitiveZones.elementAt(seg);
            
            // Find '\n' chars and split when needed
            segmentLength = segmentString.length();
            currPos = 0;
            startPos = 0;
            while (currPos < segmentLength)
            {
                // Check '\n'
                if (segmentString.charAt(currPos) == '\n')
                {
                    // Extract substring
                    tempString = segmentString.substring(startPos, currPos);

                    // Add substring to the current text line
                    currentTextLine.addString(tempString, segmentOwner, sensitiveZone);

                    // Get line width and compute the max width
                    lineWidth = currentTextLine.m_lineWidth;
                    if (lineWidth > maxWidth)
                    {
                        maxWidth = lineWidth;
                    }

                    // Create a new line
                    currentTextLine = new TextLine();
                    TMPtextLines.addElement(currentTextLine);

                    // Next char
                    startPos = currPos + 1;
                    currPos = startPos;
                }
                else
                {
                    // Next char
                    currPos++;
                }
            }

            // add the last line
            if (startPos < segmentLength)
            {
                // Extract substring
                tempString = segmentString.substring(startPos, currPos);

                // Add substring to the current text line
                currentTextLine.addString(tempString, segmentOwner, sensitiveZone);

                // Get line width and compute the max width
                lineWidth = currentTextLine.m_lineWidth;
                if (lineWidth > maxWidth)
                {
                    maxWidth = lineWidth;
                }
            }
        }
       
        for(int seg = 0 ; seg < TMPtextLines.size() ; seg++)
        {
        	if(((TextLine)(TMPtextLines.elementAt(seg))).m_strings.size() != 0)
        	{
        		int leninpxl = ((TextLine)(TMPtextLines.elementAt(seg))).m_lineWidth;
        		
        		// Get a segment string and its oproperties
	            segmentString = (String) ((TextLine)(TMPtextLines.elementAt(seg))).m_strings.elementAt(0);
	            segmentOwner = (NodeProperties) ((TextLine)(TMPtextLines.elementAt(seg))).m_stringProperties.elementAt(0);
	            sensitiveZone = (SensitiveZone) ((TextLine)(TMPtextLines.elementAt(seg))).m_stringSensitiveZones.elementAt(0);

	            TextLine tl = new TextLine();
	            TextLine tl2 = new TextLine();
	            m_textLines.addElement(tl2);
	            
        		if(leninpxl > m_Width/4096)
        		{
        			boolean bContinue = true;
        			// search the substring in width
        			segmentLength = segmentString.length();
        			currPos = 0;
    	            startPos = 0;
    	            do
    	            {
    	            	tl.addString(""+segmentString.charAt(currPos), segmentOwner, sensitiveZone);
    	            	if (tl.m_lineWidth > m_Width/4096)
    	                {
    	                    // Extract substring
    	                    tempString = segmentString.substring(startPos, currPos);
    	                    tl2.addString(tempString, segmentOwner, sensitiveZone);
    	                    
    	                    tl2 = new TextLine();
    	                    m_textLines.addElement(tl2);
    	                    
    	                    // reinit
    	                    tl.clean();
    	                    
    	                    segmentString = segmentString.substring(currPos);
    	                    segmentLength = segmentString.length();
    	                    currPos = 0;
    	    	            startPos = 0;
    	    	            
    	    	            tl.addString(segmentString, segmentOwner, sensitiveZone);
    	    	            
    	    	            if(tl.m_lineWidth < m_Width/4096)
    	    	            {
    	    	            	m_textLines.addElement(tl);
    	    	            	bContinue = false;
    	    	            }
    	    	            tl.clean();
    	                }
    	                else
    	                {
    	                    // Next char
    	                    currPos++;
    	                    if(currPos >= segmentLength)
    	                    	bContinue = false;
    	                }
    	            }while (bContinue);
    	            
    	            tl2.addString(segmentString, segmentOwner, sensitiveZone);
        			tl2 = new TextLine();
                    m_textLines.addElement(tl2);
        		}
        		else
        		{
        			tl2.addString(segmentString, segmentOwner, sensitiveZone);
        			tl2 = new TextLine();
                    m_textLines.addElement(tl2);
        		}
        	}
        }
        // return max width
        return m_Width/4096;
    }

    /**
     * Get the first child which is a string
     * @return String
     */
    public String getFirstString()
    {
        // get the first string
        for(int i=0;i<m_children.size();i++)
        {
            if (m_children.elementAt(i) instanceof String)
            {
                return (String)m_children.elementAt(i);
            }
        }

        return "";
    }

    /**
     * Set the first child which is a string with a new value
     * @param string
     */
    private void setFirstString(String string)
    {
        // set the first string
        for(int i=0;i<m_children.size();i++)
        {
            if (m_children.elementAt(i) instanceof String)
            {
                m_children.removeElementAt(i);
                m_children.insertElementAt(string, i);
            }
        }
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////// Inner class for text lines //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    private class TextLine
    {
        /** The list of string in a line **/
        public Vector m_strings = null;
        /** The list of string properties. There is a link between m_strings and m_stringProperties **/
        public Vector m_stringProperties = null;
        /** The list of sensitive zones. There is a link between m_strings and m_stringSensitiveZones **/
        public Vector m_stringSensitiveZones = null;
        /** The width in pixels of the line **/
        public int m_lineWidth = 0;
        /** The height in pixels of the line **/
        public int m_lineHeight = 0;
        /** The distance between top and baseline of the line **/
        public int m_lineBaselineHeight = 0;

        /**
         * Constructor of instances of TextLine class
         */
        public TextLine()
        {
            // Allocate vectors
            m_strings = new Vector();
            m_stringProperties = new Vector();
            m_stringSensitiveZones = new Vector();

            // Init members
            m_lineWidth = 0;
            m_lineHeight = 0;
            m_lineBaselineHeight = 0;
        }

        /**
         * Reset a text line
         */
        public void clean()
        {
            m_strings.removeAllElements();
            m_stringProperties.removeAllElements();
            m_stringSensitiveZones.removeAllElements();
            m_lineWidth = 0;
            m_lineHeight = 0;
            m_lineBaselineHeight = 0;
        }

        /**
         * Add a string to the end of the line.
         * It updates the line total width and the line height.
         *
         * @param stringToAdd The string to add at the end of the line
         * @param properties The properties of the string to add
         * @param sensitiveZone The sensitive zone of this string
         */
        public void addString(String stringToAdd, NodeProperties properties, SensitiveZone sensitiveZone)
        {
            // Check string
            if (stringToAdd.length() == 0)
            {
                return;
            }

            // Add in vectors
            m_strings.addElement(stringToAdd);
            m_stringProperties.addElement(properties);
            m_stringSensitiveZones.addElement(sensitiveZone);

            // Update line width, height and baseline
            Object font = getFont(properties.m_fontFamily, properties.m_fontStyle, properties.m_fontSize);
            int width = m_graphObjectManager.m_renderer.stringWidth(font, stringToAdd);
            m_lineWidth += width;
            int heigth = m_graphObjectManager.m_renderer.getFontHeight(font);
            if (heigth > m_lineHeight) m_lineHeight = heigth;
            int baseline = m_graphObjectManager.m_renderer.getFontBaseline(font);
            if (baseline > m_lineBaselineHeight) m_lineBaselineHeight = baseline;
        }

        /**
         * Draw the line at the specified position
         * @param x Top-Left corner where to start drawing the line
         * @param y Top-Left corner where to start drawing the line
         */
        public void drawLine(int x, int y)
        {
            // Sanity check
            if (m_strings.size() != m_stringProperties.size())
            {
                return;
            }

            // Init position to draw strings
            int posX = x;
            int posY = y + m_lineBaselineHeight; // because anchor if at baseline-left

            // Iterate through line segments
            String lineSegmentString = null;
            NodeProperties lineSegmentProperties = null;
            SensitiveZoneText lineSegmentZone = null;
            Rect2D bbox = new Rect2D(-1, -1, -1, -1);
            for (int i = 0; i < m_strings.size(); i++)
            {
                // Get a string and its properties
                lineSegmentString = (String) m_strings.elementAt(i);
                lineSegmentProperties = (NodeProperties) m_stringProperties.elementAt(i);
                lineSegmentZone = (SensitiveZoneText)m_stringSensitiveZones.elementAt(i);

                // Check visibility and draw string
                if (lineSegmentProperties.m_visibility != NodeProperties.VISIBILITY_HIDDEN)
                {
                    // Set font and color
                    Object font = getFont(lineSegmentProperties.m_fontFamily, lineSegmentProperties.m_fontStyle, lineSegmentProperties.m_fontSize);
                    m_graphObjectManager.m_renderer.setFontForGraphics(font);
                    m_graphObjectManager.m_renderer.setColor(lineSegmentProperties.m_fill, lineSegmentProperties.m_fillOpacity&0xFF);

                    // Draw the string with baseline-left anchor in order to have a convenient
                    // display when several font are used on the same line
                    m_graphObjectManager.m_renderer.drawStringAtBaselineLeft(lineSegmentString, posX, posY, bbox);
                    if (Debug.DEBUG_BBOX) System.out.println(this + " : bbox= (" + bbox.x_r + "," + bbox.y_r + "," + bbox.width_r + "," + bbox.height_r + ")");

                    // Provide the bounding box to the sensitive zone
                    if (null != lineSegmentZone)
                    {
                        lineSegmentZone.addZone(bbox);
                        if (Debug.DEBUG_BBOX) System.out.println(lineSegmentZone + " : bbox= (" + lineSegmentZone.m_rect.x_r + "," + lineSegmentZone.m_rect.y_r + "," + lineSegmentZone.m_rect.width_r + "," + lineSegmentZone.m_rect.height_r + ")");
                    }
                }

                // Get string segment width and update x position
                Object font = getFont(lineSegmentProperties.m_fontFamily, lineSegmentProperties.m_fontStyle, lineSegmentProperties.m_fontSize);
                posX += m_graphObjectManager.m_renderer.stringWidth(font, lineSegmentString);
            }
        }
    }
}