/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.node;


import iso.mpeg.streamezzo.player.BasicRendering;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.SensitiveZoneText;

import java.util.Vector;

import javax.microedition.lcdui.Font;

public abstract class TextBase extends RenderedGraphObject
{
    protected Object m_font = null;


    private byte m_previousFontFamily = Font.FACE_SYSTEM;
    private byte m_previousFontStyle = Font.STYLE_PLAIN;
    private short m_previousFontSize = Font.SIZE_MEDIUM;

    /** the sensitive zone */
    public SensitiveZoneText m_senstiveZone = new SensitiveZoneText();


    public TextBase(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
    }

    /**
     * Return a convenient font regard to font specification of the current insatnce
     * @return The convenient font
     */
    protected Object getFont()
    {
        byte fontFamily = Font.FACE_MONOSPACE;
        if ("MONOSPACE".equals(m_actualProperties.m_fontFamily)) fontFamily = Font.FACE_MONOSPACE;
        else if ("PROPORTIONAL".equals(m_actualProperties.m_fontFamily)) fontFamily = Font.FACE_PROPORTIONAL;
        else if ("SYSTEM".equals(m_actualProperties.m_fontFamily)) fontFamily = Font.FACE_SYSTEM;

        short fontSize = Font.SIZE_SMALL;
        if (m_actualProperties.m_fontSize > BasicRendering.limitBetween12And18) fontSize = Font.SIZE_LARGE;
        else if (m_actualProperties.m_fontSize > BasicRendering.limitBetween9And12) fontSize = Font.SIZE_MEDIUM;
        else if (m_actualProperties.m_fontSize > -1) fontSize = Font.SIZE_SMALL;

        byte style = m_actualProperties.m_fontStyle;
        if (style < 0 || style > 8) style = 0;

        return m_graphObjectManager.m_renderer.getFont(fontFamily, BasicRendering.fontStyles[style], fontSize);
    }

    /**
     * Return a convenient font regard to font specification
     * @param family Font family
     * @param style Font style
     * @param size Font size (20.12 fixed point)
     * @return The convenient font
     */
    protected Object getFont(String family, byte style, int size)
    {
        byte fontFamily = Font.FACE_MONOSPACE;
        if ("MONOSPACE".equals(family)) fontFamily = Font.FACE_MONOSPACE;
        else if ("PROPORTIONAL".equals(family)) fontFamily = Font.FACE_PROPORTIONAL;
        else if ("SYSTEM".equals(family)) fontFamily = Font.FACE_SYSTEM;

        short fontSize = Font.SIZE_SMALL;
        if (size > BasicRendering.limitBetween12And18) fontSize = Font.SIZE_LARGE;
        else if (size > BasicRendering.limitBetween9And12) fontSize = Font.SIZE_MEDIUM;
        else if (size > -1) fontSize = Font.SIZE_SMALL;

        byte fontstyle = style;
        if (fontstyle < 0 || fontstyle > 8) fontstyle = 0;

        return m_graphObjectManager.m_renderer.getFont(fontFamily, BasicRendering.fontStyles[fontstyle], fontSize);
    }

    /**
     *
     * @param family Font family
     * @param style Font style
     * @param size Font size (20.12 fixed point)
     */
    public void setFont(String family, byte style, int size)
    {
        byte u = m_previousFontFamily;
        if ("MONOSPACE".equals(family))
            u = Font.FACE_MONOSPACE;
        else if ("PROPORTIONAL".equals(family))
            u = Font.FACE_PROPORTIONAL;
        else if ("SYSTEM".equals(family)) u = Font.FACE_SYSTEM;
        short u1 = m_previousFontSize;
        if (size > BasicRendering.limitBetween12And18)
            u1 = Font.SIZE_LARGE;
        else if (size > BasicRendering.limitBetween9And12)
            u1 = Font.SIZE_MEDIUM;
        else if (size > -1) u1 = Font.SIZE_SMALL;
        if (style < 0 || style > 8) style = m_previousFontStyle;
        m_font = m_graphObjectManager.m_renderer.getFont(u, BasicRendering.fontStyles[style], u1);
        if (u != m_previousFontFamily || u1 != m_previousFontSize || style != m_previousFontStyle)
        {
            m_previousFontFamily = u;
            m_previousFontSize = u1;
            m_previousFontStyle = style;

            // Pre-rendering needed
            wasUpdated();
        }
    }

    /**
     * RenderedGraphObject overriding
     * Notify that some properties influencing rendering have changed.
     * This oprations set flags to ask for a pre-render.
     */
    public void wasUpdated()
    {
        m_wasUpdated = true;
        m_needPreRender = true;

        // Nofity parent if this is a textbase object
        if ( (m_parent != null) && (m_parent instanceof TextBase) )
        {
            ((TextBase)m_parent).wasUpdated();
            if(Debug.DEBUG_TRACE) System.out.println("textHasChanged : m_parent = " + m_parent );
        }
    }

    /**
     * Replace a text content
     * @param textContent new text content
     */
    public void setDOMTextContent(Vector textContent)
    {
        // Replace first find text content by the new one
        // and deletes other text content.
        // Id est new text content will replace all older text content by other child nodes are kept

        // Sanity check
        if ( m_children == null)
        {
            return;
        }

        wasUpdated();

        // Iterate through children to remove all text contents
        Object child = null;
        boolean newTextContentInserted = false;
        for (int childIndex = m_children.size() - 1; childIndex >= 0; childIndex--)
        {
            // Get a child
            child = m_children.elementAt(childIndex);

            // Unregister this node
            if (child instanceof GraphObject)
            {
                m_graphObjectManager.unRegisterNode(((GraphObject) child));
            }

            // Remove if it is text content
            if (child instanceof TextBase)
            {
                ((TextBase)child).removeAllChildren();
                m_children.removeElementAt(childIndex);
            }
            else if (child instanceof String)
            {
                m_children.removeElementAt(childIndex);
            }
        }

        // Set new text content
        Object textContentUnit = null;
        for (int childIndex = 0; childIndex < textContent.size(); childIndex++)
        {
            textContentUnit = textContent.elementAt(childIndex);
            m_children.addElement(textContentUnit);
            if (textContentUnit instanceof GraphObject)
            {
                ((GraphObject)textContentUnit).setParent(this);
            }
        }
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////// Children management operations ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * This operation add a new child to this node.
     *
     * @param child The sibling of the current node
     */
    public void appendChild(GraphObject child)
    {
        // Sanity check
        if ( (child != null) && (m_children != null) )
        {
            m_children.addElement(child);
            child.setParent(this);
            wasUpdated();
        }
    }

    /**
     * Remove the child identified by its position.
     * @param childPos The position of the child to remove
     */
    public void removeChild(int childPos)
    {
        // Sanity check
        if ( m_children == null)
        {
            return;
        }

        if (m_children.size() > childPos)
        {
            // Get the child and remove it
            GraphObject child = (GraphObject) m_children.elementAt(childPos);
            // GraphObject child = getChildGraphObjectAt(childPos);
            m_children.removeElementAt(childPos);

            if (child != null)
            {
                m_children.removeElement(child);

                if (child instanceof GraphObject)
                {
                    // Unregister this node
                    m_graphObjectManager.unRegisterNode(child);

                    // Remove all children of this child
                    child.removeAllChildren();
                }
            }

            wasUpdated();
        }
    }

    /**
     * Remove the passed-in child
     * @param child The child object to remove
     */
    public void removeChild(GraphObject child)
    {
        // Sanity check
        if ( m_children == null)
        {
            return;
        }

        if (m_children.contains(child))
        {
            // Remove it
            m_children.removeElement(child);

            // Unregister this node
            m_graphObjectManager.unRegisterNode(child);

            // Remove all children of this child
            child.removeAllChildren();

            wasUpdated();
        }
    }

    /**
     * Remove all children
     */
    public void removeAllChildren()
    {
        // Sanity check
        if ( m_children == null)
        {
            return;
        }

        wasUpdated();

        // Iterate through children
        Object child = null;
        for (int childIndex = 0; childIndex < m_children.size(); childIndex++)
        {
            // Get a child
            child = m_children.elementAt(childIndex);

            // Unregister this node
            if (child instanceof GraphObject)
            {
                m_graphObjectManager.unRegisterNode(((GraphObject) child));
            }

            // Bitmap is a special case
            if (child instanceof Bitmap)
            {
                ((Bitmap) child).m_cacheImage = null;
            }
            else if (child instanceof GraphObject)
            {
                ((GraphObject) child).removeAllChildren();
            }
        }

        //clean the vector by removing all elements
        m_children.removeAllElements();
    }

    /**
     * The children count
     * @return the number of children
     */
    public int getChildrenCount()
    {
        // Count and ignore String
        int count = 0;
        int nbChildren = m_children.size();
        Object child = null;
        for (int i = 0; i < nbChildren; i++)
        {
            if ( m_children.elementAt(i) instanceof GraphObject )
            {
                count++;
            }
        }
        return count;
    }

    /**
     * Return the child at the requested index
     * @param index the index of the child
     * @return the child at the requested index
     */
    public Object getChild(int index)
    {
        return m_children.elementAt(index);
        //return getChildGraphObjectAt(index);
    }


////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// Rendering operations ////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     *
     * @param sceneTimeInMillis The scene time in ms
     * @param parentHasChanged If parent has been updated
     * @param useContext
     */
    public void specificRender(int sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Nothing to do because its parent do the rendering
    }
}