/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.node;

import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;
import iso.mpeg.streamezzo.player.util.NodeProperties;

public class Use extends RenderedGraphObject
{
    //newly added attributes
	//ID of the target node
	public Xlink xlink = new Xlink();
		    
    // x,y coordinate
    private Point2D m_rActualXY = new Point2D(0, 0);
    private Point2D m_rPrevActualXY = new Point2D(0, 0);
    public Point2D m_rXY = new Point2D(0, 0);

    // width and height
    private Point2D m_rActualWH = new Point2D(0, 0);
    private Point2D m_rPrevActualWH = new Point2D(0, 0);
    public Point2D m_rWH = new Point2D(0, 0);

    public Matrix2D m_xyMatrix = Matrix2D.createIdentityMatrix();
    private Matrix2D m_newActualTransform = Matrix2D.createIdentityMatrix();

    public Use(GraphObjectManager graphObjectManager)
    {
        super(graphObjectManager);
    }

    /**
     * Update the generated matrix with x and y
     */
    private void updateXYMatrix()
    {
        m_xyMatrix.setTranslation(m_rActualXY);
        m_wasUpdated = false;
    }

    /**
     * Set DOM X
     * @param rX the new x value
     */
    public void setDOMX(int rX)
    {
        m_rXY.x_r = rX;
        wasUpdated();
    }

    /**
     * Update actual X
     * @param rNewX the new x value
     */
    public void updateActualX(int rNewX)
    {
        // Check changes
        if (rNewX != m_rPrevActualXY.x_r)
        {
            m_rActualXY.x_r = rNewX;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualXY.x_r = m_rPrevActualXY.x_r;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual X
     */
    public int getActualX()
    {
        return m_rActualXY.x_r;
    }

    /**
     * Set DOM CY
     * @param rY the new y value
     */
    public void setDOMY(int rY)
    {
        m_rXY.x_r = rY;
        wasUpdated();
    }

    /**
     * Update actual Y
     * @param rNewY the new y value
     */
    public void updateActualY(int rNewY)
    {
        // Check changes
        if (rNewY != m_rPrevActualXY.y_r)
        {
            m_rActualXY.y_r = rNewY;
            m_wasUpdated = true;
        }
        else
        {
            m_rActualXY.y_r = m_rPrevActualXY.y_r;
            m_wasUpdated = false;
        }
    }

    /**
     * Get the actual Y
     */
    public int getActualY()
    {
        return m_rActualXY.y_r;
    }

    public void setDOMHRef(int href)
    {
        xlink.getHref().id = href;
    }

    public void addDOMX(int r_x)
    {
        m_rXY.x_r += r_x;
    }
    public void addDOMY(int r_y)
    {
        m_rXY.y_r += r_y;
    }

    /**
     * This operation is responsible to reset presentation values with the DOM Values
     * Doesn't apply on base properties (stroke, fill, opacity, ...)
     */
    public void resetDOMValues()
    {
        // Changed ?
        if (m_rXY.x_r != m_rActualXY.x_r || m_rXY.y_r != m_rActualXY.y_r)
        {
            // Store previous values
            m_rPrevActualXY.x_r = m_rActualXY.x_r;
            m_rPrevActualXY.y_r = m_rActualXY.y_r;

            // Set DOM values
            m_rActualXY.x_r = m_rXY.x_r;
            m_rActualXY.y_r = m_rXY.y_r;

            m_wasUpdated = true;
        }
    }


    public void specificRender(int a_sceneTimeInMillis, boolean parentHasChanged, GraphObject useContext)
    {
        // Check visibility
        if ( (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_HIDDEN) ||
             (m_actualProperties.m_visibility == NodeProperties.VISIBILITY_COLLAPSE) )
        {
            return;
        }

        // get the referenced node
        Object impl = m_graphObjectManager.getNode(xlink.getHref().id);

        if ((impl != null) && (impl instanceof RenderedGraphObject))
        {
            if (m_wasUpdated)
            {
                updateXYMatrix();
                Matrix2D.Mul(m_xyMatrix, m_actualTransform, m_newActualTransform);
            }
            // render children with 'use' matrix
            ((RenderedGraphObject)impl).render(a_sceneTimeInMillis, parentHasChanged, m_newActualTransform, m_actualProperties, this);
        }
    }

    public GraphObject getTarget()
    {
        return m_graphObjectManager.getNode(xlink.getHref().id);
    }


}
