/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.tool;

import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.node.CursorManager;
import iso.mpeg.streamezzo.player.node.GraphObject;
import iso.mpeg.streamezzo.player.node.Listener;
import iso.mpeg.streamezzo.player.node.RenderedGraphObject;
import iso.mpeg.streamezzo.player.node.SVGset;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.Navigator;
import iso.mpeg.streamezzo.player.util.SensitiveZone;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class GraphObjectManager {
    public Rendering m_renderer = null;
    public SceneManager m_sceneManager = null;
    //Cached Objects
    //Single Objects (eg., images)
    protected Hashtable m_singleSceneObjects = new Hashtable();
    protected Hashtable m_singlePermanentObjects = new Hashtable();

    /** Node IDs table : key is node ID and value is a GraphObjetc */
    public Hashtable m_ids = new Hashtable();

    /** Nodes to post-render */
    protected Vector m_nodeToPostRender = new Vector();

    /** Nodes waiting for registration */
    protected Hashtable m_waitingNodes = new Hashtable();

    // list of timed nodes
    protected Vector m_timedNodes = new Vector();

    // Navigator
    private Navigator m_navigator = null;


    public  void setRenderer(Rendering a_renderer)
    {
        m_renderer = a_renderer;
        m_sceneManager = (SceneManager)a_renderer;
    }

    /**
     *  Destroy the scene by deleting all the references (i.e., DefNames). To
     * release the memory of an object, one must call deleteNode(String)
     * If no DefName is set, then the object cannot be retreived but by using
     * its parent. The memory is not released until the root node references
     * existed.
     *
     * @see     #deleteNode
     *
     */
    /*public  final void deleteScene()
    {
        m_ids.clear();
        // clear the waiting node list
        m_waitingNodes.clear();
        //Clear the local single object
        m_singleSceneObjects.clear();
        // clean the animations list
        m_timedNodes.removeAllElements();
    }*/

    /**
     * Reset the table of IDs and waiting nodes but not the table of streams
     */
    public void cleanScene()
    {
        // clear the table of IDs
        m_ids.clear();
        // clear the waiting node list
        m_waitingNodes.clear();
        // clean the animations list
        m_timedNodes.removeAllElements();
    }

    /**
     * Delete a node by using its DefName and by removing it from the children
     * list of its parent.
     * It does nothing when it is applied on the root node of a scene.
     * @param   id the DefName of this object.
     *
     * @see     iso.mpeg.streamezzo.player.node.Group#removeChild
     *
     */
    public final void deleteNode(int id)
    {
        if(Debug.DEBUG_TRACE) System.out.println("deleteNode : id="+id);
        GraphObject the_node = getNode(id);

        if (the_node == null)
        {
            if(Debug.DEBUG_ERROR)
            {
                System.out.println("Error : the node that need to delete doesn't exist : id="+id);
            }
            return;
        }

        GraphObject the_parent = the_node.getParent();

        // Remove this node
        if (the_parent != null)
        {
            the_parent.removeChild(the_node);
        }
    }



    public  void setSingleObject(int streamID, Object data, boolean isPermanent)
    {
        if (!isPermanent)
        {
            m_singleSceneObjects.put(new Integer(streamID),data);
        }
        else
        {
            m_singlePermanentObjects.put(new Integer(streamID),data);
        }
    }

    public Object getSingleObject(int streamID)
    {
        Object data = m_singleSceneObjects.get(new Integer(streamID));
        if (data == null)
        {
            data = m_singlePermanentObjects.get(new Integer(streamID));
        }
        return data;
    }

    public void removeSingleObject(int streamID)
    {
        if(m_singleSceneObjects.containsKey(new Integer(streamID)))
        {
            m_singleSceneObjects.remove(new Integer(streamID));
        }
        else if(m_singlePermanentObjects.containsKey(new Integer(streamID)))
        {
            m_singlePermanentObjects.remove(new Integer(streamID));
        }
    }

    public final GraphObject getNode(int id)
    {
        GraphObject a_object = (GraphObject) m_ids.get(new Integer(id));
        if(Debug.DEBUG_ERROR){
            if(a_object == null)
            {
                System.out.println("The needed object with id="+id+" doesn't exist");
            }
        }
        return a_object;
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////// Post rendering ///////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Adds a node in the list of nodes to post-render.
     * @param nodeToPostRender
     */
    public void registerForPostRender(GraphObject nodeToPostRender)
    {
        m_nodeToPostRender.addElement(nodeToPostRender);
    }

    /**
     * Empties the list of nodes to post-render.
     */
    public void cleanPostRenderList()
    {
        m_nodeToPostRender.removeAllElements();
    }

    /**
     * Processes post rendering operation
     */
    public void ProcessPostRender()
    {
        for (int i = 0; i < m_nodeToPostRender.size(); i++)
        {
            GraphObject node = ((GraphObject)m_nodeToPostRender.elementAt(i));
            if (node != null)
            {
                node.postRender();
            }
        }
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// Node registration ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Registers a node in the table of IDs
     * @param id the ID
     * @param node the node
     */
    public void registerNode(int id, GraphObject node)
    {
        if (Debug.DEBUG_GRAPHMANAGER) System.out.println("registerNode id = "+id);

        // check if an ID is specified
        if (id != -1)
        {
            // add the node to the table of IDs
            m_ids.put(new Integer(id), node);
        }

        // if the node is time dependant (i.e. has begin/end property),
        // add it to the timed nodes list
        if (node.isTimeDependent())
        {
            m_timedNodes.addElement(node);
        }

        // if we don't have an ID specified, we finish processing now because
        // we are sure that no node are waiting to be registered
        if (id == -1) return;

        // check if any nodes are waiting for being registered as EventListener or Animation to this node
        Vector waitingList = getWaitingNodesAndDelete(id);
        if (waitingList != null)
        {
            GraphObject waitingNode = null;

            // Go through the listeners and register them to the node
            for (Enumeration e = waitingList.elements() ; e.hasMoreElements() ;)
            {
                waitingNode = (GraphObject)e.nextElement();
                if (waitingNode instanceof SVGset)
                {
                    // attach an animation
                    node.attachAnimation((SVGset)waitingNode);
                }
                else if (waitingNode instanceof EventListener)
                {
                    // attach a trigger
                    node.attachEventListener((EventListener)waitingNode);
                }
            }
        }
    }

    /**
     * Remove a node from the table of registered nodes
     * @param node the node to unregister
     */
    public void unRegisterNode(GraphObject node)
    {
        // remove from timed nodes list
        if (node.isTimeDependent())
        {
            m_timedNodes.removeElement(node);
        }

        // unregister cursor
        if (node instanceof CursorManager)
        {
            if (node == m_cursor)
            {
                m_cursor = null;
                if (Debug.DEBUG_CURSOR) System.out.println("unregister cursor : activeCursor = null");
            }
        }

        // unregister trigger listener
        if (node instanceof Listener)
        {
            GraphObject targetNode = (GraphObject)m_ids.get(new Integer(node.getTargetID()));
            if (targetNode != null)
            {
                // remove trigger as a node listener
                targetNode.detachEventListener((EventListener)node);
                // remove trigger as a key listener 
                m_navigator.detachEventListener((EventListener)node);
            }
        }

        // unregister animation
        if (node instanceof SVGset)
        {
            GraphObject targetNode = null;
            // Animation is a direct parent i.e. href is not a node ID
            if (node.getTargetID() == SVGset.HREF_DIRECT_PARENT)
            {
                targetNode = node.getParent();
            }
            // Animation has a defined href
            else if (node.getTargetID() != SVGset.HREF_UNDEFINED)
            {
                targetNode = (GraphObject)m_ids.get(new Integer(node.getTargetID()));
            }
            if (targetNode != null)
            {
                targetNode.detachAnimation((SVGset)node);
            }
        }

        // remove from IDs table
        if (node.getID() != -1)
        {
            if (Debug.DEBUG_GRAPHMANAGER) System.out.println("unregisterNode id = "+node.getID());
            m_ids.remove(new Integer(node.getID()));
        }
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// EventListener/Animation Management ///////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Attach a listener to the target node
     * @param eventListener the EventListener to attach
     * @param targetId the id of the node to listen
     */
    public void attachListener(EventListener eventListener, int targetId)
    {
        GraphObject targetNode = getNode(targetId);
        // We've found the node
        if (targetNode != null)
        {
            targetNode.attachEventListener(eventListener);
        }
        // The node is not here yet, add it to the waiting table
        else
        {
            addToWaitingNodes(eventListener, targetId);
        }
    }

    /**
     * Attach a listener to the target node
     * @param eventListener the EventListener to attach
     * @param targetNode the node to listen
     */
    public void attachListener(EventListener eventListener, GraphObject targetNode)
    {
        if (targetNode != null)
        {
            targetNode.attachEventListener(eventListener);
        }
    }

    /**
     * add a waiting node to the table
     * @param node
     */
    protected void addToWaitingNodes(Object node, int id)
    {
        Vector vector = null;
        Integer key = new Integer(id);
        // get the existing list or create a new one
        if (m_waitingNodes.containsKey(key))
        {
            vector = (Vector)m_waitingNodes.get(key);
        }
        else
        {
            vector = new Vector(1);
            m_waitingNodes.put(key, vector);
        }
        // add the waiting node to the list
        vector.addElement(node);
    }

    /**
     * Get the list of nodes waiting to be attached as EventListener to a node and
     * remove the list from the table
     * @param id the id of the node
     * @return the list of waiting nodes
     */
    protected Vector getWaitingNodesAndDelete(int id)
    {
        Integer key = new Integer(id);

        if (m_waitingNodes.containsKey(key))
        {
            // get the list
            Vector vector = (Vector)m_waitingNodes.get(key);
            // remove the list from the table
            m_waitingNodes.remove(key);

            return vector;
        }
        else
        {
            return null;
        }
    }

    /**
     * Attach an animation to the target node
     * @param anim the animation to attach
     * @param targetId the id of the targeted node
     */
    public void attachAnimation(SVGset anim, int targetId)
    {
        GraphObject targetNode = getNode(targetId);
        // We've found the node
        if (targetNode != null)
        {
            targetNode.attachAnimation(anim);
        }
        // The node is not here yet, add it to the waiting table
        else
        {
            addToWaitingNodes(anim, targetId);
        }
    }


    /**
     * Resolve begin/end linking activations
     * @param sceneTime the current scene time
     */
    public void resolveTimedNodes(int sceneTime)
    {
        GraphObject timedNode = null;

        // Go through the listeners list and send the new event
        for (Enumeration e = m_timedNodes.elements() ; e.hasMoreElements() ;)
        {
            timedNode = (GraphObject)e.nextElement();
            timedNode.preRender(sceneTime);
        }
    }

    /**
     * Set the navigator reference
     * @param navigator the Navigator
     */
    public void setNavigator(Navigator navigator)
    {
        m_navigator = navigator;
    }

    /**
     * Register the passed-in node as navigatable
     * @param node
     */
    public void registerNodeAsNavigatable(RenderedGraphObject node)
    {
        if (m_navigator != null)
        {
            m_navigator.registerNodeAsNavigatable(node);
        }
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// Cursor Management //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////
    /** The current active cursor */
    private CursorManager m_cursor = null;

    /** A flag tellling whether the cursor is disabled. Usefull to avoid checking sensitive zone for the cursor shape itself */
    private boolean m_bCursorIsDisabled = false;

    /** The stack of current picking listeners */
    private Vector m_pickingListeners = new Vector();


    /**
     * Adds a scenegraph element at the tail of the current picking listeners stack
     * @param listener
     */
    public void pushPickingListener(GraphObject listener)
    {
        if (Debug.DEBUG_CURSOR_FULL) System.out.println("CURSOR : count=" + m_pickingListeners.size() + "   -   push listener " +  listener);
        m_pickingListeners.addElement(listener);
    }

    /**
     * Removes the last elements at the tail of the current picking listeners stack
     * @param count number of element to pop
     */
    public void popPickingListeners(int count)
    {
        int size = m_pickingListeners.size();
        if (size > 0)
        {
            if (Debug.DEBUG_CURSOR_FULL) System.out.println("CURSOR : count=" + m_pickingListeners.size() + "   -   pop " + count + " listeners " +  m_pickingListeners.lastElement());
            m_pickingListeners.setSize(size - count);
        }
    }

    /**
     * Returns a reference to the current picking listener list
     * @return a reference to the current picking listener list
     */
    public Vector getPickingListeners()
    {
        return m_pickingListeners;
    }

    /**
     * Swaps the listeners list : sets the new list and returns the old
     * @param newListeners the new listener list to set
     * @return the previous listener list
     */
    public Vector swapPickingListerners(Vector newListeners)
    {
        Vector old = m_pickingListeners;
        m_pickingListeners = newListeners;
        return old;
    }

    /**
     * Register a cursor element as active cursor
     * @param cursor the cursor element to use
     */
    public void setActiveCursor(CursorManager cursor)
    {
        m_cursor = cursor;
    }

    /**
     * Returns a reference to the element that draws the current active cursor
     * @return a reference to the element that draws the current active cursor
     */
    public GraphObject getActiveCursorView()
    {
        if ( (! m_bCursorIsDisabled) && (null != m_cursor) )
        {
            if (-1 != m_cursor.getTargetID())
            {
                return getNode(m_cursor.getTargetID());
            }
        }
        return null;
    }

    /**
     * Sets whether the cursor disabled or not. The cursor must be disabled when composing the scenegraph part that
     * represents the cursor itsef otherwise the shape that reprersents the cursor will detect that the cursor
     * is inside them
     * @param bDisable
     */
    public void disableCursor(boolean bDisable)
    {
        m_bCursorIsDisabled = bDisable;
    }

    /**
     * Tells whether it's worth checking a sensitive zone. If no listeners exists
     * sensitive zone is not sensitive :) !
     * @return
     */
    public boolean isItWorthToCheckSensitiveZone()
    {
        // Check if cursor is disabled
        if (m_bCursorIsDisabled)
        {
            return false;
        }

        // Check if there is picking listeners
        if (m_pickingListeners.size() == 0)
        {
            return false;
        }

        return true;
    }

    /**
     * Returns the mouse position.
     * @return the normalized mouse position or null if there is no mouse
     */
    public Point2D getMousePosition()
    {
        if ( (null != m_cursor) && (! m_bCursorIsDisabled) )
        {
            return new Point2D(m_cursor.getPosition().x_r, m_cursor.getPosition().y_r);
        }
        return null;
    }

    /**
     * Tells whether a mouse click occures
     * @return
     */
    public boolean isMouseClick()
    {
        if (m_bCursorIsDisabled)
        {
            return false;
        }
        return m_cursor.m_bActivated;
    }

    /**
     * Called when the mouse enter whithin the sensitive zone of the element that is currently being composed
     * @param zone the zone within the mouse is entering
     */
    public void onMouseEnter(SensitiveZone zone)
    {
        if ( (! m_bCursorIsDisabled) && (null != m_cursor) )
        {
            m_cursor.setCursorEvent(EventListener.MOUSE_OVER, zone, m_pickingListeners);
        }
    }

    /**
     * Called when the mouse exit from whithin the sensitive zone of the element that is currently being composed
     * @param zone the zone that the mouse is leaving
     */
    public void onMouseExit(SensitiveZone zone)
    {
        if ( (! m_bCursorIsDisabled) && (null != m_cursor) )
        {
            m_cursor.setCursorEvent(EventListener.MOUSE_OUT, zone, m_pickingListeners);
        }
    }

    /**
     * Called when the mouse is moving whithin the sensitive zone of the element that is currently being composed
     * @param zone the zone within which the mouse is moving
     */
    public void onMouseMoveOver(SensitiveZone zone)
    {
        if ( (! m_bCursorIsDisabled) && (null != m_cursor) )
        {
            m_cursor.setCursorEvent(EventListener.MOUSE_MOVE, zone, m_pickingListeners);
        }
    }

    /**
     * Called when a mouse click occures whithin the sensitive zone of the element that is currently being composed
     * @param zone the zone whithin which the mouse is clicked
     */
    public void onMouseClick(SensitiveZone zone)
    {
        if ( (! m_bCursorIsDisabled) && (null != m_cursor) )
        {
            m_cursor.setCursorEvent(EventListener.MOUSE_CLICK, zone, m_pickingListeners);
        }
    }
}
