/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cédric Gégout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.tool;


import iso.mpeg.streamezzo.player.BasicRendering;
import iso.mpeg.streamezzo.player.library.Matrix2D;
import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Polygon2D;
import iso.mpeg.streamezzo.player.library.Real;
import iso.mpeg.streamezzo.player.node.A;
import iso.mpeg.streamezzo.player.node.Audio;
import iso.mpeg.streamezzo.player.node.Bitmap;
import iso.mpeg.streamezzo.player.node.Circle;
import iso.mpeg.streamezzo.player.node.Conditional;
import iso.mpeg.streamezzo.player.node.CursorManager;
import iso.mpeg.streamezzo.player.node.Ellipse;
import iso.mpeg.streamezzo.player.node.Event;
import iso.mpeg.streamezzo.player.node.GraphObject;
import iso.mpeg.streamezzo.player.node.Group;
import iso.mpeg.streamezzo.player.node.LaserUpdate;
import iso.mpeg.streamezzo.player.node.Line;
import iso.mpeg.streamezzo.player.node.LinearGradient;
import iso.mpeg.streamezzo.player.node.Listener;
import iso.mpeg.streamezzo.player.node.Mpath;
import iso.mpeg.streamezzo.player.node.Multi;
import iso.mpeg.streamezzo.player.node.Path;
import iso.mpeg.streamezzo.player.node.Polygon;
import iso.mpeg.streamezzo.player.node.Polyline;
import iso.mpeg.streamezzo.player.node.RadialGradient;
import iso.mpeg.streamezzo.player.node.Rect;
import iso.mpeg.streamezzo.player.node.RenderedGraphObject;
import iso.mpeg.streamezzo.player.node.SVGanimate;
import iso.mpeg.streamezzo.player.node.SVGanimateColor;
import iso.mpeg.streamezzo.player.node.SVGanimateMotion;
import iso.mpeg.streamezzo.player.node.SVGanimateTransform;
import iso.mpeg.streamezzo.player.node.SVGset;
import iso.mpeg.streamezzo.player.node.Stop;
import iso.mpeg.streamezzo.player.node.Svg;
import iso.mpeg.streamezzo.player.node.Switch;
import iso.mpeg.streamezzo.player.node.Text;
import iso.mpeg.streamezzo.player.node.TextArea;
import iso.mpeg.streamezzo.player.node.TextContent;
import iso.mpeg.streamezzo.player.node.Tspan;
import iso.mpeg.streamezzo.player.node.URI;
import iso.mpeg.streamezzo.player.node.Use;
import iso.mpeg.streamezzo.player.node.Value;
import iso.mpeg.streamezzo.player.node.Video;
import iso.mpeg.streamezzo.player.node.Xlink;
import iso.mpeg.streamezzo.player.util.NameSpace;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.QName;

import java.io.IOException;
import java.util.Vector;


public class LASeRBinParser extends SceneLoader implements IConstants {

    private static int translateNeg = 4096;
    private static int translateNegLimit = 2047;


    public LASeRBinParser(GraphObjectManager graphObjectManager) {
        super(graphObjectManager);
    }

    private Polyline m_lastPolyline = null;
    private Polygon m_lastPolygon = null;
    private Path m_lastPath = null;
    private Text m_lastText = null;
    private Group m_lastGroup = null;
    private Use m_lastUse = null;
    private Line m_lastLine = null;
    private Rect m_lastRect = null;
    
    // TML : parseCodecInitialisations variables
    private int colorIndex = -1;
    private int fontIndex = -1;
    private int privateDataIdentifierIndex = -1;
    private int tagIndex = -1;
    
    private final static byte MOVE = (byte) 0;
    private final static byte LINE = (byte) 1;
    private final static byte CLOSE = (byte) 2;
    private final static byte CUBIC_BEZIER = (byte) 3;
    private final static byte S_CUBIC_BEZIER = (byte) 4;
    private final static byte QUAD_BEZIER = (byte) 5;
    private final static byte S_QUAD_BEZIER = (byte) 6;
    // C{0} H{1} L{2} M{3} Q{4} S{5} T{6} V{7} Z{8} c{9} h{10} l{11} m{12} q{13} s{14} t{15} v{16} z{17}
    private final static byte segTypeToPathType[] = {
            CUBIC_BEZIER, LINE, LINE, MOVE, QUAD_BEZIER, S_CUBIC_BEZIER,
            S_QUAD_BEZIER, LINE, CLOSE, CUBIC_BEZIER, LINE, LINE,
            MOVE, QUAD_BEZIER, S_CUBIC_BEZIER, S_QUAD_BEZIER, LINE, CLOSE};
    

///////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// Nodes parsing operations /////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////
    
    /**
     * Decoding of the node-level objects
     *
     * @return the decoded node
     * @throws java.lang.Exception
     */
    public GraphObject readElements(GraphObject parent) throws Exception {
        // get node type
        int i = readBits(6);
        // switch to the right decoding operation
       
        GraphObject node = readElementsA2Rect(parent, i);
        if(node != null)
        	return node;
        
        node = readElementsSameXXX(parent, i);
        if(node != null)
        	return node;
        
        node = readElementsScript2Listener(parent, i);
        if(node != null)
        	return node;
        
        node = readElementsConditional2SimpleLayout(parent, i);
        if(node != null)
        	return node;
        
        switch (i) {
            case 54:
                return readTextContent();
            default:
                readExtension();
        }
        
        return null;
    }
    
    public GraphObject readElementsTwo(GraphObject parent) throws Exception 
    {
    	int skip87 = readVlcInt(5);
    	int occ = readVlcInt(5);

    	for (int i = 0 ; i < occ ; i++) 
    	{
    		int skip2 = readBits(2);
	    	int type = readBits(4);
	    	
	        switch (type) {
				case 0:
						readAnimation(parent);
					break;
				case 1:
						readDiscard(parent);
					break;
				case 2:
						readFont(parent);
					break;
				case 3:
						readFontFace(parent);
					break;
				case 4:
						readFontFaceSrc(parent);
					break;
				case 5:
						readFontFaceUri(parent);
					break;
				case 6:
						readGlyph(parent);
					break;
				case 7:
						readHandler(parent);
					break;
				case 8:
						readHkern(parent);
					break;
				case 9:
						readMissingGlyph(parent);
					break;
				case 10:
						readPrefetch(parent);
					break;
				case 11:
						readSolidColor(parent);
					break;
				case 12:
				  		readTBreak(parent);
				  	break;
				case 13:
						return readTextArea(parent);
					//break;
				case 14:
						readAnimateScroll(parent);
					break;
				case 15:
						readSetScroll(parent);
					break;
				case 16:
						readStreamSource(parent);
					break;
	        }
	    }
		return null;
	}
    
    public GraphObject readUpdatableElements(GraphObject parent) throws Exception 
    {
        if(!readBit()) // choice case 0
        {
	    	// get node type
	        int i = readBits(6);
	        // switch to the right decoding operation
	        
	        GraphObject node = readElementsA2Rect(parent, i);
	        if(node != null)
	        	return node;
	        	        
	        node = readElementsScript2Listener(parent, i+14);
	        if(node != null)
	        	return node;
        }
        else
        {
        	int i = readBits(3);
        	GraphObject node = readElementsConditional2SimpleLayout(parent, i + 47);
	        if(node != null)
	        	return node;
        }
        return null;
    }
    
    private GraphObject readElementsA2Rect(GraphObject parent, int i) throws Exception 
    {
        // switch to the right decoding operation
        switch (i) {
            case 0:
                return readA();
            case 1:
                return readAnimate(parent);
            case 2:
                return readAnimateColor(parent);
            case 3:
                return readAnimateMotion(parent);
            case 4:
                return readAnimateTransform(parent);
            case 5:
                return readAudio();
            case 6:
                return readCircle();
            case 7:
                return readMulti("defs");
            case 8:
                return readMulti("desc");
            case 9:
                return readEllipse();
            case 10:
                return readMulti("foreignObject");
            case 11:
                m_lastGroup = readGroup();
                return m_lastGroup;
            case 12:
                return readImage();
            case 13:
                return readLine();
            case 14:
                return readLinearGradient();
            case 15:
                return readMulti("metadata");
            case 16:
                return readMpath();
            case 17:
                return readPath();
            case 18:
                return readPolygon();
            case 19:
                return readPolygon();// TML : replace >> readPolyline();
            case 20:
                return readRadialGradient();
            case 21:
                return readRect();
        }
        return null;
    }
    
    private GraphObject readElementsSameXXX(GraphObject parent, int i) throws Exception 
    {
        // switch to the right decoding operation
        switch (i) {
	        case 22:
	            return readSameGroup(m_lastGroup);
	        case 23:
	            return readSameLine(m_lastLine);
	        case 24:
	            return readSamePath(m_lastPath, false);
	        case 25:
	            return readSamePath(m_lastPath, true);
	        case 26:
	            return readSamePolygon(m_lastPolygon, false, false);
	        case 27:
	            return readSamePolygon(m_lastPolygon, true, false);
	        case 28:
	            return readSamePolygon(m_lastPolygon, false, true);
	        case 29:
	            return readSamePolyline(m_lastPolyline, false, false);
	        case 30:
	            return readSamePolyline(m_lastPolyline, true, false);
	        case 31:
	            return readSamePolyline(m_lastPolyline, false, true);
	        case 32:
	            return readSameRect(m_lastRect, false);
	        case 33:
	            return readSameRect(m_lastRect, true);
	        case 34:
	            return readSameText(m_lastText, false);
	        case 35:
	            return readSameText(m_lastText, true);
	        case 36:
	            return readSameUse(m_lastUse);
        }
        return null;
    }
    
    private GraphObject readElementsScript2Listener(GraphObject parent, int i) throws Exception 
    {
        // switch to the right decoding operation
        switch (i) {
	        case 37:
	            return readScript();
	        case 38:
	            return readSet(parent);
	        case 39:
	            return readStop();
	        case 40:
	            return readSwitch();
	        case 41:
	            return readText();
	        case 42:
	            return readMulti("title");
	        case 43:
	            return readTspan();
	        case 44:
	            return readUse();
	        case 45:
	            return readVideo();
	        case 46:
	            return readListener();
        }
        return null;
    }
    
    private GraphObject readElementsConditional2SimpleLayout(GraphObject parent, int i) throws Exception 
    {
        // switch to the right decoding operation
        switch (i) {
	        case 47:
	            return readConditional();
	        case 48:
	            return readCursorManager();
	        case 49:
	            return readUpdateAny(parent, false);
	        case 50:
	            return readPrivateElementContainer(parent); // TML ; TODO verify parent parameters
	        case 51:
	            return readRectClip();
	        case 52:
	            return readSelector();
	        case 53:
	            return readSimpleLayout();
        }
        return null;
    }

    public Svg readSvg() throws Exception 
    {    	
    	Svg au = new Svg(m_graphObjectManager);
        au.m_sceneLoader = this;
        
        readObjectStart(au);
               
        if (readBit())
		{
			readAttr_custom_paint();
		}
		if (readBit())
		{
			readAttr_custom_paint();
		}
		
		if (readBit())
			au.baseProfile = readAttr_custom_byteAlignedString();
		
		if (readBit())
			au.contentScriptType = readAttr_custom_byteAlignedString();
		
		au.m_properties.m_externalResourcesRequired = readBit();
        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
		
		int v[] = read_codec_valueWithUnits();
        au.m_rSceneHeight = v[0];
        au.m_hUnitsIsPercents = v[1] == 6;
		
		if (readBit())
			au.playbackOrder = readBits(1);
		
		if (readBit())
			readAttr_preserveAspectRatio(au);// AspectRatio : cf. plus bas pour d�tailler
		
		if (readBit())
			au.snapshotTime = readAttr_time();
		
		if (readBit()) {
			au.syncBehaviorDefault = readBits(2);
		}
		
		if (readBit()) {
			au.syncToleranceDefault = readAttr_syncTolerance();
		}
		
		if (readBit()) 
			au.timelineBegin = readBits(1);
		
		if (readBit()) 
			au.version = readAttr_custom_byteAlignedString();
		
		if (readBit()) 
		{
			au.m_rViewbox[0] = readAttr_custom_fixed_16_8();
			au.m_rViewbox[1] = readAttr_custom_fixed_16_8();
			au.m_rViewbox[2] = readAttr_custom_fixed_16_8();
			au.m_rViewbox[3] = readAttr_custom_fixed_16_8();
		}
		
		v = read_codec_valueWithUnits();
        au.m_rSceneWidth = v[0];
        au.m_wUnitsIsPercents = v[1] == 6;
		
		if(readBit())
			au.m_zoomAndPan = readBit();

        // Configure the scene manager for viewport/viewbox mapping
        setSceneSize(au.m_rSceneWidth, au.m_wUnitsIsPercents, au.m_rSceneHeight, au.m_hUnitsIsPercents);
        setViewbox(au.m_rViewbox);
        setAspectRatio(au.m_preserveAspectRatioX, au.m_preserveAspectRatioY, au.m_preserveAspectRatioScale);
        setZoomAndPan(au.m_zoomAndPan);
        setBackGroundColor(au.m_backgroundFill);
        setBackGroundOpacity(au.m_backgroundFillOpacity);

//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
    	
    	
       /*


        // Viewbox and aspectRatio
        if (readBit()) {
            au.m_rViewbox[0] = readAttr_custom_fixed_16_8();
            au.m_rViewbox[1] = readAttr_custom_fixed_16_8();
            au.m_rViewbox[2] = readAttr_custom_fixed_16_8();
            au.m_rViewbox[3] = readAttr_custom_fixed_16_8();
            au.m_preserveAspectRatioX = readBits(2);
            au.m_preserveAspectRatioY = readBits(2);
            au.m_preserveAspectRatioScale = readBits(1);
        }


        // background 
        if (readBit()) {
            au.m_backgroundFill = readAttr_custom_paint();
        }
        if (readBit()) {
            au.m_backgroundFillOpacity = (byte) read_codec_0to1float();
        }
        if (readBit()) {
            if (readBit()) {
                au.syncBehaviorDefault = readBits(2);
            }
            if (readBit()) {
                au.syncToleranceDefault = read_syncTolerance();
            }
        }
        if (readBit()) au.playbackOrder = readBits(1);
        if (readBit()) au.timelineBegin = readBits(1);
        if (readBit()) au.snapshotTime = readAttr_time();
        if (readBit()) au.version = read_codec_byteAlignedString();
        if (readBit()) au.baseProfile = read_codec_byteAlignedString();
        if (readBit()) au.contentScriptType = read_codec_byteAlignedString();
        setSceneSize(au.m_rSceneWidth, au.m_wUnitsIsPercents, au.m_rSceneHeight, au.m_hUnitsIsPercents);
        setViewbox(au.m_rViewbox);
        setAspectRatio(au.m_preserveAspectRatioX, au.m_preserveAspectRatioY, au.m_preserveAspectRatioScale);
        setZoomAndPan(au.m_zoomAndPan);
        setBackGroundColor(au.m_backgroundFill);
        setBackGroundOpacity(au.m_backgroundFillOpacity);
        System.out.println("DONE");
        //TODO deal with unused attributes and with the new ones
        /*

        // Begin time
        au.m_bufferingTime = 0;
        if (readBit())
        {
            au.m_bufferingTime = readInt();
        }

        // read the bitmap hint
        boolean untransformedBitmapsHint = readBit();
        setUntransformedBitmapsHint(untransformedBitmapsHint);

        // Provide param to scene loader
        setBufferingTime(au.m_bufferingTime);
        */
        return au;
    }

    // TML
    public void readObjectStart(GraphObject g) throws Exception {
        if (readBit())
        	readAttr_custom_ID(g);
        
//   	 TML : to register the animation node if is not
        if(g.getID() == -1)
        	g.setAutoID();
       
        if (readBit())
        	readAttr_custom_rare(g);
    }
    
    // TML
    private void readFillAndStroke (RenderedGraphObject au) throws IOException {
        // TML : has_fill
        if (readBit()) {
        	au.m_properties.m_fill = readAttr_custom_paint();
        	au.m_properties.setInheritance(Value.UPDATE_COLOR_FILL, false);
        }
        
        // TML : has_stroke
        if (readBit()) {
        	au.m_properties.m_stroke = readAttr_custom_paint();
        	au.m_properties.setInheritance(Value.UPDATE_STROKE, false);
        }    	
    }

    //  TML
    public void readObjectEnd(GraphObject g) throws Exception {
    	        
        // begin : object content
        if (readBit()) {
          	readPrivateAttributeContainer(g);
        }
        readChildren(g);       
        // end : object content
    }

    // TML : TODO : verify
    public A readA() throws Exception 
    {
        A au = new A(m_graphObjectManager);
        au.m_properties.m_focusable = (byte) 1; // default behavior exception
        readObjectStart(au);
        
        readFillAndStroke(au);
        
        au.m_properties.m_externalResourcesRequired = readBit();
        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
        
        if (readBit())
            au.target = readAttr_custom_byteAlignedString();
        
        readHref(au.xlink);
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        	  
        return au;
    }

    public SVGanimate readAnimate(GraphObject parent) throws Exception 
    {
    	SVGanimate au = new SVGanimate(m_graphObjectManager);
    	
    	readObjectStart(au);
    	 
    	if(readBit())
    	{
    		int idx = readAttr_AttributeName();
    		au.m_attribute = (byte) Value.animName[idx];
    		au.m_attributeType = Value.convertTypeFromAnimValue(Value.animType[au.m_attribute]); // TML : confusion m_attributeType par raport spec ???
    	}
    	
    	readAnimateType(au);
    	
    	readSetType(parent, au, false);

    	readBit(); // enabled

//    	 has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);

        return au;
    }
    
    public SVGanimateColor readAnimateColor(GraphObject parent) throws Exception {
        
        SVGanimateColor au = new SVGanimateColor(m_graphObjectManager);
        
        readObjectStart(au);
        
//      TML : to register the animation node if is not
        if(au.getID() == -1)
        	au.setAutoID();
   	 
        if(readBit())
    	{
    		int idx = readAttr_AttributeName();
    		au.m_attribute = (byte) Value.animName[idx];
    		au.m_attributeType = Value.convertTypeFromAnimValue(Value.animType[au.m_attribute]);
    	}
    	
    	readAnimateType(au);
    	
    	readSetType(parent, au, false);

    	readBit(); // enabled
    	
    	if (readBit())
        	readAttr_any();

    	readObjectEnd(au);

        return au;
    }

    // TODO SVGanimateMotion is not (fully) implemented
    public SVGanimateMotion readAnimateMotion(GraphObject parent) throws Exception {

    	SVGanimateMotion au = new SVGanimateMotion(m_graphObjectManager);
    	
        readObjectStart(au);
    	
    	readAnimateType(au);
    	
    	readSetType(parent, au, true);

    	readBit(); // enabled

//    	 has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);

        return au;
    }
    
    public void readAnimateScroll(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	//begin
    	if(readBit())
    		readAttr_smil_times(au, false);
    	
    	//has_by
    	if(readBit())
    	{    	
        	int x = readAttr_custom_coordinate();//readAttr_custom_fixed_16_8();
        	int y = readAttr_custom_coordinate();//readAttr_custom_fixed_16_8();
        	
        	// readAttr_scrollStopType();
       	}
    	
    	//delayAtEnd
    	if(readBit())
    		readAttr_time();
    	
    	//has_delayAtStart
    	if(readBit())
    		readAttr_time();
    	
    	//direction
    	if(readBit())
    		readAttr_directionType();
    		
    	//dur
    	if(readBit())
    		readAttr_time();
    	
    	//has fill
    	if(readBit())
    		readAttr_animFill();
    	
    	//from
    	if(readBit())
    		readAttr_scrollStopType();
    	
    	//
    	if(readBit())
    		readAttr_repeatCount();
    	
    	if(readBit())
    		readAttr_repeatDur();

    	if(readBit())
    		readAttr_restart();
    	
    	//speed
    	if(readBit())
    		readAttr_time();
    	
    	//has_to
    	if(readBit())
    		readAttr_scrollStopType();
    	
    	Xlink xlinkTMP = new Xlink();
    	readHref(xlinkTMP);
    	
    	// attr_any
    	if(readBit())
    		readAttr_any();
    	
    	readObjectEnd(au);
    }
 
    // TML
    public SVGanimateTransform readAnimateTransform(GraphObject parent) throws Exception {
        
        SVGanimateTransform au = new SVGanimateTransform(m_graphObjectManager);
        
        readObjectStart(au);
      	 
        if(readBit())
    	{
    		int idx = readAttr_AttributeName();
    		au.m_attribute = (byte) Value.animName[idx];
    		au.m_attributeType = Value.convertTypeFromAnimValue(Value.animType[au.m_attribute]);
    	}
    	
        // for animateTransform, the type is folded into the attribute index, to make processing simpler
        // TODO the new type
        au.m_attribute = (byte) (NodeProperties.TRANSFORMROTATE + readAttr_rotscatra()); 
        
    	readAnimateType(au);
    	
    	readSetType(parent, au, false);

    	readBit(); // enabled

//    	 has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);
    	
        return au;
    }
    
    public SVGanimate readAnimation(GraphObject parent) throws Exception {
    	
    	SVGanimate au = new SVGanimate(m_graphObjectManager);
    	
    	readObjectStart(au);
    	 
//    	 begin
        if(readBit())
        	readAttr_smil_times(au, false); // TML : TODO : A verifier ce qu il faut retourner
                
        if(readBit())
        	au.m_dur = readAttr_time();
        
        /*au.m_properties.m_externalResourcesRequired =*/ readBit();
		//au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
		
		if(readBit())
			au.m_freeze = readAttr_animFill();
		
    	if(readBit())
    		/*au.m_rHeight =*/ readAttr_custom_coordinate();
    	
    	if(readBit())
    		readAttr_preserveAspectRatio(au);
	
    	if(readBit())
    		au.m_repeatCount = readAttr_repeatCount();
    	
    	if(readBit())
    		au.m_repeatDur = readAttr_repeatDur();
    	
    	if(readBit())
    		/* au.restart = */readAttr_restart();

      	if (readBit())
    		/*au.syncBehavior = */readAttr_syncBehavior();
      	
      	if (readBit())
    		/*au.syncTolerance = */readAttr_syncTolerance();
      	
      	if(readBit())
      		/*au.type = */ readAttr_custom_byteAlignedString();
      	
      	if(readBit())
    		/*au.m_rWidth =*/ readAttr_custom_coordinate();
      	
      	if(readBit())
    		/*au.m_rX =*/ readAttr_custom_coordinate();
      	
      	if(readBit())
    		/*au.m_rY =*/ readAttr_custom_coordinate();
      	
    	readHref(au.xlink);
    	
    	if(readBit())
            readAttr_time(); // TML TODO

    	if(readBit())
           readAttr_time(); // TML TODO
    		  
    	if(readBit())
    		/*au.syncReference = */readAttr_custom_anyURI();
    	  
//    	 has_attr_any
          if (readBit())
          	readAttr_any();
          
      	readObjectEnd(au);
          return au;
    	}

    // TML
    public Audio readAudio() throws Exception 
    {
    	Audio au = new Audio(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	//has_begin
    	if(readBit())
        	/*au.m_beginTimes = */readAttr_smil_times(au, false); // TML : TODO : A verifier ce qu il faut retourner
                
        if(readBit())
        	au.dur = readAttr_time();
        
        au.m_properties.m_externalResourcesRequired = readBit();
		au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
        
        if(readBit())
        	au.m_repeatCount = readAttr_repeatCount();
        
        if(readBit())
        	au.m_repeatDur = readAttr_repeatDur();
        
        if(readBit())
        	/*au.m_restart = */ readAttr_restart();
        
        if(readBit())
    		au.syncBehavior = readAttr_syncBehavior();
    		
        if(readBit())
    		au.syncTolerance = readAttr_syncTolerance();
                
        if(readBit())
        	/* au.type = */ readAttr_custom_byteAlignedString();
        
        readHref(au.xlink);

    	if(readBit())
            au.setDOMBegin(readAttr_time()); // TML TODO : DOM or Not

    	if(readBit())
           au.setDOMEnd(readAttr_time()); // TML TODO : DOM or Not
    		  
    	if(readBit())
    		au.syncReference = readAttr_custom_anyURI(); 

//    	 has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);
        
    	return au;
    }

    public Circle readCircle() throws Exception {
        /*trace(STR_BEGIN + "Circle");*/

        // Instanciate
        Circle au = new Circle(m_graphObjectManager);

        readObjectStart(au);
        
        readFillAndStroke(au);

        // Read cx
        if (readBit()) {
            au.m_rCX = readAttr_custom_coordinate();
        }

        // Read cy
        if (readBit()) {
            au.m_rCY = readAttr_custom_coordinate();
        }

        // Read radius
        au.m_rRadius = readAttr_custom_coordinate();

//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        // Read base properties
        readObjectEnd(au);
        
        return au;
    }

    // TML : TODO : read height width x y xlink
    public Multi readMulti(String tag) throws Exception 
    {
        Multi au = new Multi(m_graphObjectManager, tag);

        readObjectStart(au);
        
        if(tag.equals("defs") || tag.equals("foreignObject"))
        {
        	this.readFillAndStroke(au);
        }
        
        if(tag.equals("foreignObject"))
        {
			au.m_properties.m_externalResourcesRequired = readBit();
			au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
			
			// read height
			readAttr_custom_coordinate();
			// read height
			readAttr_custom_coordinate();
	        
	        // read x
	        if (readBit())
	            /*au.m_rX = */readAttr_custom_coordinate();
	        
	        // read y
	        if (readBit())
	            /*au.m_rY = */readAttr_custom_coordinate();
  		
			if(readBit()) // TML : TODO : >> readHref(au.xlink);
				/*au.m_link.setHref(*/readAttr_custom_anyURI()/*)*/;
        }
        
        if(!tag.equals("foreignObject"))
        {
//        	 has_attr_any
            if (readBit())
            	readAttr_any();
            
        	readObjectEnd(au);
        }
        else
        {
        	// has_attr_any
            if (readBit())
            	readAttr_any();
            
            // begin : object content
            if (readBit()) 
            {
            	int len = readBits(5);
                for(int i = 0 ; i < len ; i++) {
                	readPrivateElementContainer(au);
                }
            }
        }
        
        if (tag.equals("title")) 
        	System.out.println("title:" + ((TextContent) au.getChild(0)).getString());
        
        return au;
    }
    
    public Ellipse readEllipse() throws Exception {
        /*trace(STR_BEGIN + "Ellipse");*/

        // Instanciate
        Ellipse au = new Ellipse(m_graphObjectManager);

        // Read base properties
        readObjectStart(au);

        readFillAndStroke(au);
 
        // Read matrix : TML : REMOVED
//        au.m_definedTransform = readAndCreateMatrix();

        // Read cx
        if (readBit()) {
            au.m_rCX = readAttr_custom_coordinate();
        }

        // Read cy
        if (readBit()) {
            au.m_rCY = readAttr_custom_coordinate();
        }

        // Read X and Y radius
        au.m_rXRadius = readAttr_custom_coordinate();
        au.m_rYRadius = readAttr_custom_coordinate();

//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }
   
    //  TML
    public Group readGroup() throws Exception {

        Group au = new Group(m_graphObjectManager);
        
        readObjectStart(au);
        
        readFillAndStroke(au);
        
        au.m_properties.m_externalResourcesRequired = readBit();
        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);

        return au;
    }
    
    public void readHandler(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
        
        /*au.m_properties.m_externalResourcesRequired =*/ readBit();
        //au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
        
        // has type
        if(readBit())
        	readAttr_script();
        
//      has_event
        if(readBit())
        	readAttr_custom_event();
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        // mixed_text
        if(readBit())
        	readAttr_custom_byteAlignedString();
    }
    
    public void readHkern(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
//    	 has_g1
        if(readBit())
        	readAttr_custom_byteAlignedString();
        
//   	 has_g2
        if(readBit())
        	readAttr_custom_byteAlignedString();
        
//   	 has_k
        if(readBit())
        	readAttr_custom_fixed_16_8();
        
//   	 has_u1
        if(readBit())
        	readAttr_custom_byteAlignedString();
        
//   	 has_u2
        if(readBit())
        	readAttr_custom_byteAlignedString();
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
    }

    
    // TML
    public Bitmap readImage() throws Exception 
    {
    	Bitmap au = new Bitmap(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	au.m_properties.m_externalResourcesRequired = readBit();
        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
    	
    	if(readBit())
    		au.m_rHeight = readAttr_custom_coordinate();
    	
    	if(readBit())
    		au.opacity = readAttr_custom_0to1float();
		  
    	if(readBit())
    		readAttr_preserveAspectRatio(au);
		  
    	if(readBit())
    		/* au.type = */ readAttr_custom_byteAlignedString();

    	if (readBit())
    		au.m_rWidth = readAttr_custom_coordinate();
  
    	if (readBit())
    		au.m_rX = readAttr_custom_coordinate();
    
    	if (readBit())
    		au.m_rY = readAttr_custom_coordinate();
    
    	readHref(au.m_xlink);

    	if (readBit())
    		au.transformBehavior = readAttr_transformBehavior();

//    	 has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);
   
        return au;
    }

    public Line readLine() throws Exception 
    {
        // Instanciate
        Line au = new Line(m_graphObjectManager);

//      Read base properties
        readObjectStart(au);

        readFillAndStroke(au);

        // Read matrix // TML : REMOVED
        // au.m_definedTransform = readAndCreateMatrix();

        // Read x
        if (readBit()) {
            au.m_X1Y1.x_r = readAttr_custom_coordinate();
        }
        au.m_X2Y2.x_r = readAttr_custom_coordinate();

        // Read y
        if (readBit()) {
            au.m_X1Y1.y_r = readAttr_custom_coordinate();
        }
        au.m_X2Y2.y_r = readAttr_custom_coordinate();

//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }

//  TML : TODO : read value + verify LinearGradient object
    public LinearGradient readLinearGradient() throws Exception 
    {
    	LinearGradient au = new LinearGradient(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	//readFillAndStroke(au);
    	if(readBit()) // has_fill
    		readAttr_custom_paint();
    	
    	if(readBit()) // has_stroke
    		readAttr_custom_paint();
    	
    	if(readBit()) // has_gradientUnits
    		readBit();

		if(readBit()) // x1
			au.addDOMX(readAttr_custom_coordinate());
		  
		if(readBit()) // x2
			au.addDOMX2(readAttr_custom_coordinate());
		
		if(readBit()) // y1
			au.addDOMY(readAttr_custom_coordinate());
		  
		if(readBit())// y2
			au.addDOMY2(readAttr_custom_coordinate());
		
//		 has_attr_any
        if (readBit())
        	readAttr_any();
        
		readObjectEnd(au);
    	
    	return au;
    }

    // TML TODO : remove
/*  
    public CursorManager readCursor() throws Exception {
        
        if (Debug.DEBUG_PARSER) {
            System.out.println("------------------------------------------ Cursor ");
        }
        CursorManager au = new CursorManager(m_graphObjectManager);
        readBasics(au);
        readChildren(au);
        // target
        readHref(au.m_xlink);

        // Sensitive point
        if (readBit())
            au.m_pointer.x_r = translate(readBits(m_decodingContext.m_coordBits), m_decodingContext.m_coordBits);
        if (readBit())
            au.m_pointer.y_r = translate(readBits(m_decodingContext.m_coordBits), m_decodingContext.m_coordBits);
        if (Debug.DEBUG_PARSER) {
            System.out.println("endReadCursor ");
        }
        
        return au;
    }
*/

//  TML TODO : remove
    /*public Defs readDefs() throws Exception
    {
        if (Debug.DEBUG_PARSER)
        {
            System.out.println("------------------------------------------ Defs " + (depth++));
        }
        Defs au = new Defs(m_graphObjectManager);
        //readBase(au);
        int len = readVlcInt(5, "len"); // Number of nodes
        for (int i = 0; i < len; i++)
        {
            if (Debug.DEBUG_PARSER)
            {
                System.out.println("readChild " + i + " of " + len);
            }
            au.appendChild(readElements());
        }
        readBit(); // extension bit
        if (Debug.DEBUG_PARSER)
        {
            System.out.println("endReadDefs " + (--depth));
        }
        return au;
    }*/

    // TML TODO : PTETRE REMOVE
/*    
    private int GetCodeNum(int k, int Mvalue, int INFO)
    {
        return 2^(k+Mvalue) + INFO - 2^k ;
    }
*/  
    
//  TML : TODO : verify
    public Mpath readMpath() throws Exception 
    {
        Mpath au = new Mpath(m_graphObjectManager);
        
        readObjectStart(au);
        
        readHref(au.xlink);

//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }
    
    public Path readPath() throws Exception 
    {
        Path au = new Path(m_graphObjectManager);
        
        readObjectStart(au);
        
        readFillAndStroke(au);
                
        au.m_polygon = new Polygon2D();
        
        // read points sequence
        readAttr_custom_pointSequence(au.m_polygon);
        
        au.m_wasUpdated = true;
        au.m_types = new byte[readVlcInt(5)];
        for (int i = 0; i < au.m_types.length; i++) 
        {
        	int idx = readBits(5);
            au.m_types[i] = segTypeToPathType[idx];
        }
        
        if (readBit())
        	readAttr_custom_fixed_16_8();

        // Create the presentation value value
        au.m_actualPolygon = au.m_polygon.getCopy();

//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }

    public Polygon readPolygon() throws Exception 
    {
        // TML
        Polygon au = new Polygon(m_graphObjectManager);
        
        readObjectStart(au);
        
        readFillAndStroke(au);
        
        // TML : REMOVE
        //au.m_definedTransform = readAndCreateMatrix();
        
        au.m_polygon = new Polygon2D();
        
        // read points sequence
        readAttr_custom_pointSequence(au.m_polygon);
        
        // it's a polygon so close the shape
        Point2D initialPoint = (Point2D) au.m_polygon.m_points.elementAt(0);
        
        if (initialPoint != null) {
            au.m_polygon.m_points.addElement(new Point2D(initialPoint.x_r, initialPoint.y_r));
        }

        // Create the presentation value value
        au.m_actualPolygon = au.m_polygon.getCopy();

//      has_attr_any
        if (readBit())
        	readAttr_any();
        
       	readObjectEnd(au);
       	
        return au;
    }
    
    public void readPrefetch(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	//has_bandwidth
    	if(readBit())
    		readVlcInt(5);
    	
    	//has_mediaCharacterEncoding
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
    	//has_mediaContentEncodings
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    		
    	//has_mediaSize
    	if(readBit())
    		readVlcInt(5);
    	
//    	has_mediaTime
    	if(readBit())
    		readVlcInt(5);
    	
    	Xlink TMP = new Xlink();
    	readHref(TMP);
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
       	readObjectEnd(au);
    }


//  TML : TODO
    public RadialGradient readRadialGradient() throws Exception {
    	
    	RadialGradient au = new RadialGradient(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	//readFillAndStroke(au);
    	if(readBit()) // has_fill
    		readAttr_custom_paint();
    	
    	if(readBit()) // has_stroke
    		readAttr_custom_paint();

		if(readBit()) // cx
			au.addDOMX(readAttr_custom_coordinate());
		  
		if(readBit()) // cy
			au.addDOMX2(readAttr_custom_coordinate());
		
		if(readBit()) // has_gradientUnits
    		readBit();
		
		if(readBit()) // has_r;
			au.addDOMRadius(readAttr_custom_coordinate());		
		
//		 has_attr_any
        if (readBit())
        	readAttr_any();
        
		readObjectEnd(au);
    	
    	return au;

    }
    
    public Rect readRect() throws Exception {
        /*trace(STR_BEGIN + "Rect");*/

        // Instanciate
        Rect au = new Rect(m_graphObjectManager);

//      Read base properties
        readObjectStart(au);

        readFillAndStroke(au);
 
        // Read matrix : TML : REMOVED
//        au.m_definedTransform = readAndCreateMatrix();

//      Read Height
        au.m_rHeight = readAttr_custom_coordinate();
        
        // read rx
        if (readBit())
            au.m_rx = readAttr_custom_coordinate();

        // read ry
        if (readBit())
            au.m_ry = readAttr_custom_coordinate();

        // Read width
        au.m_rWidth = readAttr_custom_coordinate();

        // read x
        if (readBit()) {
            au.m_x = readAttr_custom_coordinate();
        }
        // read y
        if (readBit()) {
            au.m_y = readAttr_custom_coordinate();
        }
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }

    public Group readSameGroup(Group lastGroup) throws Exception {
        
        Group au = new Group(m_graphObjectManager);
/*        readBasics(au);
        readChildren(au);
        if (lastGroup != null) {
            lastGroup.m_properties.copyTo(au.m_properties);
            if (lastGroup.m_definedTransform != null)
                au.setTransformMatrixValues(new Matrix2D(lastGroup.m_definedTransform));
            if (lastGroup.m_size != null)
                au.m_size = new Point2D(lastGroup.m_size.x_r, lastGroup.m_size.y_r);
        }

        */
        return au;
    }
    
    public Line readSameLine(Line line) throws Exception {
        return null;
    }

    //TODO make all same/FS like this
    public Path readSamePath(Path lastPath, boolean f) throws Exception {
        Path au = new Path(m_graphObjectManager);
        /*trace(STR_BEGIN + "readSamePath" + (fs ? "FS" : ""));*/

        //copy the values
        if (lastPath != null) {
            lastPath.m_properties.copyTo(au.m_properties);
            if (lastPath.m_definedTransform != null)
                au.m_definedTransform = new Matrix2D(lastPath.m_definedTransform);
        }

        //for samePathFS
        if (f) {
            if (readBit())
                if (readBit())
                    au.m_properties.m_fill = readAttr_custom_paint();
                else
                    au.m_properties.m_fill = 0;
        }
        au.m_polygon = new Polygon2D();
        readAttr_custom_pointSequence(au.m_polygon);
        au.m_wasUpdated = true;
        au.m_types = new byte[readVlcInt(5)];
        for (int i = 0; i < au.m_types.length; i++) {
            au.m_types[i] = segTypeToPathType[readBits(5)];
        }

        // Create the presentation value value
        au.m_actualPolygon = au.m_polygon.getCopy();
        /*trace(STR_END + "readSamePath" + (fs ? "FS" : ""));*/
        return au;
    }
    
    public Polygon readSamePolygon(Polygon poly, boolean f, boolean s) throws Exception {
        // Clones a polygon with its properties and read the new points sequence

        /*trace(STR_BEGIN + "readPolygon" + (fs ? "FS" : ""));*/
        Polygon au = new Polygon(m_graphObjectManager);
        if (poly != null) {
            poly.m_properties.copyTo(au.m_properties);
            Matrix2D.copyMatrixVariables(au.m_definedTransform, poly.m_definedTransform);
        }
        if (f) {
            if (readBit())
                if (readBit())
                    au.m_properties.m_fill = readAttr_custom_paint();
                else
                    au.m_properties.m_fill = 0;
        }
        if (s) {
            if (readBit())
                if (readBit())
                    au.m_properties.m_stroke = readAttr_custom_paint();
                else
                    au.m_properties.m_stroke = 0;
        }
        au.m_polygon = new Polygon2D();
        // read points sequence
        readAttr_custom_pointSequence(au.m_polygon);
        // it's a polygon so close the shape
        Point2D initialPoint = (Point2D) au.m_polygon.m_points.elementAt(0);
        if (initialPoint != null) {
            au.m_polygon.m_points.addElement(new Point2D(initialPoint.x_r, initialPoint.y_r));
        }

        // Create the presentation value value
        au.m_actualPolygon = au.m_polygon.getCopy();

        /*trace(STR_END + "readPolygon" + (fs ? "FS" : ""));*/
        return au;
    }

    public Polyline readSamePolyline(Polyline poly, boolean f, boolean s) throws Exception {
        // Clones a polyline with its properties and read the new points sequence
        /*trace(STR_BEGIN + "SamePolyline");*/
        Polyline au = new Polyline(m_graphObjectManager);
        if (poly != null) {
            poly.m_properties.copyTo(au.m_properties);
            Matrix2D.copyMatrixVariables(au.m_definedTransform, poly.m_definedTransform);
        }
        if (f) {
            if (readBit())
                if (readBit())
                    au.m_properties.m_fill = readAttr_custom_paint();
                else
                    au.m_properties.m_fill = 0;
        }
        if (s) {
            if (readBit())
                if (readBit())
                    au.m_properties.m_stroke = readAttr_custom_paint();
                else
                    au.m_properties.m_stroke = 0;
        }
        au.m_polygon = new Polygon2D();
        // read points sequence
        readAttr_custom_pointSequence(au.m_polygon);
        // Create the presentation value
        au.m_actualPolygon = au.m_polygon.getCopy();
        /*trace(STR_END + "SamePolyline");*/
        return au;
    }

    public Polyline readSamePolylineFS(Polyline poly) throws Exception {
        
        Polyline au = new Polyline(m_graphObjectManager);
        if (poly != null) {
            poly.m_properties.copyTo(au.m_properties);
            Matrix2D.copyMatrixVariables(au.m_definedTransform, poly.m_definedTransform);
        }
        au.m_polygon = new Polygon2D();
        if (readBit())
            if (readBit())
                au.m_properties.m_fill = readAttr_custom_paint();
            else
                au.m_properties.m_fill = 0;
        if (readBit())
            if (readBit())
                au.m_properties.m_stroke = readAttr_custom_paint();
            else
                au.m_properties.m_stroke = 0;
        // read points sequence
        readAttr_custom_pointSequence(au.m_polygon);
        // Create the presentation value
        au.m_actualPolygon = au.m_polygon.getCopy();
        
        return au;
    }

    // TML : TODO : REMOVE
/*    
    public Polyline readPolyline() throws Exception {
        
        Polyline au = new Polyline(m_graphObjectManager);
        readObject(au);
        au.m_definedTransform = readAndCreateMatrix();
        au.m_polygon = new Polygon2D();
        // read points sequence
        read_codec_pointSequence(au.m_polygon);

        // Create the presentation value
        au.m_actualPolygon = au.m_polygon.getCopy();

        
        return au;
    }
*/
     
//  TML : TODO : Verify
    public Rect readSameRect(Rect rect, boolean f) throws Exception 
    {
    	Rect au = new Rect(m_graphObjectManager);
    	
    	if(rect != null)
    	{
	    	rect.m_properties.copyTo(au.m_properties); 
	    	au.m_rx = rect.m_rx;
	        au.m_ry = rect.m_ry;
    	}

    	if(readBit())
    		readAttr_custom_ID(au);
    	
    	if(f)
    	{
    		if (readBit()) {
            	au.m_properties.m_fill = readAttr_custom_paint();
            	au.m_properties.setInheritance(Value.UPDATE_COLOR_FILL, false);
            }
    	}
 
//      Read Height
        au.m_rHeight = readAttr_custom_coordinate();
        
//      Read width
        au.m_rWidth = readAttr_custom_coordinate();
        
        // read x
        if (readBit())
            au.m_x = readAttr_custom_coordinate();

        // read y
        if (readBit())
            au.m_y = readAttr_custom_coordinate();
                        
        readChildren(au);
        
        return au;
    }
    
    public Text readSameText(Text text, boolean f) throws Exception {
        /*trace(STR_BEGIN + "sameText");*/
        // Instanciate
        Text au = new Text(m_graphObjectManager);
        if(text != null)
        {
        	text.m_properties.copyTo(au.m_properties);
        	if (text.m_definedTransform != null)
        		au.m_definedTransform = new Matrix2D(text.m_definedTransform);
        	
        	au.m_rX = text.m_rX;
        	au.m_rY = text.m_rY;
        	au.m_editable = text.m_editable;
        }

        // Read properties
        if(readBit())
        	readAttr_custom_ID(au);
        
        if (readBit()) {
        	int[] xlist = readAttr_coordinateList(); // TML : TODO : Change the text class to take this list into account
            au.m_rX = xlist[0];
        }
        
        if (readBit()) {
        	int[] ylist = readAttr_coordinateList(); // TML : TODO : idem than 5 lines above
            au.m_rY = ylist[0];;
        }
        
        readChildren(au);
            
        // Done
        /*trace(STR_END + "sameText");*/
        return au;
    }
    
//  TML
    public Use readSameUse(Use lastUse) throws Exception 
    {
        Use au = new Use(m_graphObjectManager);
        
        if (lastUse != null) 
        {
            lastUse.m_properties.copyTo(au.m_properties);
            
            au.m_rXY.x_r = lastUse.m_rXY.x_r;
            au.m_rXY.y_r = lastUse.m_rXY.y_r;
            au.m_rWH.x_r = lastUse.m_rWH.x_r;
            au.m_rWH.y_r = lastUse.m_rWH.y_r;
        }
        
        if(readBit())
        	readAttr_custom_ID(au);
        
        readHref(au.xlink);

        readChildren(au);
        
        return au;
    }
    
    public Conditional readScript() throws Exception 
    {
        Conditional au = new Conditional(m_graphObjectManager);

        // Set the decoder
        au.setElementParser(this);

        readObjectStart(au);
        
        au.m_properties.m_externalResourcesRequired = readBit();
        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
        
        if(readBit())
        	au.contentType = readAttr_script();
        
        if(readBit()) // TML : TODO : >> readHref(au.xlink);
        	/*au.xlink.setHref(*/readAttr_custom_anyURI()/*)*/;
        	  
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);

        return au;
    }
    
//  TML
    public SVGset readSet(GraphObject parent) throws Exception {
       
        SVGset au = new SVGset(m_graphObjectManager);
        
        readObjectStart(au);
      	 
        if(readBit())
    	{
    		int idx = readAttr_AttributeName();
    		au.m_attribute = (byte) Value.animName[idx];
    		au.m_attributeType = Value.convertTypeFromAnimValue(Value.animType[au.m_attribute]);
    	}
        
    	readSetType(parent, au, false);
    	
    	readBit(); // enabled

//    	 has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);

        return au;

    }
    
    public void readSetScroll(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	//begin
    	if(readBit())
    		readAttr_smil_times(au, false);
    	
    	//direction
    	if(readBit())
    		readAttr_directionType();
    	
    	//increment
    	if(readBit())
    		readAttr_scrollStopType();
    	
    	//has to
    	if(readBit())
    		readAttr_scrollStopType();
    	
    	Xlink TMP = new Xlink();
    	readHref(TMP);
    	
    	if(readBit())
    		readAttr_any();
    	
    	readObjectEnd(au);
    }
    
//  TML : TODO : read value + verify Stop object 
    public Stop readStop() throws Exception 
    {
    	Stop au = new Stop(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	//readFillAndStroke(au);
    	if(readBit()) // has_fill
    		readAttr_custom_paint();
    	
    	if(readBit()) // has_stroke
    		readAttr_custom_paint();
    	
    	// offset
    	readAttr_custom_fixed_16_8();
		
//    	 has_attr_any
        if (readBit())
        	readAttr_any();
        
		readObjectEnd(au);
    	
    	return au;
    }
    
    public void readStreamSource(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	//height
    	if(readBit())
    		readAttr_custom_coordinate();
    	
    	// has_mode
    	if(readBit())
    		readAttr_streamSourceModeType();
    	
    	//has_sourceIndex
    	if(readBit())
    		readVlcInt(5);
    	
    	if(readBit())
    		readAttr_sourcesType();
    	
//    	width
    	if(readBit())
    		readAttr_custom_coordinate();
    	
//   	 has_attr_any
        if (readBit())
        	readAttr_any();
        
		readObjectEnd(au);
    }

    public Switch readSwitch() throws Exception 
    {
        Switch au = new Switch(m_graphObjectManager);
        
        readObjectStart(au);
        
        readFillAndStroke(au);

        au.m_properties.m_externalResourcesRequired = readBit();
        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);

        // Get values to evaluate
        au.m_systemLanguage = m_systemLanguage;
        // TODO other values (this todo was here in the first place)

        return au;
    }
    
    public void readTBreak(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);
    }
    
    public Text readText() throws Exception 
    {   
        Text au = new Text(m_graphObjectManager);
        
        readObjectStart(au);
        
        readFillAndStroke(au);
        
        au.m_editable = readAttr_editable();

        // rotate
        if(readBit())
        {
        	int[] item = readAttr_floatList();
        }
        	
        if (readBit()) {
        	int[] xlist = readAttr_coordinateList(); // TML : TODO : Change the text class to take this list into account
            au.m_rX = xlist[0];
        }
        
        if (readBit()) {
        	int[] ylist = readAttr_coordinateList(); // TML : TODO : idem than 5 lines above
            au.m_rY = ylist[0];;
        }
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }
    
    public TextArea readTextArea(GraphObject parent) throws Exception
    {
    	TextArea au = new TextArea(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	readFillAndStroke(au);
    	
    	if(readBit())
    		au.m_editable = readAttr_editable();
    	
    	//height
    	if(readBit())
    		au.m_Height = readAttr_custom_coordinate();
  
//    	width
    	if(readBit())
    		au.m_Width = readAttr_custom_coordinate();
    	
//    	x
    	if(readBit())
    		au.m_rX = readAttr_custom_coordinate();
    	au.m_rX = 4096;
    	
//    	y
    	if(readBit())
    		au.m_rY = readAttr_custom_coordinate();
    	au.m_rY = 4096;
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }
    

//  TML
    public Tspan readTspan() throws Exception 
    {
        Tspan au = new Tspan(m_graphObjectManager);
        
        readObjectStart(au);
        
        readFillAndStroke(au);
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }
    
    public void readUpdate_Activate() throws IOException
    {
    	LaserUpdate au = new LaserUpdate();
    	
    	au.ref = readAttr_custom_IDREF();
    	
    	if(readBit())
    		readAttr_any();
    	
    	processActivate(au.ref);
    }
    
    public void readUpdate_Deactivate() throws IOException
    {
    	LaserUpdate au = new LaserUpdate();
    	
    	au.ref = readAttr_custom_IDREF();
    	
    	if(readBit())
    		readAttr_any();
    	
    	processDesactivate(au.ref, true, true);
    }
    
//  TML
    public Use readUse() throws Exception {
    	
        Use au = new Use(m_graphObjectManager);
        
        readObjectStart(au);
        
        readFillAndStroke(au);
        
//        au.m_definedTransform = readAndCreateMatrix();

		au.m_properties.m_externalResourcesRequired = readBit();
        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
        
        if (readBit()) {
        	au.m_properties.m_overflow = (byte) readBits(2); // TML : TODO : v�rifier que l'on a bien visible(0)
        	au.m_properties.setInheritance(Value.UPDATE_OVERFLOW, false);
        }
        
        // Read x
        if (readBit()) {
            au.m_rXY.x_r = readAttr_custom_coordinate();
        }

        // Read y
        if (readBit()) {
            au.m_rXY.y_r = readAttr_custom_coordinate();
        }
        
        readHref(au.xlink);
        
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
        
        return au;
    }
    
//  TML
    public Video readVideo() throws Exception 
    {
    	Video au = new Video(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	// begin
    	if(readBit())
        	/*au.m_beginTimes = */readAttr_smil_times(au, false); // TML : TODO : A verifier ce qu il faut retourner
                
        if(readBit())
        	au.dur = readAttr_time();
        
        au.m_properties.m_externalResourcesRequired = readBit();
        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
        
        if(readBit())
    		au.m_rHeight = readAttr_custom_coordinate();
        
        if(readBit())
        	/* au.m_overlay = */readAttr_overlay();
        
        if(readBit())
        	readAttr_preserveAspectRatio(au);
        
        if(readBit())
        	au.m_repeatCount = readAttr_repeatCount();
        
        if(readBit())
        	au.m_repeatDur = readAttr_repeatDur();
        
        if(readBit())
        	/* au.m_restart = */ readAttr_restart();
        
        if(readBit())
    		au.syncBehavior = readAttr_syncBehavior();
    		
        if(readBit())
    		au.syncTolerance = readAttr_syncTolerance();
        
        if(readBit())
        	au.transformBehavior = readAttr_transformBehavior();
        
        if(readBit())
        	/* au.type = */ readAttr_custom_byteAlignedString();
        
        if(readBit())
        	au.m_rWidth = readAttr_custom_coordinate();
        	
        if(readBit())
        	au.m_rX = readAttr_custom_coordinate();

        if(readBit())
        	au.m_rY = readAttr_custom_coordinate();

        
        readHref(au.xlink);

    	if(readBit())
            readAttr_time();

    	if(readBit())
           readAttr_time();
    	
    	// fullscreen
    	if(readBit())
    	{
    		/* au.m_fullscreen = */ readBit();
    	}
    		  
    	if(readBit())
    		/*au.m_syncReference.setxxxx(*/readAttr_custom_anyURI(); 
    		  
//    	 has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);

        return au;
    }

    public Listener readListener() throws Exception 
    {
        Listener au = new Listener(m_graphObjectManager);
        
        readObjectStart(au);
        
        if(readBit())
            au.defaultAction = readBits(1);

//      hasEvent
        if (readBit())
            au.event = readAttr_custom_event();
        else
            au.event = new Event();
        	  
        if (readBit())
            au.handler = readAttr_custom_anyURI();
        	 
        if (readBit()) 
        {
            au.m_observer = readAttr_custom_IDREF();
            m_graphObjectManager.attachListener(au, au.m_observer);
        }
        
        readBits(1); /* au.phase = */
  
        if (readBit())
        	au.propagate = readBits(1);
  
        if (readBit())
        	au.target.id = readAttr_custom_IDREF();

        au.m_enabled = readBit();
  
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);

        // attach trigger to the Navigator
        m_navigator.attachEventListener(au);
        if (Debug.DEBUG_PARSER) {
            System.out.println("trigger event:" + au.event.event + " observer:" + au.m_observer + " href:" + au.handler.id + " beginOrEnd:" + au.m_beginOrEnd + " delay:" + au.m_delay);
        }

        return au;
    }
    
    public void readMissingGlyph(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	//has_d
    	if(readBit())
    	{
    		Value av = new Value();
    		readAttr_custom_path(av);
    	}
    	
    	//has_horizAdvX
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    		
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        readObjectEnd(au);
    }
    
//  TML : TODO
    public Conditional readConditional() throws Exception 
    {
    	Conditional au = new Conditional(m_graphObjectManager);
    	
//    	 Set the decoder
        au.setElementParser(this);
    	
    	readObjectStart(au);
    	
    	// begin
    	if(readBit())
    		readAttr_smil_times(au, false);
    	    	
    	au.m_properties.m_externalResourcesRequired = readBit();
    	au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
    	
    	/*au.enabled = */readBit();
    	
    	if (readBit())
    		readAttr_any();
    	   	
    	readUpdateListType_content(au);
    	
    	if(readBit())
          	readPrivateAttributeContainer(au);
    	
        return au;
    }
    
//  TML : TODO :Verify
    public CursorManager readCursorManager() throws Exception 
    {
    	CursorManager au = new CursorManager(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	// has_x
    	if(readBit())
    		au.m_pointer.x_r = readAttr_custom_coordinate();
    	
    	// has_y
    	if(readBit())
    		au.m_pointer.y_r = readAttr_custom_coordinate();
    	
    	readHref(au.m_xlink);
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);

        return au;
    }
    
    public GraphObject readUpdateAny (GraphObject node, boolean Is4UpdateElement) throws Exception 
    {
    	Group au = new Group(m_graphObjectManager);
    	
    	int reserved = readBits(m_decodingContext.m_extensionIDbits); // reserved
    	int skiplen = readVlcInt(5);
    	
    	if(reserved == 2)
    	{
    		if(Is4UpdateElement)
    			readLASeRAmd1Command(node);
    		else
    			au.appendChild(readElementsTwo(node));  // TML : TODO finish
    	}
    	else
    	{
	    	if (skiplen % 8 != 0)
	        	readBits(skiplen % 8);
	        if (skiplen >> 3 != 0)
	        	readByteArray(skiplen >> 3);
    	}
        return au;
    }
    
//  TML : TODO : verify return new multi
    private GraphObject readPrivateElementContainer(GraphObject node) throws Exception 
    {
        int ch4 = readBits(4);
        switch (ch4) {
            case 0:
            	readPrivateAnyXMLElement(node);
                break;
            case 1:
            	readPrivateOpaqueElement();
                break;
            case 2:
            	readElement_any();
                break;
            default:
            	readAttr_custom_extension();
                break;
        }
        return new Multi(m_graphObjectManager, "");
    }
    
//  TML : TODO : verify
    public Group readRectClip() throws Exception 
    {
    	Group au = new Group(m_graphObjectManager);
    	au.m_choicePosition = -2; // TML : indicate to the group node that it is a rectClip
    	
    	readObjectStart(au);
    	
    	readFillAndStroke(au);
    	
    	au.m_properties.m_externalResourcesRequired = readBit();
    	au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
    	
//    	au.m_size = new Point2D(206*4096, 254*4096); // TML : is it the default value ?
    	if(readBit()) {
    		au.m_size = new Point2D(0, 0);
    		readAttr_Point(au.m_size);
    	}
    		
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);

        return au;
    }
    
    public void readRefreshScene() throws Exception
    {
    	LaserUpdate au = new LaserUpdate();
    	
    	/*au.time = */readVlcInt(5);
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        Object node = readSvg();
        if (! (node instanceof Svg)) {
            return;
        }
        
        Svg svgNode = (Svg) node;
        
        // processRefreshScene(svgNode);
    }

    public Group readSelector() throws Exception 
    {
    	Group au = new Group(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	readFillAndStroke(au);
    	
    	au.m_properties.m_externalResourcesRequired = readBit();
    	au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
    	
    	if(readBit())
    		au.m_choicePosition = readAttr_choice();
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);

        return au;
    }
    
//  TML : TODO : Verify
    public Group readSimpleLayout() throws Exception 
    {
    	Group au = new Group(m_graphObjectManager);
    	
    	readObjectStart(au);
    	
    	readFillAndStroke(au);
    	
    	au.m_size = new Point2D(0,0); // TML : is it the default value ?
    	if(readBit())
    		readAttr_Point(au.m_size);
    	
    	au.m_properties.m_externalResourcesRequired = readBit();
    	au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);

        return au;
    }
    
    public void readSolidColor(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
    	readObjectEnd(au);
    }
    
    public TextContent readTextContent() throws IOException {
        return new TextContent(m_graphObjectManager, readAttr_custom_byteAlignedString());
    }
    
//  TML
    public void readExtension() throws Exception 
    {
    	readAttr_custom_extension();
    }
    
    public void readLASeRAmd1Command(GraphObject node) throws Exception
    {
    	int skip87 = readVlcInt(5);
    	int occ = readVlcInt(5);

    	for (int i = 0 ; i < occ ; i++) 
    	{
    		int skip2 = readBits(2);
	    	int type = readBits(2);
	    	switch(type)
	    	{
	    		case 0:
    					readPrivateElementContainer(node);
    				break;
		    	case 1:
		    			readUpdate_Activate();
		    		break;
		    	case 2:
		    			readUpdate_Deactivate();
		    		break;
	    	}
    	}
    }
    
///////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////// END Nodes parsing operations ///////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////    


    
///////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// Attributes parsing operations /////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////
	
    private boolean readAttr_animFill() throws IOException {
        // Enumeration: freeze{0} remove{1}
        return readBit();
    }
    
    private void readAttr_any() throws IOException { // TML
    	boolean hasMore = false;
//    	if(readBit()) // ????????
//    	{
	    	do {
	    		readBits(m_decodingContext.m_extensionIDbits);
	    		int len = readVlcInt(5);
	//    		readBits(len); TML : TODO : � �claircir avec les 2 lignes ci-dessous qui sont invers�es par rapport au J2SE + conditions manquantes
	            if (len % 8 != 0)
	            	readBits(len % 8);
	            if (len >> 3 != 0)
	            	readByteArray(len >> 3);
	            /*readByteArray(len >> 3);
	    		readBits(len % 8);*/
	    		hasMore = readBit();
	    	} while (hasMore);
//    	}
    }
    
//  TML : TODO : if already exist + read and retun value + case inversion like spec
    private int readAttr_AttributeName() throws Exception 
    {
    	if(readBit()) // choice case 1
    	{
    		int max = 2;
    		int item[] = new int[max];
            for (int t = 0 ; t < max ; t++)
            {
            	item[t] = readVlcInt(5);
            }
            return item[0];
    	}
    	else
    	{
//    		 Enumeration: a.target{0} accumulate{1} additive{2} audio-level{3} bandwidth{4} begin{5} calcMode{6} children{7} choice{8} clipBegin{9} clipEnd{10} color{11} color-rendering{12} cx{13} cy{14} d{15} delta{16} display{17} display-align{18} dur{19} editable{20} enabled{21} end{22} event{23} externalResourcesRequired{24} fill{25} fill-opacity{26} fill-rule{27} focusable{28} font-family{29} font-size{30} font-style{31} font-variant{32} font-weight{33} fullscreen{34} gradientUnits{35} handler{36} height{37} image-rendering{38} keyPoints{39} keySplines{40} keyTimes{41} line-increment{42} listener.target{43} mediaCharacterEncoding{44} mediaContentEncodings{45} mediaSize{46} mediaTime{47} nav-down{48} nav-down-left{49} nav-down-right{50} nav-left{51} nav-next{52} nav-prev{53} nav-right{54} nav-up{55} nav-up-left{56} nav-up-right{57} observer{58} offset{59} opacity{60} overflow{61} overlay{62} path{63} pathLength{64} pointer-events{65} points{66} preserveAspectRatio{67} r{68} repeatCount{69} repeatDur{70} requiredExtensions{71} requiredFeatures{72} requiredFormats{73} restart{74} rotate{75} rotation{76} rx{77} ry{78} scale{79} shape-rendering{80} size{81} solid-color{82} solid-opacity{83} stop-color{84} stop-opacity{85} stroke{86} stroke-dasharray{87} stroke-dashoffset{88} stroke-linecap{89} stroke-linejoin{90} stroke-miterlimit{91} stroke-opacity{92} stroke-width{93} svg.height{94} svg.width{95} syncBehavior{96} syncBehaviorDefault{97} syncReference{98} syncTolerance{99} syncToleranceDefault{100} systemLanguage{101} text-align{102} text-anchor{103} text-decoration{104} text-display{105} text-rendering{106} textContent{107} transform{108} transformBehavior{109} translation{110} vector-effect{111} viewBox{112} viewport-fill{113} viewport-fill-opacity{114} visibility{115} width{116} x{117} x1{118} x2{119} xlink:actuate{120} xlink:arcrole{121} xlink:href{122} xlink:role{123} xlink:show{124} xlink:title{125} xlink:type{126} xml:base{127} xml:lang{128} y{129} y1{130} y2{131} zoomAndPan{132}
    		return readBits(8);
    	}
    }
    
    private int readAttr_choice() throws IOException
    {
	    if (!readBit()) // choice = "value"
		    return readBits(8);
	    else if (readBit()) // choice = none
	    		return -1;
	    else return -3; // choice = all
	}
    
    //  TML
    private int[] readAttr_coordinateList() throws IOException
    {
    	int len = readVlcInt(5);
    	int res[] = new int[len];
        for (int i = 0; i < len; i++) 
        	res[i] = readAttr_custom_coordinate();
        return res;
    }
    
    //  TODO error message saying that we treat only the uri.id
    private URI readAttr_custom_anyURI() throws IOException {
        URI uri = new URI();
        if (readBit()) {
            uri.uri = readAttr_custom_byteAlignedString();
            if (readBit()) {
                uri.data = readByteArray(readVlcInt(5));
            }
        }
        if (readBit())
            uri.id = readAttr_custom_IDREF();
        if (readBit())
            uri.streamID = readAttr_custom_IDREF();
        return uri;
    }
    
    // TML
    private int readAttr_custom_coordinate() throws IOException
    {
    	return translate(readBits(m_decodingContext.m_coordBits));
    }
    
    private void readAttr_custom_ElementAttributeList() throws Exception
    {
    	int nbIds = readVlcInt(5);
    	for (int i = 0; i < nbIds; i++)
    	{
    		readAttr_custom_IDREF();
    		readAttr_AttributeName();
    	}
    }
    
    private void readAttr_custom_AnimatedValues(SVGanimate au) throws Exception {
		int type = readBits(4);
		int nbValue = readVlcInt(5);
		au.m_values = new Value[nbValue];
	
		for(int i = 0 ; i < nbValue ; i++)
		{
			Value av = new Value();
			av.m_animValueType = type;
	
			readAttr_custom_AnimatedValue(av);
	       
			au.m_values[i] = av;
		}
	}
	
    private void readAttr_custom_AnimatedValue(Value av) throws Exception 
	{
		// escapeFlag
		if (readBit()) {
			av.m_escapeEnum = readBits(2);
		} 
	    else {
	    	switch(av.m_animValueType) {
	            	case Value.ANIMTYPE_string:
	                    av.m_string = readAttr_custom_byteAlignedString();
	                break;
	            case Value.ANIMTYPE_float: // float
	            		av.m_enumOrColorOrRealOrID = readAttr_custom_fixed_16_8();
	                break;
	            case Value.ANIMTYPE_uri:
	            		av.m_uri = readAttr_custom_anyURI();
	                break;
	            case Value.ANIMTYPE_path: // path
	            		readAttr_custom_path(av);
	                break;
	            case Value.ANIMTYPE_pointSeq: // pointSeq
	                	av.m_polygon = new Polygon2D();
	                	readAttr_custom_pointSequence(av.m_polygon);
	                break;
	            case Value.ANIMTYPE_fraction:
	                	av.m_enumOrColorOrRealOrID = readAttr_custom_0to1float();
	                break;
	            case Value.ANIMTYPE_color:
	                	av.m_enumOrColorOrRealOrID = readAttr_custom_paint();
	                break;
	            case Value.ANIMTYPE_enum: // enum
	            case Value.ANIMTYPE_id: // id
	            		av.m_enumOrColorOrRealOrID = readVlcInt(5);
	                break;
	            case Value.ANIMTYPE_font: // font
	            		av.m_string = (String) m_decodingContext.m_fontArray.elementAt(readVlcInt(5));
	                break;
	            case Value.ANIMTYPE_ints: // ints
	                	int ma = readVlcInt(5);
	                    av.m_floatsOrInts = new int[ma];
	                    for (int j = 0 ; j < ma ; j++) {
	                        av.m_floatsOrInts[j] = readVlcInt(5);
	                    }
	                break;
	            case Value.ANIMTYPE_floats: // floats
	                	int max = readVlcInt(5);
	                    av.m_floatsOrInts = new int[max];
	                    for (int j = 0; j < max; j++) {
	                        av.m_floatsOrInts[j] = readAttr_custom_fixed_16_8();
	                    }
	                break;
	            case Value.ANIMTYPE_point: // point
	            		av.m_point = new Point2D(readAttr_custom_coordinate(), readAttr_custom_coordinate());
	                break;
	            default:
	            		readAttr_custom_extension();
	                break;
	    	}
	    }
	}
	
	private void readAttr_custom_ID (GraphObject node) throws IOException {

    	// read the provided id
         node.setID(readVlcInt(5));
         // reading the codec_ID field
         if (m_decodingContext.m_hasStringIDs) {
             node.stringId = readAttr_custom_byteAlignedString();
             System.out.println("string ids not managed in this version");
         }
         if (readBit()) { // TML : TODO : Inversion des 2 lignes par rapport � l'encodeur ??
             int len = readVlcInt(5);
             if (len % 8 > 0) readBits(len%8);
             if (len >> 3 > 0) readByteArray(len>>3);
         }
    }
	
	private int readAttr_custom_IDREF() throws IOException
	{
		int href = readVlcInt(5);
		
		if(readBit())
			readByteArray(readVlcInt(5)); // reserved
		
		return href;
	}
    	
//	 TML 
	private Event readAttr_custom_event() throws IOException
    {
    	Event event = new Event();
        if (readBit()) {
            event.event = readBits(6);
            switch (event.event) {
            case 1: //accessKey
            case 12: //longAccessKey
                event.keyCode = readVlcInt(5);
                break;
            }
        } else {
            event.eventString = readAttr_custom_byteAlignedString();
        }
        return event;
    }
    
	private int readAttr_custom_paint() throws java.io.IOException {
		/*if (br.readBit("isColorIndex")) { // more explicit log message
            return man.getColor(br.readBits(colorIndexBits, msg));
        } else {
            switch(br.readBits(2, "ch2")) {
              case 0:
                  //enum
                  switch (br.readBits(2, msg)) {
                      case 0:
                          return -3;
                      case 1:
                          return -1;
                      case 2:
                          return -2;
                  }
                  break;
              case 1:
                  // URI
                  return -4 - currentScene.idMan.registerPaintURI(class_anyURI(br, msg));
              case 2:
                  //System Paint Server
                  class_byteAlignedString(br, msg);
                  break;
              case 3:
                  // extension
                  class_extension(br, msg);
                  break;
            }
        }*/if (readBit()) 
        {
            int index = readBits(m_decodingContext.m_colorIndexBits);
            return ((Integer) m_decodingContext.m_colorsArray.elementAt(index)).intValue();
        } 
        else 
        {
        	switch (readBits(2)) 
            {
        	case 0:
                switch (readBits(2)) 
                {
	                case 0:
	                    return -3;
	                case 1:
	                    return -1;
	                case 2:
	                    return -2;
                }
                break;
        	case 1:
                return -4 - m_decodingContext.registerPaintURI(readAttr_custom_anyURI()); // TML : TODO : v�rifier
        	case 2:
                //System Paint Server
        		readAttr_custom_byteAlignedString();
                break;
            case 3:
                // extension
                readAttr_custom_extension();
                break;
            }
        }
        return Integer.MIN_VALUE;
    }
	
	private void readAttr_custom_path(Value v) throws Exception
	{
		v.m_polygon = new Polygon2D();
	
		readAttr_custom_pointSequence(v.m_polygon);
	    
	    v.m_types = new byte[readVlcInt(5)];
	    for (int i = 0; i < v.m_types.length; i++) 
	    {
	        v.m_types[i] = segTypeToPathType[readBits(5)];
	    }
	}	
	
//	 TML : TODO : finish !
	private void readAttr_custom_rare(GraphObject node) throws IOException { // TML

        int nb = readBits(6);
        for (int i = 0; i < nb; i++) {
            int rareId = readBits(6);
            switch (rareId) {
                case /* RARE_CLASS */ 0:
                    // ((XpropElem) node.classe = TML : TODO 
                	readAttr_custom_byteAlignedString();
                    break;
                case /* RARE_AUDIO_LEVEL */ 1:
                    // ((XpropElem) b).audio_level = TML : TODO
                	/*((RenderedGraphObject)node).m_properties.audio_level = */ 
                	/*((Audio)node).audio_level =*/ readAttr_custom_0to1float();
                    break;
                case /* RARE_COLOR */ 2:
                    ((RenderedGraphObject)node).m_properties.m_color = readAttr_custom_paint();
                    ((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_COLOR, false);
                    break;
                case /* RARE_COLOR_RENDERING */ 3:
                    // ((XpropElem) b).color_rendering =  TML : TODO
                    readBits(2);
                    break;
                case /* RARE_DISPLAY */ 4:
                	((RenderedGraphObject)node).m_properties.m_display = (byte) readBits(5);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_DISPLAY, false);
                    break;
                case /* RARE_DISPLAY_ALIGN */ 5:
                	((RenderedGraphObject)node).m_properties.m_displayAlign = (byte) readBits(3);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_DISPLAYALIGN, false);
                    break;
                case /* RARE_FILL_OPACITY */ 6:
                	((RenderedGraphObject)node).m_properties.m_fillOpacity = (byte) readAttr_custom_0to1float();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_FILLOPACITY, false);
                    break;
                case /* RARE_FILL_RULE */ 7:
                	((RenderedGraphObject)node).m_properties.m_fillRule = (byte) readBits(2);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_FILLRULE, false);
                    break;
                case /* RARE_IMAGE_RENDERING */ 8:
                    // ((XpropElem) b).image_rendering = TML : TODO
                    readBits(2);
                    break;
                case /* RARE_LINE_INCREMENT */ 9:
                    // ((XpropElem) b).line_increment = SDLClasses.lineIncrementType(br, "line-increment"); TML : TODO 
                	readLineIncrementType();
                	break;
                case /* RARE_POINTER_EVENTS */ 10:
                	((RenderedGraphObject)node).m_properties.m_pointerEvents = (byte) readBits(4);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_POINTEREVENTS, false);
                    break;
                case /* RARE_SHAPE_RENDERING */ 11:
                    // ((XpropElem) b).shape_rendering = TML : TODO
                    readBits(3);
                    break;
                case /* RARE_SOLID_COLOR */ 12:
                    // ((XpropElem) b).solid_color = TML : TODO
                    readAttr_custom_paint();
                    break;
                case /* RARE_SOLID_OPACITY */ 13:
                    // ((XpropElem) b).solid_opacity = TML : TODO
                    readAttr_custom_0to1float();                    
                    break;
                case /* RARE_STOP_COLOR */ 14:
                	readAttr_custom_paint();
//                	((RenderedGraphObject)node).m_properties.m_stopColor = readAttr_custom_paint();
//                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STOPCOLOR, false);
                    break;
                case /* RARE_STOP_OPACITY */ 15:
                	((RenderedGraphObject)node).m_properties.m_stopOpacity = (byte) readAttr_custom_0to1float();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STOPOPACITY, false);
                    break;
                case /* RARE_STROKE_DASHARRAY */ 16:
                    if (!readBit()) {
                        int len = readVlcInt(5);
                        ((RenderedGraphObject)node).m_properties.m_strokeDasharray = new int[len];
                        for (int ii = 0; ii < len; ii++) {
                        	((RenderedGraphObject)node).m_properties.m_strokeDasharray[ii] = readAttr_custom_fixed_16_8();
                        }
                    }
                    ((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STROKEDASHARRAY, false);
                    break;
                case /* RARE_STROKE_DASHOFFSET */ 17:
                	((RenderedGraphObject)node).m_properties.m_strokeDashoffset = readAttr_custom_fixed_16_8i();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STROKEDASHOFFSET, false);
                    break;
                case /* RARE_STROKE_LINECAP */ 18:
                	((RenderedGraphObject)node).m_properties.m_strokeLinecap = (byte) readBits(2);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STROKELINECAP, false);
                    break;
                case /* RARE_STROKE_LINEJOIN */ 19:
                	((RenderedGraphObject)node).m_properties.m_strokeLinejoin = (byte) readBits(2);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STROKELINEJOIN, false);
                    break;
                case /* RARE_STROKE_MITERLIMIT */ 20:
                	((RenderedGraphObject)node).m_properties.m_strokeMiterlimit = readAttr_custom_fixed_16_8i();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STROKEMITERLIMIT, false);
                    break;
                case /* RARE_STROKE_OPACITY */ 21:
                	((RenderedGraphObject)node).m_properties.m_strokeOpacity = (byte) readAttr_custom_0to1float();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STROKEOPACITY, false);
                    break;
                case /* RARE_STROKE_WIDTH */ 22:
                	((RenderedGraphObject)node).m_properties.m_strokeWidth = readAttr_custom_fixed_16_8i();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_STROKEWIDTH, false);
                    break;
                case /* RARE_TEXT_ANCHOR */ 23:
                	((RenderedGraphObject)node).m_properties.m_text_anchor = (byte) readBits(2);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_TEXTANCHOR, false);
                    break;
                case /* RARE_TEXT_RENDERING */ 24:
                    // ((XpropElem) b).text_rendering = TML : TODO
                    readBits(3);
                    break;
                case /* RARE_VIEWPORT_FILL */ 25:
                	int color = readAttr_custom_paint();
                	if (node instanceof Svg)
                		((Svg)node).m_backgroundFill = color;
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_VIEWPORTFILL, false);
                    break;
                case /* RARE_VIEWPORT_FILL_OPACITY */ 26:
                	byte fill = (byte) readAttr_custom_0to1float();
                	if (node instanceof Svg)
                		((Svg)node).m_backgroundFillOpacity = fill;
                	 ((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_VIEWPORTFILLOPACITY, false); 
                    break;
                case /* RARE_VECTOR_EFFECT */ 27:
                    // ((XpropElem) b).vector_effect = TML : TODO 
                    readBits(4);
                    break;
                case /* RARE_VISIBILITY */ 28:
                	((RenderedGraphObject)node).m_properties.m_visibility = (byte) readBits(2);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_VISIBILITY, false);
                    break;
                case /* RARE_REQUIREDEXTENSIONS */ 29:
                	((RenderedGraphObject)node).m_properties.m_requiredExtensions = readAttr_custom_byteAlignedString();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_REQUIREDEXTENSIONS, false);
                    break;
                case /* RARE_REQUIREDFEATURES */ 30:
                    int len = readVlcInt(5);
                    ((RenderedGraphObject)node).m_properties.m_requiredFeatures = new byte[len];
                    for (int ii = 0; ii < len; ii++) {
                    	((RenderedGraphObject)node).m_properties.m_requiredFeatures[ii] = (byte) readBits(6);
                    }
                    ((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_REQUIREDFEATURES, false);
                    break;
                case /* RARE_REQUIREDFORMATS */ 31:
                	((RenderedGraphObject)node).m_properties.m_requiredFormats = readAttr_custom_byteAlignedString();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_REQUIREDFORMATS, false);
                    break;
                case /* RARE_SYSTEMLANGUAGE */ 32:
                	((RenderedGraphObject)node).m_properties.m_systemLanguage = readAttr_custom_byteAlignedString();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_SYSTEMLANGUAGE, false);
                    break;
                case /* RARE_XML_BASE */ 33:
                	((RenderedGraphObject)node).m_properties.m_xmlbase = readAttr_custom_byteAlignedString();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_XML_BASE, false);
                    break;
                case /* RARE_XML_LANG */ 34:
                	((RenderedGraphObject)node).m_properties.m_xmllang = readAttr_custom_byteAlignedString();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_XML_LANG, false);
                    break;
                case /* RARE_XML_SPACE */ 35:
                	readBits(1); // TML : TODO : A utiliser !!
//                    ((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_XML_SPACE, false);
                    break;
                case /* RARE_FOCUSNEXT */ 36:
                    // b.navNext = SDLClasses.focusType(br, "navNext"); TODO : TML
                    readCustom_focus("navNext");
                    break;
                case /* RARE_FOCUSNORTH */ 37:
                    // b.navUp = SDLClasses.focusType(br, "navUp"); TODO : TML
                    readCustom_focus("navUp");
                    break;
                case /* RARE_FOCUSNORTHEAST */ 38:
                    // b.navUpRight = SDLClasses.focusType(br, "navUpRight"); TODO : TML
                    readCustom_focus("navUpRight");
                    break;
                case /* RARE_FOCUSNORTHWEST */ 39:
                    // b.navUpLeft = SDLClasses.focusType(br, "navUpLeft"); TODO : UML
                    readCustom_focus("navUpLeft");
                    break;
                case /* RARE_FOCUSPREV */ 40:
                    // b.navPrev = SDLClasses.focusType(br, "navPrev"); TODO : UML
                    readCustom_focus("navPrev");
                    break;
                case /* RARE_FOCUSSOUTH */ 41:
                    // b.navDown = SDLClasses.focusType(br, "navDown"); TODO : UML
                    readCustom_focus("navDown");
                    break;
                case /* RARE_FOCUSSOUTHEAST */ 42:
                    // b.navDownRight = SDLClasses.focusType(br, "navDownRight"); TODO : UML
                    readCustom_focus("navDownRight");
                    break;
                case /* RARE_FOCUSSOUTHWEST */ 43:
                    // b.navDownLeft = SDLClasses.focusType(br, "navDownLeft"); TODO : UML
                    readCustom_focus("navDownLeft");
                    break;
                case /* RARE_FOCUSWEST */ 44:
                    // b.navLeft = SDLClasses.focusType(br, "navLeft"); TODO : UML
                    readCustom_focus("navLeft");
                    break;
                case /* RARE_FOCUSABLE */ 45:
                	((RenderedGraphObject)node).m_properties.m_focusable = (byte) readBits(2);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_FOCUSABLE, false);
                    break;
                case /* RARE_FOCUSEAST */ 46:
                    // b.navRight = SDLClasses.focusType(br, "navRight"); TODO : TML
                    readCustom_focus("navRight");
                    break;
                case /* RARE_TRANSFORM */ 47:
                    // TODO : TML: v�rifier que c'est bon
                    try {
                    	((RenderedGraphObject)node).m_definedTransform = readCustom_Matrix();
					} catch (Exception e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
                    break;
                    /*case 48: TML : TODO 
                    case 49:*/
                case /* RARE_FONT_VARIANT */ 50:
                	((RenderedGraphObject)node).m_properties.m_fontVariant = (byte) readBits(2);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_FONTVARIANT, false);
                    break;
                case /* RARE_FONT_FAMILY */ 51:
                	((RenderedGraphObject)node).m_properties.m_fontFamily = readFontFamily();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_FONTFAMILY, false);
                    break;
                case /* RARE_FONT_SIZE */ 52:
                	((RenderedGraphObject)node).m_properties.m_fontSize = readAttr_custom_fixed_16_8i();
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_FONTSIZE, false);
                    break;
                case /* RARE_FONT_STYLE */ 53:
                	((RenderedGraphObject)node).m_properties.m_fontStyle = (byte) readBits(3);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_FONTSTYLE, false);
                    break;
                case /* RARE_TEXT_DECORATION */ 48:
                    // ((XpropElem) b).text_decoration = TML : TODO
                    readBits(4);
                    break;
                case /* RARE_FONT_WEIGHT */ 54:
                	((RenderedGraphObject)node).m_properties.m_fontWeight = (byte) readBits(4);
                	((RenderedGraphObject)node).m_properties.setInheritance(Value.UPDATE_FONTWEIGHT, false);
                    break;
                case /* RARE_HREF_TITLE */ 55:
                    // ((Ixlink) b).setTitle(SDLClasses.class_byteAlignedString(br, "title")); TML : TODO
                	readAttr_custom_byteAlignedString();
                    break;
                case /* RARE_HREF_TYPE */ 56:
                    // ((Ixlink) b).setXlinkType( TML : TODO
                    readBits(3);
                    break;
                case /* RARE_HREF_ROLE */ 57:
                    // ((Ixlink) b).setRole(SDLClasses.class_byteAlignedString(br, "role")); TML : TODO
                	readAttr_custom_anyURI();
                    break;
                case /* RARE_HREF_ARCROLE */ 58:
                    // ((Ixlink) b).setArcrole(SDLClasses.class_byteAlignedString(br, "arcrole")); TML : TODO
                	readAttr_custom_anyURI();
                    break;
                case /* RARE_HREF_ACTUATE */ 59:
                    // ((Ixlink) b).setActuate( TML : TODO
                    readBits(2);
                    break;
                case /* RARE_HREF_SHOW */ 60:
                    // ((Ixlink) b).setShow TML : TODO 
                    readBits(3);
                    break;
                case /* RARE_END */ 61: 
                	/* TML : TODO 
                    if (b instanceof Xset) {
                        ((Xset)b).end = SDLClasses.timesType(br, "end");
                    } else {
                        ((Xaudio)b).end = SDLClasses.timesType(br, "end");                        
                    } */
                    readAttr_smil_times(node, true);
                    break;
                case /* RARE_MAX */ 62:
                    // ((Xset) b).max = SDLClasses.class_time(br, "max", true); TML : TODO 
                	readTime();
                    break;
                case /* RARE_MIN */ 63:
                    // ((Xset) b).min = SDLClasses.class_time(br, "min", true); TML : TODO 
                	readTime();
                    break;
            }
        }
    }
	
	private void readAttr_custom_rareTwo(GraphObject node) throws IOException { // TML

        int nb = readBits(2);
        for (int i = 0; i < nb; i++) {
            int rareId = readBits(2);
            switch (rareId) {
                case /* RARE_SYNCMASTER */ 0:
                	boolean b = false;
                	break;
                case /* RARE_FOCUSHIGHLIGHT */ 1:
                	// 	Enumeration: auto{0} none {1}
                		/*int focusHighlight =*/ readBits(2);
                	break;
                case /* RARE_INITIALVISIBILITY */ 2:
                		// Enumeration: whenStarted{0} always {1}
                    	/*int focusHighlight =*/ readBits(2);
                    break;
            }
        }
	}
	
	public Value readAttr_custom_updateValue(int type, int index) throws Exception {
        Value v = new Value();
        v.m_updateValueType = type;
        if (index >= 0) {
            switch (type) {
                case Value.ITYPE_point:
                	v.m_point = new Point2D(readAttr_custom_coordinate(), readAttr_custom_coordinate());
                    break;
                case Value.ITYPE_int:
                    // small int
                    v.m_enumOrColorOrRealOrID = readBits(8);
                    break;
                case Value.ITYPE_float:
                    // float
                    v.m_enumOrColorOrRealOrID = readAttr_custom_fixed_16_8();
                    break;
                    
                case Value.ITYPE_keyTime: // TML : TODO : timevalue
                	if(readBit()) {
                		if(readBit()) {
                			/*v.timeValue = 0;*/
                		}
                		else {
                			/*v.timeValue = 1;*/
                		}
                	}
                	else {
                		/*v.timeValue = */readBits(12);
                	}
                    break;
                case Value.ITYPE_smil_time:
                	v.m_floatsOrInts = readAttr_smil_times(null, false);
                	break;
                case Value.ITYPE_0to1:
                    /*v.m_valueX = */readAttr_custom_0to1float();
                	/*v.m_valueX = */readAttr_custom_0to1float();
                    break;
                case -1: // error
                    throw new RuntimeException("type error in update");
                default:
                	readAttr_custom_extension();
                    break;
            }
            return v;
        } else {
            switch (type) {
                case Value.TYPE_boolean:
                    v.m_boolean = readBit();
                    break;
                case Value.TYPE_enum:
                	if (!readBit()) {
                		v.m_enumOrColorOrRealOrID = readVlcInt(5);
                    }
                    break;
                case Value.TYPE_index: // TML : TODO : affectation
                	if(!readBit()){
                    	if (readBit()) {
                            v.m_escapeEnum = readBits(2);
                        }
                    	else {
                            v.m_enumOrColorOrRealOrID = readVlcInt(5);
                    	}
                    }
                	break;
                case Value.TYPE_ID:
                	if(!readBit()){
                    	if (readBit()) {
                            v.m_escapeEnum = readBits(2);
                        }
                    	else {
                            v.m_enumOrColorOrRealOrID = readVlcInt(5);
                    	}
                    }
                    break;
                case Value.TYPE_color:
                    v.m_enumOrColorOrRealOrID = readAttr_custom_paint();
                    break;
                case Value.TYPE_fraction:
                	if(!readBit()){
                		v.m_enumOrColorOrRealOrID = readAttr_custom_0to1float();
                	}
                    break;
                case Value.TYPE_float:
                	if(!readBit()){
                    	if (readBit()) {
                            v.m_escapeEnum = readBits(2);
                        }
                    	else {
                    		v.m_enumOrColorOrRealOrID = readAttr_custom_fixed_16_8();
                    	}
                    }
                    break;
                case Value.TYPE_time:
                    v.m_enumOrColorOrRealOrID = readAttr_time();
                    break;
                case Value.TYPE_point:
                    v.m_point = new Point2D(readAttr_custom_coordinate(), readAttr_custom_coordinate());
                    break;
                case Value.TYPE_matrix:
                    v.m_matrix = readCustom_Matrix();
                    break;
                case Value.TYPE_string:
                    v.m_string = readAttr_custom_byteAlignedString();
                    break;
                case Value.TYPE_points:
                    v.m_polygon = new Polygon2D();
                    readAttr_custom_pointSequence(v.m_polygon);
                    break;
                case Value.TYPE_path:
                    /*v.m_path =*/ 
                	v.m_path.m_polygon = new Polygon2D();
                	readAttr_custom_pointSequence(v.m_path.m_polygon);
            		v.m_types = new byte[readVlcInt(5)];
            		for (int i = 0; i < v.m_types.length; i++) {
            			v.m_types[i] = (byte) readBits(5);
            		}
                    break;
                case Value.TYPE_ints:
                    int nb = readVlcInt(5);
                    v.m_floatsOrInts = new int[nb];
                    for (int i = 0; i < nb; i++) {
                        v.m_floatsOrInts[i] = readBits(8);
                    }
                    break;
                case Value.TYPE_floats: // floatValues
                    if(!readBit()){
                    	if (readBit()) {
                            v.m_escapeEnum = readBits(2);
                        }
                    	else {
                    		int nb1 = readVlcInt(5);
                            v.m_floatsOrInts = new int[nb1];
                            for (int i = 0; i < nb1; i++) {
                                v.m_floatsOrInts[i] = readAttr_custom_fixed_16_8();
                            }
                    	}
                    }
                    break;
                case Value.TYPE_keyTimes:
                	/*v.m_timeValue = */read_float12ListType("keyTimes");
                	break;
                case Value.TYPE_smil_times:
                    v.m_floatsOrInts = readAttr_smil_times(null, false);
                    break;
                case Value.TYPE_unit: // unitValue
                    v.m_floatsOrInts = read_codec_valueWithUnits();
                    break;
                case Value.TYPE_URI:
                    v.m_uri = readAttr_custom_anyURI();
                    break;
                case Value.TYPE_event:
                    v.m_event = readAttr_custom_event();
                    break;
                case Value.TYPE_scale:
                	/*v.m_scaleX = */readAttr_custom_fixed_16_8(); // TML : TODO : affectation
                	/*v.m_scaleY = */readAttr_custom_fixed_16_8(); // TML : TODO : affectation
                	break;
                case Value.TYPE_coordinate:
                	/* v.m_coord = */readAttr_custom_coordinate(); // TML : TODO : affectation
                	break;
                case -1: // error
                    throw new RuntimeException("type error in update");
                default:
                    readAttr_custom_extension();
            }
            return v;
        }
    }
	
	private int readAttr_directionType() throws IOException
	{
//		 Enumeration: down{0} left{1} right{2} up{3}
		  return readBits(4);
	}
    
    private boolean readAttr_editable() throws IOException
    {
    	return readBit();
    }
    
    private int[] readAttr_floatList() throws IOException
    {
    	int max = readVlcInt(5);
    	int[] item = new int[max];
    	
    	for(int i = 0 ; i < max ; i++)
    	{
    		item[i] = readAttr_custom_fixed_16_8();
    	}
    	return item;
    }
	
//	 TML : TODO : return value ??
    private void readAttr_overlay() throws IOException
    {
        if(!readBit()) // choice case 0
        {
            /* return */ readAttr_custom_extension();
        }
       	/* return */ readBit();
    }
    
//  TML : TODO : r�cup�rer les valeurs : cf. ci-dessus
	private void readAttr_preserveAspectRatio(GraphObject node) throws IOException // TML : TODO : recuperer les valeurs de l'aspect
	{
		if(!readBit())
		{
			if(!readBit())
			{
				readBits(4);
			}
			else
			{
				readBits(4);
			}
		}
		else
		{
			readBits(5);
		}
	}
    
//	 TML : TODO : return value
    private void readAttr_Point(Point2D size) throws IOException
    {  
    	size.x_r = readAttr_custom_coordinate();
    	size.y_r = readAttr_custom_coordinate();
    }
    
//	 TML : TODO : Verify with spec pb
    private void readAttr_custom_pointSequence(Polygon2D polygon) throws Exception {
        int len = readVlcInt(5);
        polygon.m_points.ensureCapacity(len);
        if (Debug.DEBUG_SHAPE_POLYGON) System.out.println("    number of points of polygon : " + len);
        if (!readBit()) {
            /*
               If the number of points is less than 3, a simple encoding is used.
               A number of 5 bits give the number of bits on which the coordinates are encoded.
               */
            if (len < 3) {
                int bits = readBits(5); // number of bits
                for (int i = 0; i < len; i++) {
                    polygon.m_points.addElement(new Point2D(translate(readBits(bits), bits), translate(readBits(bits), bits)));
                }
            } else {
                /*
                    If the number of points LEN in the polygon are more than 3, then:
                    - the number N1 is read on 5 bits
                    - the first two coordinates are read from the stream using N1 bits
                    - the number NX and NY are read on 5 bits
                    - then LEN-1 points are read, using NX points for the X value and NY bits for the Y value, both X and Y
                    being differences from the previous point
                    */
                int bits = readBits(5); // number of bits for the first
                int pointX = translate(readBits(bits), bits);
                int pointY = translate(readBits(bits), bits);
                polygon.m_points.addElement(new Point2D(pointX, pointY));
                int bitsX = readBits(5); // number of bits for X
                int bitsY = readBits(5); // number of bits for Y
                for (int i = 1; i < len; i++) {
                    pointX = Real.applyAdd(pointX, translate(readBits(bitsX), bitsX));
                    pointY = Real.applyAdd(pointY, translate(readBits(bitsY), bitsY));
                    polygon.m_points.addElement(new Point2D(pointX, pointY));
                }
            }
        } 
        else 
        {
        	if (this.m_decodingContext.m_pointsCodec == 0)
        	{// pointsCodec is a LASeRHeader attribute
/*        		int kvalue = readBits(4);
        		int bits = readBits(5);
        		
        		int pointX = translate(readBits(bits), bits);
                int pointY = translate(readBits(bits), bits);
                polygon.m_points.addElement(new Point2D(pointX, pointY));
                
                int XMvalue = 0;
                int YMvalue = 0;
                int CodeNum = 0;
                int Diff = 0;
                for(int i = 1 ; i < len ; i++) 
                {
                	// to calculate X point
                	do 
                	{
                		//bitX = readBit();
                		XMvalue ++;
                	} while (!readBit());
                	
                	readBit(); 
                	int endX = 1;
                	
                	int INFO_dx = readBits(XMvalue + kvalue);
                
                	CodeNum = GetCodeNum(kvalue, XMvalue, INFO_dx);
                	Diff = GetDiff(CodeNum);
                	((Point2D)polygon.m_points.elementAt(i)).x_r = ((Point2D)polygon.m_points.elementAt(i-1)).x_r + Diff;
                	//x[i] = x[i-1] + Diff;
                	// to calculate Y point
                	do 
                	{
                		YMvalue ++;
                	} while (!readBit());
            
                	readBit();
                	
                	int endY = 1;
                
                	int INFO_dy = readBits(YMvalue + kvalue);
                
                	CodeNum = GetCodeNum(kvalue, YMvalue, INFO_dy);
                	Diff = GetDiff(CodeNum);
                	((Point2D)polygon.m_points.elementAt(i)).y_r = ((Point2D)polygon.m_points.elementAt(i-1)).y_r + Diff;
                }*/
        	} 
        	else 
        	{
        		readAttr_custom_extension();//attr_custom_extension ext; 
        		//throw new RuntimeException("this kind of point is unimplemented");
        	}
        }
    }
    
    private int readAttr_repeatCount() throws IOException
    {
    	if(!readBit())
    		return this.readAttr_custom_fixed_16_8();
    		
    	return SVGset.REPEAT_INDEFINITLY;
    }
    
    // TML NEW METHODE
    private int readAttr_repeatDur() throws IOException {
    	if (!readBit()) { // case 0
            return translateTimeUponReading(readVlcInt(5), true); // TML : TODO : verify 2nd param always true in the decoder
        }
    	return BasicRendering.TIME_INDEFINITE;
    }
    
    private int readAttr_restart() throws IOException
    {
    	return readBits(2);
    }
    
    private int readAttr_rotate() throws IOException
    {
    	if (readBit()) {
			 return (readBit() ? 0 : 1); // TML : TODO : verify the 0 or 1 value to return
		} else {
			return readAttr_custom_fixed_16_8();
		}
    }
    
    private int readAttr_rotscatra() throws IOException
    {
//    	 Enumeration: rotate{0} scale{1} skewX{2} skewY{3} translate{4}
    	return readBits(3);
    }
    
//  TML : TODO ; retrun value
    private String readAttr_script () throws IOException
    {
	    if(!readBit()) // choice cas 0
	    	return readAttr_custom_byteAlignedString();
	    
	    readBit();
	    return "";
	}
    
    private int readAttr_scrollStopType() throws IOException
    {
    	if(!readBit()) // choice 0
    	{
    		return readAttr_custom_IDREF();
    	}
    	
    	int x = readAttr_custom_fixed_16_8();
    	int y = readAttr_custom_fixed_16_8();
    	
    	return x; // TML : TODO : return value ?
    }
    
    private int[] readAttr_smil_times (GraphObject node, boolean isForRareEnd) throws IOException
    {
    	int SmilTimes[];
    	
    	if(!readBit()) // choice case 0
    	{
    		int max = readVlcInt(5);
    		SmilTimes = new int[max];
    		for(int i = 0 ; i < max ; i++)
    		{
    			SmilTimes[i] = readCustom_smil_time(node, isForRareEnd);
    		}
    		return SmilTimes;
    	}
    	else
    	{
    		SmilTimes = new int[1];
    		
    		if(node instanceof Conditional)
    			((Conditional)node).m_begin = BasicRendering.TIME_INDEFINITE;
    		else if(node instanceof Video)
    			((Video)node).m_beginTimes.addElement(new Integer(BasicRendering.TIME_INDEFINITE));
    		else if(node instanceof Audio)
    			((Audio)node).m_beginTimes.addElement(new Integer(BasicRendering.TIME_INDEFINITE));
    		else if(node instanceof SVGset)
    			((SVGset)node).m_begin = BasicRendering.TIME_INDEFINITE;
    		
    		SmilTimes[0] = 0;//indefinite{0}
    		
    		return SmilTimes;
    	}
    }
    
    private URI[] readAttr_sourcesType() throws IOException
    {
    	int occ = readVlcInt(5);
    	URI[] sources = new URI[occ];

    	for(int i = 0 ; i < occ ; i++)
    	{
    		sources[i] = readAttr_custom_anyURI();
    	}
    	return sources;
    }
    
    private int readAttr_streamSourceModeType() throws IOException
    {
    	// Enumeration: keepOld{0} playList{1} replace{2} useOld{3}
    	return readBits(3);
    }

    private int readAttr_syncBehavior() throws IOException
    {
    	return readBits(2);
    }
       
    private int readAttr_syncTolerance() throws IOException {
        int sw = readBits(1);
        switch (sw) {
            case 0:
                return readVlcInt(5);
            case 1:
                return -1;
        }
        return Integer.MIN_VALUE;
    }     
    
    private int readAttr_time () throws IOException {
        if (!readBit()) { // case 0
            boolean b = readBit();
            int a = translateTimeUponReading(readVlcInt(5), true); // TML : TODO : verify 2nd param always true in the decoder
            return (b ? -a : a);
        }
        switch(readBits(2)) {
        //Enumeration: auto{0} indefinite{1} media{2} none{3}
        	case 0:
        		return Integer.MIN_VALUE+1; // auto
            case 1:
	            return Integer.MIN_VALUE; // indefinite
	        case 2:
	            return Integer.MAX_VALUE; // media
	        case 3:
	        	return Integer.MAX_VALUE; // none TML TODO : VALUE TO RETURN ?
	        
	    }
        return Integer.MIN_VALUE;
    }
    
    private int readAttr_transformBehavior() throws IOException
    {
        // Enumeration: geometric{0} pinned{1} pinned_180{2} pinned_270{3} pinned_90{4}
        return readBits(4);
    }
    
    private int readCustom_focus (String msg) throws IOException {
        int sw = readBits(1);
        switch (sw) { // TML : TODO : invers� !?
            case 1:
                return Integer.MAX_VALUE - readBits(1);
            case 0:
                return readAttr_custom_IDREF();
        }
        return Integer.MIN_VALUE;
    }
    
    public Matrix2D readCustom_Matrix() throws Exception {
    	int xx_r, yy_r, xy_r, yx_r, xz_r, yz_r;
        if (!readBit()) {
            if (readBit()) {
                xx_r = translateScale(readBits(m_decodingContext.m_coordBits + m_decodingContext.m_scaleBits_minus_coordBits));
                yy_r = translateScale(readBits(m_decodingContext.m_coordBits + m_decodingContext.m_scaleBits_minus_coordBits));
            } else {
                xx_r = Real.convertIntToRealInt(1);
                yy_r = Real.convertIntToRealInt(1);
            }
            if (readBit()) {
                xy_r = translateScale(readBits(m_decodingContext.m_coordBits + m_decodingContext.m_scaleBits_minus_coordBits));
                yx_r = translateScale(readBits(m_decodingContext.m_coordBits + m_decodingContext.m_scaleBits_minus_coordBits));
            } else {
                xy_r = Real.convertIntToRealInt(0);
                yx_r = Real.convertIntToRealInt(0);
            }
            if (readBit()) {              
                xz_r = translate(readBits(m_decodingContext.m_coordBits + m_decodingContext.m_scaleBits_minus_coordBits));//, m_decodingContext.m_coordBits + m_decodingContext.m_scaleBits_minus_coordBits);
                yz_r = translate(readBits(m_decodingContext.m_coordBits + m_decodingContext.m_scaleBits_minus_coordBits));//, m_decodingContext.m_coordBits + m_decodingContext.m_scaleBits_minus_coordBits);
               
            } else {
                xz_r = Real.convertIntToRealInt(0);
                yz_r = Real.convertIntToRealInt(0);
            }
        } else {
        	readAttr_custom_extension();
            return null;
        }
        return new Matrix2D(xx_r, xy_r, xz_r, yx_r, yy_r, yz_r);
    }
    
//  TML : TODO : affecter les valeurs ? + return
    private int readCustom_smil_time(GraphObject node, boolean isForRareEnd) throws IOException
    {
    	Listener listener = null;
    	if(readBit()) // has_event
    	{
     		URI listenerURI = new URI();
    		if(readBit()) // has_indentifier
    		{
    			listenerURI.id = readAttr_custom_IDREF(); // attr_custom_idref
    		}
    		listener = CreateListener(readAttr_custom_event(), listenerURI, node);
    		
    		if(node instanceof Conditional)
    		{
    			if(listener != null)
    			{
    				GraphObject initiator = this.m_graphObjectManager.getNode(listenerURI.id);
    				if(initiator instanceof Conditional)
    					((Conditional)initiator).m_EventPost = (byte)listener.event.event;
    			}
    		}
    		/*
    		if(node instanceof Conditional)
    			((Conditional)node).m_begin = (((Conditional)node).m_begin <= 0)? BasicRendering.TIME_INDEFINITE : ((Conditional)node).m_begin;
    		else if(node instanceof Video)
    			((Video)node).m_beginTimes.addElement(new Integer((((Video)node).getBeginTime() <= 0)? BasicRendering.TIME_INDEFINITE : ((Video)node).getBeginTime()));
    		else if(node instanceof Audio)
    			((Audio)node).m_beginTimes.addElement(new Integer((((Audio)node).getBeginTime() <= 0)? BasicRendering.TIME_INDEFINITE : ((Audio)node).getBeginTime()));
    		else if(node instanceof SVGset)
    			((SVGset)node).m_begin = (((SVGset)node).m_begin <= 0)? BasicRendering.TIME_INDEFINITE : ((SVGset)node).m_begin;
*/
/*    		if(node instanceof Conditional)
    			((Conditional)node).m_begin = BasicRendering.TIME_INDEFINITE;
    		else if(node instanceof Video)
    			((Video)node).m_beginTimes.addElement(new Integer(BasicRendering.TIME_INDEFINITE));
    		else if(node instanceof Audio)
    			((Audio)node).m_beginTimes.addElement(new Integer(BasicRendering.TIME_INDEFINITE));
    		else if(node instanceof SVGset)
    			((SVGset)node).m_begin = BasicRendering.TIME_INDEFINITE;
  */  			
    	}
    	if (readBit()) // has_clock
    	{
    		int time;
    		if(readBit()) // sign
    			time = -readVlcInt(5);
    		else
    			time = readVlcInt(5);
    		
    		time = translateTimeUponReading(time, true); // decoder false;
    		
    		if(node instanceof Conditional)
    		{
    			if(listener != null)
    			{
    				listener.m_delay = time;
    				((Conditional)node).m_EventPost = (byte)listener.event.event;
    			}
    			else
    			{
        			((Conditional)node).m_begin = time; // clock
        		}
    		}
    		else if(node instanceof Video)
    			((Video)node).m_beginTimes.addElement(new Integer(time));
    		else if(node instanceof Audio)
    			((Audio)node).m_beginTimes.addElement(new Integer(time));
    		else if(node instanceof SVGset)
    		{
    			if(listener != null)
    			{
    				listener.m_delay = time;
    				if(isForRareEnd)
    					listener.m_beginOrEnd =  false;
    			}
    			else
    			{
        			if(isForRareEnd)
        				((SVGset)node).m_end = time;
        			else
        				((SVGset)node).m_begin = time; // clock
        		}
    		}
    		else
    			//throw new RuntimeException("smil type cast error (read_CustomSmilTime)");
    			;
    	}
    	return 0;
    }
    
///////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////// END Attributes parsing operations ///////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////
	
    protected void readAnimateType(SVGanimate au) throws Exception 
    {
    	// TML
    	if(readBit())
    		au.m_accumulate = readBit();
    	
    	if(readBit())
    		au.m_additive = readBit();
    	
    	// by
    	if (readBit()) 
    	{
    		au.m_by = new Value();
    		au.m_by.m_animValueType = readBits(4);
    		readAttr_custom_AnimatedValue(au.m_by);
    	}
    	
    	// calcmode
    	if (readBit())
    		au.m_calcMode = (byte) readBits(2);
    	
    	// from
    	if (readBit()) 
    	{
			au.m_from = new Value();
			au.m_from.m_animValueType = readBits(4);
			readAttr_custom_AnimatedValue(au.m_from);
		}
    	
//    	 Key splines
        if (readBit())
        	au.m_keySplines = read_float12ListType("keySplines");
    	
        // Key times
        if (readBit())
        	au.m_keyTimes = read_float12ListType("keyTimes");
        
        // has values
        if (readBit()) 
        {
        	readAttr_custom_AnimatedValues(au);
        }
    }
    
//  TML : TODO : must be verify 
    private void readSetType(GraphObject parent, SVGset au, boolean isForAnimateMotion) throws Exception 
    {
    	// attr_attributeType
    	if(readBit())
    		readBits(2);
    	
    	// begin
        if(readBit())
        	readAttr_smil_times(au, false); // TML : TODO : A verifier ce qu il faut retourner
                
        if(readBit())
        	au.m_dur = readAttr_time();
        
        if(readBit()) // attr_fill
        	au.m_freeze = !readBit();
        
        if(readBit())
        	au.m_repeatCount = readAttr_repeatCount();
        // TML : TODO :  
        // -1 because m_simpleAnimCount = 0; after m_simpleAnimCount++ in genericStartAnimation 
        // -1 
/*        
        else
            au.m_repeatCount = SVGset.REPEAT_INDEFINITLY;
*/
        // TML : TODO : Verify with spec no readAttr_repeatDur specify
        if(readBit())
        	au.m_repeatDur = readAttr_repeatDur();
/*
        else
        	au.m_repeatDur = BasicRendering.TIME_INDEFINITE;
*/      
        if(readBit())
        	au.m_restart = (byte) readBits(2);
        if (au.m_restart == 3)
            au.m_restart = SVGset.RESTART_ALWAYS;
        
        if (readBit()) 
        {
            au.m_to = new Value();
            au.m_to.m_animValueType = readBits(4);
            this.readAttr_custom_AnimatedValue(au.m_to);
        }
        
        if (isForAnimateMotion) 
        {
        	// <animateMotion from="15,43" to="65,30"...
        	au.m_attribute = NodeProperties.TRANSFORMTRANSLATE;
        	au.m_attributeType = Value.FLOATS;
        	
        	if (readBit()) {
        		((SVGanimateMotion)au).keyPoints = new Polygon2D();
        		readAttr_custom_pointSequence(((SVGanimateMotion)au).keyPoints);
        		// TML : TODO verify in spec attr_floatList not a readAttr_custom_pointSequence but keyPoints i a polygone not a int[] 
        	}
        	
        	if (readBit()) { // TML TODO : Verify
//        		au.m_attribute = Value.ANIM_D;
        		
        		((SVGanimateMotion)au).polygon = new Polygon2D();
        		readAttr_custom_pointSequence(((SVGanimateMotion)au).polygon);
        		((SVGanimateMotion)au).types = new byte[readVlcInt(5)];
        		for (int i = 0; i < ((SVGanimateMotion)au).types.length; i++) {
        			((SVGanimateMotion)au).types[i] = segTypeToPathType[readBits(5)];
        		}
        	}
        	// has_rotate
        	if (readBit()) {
        		((SVGanimateMotion)au).rotate = readAttr_rotate();
        	}
        }
        
        readHref(au.xlink);
        
        int attach = -1;
        if (au.xlink.getHref() == null) {
            if (parent.getID() == -1) {
                parent.setAutoID();
            }
            attach = parent.getID();
        } else {
            attach = au.xlink.getHref().id;
        }
                	
        m_graphObjectManager.attachAnimation(au, attach);       
    }

    public int[] read_float12ListType(String msg) throws IOException {
        int nb = readVlcInt(5);
        int u[] = new int[nb];
        boolean shortcut;
        for (int i = 0; i < nb; i++) {
            shortcut = readBit();
            if (shortcut) {
                if (readBit())
                    u[i] = 0;
                else
                    u[i] = 4096;
            } else
                u[i] = readBits(12);
        }
        return u;
    }
    
    private void readUpdateListType_content(Conditional node) throws Exception 
    {
    	int len = readVlcInt(5);
    	
    	flush(); // attr_custom_align encoding-align-before; = dummy
/*
     	int occ = readVlcInt(5);
    	for(int i = 0 ; i < occ ; i++)
    	{
    		readUpdates(i);
    	}
*/    	
     	node.m_buffer = readByteArray(len);
    	flush(); // attr_custom_align encoding-align-after; = dummy
    	
    }
       
    private Listener CreateListener(Event ev, URI uri, GraphObject node)
    {
      
    	Listener au = new Listener(m_graphObjectManager);
    	
    	au.setAutoID();
    	
    	au.event = ev;
         
    	URI u = new URI();
    	u.id = node.getID();
 
      	au.handler = u;
  	
      	if(uri.id != Integer.MIN_VALUE)
      	{
	    	au.m_observer = uri.id;
	        m_graphObjectManager.attachListener(au, au.m_observer);
      	}
        
        m_navigator.attachEventListener(au);
        
        au.m_enabled = true;
                
        return au;
    }
   
    

    
    
    
    

    

    

    

    

    
    
    

    public GraphObject readExtend() throws Exception {
        /*trace(STR_BEGIN + "extend");*/
        readBits(m_decodingContext.m_extensionIDbits);
        int len = readVlcInt(5);
        readByteArray(len);
        /*trace(STR_END + "extend");*/
        return null;
    }

    

    // TML
    private void readPrivateElement(Multi priv) throws Exception {
        int tagID = readBits(m_decodingContext.tagIndexBits);
        QName qn = (QName) m_decodingContext.privateTags.elementAt(tagID);
        priv.tag = qn.name;
        priv.ns = qn.ns;
        if(readBit())
        	readAttr_custom_ID(priv);
        int max = readVlcInt(5);
        priv.privateAttributes = new QName[max];
        
        for (int j = 0; j < max; j++) {
            if (!readBit()) {
            	int hasNSbits = log2sup(Math.max(0, m_decodingContext.nbOfAttrs(0) - 1));
                int attrIndex = readBits(hasNSbits);
                priv.privateAttributes[j] = new QName();
                priv.privateAttributes[j].set(m_decodingContext.getAttr(0, attrIndex));
                priv.privateAttributes[j].value = readAttr_custom_byteAlignedString();
            } else {
            	int noNSbits = log2sup(Math.max(0, m_decodingContext.nbOfAttrs(tagID) - 1));
                int attrIndex = readBits(noNSbits);
                priv.privateAttributes[j] = new QName();
                priv.privateAttributes[j].set(m_decodingContext.getAttr(tagID, attrIndex));
                priv.privateAttributes[j].value = readAttr_custom_byteAlignedString();
            }
        }
        if (readBit()) {
            readPrivateChildren(priv);
        }
    }

    

    private void readTime() throws IOException // TML : TODO : on a juste fait la lecture
    {
       if (!readBit()) {
               readBit();
               readVlcInt(5);
           }
       else readBits(2);
    }

    private void readLineIncrementType() throws IOException // TML : TODO : on a juste fait la lecture
    {
           if (readBit())
        	   readAttr_custom_fixed_16_8();
           else
              readBit();
    } 
    
    // TML : TODO : verify
    private void readPrivateAttributeContainer(GraphObject node) throws IOException {

            int ch4 = readBits(4);
            switch (ch4) {
                case 0:
                	readPrivateAnyXMLAttribute(node);
                    break;
                case 1:
                	readPrivateOpaqueAttribute();
                    break;
                case 2:
                	readAttr_any();
                    break;
                default:
                	readAttr_custom_extension();
                    break;
            }
    }
    
    private void readPrivateAnyXMLAttribute(GraphObject node) throws IOException {
    	int skipLen = readVlcInt(5);
    	flush();
    	int count = readVlcInt(5); // TML
    	for (int i = 0; i < count; i++) {
    		 readPrivateAttribute(0, node);
    	  }
    	flush();
    	}
    
    private void readPrivateAttribute(int index, GraphObject node) throws IOException 
    {
    	QName qn = new QName();
		qn = m_decodingContext.getAttr(0, readBits(log2sup(m_decodingContext.nbOfAttrs(0))));
		node.privateAttributes[index].set(qn);
		readAttr_custom_byteAlignedString(); // node.value
    }

    private void readPrivateOpaqueAttribute() throws IOException {
    	readPrivateOpaqueElement();
    	}
    
    // TML
    private void readPrivateAnyXMLElement(GraphObject node) throws Exception 
    {
    	int skipLen = readVlcInt(5);
    	flush();
    	readPrivateChildren(node);
    	flush();
    }
    
    private void readPrivateChildren(GraphObject asc) throws Exception 
    {
        int count = readVlcInt(5);
        for (int i = 0; i < count; i++) {
            if (readBit()) {
                if (readBit()) {
                    asc.appendChild(new TextContent(m_graphObjectManager, readAttr_custom_byteAlignedString()));
                } else {
                    Multi pe = new Multi(m_graphObjectManager, null);
                    readPrivateElement(pe);
                    asc.appendChild(pe);
                }
            } else {
                asc.appendChild(readUpdatableElements(asc));
            }
        }
    }
    
    // TML
    private void readPrivateOpaqueElement() throws IOException
    {
    	int skipLen = readVlcInt(5);
        flush();
        readBits(m_decodingContext.nameSpaceIndexBits);
        flush();
        readBits(skipLen - ((m_decodingContext.nameSpaceIndexBits + 7) >> 3));
    }
    
    // TML
    private void readElement_any() throws IOException
    {
    	readBits(m_decodingContext.m_extensionIDbits);
    	int skipLen = readVlcInt(5);  
		readByteArray(skipLen);
    }
  
     /**
     * Reads the "children" element from the node
     *
     * @param node
     * @throws Exception
     */
    private void readChildren(GraphObject node) throws Exception {
        // hasChildren
        if (readBit()) {
            int len = readVlcInt(5); //vluimsbf
            for (int i = 0; i < len; i++) {
                node.appendChild(readElements(node));
            }
        }
    }

    private int readVlcAttributeIndex() throws IOException {
        int i = readBits(1);
        int nbBits = 1;
        if (i > GENIND_stroke) {
            i = readBits(1) + (i << 1);
            nbBits = 2;
            if (i > GENIND_fill) {
                i = readBits(2) + (i << 2);
                nbBits = 4;
                if (i > GENIND_stroke_width) {
                    i = readBits(1) + (i << 1);
                    nbBits = 5;
                    if (i > GENIND_font_family) {
                        i = readBits(1) + (i << 1);
                        nbBits = 6;
                        if (i > GENIND_stroke_linecap) {
                            i = readBits(1) + (i << 1);
                            nbBits = 7;
                            if (i > GENIND_xml_lang) {
                                i = readBits(1) + (i << 1);
                                nbBits = 8;
                                if (i > GENIND_systemLanguage) {
                                    i = readBits(1) + (i << 1);
                                    nbBits = 9;
                                    if (i > GENIND_xml_space) {
                                        i = readBits(2) + (i << 2);
                                        nbBits = 11;
                                        if (i > GENIND_audio_level) {
                                            i = readBits(1) + (i << 1);
                                            nbBits = 12;
                                            if (i > GENIND_vector_effect) {
                                                i = readBits(1) + (i << 1);
                                                nbBits = 13;
                                                if (i > GENIND_focusable) {
                                                    i = readBits(1) + (i << 1);
                                                    nbBits = 14;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        /*traceBits(nbBits, i, "Generic attribute");*/
        return i;
    }


///////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// Updates parsing operations ///////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * parses the lsr:LASeRHeader element
     *
     * @throws Exception
     */
    public void parseLASeRHeader() throws Exception 
    { // TML
        //TODO add members (DecodingContext) for fields in LASeRHeader
    	m_decodingContext.m_profile = (byte) readBits(8);
    	m_decodingContext.m_level = (byte) readBits(8);
        readBits(3); // TML : reserved
        m_decodingContext.m_pointsCodec = readBits(2);
        m_decodingContext.m_pathComponents = (byte) readBits(4);  // TML : cf. doc, not SDL which was wrong.
        if (m_decodingContext.m_pathComponents == 15) 
        	m_decodingContext.m_pathComponents = -1;
        m_decodingContext.m_useFullRequestHost = readBit(); 
        if (readBit()) 
        	m_decodingContext.m_timeResolution = readBits(16);
        m_decodingContext.m_colorDimBits = readBits(4) + 1;
        m_decodingContext.m_resolution = readBits(4);
        if (m_decodingContext.m_resolution > 8) {
            m_decodingContext.m_resolution -= 16;
        }
        m_decodingContext.m_coordBits = readBits(5); // TML : invert this line and the following one
        m_decodingContext.m_scaleBits_minus_coordBits = readBits(4);
        translateNeg = (1 << m_decodingContext.m_coordBits);
        translateNegLimit = (1 << (m_decodingContext.m_coordBits - 1)) - 1;
        // TML
        m_decodingContext.m_newSceneIndicator = readBit(); // TML : changed from append to newSceneIndicator
        // String IDs
        m_decodingContext.m_hasStringIDs = readBit();
        readBits(2); // TML : reserved
        m_decodingContext.m_extensionIDbits = readBits(4);
        
        // m_decodingContext.m_hasPrivateData = readBit();
        // m_decodingContext.m_hasExtendedAttributes = readBit();
        // Append mode
        //m_decodingContext.m_appendMode = readBit();
        //if (Debug.DEBUG_APPEND_MODE) {
        //    System.out.println("m_appendMode = " + m_decodingContext.m_appendMode);
        //}
       
        // BiM extensions
        if (readBit()) {
        	int len = readVlcInt(5);
        	for (int i = 0; i < len; i++) {
        		readBits(8);
            }
        	//readByteArray(len); // TODO : ExtensionDescriptor
        }
        if (readBit()) {
        	// TML : TODO : hasExtension : is there anything to do with this ? Attention, l'encodeur n'est pas bon !!
        	readAttr_custom_extension();
        }
    }

    /**
     * Read the lsru:InsertIndexed command and processes it.
     *
     * @throws Exception
     */
/*    public void readInsertIndexed() throws Exception {
        
        LaserUpdate update = new LaserUpdate();

        // read the target id
        if (readBit()) {
            update.ref = readVlcInt(5);
        }

        // read the attribute index
        update.index = readIndex();

        // Get attribute type
        update.attributeTag = readBits(8);
        if (update.attributeTag == Value.UPDATE_CHILDREN) {
            // Add Children
            update.objectToAdd = readElements(null);
        } else {
            // read the attribute
            update.updateValue = readAttr_custom_updateValue(Value.updateValueType(update.attributeTag, update.index), update.index);
        }

        // Process
        processInsertIndexed(update);
    }
*/
    /**
     * Read the lsru:SendEvent command and processes it.
     *
     * @throws Exception
     */
    // TML : TODO
    public void readSendEvent() throws Exception {

        // read the event
    	Event event = readAttr_custom_event();
  	
        if (readBit()) {
            boolean neg = readBit();
            int i = readVlcInt(5);
            if (neg) i = -i;
            event.x = i; // TML : pas sur qu'il faille mettre ce r�sultat dans event.x !
        }
    	
        if (readBit())//HasPointValue
        {
            // point value
            event.x = translate(readBits(m_decodingContext.m_coordBits));
            event.y = translate(readBits(m_decodingContext.m_coordBits));
        }
        // read the target id
        int ref = readAttr_custom_IDREF();

        if (readBit())//hasStringValue
        {
            event.string = readAttr_custom_byteAlignedString();
        }
        
    	// has_attr_any
        if (readBit())
        	readAttr_any();

        // Process
        processSendEvent(ref, event);
    }

    /**
     * Read the lsru:Restore command and processes it.
     *
     * @throws Exception
     */
    public void readRestore() throws Exception {

    	String groupID = readAttr_custom_byteAlignedString();
    	
//      has_attr_any
        if (readBit())
        	readAttr_any();
        
        // TML TODO : Need processReplace
        
    }

    /**
     * Read the lsru:Clean command and processes it.
     *
     * @throws Exception
     */
    public void readClean() throws Exception {

    	String groupID = readAttr_custom_byteAlignedString();
    	
    	//    	 has_attr_any
        if (readBit())
        	readAttr_any();
    	
    	// TML : TODO : Need processClean
    }

    /**
     * Read the lsru:save command and processes it.
     *
     * @throws Exception
     */
    public void readSave() throws Exception 
    {
    	String groupID = readAttr_custom_byteAlignedString();
    	
    	readAttr_custom_ElementAttributeList();
        
//   	 has_attr_any
        if (readBit())
        	readAttr_any();
        
//      TML : TODO : Need processSave
    }

    /**
     * Read the lsru:add command and processes it.
     *
     * @throws Exception
     */
    public LaserUpdate readAdd() throws Exception {
        
        LaserUpdate update = new LaserUpdate();

        if(readBit())
        	update.attributeTag = readAttr_AttributeName();
        
        if(readBit())
        	/*au.operandAttribute*/ readAttr_AttributeName();

        if(readBit())
    	    /*au.operandElementId = */readAttr_custom_IDREF();
    		  
        update.ref = readAttr_custom_IDREF();
        
        if(readBit())
        	update.updateValue = readAttr_custom_updateValue(Value.updateValueType(update.attributeTag, -1), -1);
        
        if (readBit())
        	readAttr_any();

        // Process
        processAdd(update);

        return update;
    }

    /**
     * Read the lsru:ReplaceAttribute command and processes it.
     *
     * @throws Exception
     */
/*    public void readReplaceAttribute() throws Exception {
        
        LaserUpdate au = new LaserUpdate();
        au.ref = readVlcInt(5);
        // read the target id

        // read the optionnal index
        au.index = Value.NO_INDEX;
        if (readBit()) {
            au.index = readIndex();
        }

        // read the attribute to replace
        au.attributeTag = readBits(8);

        // Get the target node before reading the value because of the folowing case :
        // if a new node has the same ID than the node to replace the getNode opration
        // doesn't return the proper ID.
        //GraphObject targetNode = m_graphObjectManager.getNode(target);

        // read the attribute value
        au.updateValue = readAttr_custom_updateValue(Value.updateValueType(au.attributeTag, au.index), au.index);

        // Process
        processReplaceAttribute(au);
        
    }
*/
    /**
     * Read the lsru:DeleteIndexed command and processes it.
     *
     * @throws Exception
     */
/*    public void readDeleteIndexed() throws Exception {
        
        LaserUpdate au = new LaserUpdate();
        au.ref = readVlcInt(5);

        // read index
        au.index = readIndex();
        // process
        processDeleteIndexed(au.ref, au.index);
        
    }
*/    
    public void readDiscard(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	if(readBit())
    	{
    		readAttr_smil_times(null, false);
    	}
    	
    	Xlink TMP = new Xlink();
    	readHref(TMP);
    	  
//        has_attr_any
    	if (readBit())
    		readAttr_any();
          
    	// Read base properties
    	readObjectEnd(au);
	}


    /**
     * Read the lsru:ReplaceIndexed command and processes it.
     *
     * @throws Exception
     */
/*    public void readReplaceIndexed() throws Exception {
        if (Debug.DEBUG_PARSER) {
            System.out.println("readReplaceIndexed");
        }
        // read the index
        int index = readIndex();
        // read the target
        int ref = readVlcInt(5);

        // Get the target node before reading the value because of the folowing case :
        // if a new node has the same ID than the node to replace the getNode opration
        // doesn't return the proper ID.
        GraphObject targetNode = m_graphObjectManager.getNode(ref);

        // read the new node
        Object a_node = readElements(null);

        // process the command
        if (targetNode != null) {
            processReplaceIndexed(targetNode, index, a_node);
        }
    }
*/
    /**
     * Read the lsru:Replace command
     *
     * @throws Exception
     */
    // TML : TODO : return values
    public void readReplace() throws Exception 
    {
    	LaserUpdate au = new LaserUpdate();

    	if(readBit())
    		au.attributeTag = readAttr_AttributeName();
    	
        if(readBit())
            au.index = readVlcInt(5);
        
        if(readBit())
        	/*au.operandAttribute*/ readAttr_AttributeName();

        if(readBit())
    	    /*au.operandElementId = */readAttr_custom_IDREF();
    		  
        au.ref = readAttr_custom_IDREF();
    	    	
        if(readBit())
        	au.updateValue = readAttr_custom_updateValue(Value.updateValueType(au.attributeTag, au.index), au.index);
    		  
        if (readBit())
        	readAttr_any();
        
        if(readBit()) 
    	{
    		int len = readVlcInt(5);
    		for(int t = 0 ; t < len ; t++)
    		{
    			au.objectToAdd = readUpdatableElements(null);
    		}
    		processReplace(au);
    	}
        if(au.index != Integer.MIN_VALUE)
        {
        	GraphObject targetNode = m_graphObjectManager.getNode(au.ref);

            // read the new node
            Object a_node = readElements(null);

            // process the command
            if (targetNode != null) {
                processReplaceIndexed(targetNode, au.index, a_node);
            }
        }
        if(au.attributeTag != Integer.MIN_VALUE)
        {
        	processReplaceAttribute(au);
        }
    }

    /**
     * Read the lsru:Delete command and processes it.
     *
     * @throws Exception
     */
    // TML
    public void readDelete() throws Exception 
    {
    	LaserUpdate au = new LaserUpdate();
    	
    	if(readBit())
    		au.attributeTag = readAttr_AttributeName();
    	
    	if(readBit())
    		au.index = readVlcInt(5);
    	 
		au.ref = readAttr_custom_IDREF();
    	
    	if (readBit())
    		readAttr_any();
    	
    	if(au.index != Integer.MIN_VALUE)
    		processDeleteIndexed(au.ref, au.index);
    	else
    		processDelete(au);
    }

    /**
     * Read the lsru:Insert command and processes it.
     *
     * @throws Exception
     */
    public void readInsert() throws Exception 
    {
    	LaserUpdate au = new LaserUpdate();
    	
    	if(readBit())
    		au.attributeTag = readAttr_AttributeName();
    	
    	if(readBit())
    		au.index = readVlcInt(5);
    	 
		au.ref = readAttr_custom_IDREF();
    	
		if(readBit())
			au.updateValue = readAttr_custom_updateValue(Value.updateValueType(au.attributeTag, -1), -1);
			
    	if (readBit())
    		readAttr_any();
    		  
		if(readBit())
		{
			au.objectToAdd = readUpdatableElements(null); // node.appendChild(readElements(node));
			au.attributeTag = Value.UPDATE_CHILDREN;
		}

		if(au.index != Integer.MIN_VALUE)
    		processInsertIndexed(au);
    	else
    		processInsert(au);
    }

    public void readExtendUpdate() throws Exception {
        /*trace(STR_BEGIN + "extendUpdate");*/
        readBits(m_decodingContext.m_extensionIDbits);
        int len = readVlcInt(5);
        readByteArray(len);
        /*trace(STR_END + "extendUpdate");*/
    }

    /**
     * Read the lsru:NewScene command and processes it.
     * Inserts a new scene in the browser. Any scene already present is stopped,
     * its streams closed and its resources reclaimed.
     *
     * @param timestamp The timestamp of the unit which contains this newScene commend
     * @throws Exception
     */
    public void readNewScene(int timestamp) throws Exception {
        // Pre process
        preProcessNewScene(timestamp);

        if (readBit())
        	readAttr_any();
        
        // Read svg element
//        Object node = readElements(null); TML
        Object node = readSvg();
        if (! (node instanceof Svg)) {
            return;
        }
        
        Svg svgNode = (Svg) node;
        
       // Process command
        processNewScene(svgNode);
    }

    /**
     * Reads lsru:ScriptContent.
     * TODO Model elements for ScriptContent
     *
     * @throws IOException
     */
    public void readScriptContent() throws IOException {
        /*string*/ readAttr_custom_byteAlignedString();//TODO use it!
    }

    /**
     * Decoding of the update-level objects
     *
     * @throws java.lang.Exception
     */
    public void readUpdates(int timestamp) throws Exception {
        // read update type
        int i = readBits(4);
        if (Debug.DEBUG_PARSER) {
            System.out.println("readUpdate value : " + i);
        }

        // TML
        // switch the update type 
        switch (i) {// readInsertIndexed();readReplaceAttribute();readDeleteIndexed();
            case 0:
            	readAdd();
                return;
            case 1:
            	readClean();
                return;
            case 2:
                readDelete();
                return;
            case 3:
            	readInsert();
                return;
            case 4:
            	readNewScene(timestamp);
                return;
            case 5:
            	readRefreshScene();
                return;
            case 6:
            	readReplace();
                return;
            case 7:
            	readRestore();
                return;
            case 8:
                readSave();
                break;
            case 9:
            	readSendEvent();
                break;
            case 10:
            	readUpdateAny(null, true); 
                break;
            case 11:
            	readAttr_custom_byteAlignedString();
                break;
           default:
                if (Debug.DEBUG_PARSER) {
                    System.out.println("ERROR in readUpdate, type '" + i + "'is wrong");
                }
                break;
        }
        return; // TML
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// Attributes parsing operations /////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////

    public void readOptionalChildren(GraphObject go) throws Exception {
        if (readBit()) {
            int len = readVlcInt(5); // Number of nodes
            for (int i = 0; i < len; i++) {
                if (Debug.DEBUG_PARSER) {
                    System.out.println("readChild " + i + " of " + len);
                }
                go.appendChild(readElements(go));
            }
        }
    }
        
    public Matrix2D readAndCreateMatrix() throws Exception {
        if (!readBit())
            return null;
        else
            return readCustom_Matrix();
    }

    public void readUpdateList(byte data[], int timestamp) throws Exception {
        setDataForBitReader(data, false);
        // Number of updates
        int len = readVlcInt(5);
        for (int i = 0; i < len; i++) {
            readUpdates(timestamp);
        }
        releaseDataFromBitReader();
    }

    

///////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////// AU operations ///////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////
    public final static int m_normalizer[] = {1, 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047};

    //TODO fields for LASeRUnitHeader

    /**
     * parses the LASeRUnitHeader from the LASeRUnit
     *
     * @throws IOException
     */
    private void parseLASeRUnitHeader() throws IOException {
        /*resetEncodingContext*/ boolean resetEncodingContext = readBit();
        //hasExtension
        if (readBit()) {
            /*privateData*/ readByteArray(readVlcInt(5));
        }
    }

    /**
     * parses the codec_initialisations from the LASeRUnit
     *
     * @throws IOException
     */
    private void parseCodecInitialisations() throws IOException {
        int nbColors = 0;
        int nbFonts = 0;
        int nbNameSpaces = 0;
        int nbTags = 0;

        /*trace(STR_BEGIN + "codec_initialisations");*/

        // colorInitialisation  read colors
        if (readBit()) {
            nbColors = readVlcInt(5);
            for (int i = 0; i < nbColors; i++) {
            	colorIndex++;
                int co = m_graphObjectManager.m_renderer.shiftRed(((255 * readBits(m_decodingContext.m_colorDimBits)) / m_normalizer[m_decodingContext.m_colorDimBits]));
                co += m_graphObjectManager.m_renderer.shiftGreen(((255 * readBits(m_decodingContext.m_colorDimBits)) / m_normalizer[m_decodingContext.m_colorDimBits]));
                co += m_graphObjectManager.m_renderer.shiftBlue(((255 * readBits(m_decodingContext.m_colorDimBits)) / m_normalizer[m_decodingContext.m_colorDimBits]));
                m_decodingContext.m_colorsArray.addElement(new Integer(co));
            }
            m_decodingContext.m_colorIndexBits = log2sup(colorIndex); // TML : TODO : Spec
        }

        // fontInitialisation 
        if (readBit()) {
            nbFonts = readVlcInt(5);
            for (int i = 0; i < nbFonts; i++) {
            	fontIndex++;
                String f = readAttr_custom_byteAlignedString();
                m_decodingContext.m_fontArray.addElement(f);
                m_decodingContext.m_definedFonts++;
            }
            m_decodingContext.m_fontIndexBits = log2sup(fontIndex);
        }
        //privateDataIdentifierInitialisation 
        if (readBit()) {
            nbNameSpaces = readVlcInt(5);
            for (int i = 0; i < nbNameSpaces; i++) {
            	privateDataIdentifierIndex++;
                m_decodingContext.registerNamespace(NameSpace.get(readAttr_custom_byteAlignedString()));
            }
            m_decodingContext.nameSpaceIndexBits = log2sup(privateDataIdentifierIndex);
        }
        // anyXMLInitialisation
        if (readBit()) {
            nbTags = readVlcInt(5);
            for (int i = 0; i < nbTags; i++) {
                QName t = new QName();
                m_decodingContext.registerTag(t);
                if (i != 0) {
                    t.ns = (NameSpace) NameSpace.nameSpaces.elementAt(readBits(m_decodingContext.nameSpaceIndexBits));
                    t.name = readAttr_custom_byteAlignedString();
                }
                if (readBit()) {
                    t.attrs = new Vector();
                    int nbAttributes = readVlcInt(5);
                    for (int j = 0; j < nbAttributes; j++) {
                        QName a = new QName();
                        t.attrs.addElement(a);
                        if (j == 0) {
                            a.ns = (NameSpace) NameSpace.nameSpaces.elementAt(readBits(m_decodingContext.nameSpaceIndexBits));
                        }
                        a.name = readAttr_custom_byteAlignedString();
                    }
                }
            }
            m_decodingContext.tagIndexBits = log2sup(m_decodingContext.privateTags.size()-1);
        }
        
        // extendedInitialisation 
        // TML
    	int countG = readVlcInt(5);
    	for (int i = 0; i < countG; i++) 
    	{
    		/*binaryIdForThisStringID[[i]] = */readVlcInt(5);// TML : TODO : Finish
    	    /*stringID[[i]] = */readAttr_custom_byteAlignedString();
    	}
    	// hasExtension;
    	if(readBit())
    	{
    		int len = readVlcInt(5);
    		int countGS = readVlcInt(5);
    		
    		for (int i = 0; i < countG; i++) 
        	{
        		/*localStreamIdForThisGlobal[[i]] = */readVlcInt(5);// TML : TODO : Finish
        	    /*globalName[[i]] = */readAttr_custom_byteAlignedString();
        	}
    		readByteArray(len - 0); // skip // TML : TODO : REPLACE THE 0
    	}
        
        if (Debug.DEBUG_TRACE) {
            System.out.println("parseCodecInitialisations: " +
                    " nb of Colors= " + nbColors + ", nb of fonts= " + nbFonts +
                    ", nb of tameSpaces= " + nbNameSpaces + ", nb of tags= " + nbTags);
        }
        /*trace(STR_END + "codec_initialisations");*/
    }

    /**
     * Computes the base 2 log of the absolute value of the parameter, ie the number of bits necessary to
     * represent the absolute value of the parameter
     * <p/>
     * Used for determining the size of the colors index
     */
    private static int log2sup(int i) {
        int n = 0;
        for (; i > 0; i >>>= 1) {
            n++;
        }
        return n;
    }

    /**
     * parses and executes a LASeR access unit
     *
     * @param unitTimestamp the
     * @throws Exception
     */
    public void parseLASeRUnit(int unitTimestamp) throws Exception {
        parseLASeRUnitHeader();
        parseCodecInitialisations();

        // parse and execute the laser commands
        int nbUpdates = readVlcInt(5);
        if (Debug.DEBUG_TRACE)
            System.out.println("executeLASeRUnit at time=" + m_renderingTime + " , nb of Updates=" + nbUpdates);
        for (int i = 0 ; i < nbUpdates+1 ; i++) {
            readUpdates(unitTimestamp);
        }
        if (readBit()) { // TML
        	readAttr_custom_extension();
        }
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Misc parsing operations ///////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Translates the integer N read from the stream to its float value.
     * <p/>
     * The stream contains a "p"-bits, 2's complement integer. When read to an int, the first operation is to restore
     * the sign. If N is over 2^(bits-1)-1, it is translated to 2^(bits)-N.
     * Then the correction for the resolution is applied. If resolution is 2, resol2 is 4, and the unit is 1/4.
     * N (or 4096-N) is then divided by resol2.
     *
     * @param n    the integer picked up from the stream
     * @param bits the number of bits decoded from the stream
     * @return the float equivalent of the integer number decoded from the stream
     */
    public int translate(int n, int bits) {
        // return a integer with a real value
        return translateU(negp(n, bits));
    }

    public int translateU(int n) {
        int res_r = Real.convertIntToRealInt(n);
        res_r = Real.divByPowerOf2(res_r, m_decodingContext.m_resolution);
        return res_r;
        //return n << (m_decodingContext.m_coordBits - m_decodingContext.m_resolution);
    }

    public int readIndex() throws java.io.IOException 
    {
        int index;
        if (readBit())//atEnd
        {
            index = -1; // At end
        } else {
            index = readVlcInt(5);
        }
        return index;
    }
    
    public void readFont(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	/*au.m_properties.m_externalResourcesRequired =*/ readBit();
//        au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);
    	
    	//horizAdvX
    	if(readBit())
    		this.readAttr_custom_fixed_16_8();
    	
//    	horizOriginX
    	if(readBit())
    		this.readAttr_custom_fixed_16_8();
    	
    	if(readBit())
    		readAttr_any();
    	
    	readObjectEnd(au);
    }
    
    public void readFontFace(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	//has_accentHeight
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_alphabetic
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_ascent
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_bbox
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_capHeight
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_descent
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
    	/*au.m_properties.m_externalResourcesRequired =*/ readBit();
//      au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);

//    	has_fontFamily
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_fontStretch
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_fontStyle
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_fontVariant
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_fontWeight
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_hanging
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_ideographic
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_mathematical
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_overlinePosition
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_overlineThickness
    	if(readBit())
    		readAttr_custom_fixed_16_8();

//    	has_panose1
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_slope
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_stemh
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_stemv
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_strikethroughPosition
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_strikethroughThickness
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_underlinePosition
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_underlineThickness
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_unicodeRange
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_unitsPerEm
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_widths
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
//    	has_xHeight
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
    	if(readBit())
    		readAttr_any();
    	
    	readObjectEnd(au);
    }
    
    public void readFontFaceSrc(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	if(readBit())
    		readAttr_any();
    	
    	readObjectEnd(au);
    }
    
    public void readFontFaceUri(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	/*au.m_properties.m_externalResourcesRequired =*/ readBit();
//      au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);

    	Xlink TMP = new Xlink();
    	readHref(TMP);
    	
    	if(readBit())
    		readAttr_any();
    	
    	readObjectEnd(au);
    }

    public String readFontFamily() throws java.io.IOException {
        if (readBit())
            return null; //TODO value for inherit
        else
        	// TML : TODO : to fix
        	//            return (String) m_decodingContext.m_fontArray.elementAt(readBits(m_decodingContext.m_fontIndexBits));
        	{
        	readBits(m_decodingContext.m_fontIndexBits);
        	return null;
        	}
        	
    }
    
    public void readGlyph(GraphObject parent) throws Exception
    {
    	Group au = new Group(m_graphObjectManager); // TML : TODO : change group by font object
    	
    	readObjectStart(au);
    	
    	//arabicForm
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
    	if(readBit())
    	{
    		Value av = new Value();
    		readAttr_custom_path(av);
    	}
    	
    	/*au.m_properties.m_externalResourcesRequired =*/ readBit();
//      au.m_properties.setInheritance(Value.UPDATE_EXTERNALRESOURCESREQUIRED, false);

    	//has_glyphName;
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
    	//has_horizAdvX
    	if(readBit())
    		readAttr_custom_fixed_16_8();
    	
//    	has_lang;
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
    	//has_unicode
    	if(readBit())
    		readAttr_custom_byteAlignedString();
    	
    	if(readBit())
    		readAttr_any();
    	
    	readObjectEnd(au);
    }

    public String getFontFamily(int index) {
        if ((index > 0) && index < m_decodingContext.m_fontArray.size()) {
            return (String) m_decodingContext.m_fontArray.elementAt(index);
        }
        return null;
    }

    /**
     * Translates the integer N read from the stream to its float value.
     * <p/>
     * The stream contains a 12-bits, 2's complement integer. When read to an int, the first operation is to restore
     * the sign. If N is over 2047, it is translated to 4096-N.
     * Then the correction for the resolution is applied. If resolution is 2, resol2 is 4, and the unit is 1/4.
     * N (or 4096-N) is then divided by resol2.
     *
     * @param n the 12-bit integer number picked up from the stream
     * @return the float equivalent of the 12-bit integer number decoded from the stream
     */
    
    public int translate(int n) {
        if (n > translateNegLimit) {
            n -= translateNeg;
        }
        int res_r = Real.convertIntToRealInt(n);
        res_r = Real.divByPowerOf2(res_r, m_decodingContext.m_resolution);
        return res_r;
//        return n << (m_decodingContext.m_coordBits - m_decodingContext.m_resolution);
    }
    
    // TML : translate the integer (*4096)
    public int translated12(int n) {
    	return n << (12 - m_decodingContext.m_resolution);
    }

    /**
     * Translates the integer N read from the stream to its float value, when the number is a scale factor.
     * <p/>
     * The stream contains a 12-bits, 2's complement integer. When read to an int, the first operation is to restore
     * the sign. If N is over 2047, it is translated to 4096-N.
     * Then the correction for the resolution is applied. For a scale factor, the resolution is constant and equal
     * to 1/256.
     *
     * @param n the 12-bit integer number picked up from the stream
     * @return the float equivalent of the 12-bit integer number decoded from the stream
     */
    public int translateScale(int n) {
        if (n >= scaleTab[m_decodingContext.m_scaleBits_minus_coordBits]) {
            n -= scaleTab[m_decodingContext.m_scaleBits_minus_coordBits + 1];
        }
        int res_r = Real.convertIntToRealInt(n);
        res_r = Real.divByPowerOf2(res_r, 8);
        return res_r;
    }

    private final static int scaleTab[] = {2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576,
            2097152, 4194304, 8388608, 16777216, 33554432, 67108864, 134217728
    };

    private void readHref(Xlink o) throws IOException {
        if (readBit())
            o.setHref(readAttr_custom_anyURI());
    }

  
    private int GetDiff(int codeNum)
    {
    	int diff;
        if (codeNum == 0) {
            diff = 0;
            return diff;
        } else {
              if ((codeNum%2) == 0 ) {
                  diff = -codeNum/2;
                  return diff;
              }
              else {
                  diff = (codeNum+1)/2;
                  return diff;
              }
        }
    }

/*   TML : TODO : Perapse not use 
    public String contentType() throws Exception {
        if (readBit()) {
            if (readBit()) {
                int i = readBits(2);
                switch (i) {
                    case 0:
                        return "application/laserScript";
                    case 1:
                        return "application/ecmaScript";
                    case 2:
                        return "application/java-archive";
                    default:
                        throw new RuntimeException("decoding error: unknown content type " + i);
                }
            } else {
                return read_codec_byteAlignedString();
            }
        } else {
            return null;
        }
    }
*/
}