/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.tool;

import java.io.ByteArrayOutputStream;

/**
 * The MediaBuffer class is used as a buffer for storing media coming
 * from media units.
 * Every units are stored sequentially in the ByteArrayOutputStream.
 * As soon as the media is fully received, a raw array is created from it and
 * the ByteArrayOutputStream is destroyed.
 */

public class MediaBuffer
{
    // flag indicating when media is fully written
    private boolean m_isMediaReceived = false;
    // for storing the media in a temporary buffer
    private ByteArrayOutputStream m_baos = new ByteArrayOutputStream();
    // the media in a raw array
    private byte[] m_data = null;

    /**
     * Returns true if the media is fully received
     * @return the media received state
     */
    public synchronized boolean isMediaReceived()
    {
        return m_isMediaReceived;
    }

    /**
     * Sets the media as fully received
     */
    public synchronized void setMediaReceived()
    {
        m_isMediaReceived = true;
        // create a byte array from the ByteArrayOutputStream
        m_data = m_baos.toByteArray();
        // ByteArrayOutputStream is no longer needed
        m_baos = null;
    }

    /**
     * Writes media data in the buffer
     * @param data The media data to write
     */
    public synchronized void write(byte[] data) throws java.io.IOException
    {
        if (!m_isMediaReceived)
        {
            m_baos.write(data);
        }
    }

    /**
     * Returns the raw media data
     * @return the byte array of data
     */
    public synchronized byte[] getRawData()
    {
        if (m_isMediaReceived)
        {
            return m_data;
        }
        else
        {
            return null;
        }
    }
}
