/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cédric Gégout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.tool;

import iso.mpeg.streamezzo.player.BasicRendering;
import iso.mpeg.streamezzo.player.library.Real;
import iso.mpeg.streamezzo.player.node.Event;
import iso.mpeg.streamezzo.player.node.GraphObject;
import iso.mpeg.streamezzo.player.node.Group;
import iso.mpeg.streamezzo.player.node.LaserUpdate;
import iso.mpeg.streamezzo.player.node.Listener;
import iso.mpeg.streamezzo.player.node.Multimedia;
import iso.mpeg.streamezzo.player.node.RenderedGraphObject;
import iso.mpeg.streamezzo.player.node.SVGset;
import iso.mpeg.streamezzo.player.node.Svg;
import iso.mpeg.streamezzo.player.node.TextArea;
import iso.mpeg.streamezzo.player.node.URI;
import iso.mpeg.streamezzo.player.node.Value;
import iso.mpeg.streamezzo.player.util.EventListener;
import iso.mpeg.streamezzo.player.util.NameSpace;
import iso.mpeg.streamezzo.player.util.Navigator;
import iso.mpeg.streamezzo.player.util.NodeProperties;
import iso.mpeg.streamezzo.player.util.QName;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public abstract class SceneLoader implements Runnable {

    public static final String INIT_LASER_URL = "lres://top.lsr";
    public static final String ERROR_LASER_URL = "lres://err.lsr";

    private final static int UNDEFINED_TIMESTAMP = -1;


    private PrintStream traceFile = System.out;
    /*public static final String STR_BEGIN = "---------------------------------------- ";
    public static final String STR_END 	 = "++++++++++++++++++++++++++++++++++++++++ ";*/

    /**
     * Test memory
     private long memory = 0;
     */
    /**
     * Temporary variables for reading stream
     private int tmp_len = 0;
     */
    /**
     * Error variable
     */
    private boolean errorLASeR = false;
    /**
     * Time of the scene in milliseconds
     */
    protected int m_renderingTime;
    private long m_offsetRenderingTime;
    /**
     * Background of the scene
     */
    private int m_backColor = Value.COLOR_NONE;
    private byte m_backOpacity = 0;
    private Object m_backImage = null;
    /**
     * We are using SVG relative metrics mode
     */
    private int m_screenWidth, m_screenHeight;
    private int m_screenSceneWidth, m_screenSceneHeight;

    /**
     * Infos required when loading a new scene
     */
    private boolean m_needLoadScene = false;
    private String m_url = null;
    private String m_newSceneURL = null;
    public Group rootGroup = null;
    private NodeProperties rootProperties = new NodeProperties();

    /**
     * Thread used for reading the m_content
     */
    private Thread m_thread = null;
    private boolean m_isPaused;

    /**
     * Used for synchronized Render bloc code and ResetRootNode bloc code
     */
    private Object lock = new Object();
    public DataInputStream m_byteIS = null;

    private int keyboardPolicy = EXPLICIT;

    private final static int EXPLICIT = 0;
    private final static int ACTIV_SWITCHING = 1;
    private final static int ZOOM_PAN = 2;

    /** The navigator that manages the navigation */
    public Navigator m_navigator = new Navigator();

    public GraphObjectManager m_graphObjectManager = null;

    /** System language */
    protected String m_systemLanguage = null;

    public abstract void readUpdateList(byte data[], int timestamp) throws Exception;

    public SceneLoader(GraphObjectManager graphObjectManager)
    {
        m_graphObjectManager = graphObjectManager;
        m_renderingTime = -1;  // not set
        m_offsetRenderingTime = -1;
        m_decodingContext.m_offsetAppendSceneTime = 0;
        rootGroup = new Group(graphObjectManager);
        rootGroup.setID(0);
        graphObjectManager.setNavigator(m_navigator);
    }

    public void reLoadScene(String newURL)
    {
        m_newSceneURL = new String(newURL);
        m_needLoadScene = true;
    }

    public void loadScene(String url)
    {
        // Get the system language
        m_systemLanguage = m_graphObjectManager.m_renderer.getSystemLanguage();


        if (url != null)
        {
            m_url = url;
        }
        m_screenWidth = m_graphObjectManager.m_renderer.getCanvasWidth();
        m_screenHeight = m_graphObjectManager.m_renderer.getCanvasHeight();
        if (m_thread != null)
        {
            if (!m_thread.isAlive())
            {
                synchronized(this)
                {
                    // Wake up the thread to change the timing
                    this.notify();
                    m_thread = null;
                }
            }
            else
            {
                try
                {
                    m_graphObjectManager.m_renderer.closeConnector();
                }
                catch (Exception e){ }
            }
        }
        //free BitReader semaphor if used
        if (dataIsUsed && isSceneLoaderThread)
        {
            dataIsUsed = false;
        }
        m_thread = new Thread(this);
        if (!m_thread.isAlive())
        {
            m_thread.start();
        }
    }

    public void run()
    {
        synchronized(this)
        {
            try
            {
                if (Debug.DEBUG_RUN_SCENE_LOADER)
                {
                    System.out.println("T2 : Run SceneLoader object, URL : " + m_url);
                }
                getSceneContent(m_url);
            }
            catch (Exception ee)
            {
                ee.printStackTrace();
            }
        }
    }


    /**
     * The function sets the background image. The m_content is defined by this
     * object that is the root of the scene. Only the public method render()
     * draws the background.
     * The root node shall be a Group node.
     *
     * @param backImage the Image of the background
     * @see iso.mpeg.streamezzo.player.node.GraphObject#render
     */
    public void setBackGroundImage(Object backImage)
    {
        m_backImage = backImage;
    }

    public int getRenderingTime()
    {
        return m_renderingTime;
    }

    public int pauseScene()
    {
        m_isPaused = true;
        return m_renderingTime;
    }

    public void resumeScene()
    {
        m_isPaused = false;
    }

    /**
     * The function that enables the Rendering to play at
     * a specific time the m_content
     */
    public void render()
    {
        if (m_isPaused || (m_bufferingEndDate - System.currentTimeMillis() > 0))
        {
            m_offsetRenderingTime = System.currentTimeMillis() - m_renderingTime;
            return;
        }
        if (m_backImage != null)
        {
            m_graphObjectManager.m_renderer.drawImageAtTopLeft(m_backImage, 0, 0);
        }
        else if (m_backColor != Value.COLOR_NONE)
        {
            m_graphObjectManager.m_renderer.setColor(m_backColor, m_backOpacity);
            m_graphObjectManager.m_renderer.fillRect(
                    0,
                    0,
                    m_graphObjectManager.m_renderer.getCanvasWidth(),
                    m_graphObjectManager.m_renderer.getCanvasHeight());
        }
        else
        {
            m_graphObjectManager.m_renderer.clearImage();
        }
        try
        {
            synchronized (lock)
            {
                if (errorLASeR)
                {
                    //set size of the scene
                    m_sceneWidth = m_screenWidth;
                    m_sceneHeight = m_screenHeight;
                    //set normalization mode
                    m_graphObjectManager.m_renderer.setNormalizingMode(false);
                    //delete the current scene
                    cleanRootNode();
                    //Remove all units in m_LASeRUnitBuffer
                    reinitializeLASeRUnitBuffer();
                    //compute new one with error message
                    TextArea errorText = new TextArea(this.m_graphObjectManager);
                    errorText.m_properties.m_display = NodeProperties.DISPLAY_INLINE;
                    errorText.setString(BasicRendering.ERROR_MESSAGE);
                    errorText.setFont("SYSTEM", (byte)2, (short)0x120);
                    errorText.m_properties.m_text_anchor = NodeProperties.TEXTALIGN_CENTER;
                    errorText.m_properties.m_displayAlign = NodeProperties.DISPLAYALIGN_CENTER;
                    errorText.m_properties.m_fill = BasicRendering.DRAW_COLOUR;
                    errorText.m_properties.m_stroke = BasicRendering.DRAW_COLOUR;
                    rootGroup.appendChild(errorText);
                    //set error flags, error is catched
                    errorLASeR = false;
                }

                // Execute the stored Unit for scene description
                executeLASeRUnitIndexBeforeTime(m_renderingTime);

                // Event processing
                m_navigator.processKeyEvents(m_renderingTime);
                m_graphObjectManager.resolveTimedNodes(m_renderingTime);
                if (m_navigator.isKeyPressed())
                {
                    m_graphObjectManager.m_renderer.processZoomAndPan(m_navigator.getPressedKeyCode());
                }

                //Render the scene
                rootGroup.render(m_renderingTime, false, null, rootProperties, null);

                //Post-render
                m_navigator.processNavigation(m_renderingTime);
                m_graphObjectManager.ProcessPostRender();
                m_graphObjectManager.cleanPostRenderList();
            }

            //Checking if a new scene needs to be load due to an user event
            if (m_needLoadScene && m_byteIS == null)
            {
                loadScene(m_newSceneURL);
                m_needLoadScene = false;
                m_newSceneURL = null;
            }
        }
        catch (Exception ee)
        {
            ee.printStackTrace();
        }
        if (m_offsetRenderingTime == -1)
        {
            m_renderingTime = 0;
            m_offsetRenderingTime = System.currentTimeMillis();
        }
        m_renderingTime = (int)(System.currentTimeMillis() - m_offsetRenderingTime);
    }

    public int getSceneWidth()
    {
        return m_sceneWidth;
    }

    public int getSceneHeight()
    {
        return m_sceneHeight;
    }

    /**
     * Viewbox getter
     *
     * @return the viewbox as {x, y, width, height}
     */
    public int[] getViewbox()
    {
        return m_rViewBox;
    }

    /**
     * ******************************************************************
     * PRIVATE LIBRARY
     * *******************************************************************
     */
    protected void cleanRootNode()
    {
        synchronized (lock)
        {
            rootGroup.removeAllChildren();
            rootGroup.cleanScene();
            rootGroup.setID(0);
            m_renderingTime = 0;
            m_offsetRenderingTime = System.currentTimeMillis();
            // Remove every key listeners
            m_navigator.cleanEventListeners();
        }
    }

    public void init(String start)
    {
        m_newSceneURL = (start != null ? start : INIT_LASER_URL);
        loadScene(m_newSceneURL);
        m_needLoadScene = false;
        m_newSceneURL = null;
    }

    /**
     * Read the HTTP headers and the data using HttpConnection.
     * Check the response code to insure successful retrieval.
     * <p/>
     * Connector.open is used to open url and a HttpConnection is returned.
     * The HTTP headers are read and processed.
     * If the length is available, it is used to read the data in bulk.
     * From the HttpConnection the InputStream is opened.
     * It is used to read every character until end of file (-1).
     *
     * @param url the URL to process.
     */
    void getSceneContent(String url) throws IOException, OutOfMemoryError
    {
        boolean goOn = true;
        try
        {
            if (BasicRendering.SYNCHRONIZE_HTTP)
            {
                if (BasicRendering.SYNCHRONIZE_HTTP_TRACE)
                {
                    System.out.println("T2 : BEGIN : openURL");
                }
                synchronized (this)
                {
                    m_graphObjectManager.m_renderer.cleanMemory();
                    m_byteIS = m_graphObjectManager.m_renderer.openURL(url);
                }
                if (BasicRendering.SYNCHRONIZE_HTTP_TRACE) {
                    System.out.println("T2 : END : openURL");
                }
            }
            else
            {
                m_byteIS = m_graphObjectManager.m_renderer.openURL(url);
            }
            if (m_byteIS != null)
            {
                while(goOn) // Main parsing loop
                {
                    goOn = decodeLASeRPacket();
                    this.wait(5);
                }
            }
            if (Debug.DEBUG_TRACE)
            {
                System.out.println("Close : m_byteIS because it's end of stream in getSceneContent");
            }
            m_graphObjectManager.m_renderer.closeConnector();
        }
        catch (Exception e)
        {
            e.printStackTrace();
            if (Debug.DEBUG_RAW_DATA_RECEIVED)
            {
                System.out.print("Tn : LASeRPacketHeader : ");
                for (int i = 0; i < 11; i++)
                {
                    System.out.print(Integer.toHexString(0xFF & LASeRPacketHeader[i]));
                    System.out.print(" ");
                }
                System.out.println();
            }
            if (m_byteIS != null)
            {
                m_byteIS.close();
            }
            if (!errorLASeR)
            {
                errorLASeR = true;
                getSceneContent(ERROR_LASER_URL);
            }
        }
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////// Scene properties  //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////
    private int m_bufferingTime;
    private long m_bufferingEndDate;
    private int m_sceneWidth = 100;
    private int m_sceneHeight = 100;
    /*private boolean m_untransformedBitmapsHint;    */
    private String m_baseURL = "";
    private int m_lastTime = -100;
    /** viewbox {x, y, w, h}. Values are fiwed point 20.12 */
    private int[] m_rViewBox = {0, 0, 0, 0};

    public void setBufferingTime(int bufferingTime)
    {
        m_bufferingTime = bufferingTime;
    }

    public void setUntransformedBitmapsHint(boolean untransformedBitmapsHint)
    {
        /*m_untransformedBitmapsHint = untransformedBitmapsHint; */
    }

    public void setSceneSize(int rSceneWidth, boolean wUnitsIsPercents, int rSceneHeight, boolean hUnitsIsPercents)
    {
        if (wUnitsIsPercents)
        {
            // Get canvas width
            int canvasWidth = m_graphObjectManager.m_renderer.getCanvasWidth();

            // Compute scene width
            m_sceneWidth = Real.round(Real.applyDiv(Real.applyMul(Real.convertIntToRealInt(canvasWidth), rSceneWidth) , Real.convertIntToRealInt(100)));
        }
        else
        {
            m_sceneWidth = Real.round(rSceneWidth);
        }

        if (hUnitsIsPercents)
        {
            // Get canvas height
            int canvasHeight = m_graphObjectManager.m_renderer.getCanvasHeight();

            // Compute scene width
            m_sceneHeight = Real.round(Real.applyDiv(Real.applyMul(Real.convertIntToRealInt(canvasHeight), rSceneHeight) , Real.convertIntToRealInt(100)));
        }
        else
        {
            m_sceneHeight = Real.round(rSceneHeight);
        }

        // Update renderer properties
        if (m_graphObjectManager.m_renderer != null)
        {
            m_graphObjectManager.m_renderer.updateNormalizationFactors(m_rViewBox, m_sceneWidth, m_sceneHeight);
            m_screenSceneWidth = m_graphObjectManager.m_renderer.normalizeW(m_sceneWidth << 12);
            m_screenSceneHeight = m_graphObjectManager.m_renderer.normalizeH(m_sceneHeight << 12);
        }
    }

    public void setViewbox(int[] rViewbox)
    {
        // Set new viewbox
        m_rViewBox[0] = rViewbox[0];
        m_rViewBox[1] = rViewbox[1];
        m_rViewBox[2] = rViewbox[2];
        m_rViewBox[3] = rViewbox[3];

        // Update renderer properties
        if (m_graphObjectManager.m_renderer != null)
        {
            m_graphObjectManager.m_renderer.updateNormalizationFactors(m_rViewBox, m_sceneWidth, m_sceneHeight);
            m_screenSceneWidth = m_graphObjectManager.m_renderer.normalizeW(m_sceneWidth << 12);
            m_screenSceneHeight = m_graphObjectManager.m_renderer.normalizeH(m_sceneHeight << 12);
        }
    }

    public void setAspectRatio(int preserveAspectRatioX, int preserveAspectRatioY, int preserveAspectRatioScale)
    {
        // Update renderer properties
        if (m_graphObjectManager.m_renderer != null)
        {
            m_graphObjectManager.m_renderer.setAspectRatio(preserveAspectRatioX, preserveAspectRatioY, preserveAspectRatioScale);
            m_graphObjectManager.m_renderer.updateNormalizationFactors(m_rViewBox, m_sceneWidth, m_sceneHeight);
            m_screenSceneWidth = m_graphObjectManager.m_renderer.normalizeW(m_sceneWidth << 12);
            m_screenSceneHeight = m_graphObjectManager.m_renderer.normalizeH(m_sceneHeight << 12);
        }
    }

    public void setZoomAndPan(boolean zoomAndPan)
    {
        m_graphObjectManager.m_renderer.enableZoomAndPan(zoomAndPan);
    }

    public void setBackGroundColor(int backColor)
    {
        m_backColor = backColor;
    }

    public void  setBackGroundOpacity(byte backOpacity)
    {
        m_backOpacity = backOpacity;
    }

    public void setBaseURL(String baseURL)
    {
        m_baseURL = baseURL;
    }


//////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////Stream header management ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * A stream header
     */
    private class StreamHeader
    {
        //newly added attributes
    	public int timeStampResolution = -1;
    	public String mimeType = null;
    	public String url = null;

    	public int m_streamID = -1;
        public int m_objectTypeIndication = -1;
        public int m_streamType = -1;
        public boolean m_isPermanent = false;
        public boolean m_isRemote = false;
        public int m_bufferSize = 0;
        public int m_time = UNDEFINED_TIMESTAMP;
        public byte m_decoderSpecificInfo[] = null;
        public Object m_data = null;
    }

    /** The stream header list to be able to get stream properties from its streamID */
    private Hashtable m_streamHeaders = new Hashtable(1);

    protected void addStreamHeader(StreamHeader header)
    {
        // Get streamID
        Integer streamID = new Integer(header.m_streamID);

        // If this streamID is already stored then remove it
        if (m_streamHeaders.containsKey(streamID))
        {
            m_streamHeaders.remove(streamID);
        }

        // Add
        m_streamHeaders.put(streamID, header);
    }

    /**
     * Finds the header linked to the passed-in streamID
     * @param streamID The streamID of the header to get
     * @return The header linked to the passed-in streamID (or null)
     */
    protected StreamHeader getStreamHeaderFromID(int streamID)
    {
        return (StreamHeader)m_streamHeaders.get(new Integer(streamID));
    }

    /**
     * Clears the stream headers list
     */
    protected void clearStreamHeaders()
    {
        m_streamHeaders.clear();
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Access unit management ///////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////
    public abstract void parseLASeRUnit(int unitTimestamp) throws Exception;


    public Vector m_units = new Vector();

    private class Unit
    {
        int m_time;
        byte m_buf[];

        public Unit(int time, byte buf[])
        {
            m_time = time;
            m_buf = buf;
        }
    }

    public void addUnit(int time, byte buf[])
    {
        if(Debug.DEBUG_TRACE) System.out.println("addUnit : time="+time);
        // store the unit
        Unit a_u = new Unit(time, buf);
        m_units.addElement(a_u);
        if (m_lastTime <= time)
        {
            m_lastTime = time;
        }
        // have the gc run
        Runtime.getRuntime().gc();
    }

    public void executeLASeRUnitIndexBeforeTime(int sceneTime)
    {
        int ii = 0;
        Unit a_u = null;
        while (ii < m_units.size())
        {
            a_u = (Unit)(m_units.elementAt(ii));
            if (a_u.m_time <= sceneTime)
            {
                // parse and execute the AU on the scene
                try
                {
                    setDataForBitReader(a_u.m_buf, false);
                    parseLASeRUnit(a_u.m_time);
                    releaseDataFromBitReader();
                }
                catch (Exception ee)
                {
                    ee.printStackTrace();
                }
                //remove the LASeRUnit from the buffer
                m_units.removeElementAt(ii);
            }
            else
            {
                return;
            }
        }
    }

    public void reinitializeLASeRUnitBuffer()
    {
        m_lastTime = -100;
        m_units.removeAllElements();
        System.gc();
    }

    /**
     * All units with timestamp before the specified time will be deleted
     * @param timeBeforeDeleteUnits
     */
    public void reinitializeLASeRUnitBuffer(int timeBeforeDeleteUnits)
    {
        if (Debug.DEBUG_TRACE) System.out.println("reinitializeLASeRUnitBuffer : timeBefore = " + timeBeforeDeleteUnits);
        if (Debug.DEBUG_TRACE) System.out.println("reinitializeLASeRUnitBuffer : m_units.size() = " + m_units.size());
        m_lastTime = -100;

        // Remove units arrived before the specified time
        Unit unit = null;
        int nbUnits = m_units.size();
        for (int i = nbUnits - 1; i >= 0; i--)
        {
            unit = (Unit)m_units.elementAt(i);
            if (Debug.DEBUG_TRACE) System.out.println("reinitializeLASeRUnitBuffer : unit.m_time = " + unit.m_time);
            if (unit.m_time < timeBeforeDeleteUnits)
            {
                m_units.removeElementAt(i);
            }
        }
        if (Debug.DEBUG_TRACE) System.out.println("reinitializeLASeRUnitBuffer : m_units.size() = " + m_units.size());
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// Decoding context ///////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////
    public DecodingContext m_decodingContext = new DecodingContext();

    public class DecodingContext
    {
    	// TML 
    	public int m_profile = 0;
    	public int m_level = 0;
    	public int m_pathComponents = 0;
    	public boolean m_useFullRequestHost = false;
        public int m_colorDimBits = 8;
        public int m_resolution = 0;
        public int m_scaleBits_minus_coordBits = 0;
        public int m_coordBits = 0;
        public int m_extensionIDbits= 0;
        public boolean m_hasStringIDs = false;
        public boolean m_hasPrivateData = false;
        public boolean m_hasExtendedAttributes = false;
        public boolean m_newSceneIndicator = false;
        public Vector m_colorsArray = new Vector();  // the array of colors
        public int m_colorIndexBits = 0;             // the encoding length of color indexes
        public Vector m_fontArray = new Vector();
        public int m_definedFonts = 0;
        public int m_fontIndexBits = 0;
        public int m_offsetAppendSceneTime;
        public int m_pointsCodec;
        public int m_timeResolution = 1000;
        public Vector nameSpaces = new Vector();
        public Vector privateTags = new Vector();
        public Vector paintURI = new Vector();
        public int nameSpaceIndexBits = 0;
        public int tagIndexBits = 0;
        public void registerNamespace(NameSpace n) {
            if (nameSpaces.indexOf(n) >= 0) return;
            nameSpaces.addElement(n);
        }
        public void registerTag(QName name) {
            if (privateTags.size() == 0 && name.name != null) {
                // if the first registered QName has a non null name, then we are parsing an XML
                // and we should add the dummy
                privateTags.addElement(new QName());
            }
            QName t = null;
            for (int i = 0; i < privateTags.size(); i++) {
                QName tag = (QName)privateTags.elementAt(i);
                if (tag.equals(name)) {
                    t = tag;
                    break;
                }
            }
            if (t == null) {
                privateTags.addElement(name);
            }
        }
        public QName getAttr(int i, int j) {
            return ((QName)((QName)privateTags.elementAt(i)).attrs.elementAt(j));
        }
        public int nbOfAttrs(int i) {
            return ((QName)privateTags.elementAt(i)).attrs.size();
        }
        public int registerPaintURI(URI u) {
            if (paintURI == null) paintURI = new Vector();
            //System.out.println("register "+u);
            for (int i = 0; i < paintURI.size(); i++) {
                URI v = (URI)paintURI.elementAt(i);
                if (v.equals(u)) return i;
            }
            paintURI.addElement(u);
            return paintURI.size()-1;
        }
    };

///////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////// Command processing operations ///////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * executes the lsr:LASeRHeader element
     * @param header
     */
    public void executeLASeRHeader(StreamHeader header)
    {
        // reset scene and scene clock
        if (!m_decodingContext.m_newSceneIndicator)
        {
            // reset scene
            cleanRootNode();

            // reset scene color table
            m_decodingContext.m_colorIndexBits = 0;
            m_decodingContext.m_fontArray.removeAllElements();
            m_decodingContext.m_colorsArray.removeAllElements();

            // remove all units in m_LASeRUnitBuffer
            reinitializeLASeRUnitBuffer();
            // reset time offset
            m_decodingContext.m_offsetAppendSceneTime = 0;

            // remove all stream headers
            clearStreamHeaders();
        }
        else
        {
            // Remove all units arrived before because decoding parameters have changed
            reinitializeLASeRUnitBuffer(header.m_time);

            // get now time and apply to appened scene AUs
            m_decodingContext.m_offsetAppendSceneTime = m_renderingTime;
        }
    }
    
    public void processActivate(int ActivateNodeIdRef)
    {
    	boolean bIsRenderedGraphObject = true;
    	
    	GraphObject targetNode = m_graphObjectManager.getNode(ActivateNodeIdRef);
    	if(targetNode == null)
    	{
    		System.out.println("WARNING: node with id : " + ActivateNodeIdRef + " can't be find in the scenegraph (processActivate)");
    		return;
    	}
    	
    	if(targetNode instanceof RenderedGraphObject)
    	{
    		if(((RenderedGraphObject)targetNode).m_properties.m_Deactivated == false)
    			return;
    	}
    	else if(targetNode instanceof Multimedia)
    	{
    		bIsRenderedGraphObject = false;
    		
    		if(((Multimedia)targetNode).getMultimediaPlayingState() != -10)
    			return;
    		else
    			((Multimedia)targetNode).setMultimediaPlayingState(SVGset.ANIMSTATE_IDLE);
    	}
    	else if(targetNode instanceof SVGset)
    	{
    		bIsRenderedGraphObject = false;
    		
    		if(((SVGset)targetNode).getAnimationState() != -10)
    			return;
    		else
    			((SVGset)targetNode).setAnimationState(SVGset.ANIMSTATE_IDLE);
    	}
    	
    	
    	if(bIsRenderedGraphObject)
    	{
	    	((RenderedGraphObject)targetNode).m_properties.m_Deactivating = false;
	    	((RenderedGraphObject)targetNode).m_properties.m_DeactivatedOwner = false;
	    	((RenderedGraphObject)targetNode).m_properties.m_Deactivated = false;
    	}
    	if(targetNode instanceof Listener)
    	{
    		((Listener)targetNode).m_enabled = true;
    	}
    	
    	
    	GraphObject ParentTargetNode = targetNode.getParent();
    	if(ParentTargetNode == null)
    	{
    		return;
    	}
    	
        processActivate(ParentTargetNode.getID());
        
        targetNode.notifyEvent(EventListener.UI_ACTIVATE_EVENT, new Integer(0));
    }
    
    public void processDesactivate(int DesactivateNodeIdRef, boolean IsDesactivateAncestor, boolean IsOwner)
    {
    	boolean bIsRenderedGraphObject = true;
    	
    	GraphObject targetNode = m_graphObjectManager.getNode(DesactivateNodeIdRef);
    	if(targetNode == null)
    	{
    		System.out.println("WARNING: node with id : " + DesactivateNodeIdRef + " can't be find in the scenegraph (processDesactivate)");
    		return;
    	}
    	
    	if(targetNode instanceof RenderedGraphObject)
    	{
    		if(((RenderedGraphObject)targetNode).m_properties.m_Deactivating || ((RenderedGraphObject)targetNode).m_properties.m_Deactivated)
    			return;
    	}
    	else if(targetNode instanceof Multimedia)
    	{
    		bIsRenderedGraphObject = false;
    		
    		if(((Multimedia)targetNode).getMultimediaPlayingState() == -10)
    			return;
    		else
    			((Multimedia)targetNode).setMultimediaPlayingState(-10);
    	}
    	else if(targetNode instanceof SVGset)
    	{
    		bIsRenderedGraphObject = false;
    		
    		if(((SVGset)targetNode).getAnimationState() == -10)
    			return;
    		else
    			((SVGset)targetNode).setAnimationState(-10);
    	}
    	
    	if(bIsRenderedGraphObject)
    	{
    		((RenderedGraphObject)targetNode).m_properties.m_Deactivating = true;
    		((RenderedGraphObject)targetNode).m_properties.m_DeactivatedOwner = IsOwner;
    		((RenderedGraphObject)targetNode).m_properties.m_Deactivated = IsDesactivateAncestor;
    	}
    	if(targetNode instanceof Listener)
    	{
    		((Listener)targetNode).m_enabled = false;
    	}
    	
        int NbChildren = targetNode.getChildrenCount();
        Object child;
        for (int i = 0 ; i < NbChildren  ; i++)
        {
            child = targetNode.getChild(i);
            if (child != null)
            {
                if (child instanceof GraphObject)
                {
                	processDesactivate(((GraphObject)child).getID(), true, false);
                }
            }
        }
        
        if(bIsRenderedGraphObject)
        	((RenderedGraphObject)targetNode).m_properties.m_Deactivating = false;
        
        targetNode.notifyEvent(EventListener.UI_DESACTIVATE_EVENT, new Integer(0));
    }

    protected void preProcessNewScene(int timestamp)
    {
        // Reset the scene
        cleanRootNode();

        // reset scene color table
        /*m_colorIndexBits = 0;
        m_definedSceneColors = 0;
        m_fontArray.removeAll();
        m_colorsArray.removeAll();*/

        //Remove all units arrived before this newScene command
        reinitializeLASeRUnitBuffer(timestamp);

        // reset time offset
        m_decodingContext.m_offsetAppendSceneTime = 0;
        StreamHeader streamHeader = null;

        // Go through the stream header list and remove them if their time stamp is befor the new scene time stamp
        for (Enumeration e = m_streamHeaders.elements() ; e.hasMoreElements() ;)
        {
            streamHeader = (StreamHeader)e.nextElement();
            if (streamHeader.m_time < timestamp)
            {
                m_streamHeaders.remove(new Integer(streamHeader.m_streamID));
                m_graphObjectManager.removeSingleObject(streamHeader.m_streamID);
            }
        }

        System.gc();
    }

    protected void processNewScene(GraphObject rootNode)
    {
        // Insert it into the scene graph
        rootGroup.appendChild((GraphObject)rootNode);

        // Compute buffering end time
        m_bufferingEndDate = System.currentTimeMillis() + m_bufferingTime;

        // Update renderer properties
        if (m_graphObjectManager.m_renderer != null)
        {
            m_graphObjectManager.m_renderer.updateNormalizationFactors(m_rViewBox, m_sceneWidth, m_sceneHeight);
            m_screenSceneWidth = m_graphObjectManager.m_renderer.normalizeW(m_sceneWidth << 12);
            m_screenSceneHeight = m_graphObjectManager.m_renderer.normalizeH(m_sceneHeight << 12);
        }
    }

    /**
     * process the lsru:Replace command
     */
    public void processReplace(LaserUpdate au)
    {
        // get the replaced node immediatly as it can be removed from nodes table when reading new node if
        // new node.id == replacednode.id
        GraphObject the_ReplacedNode = m_graphObjectManager.getNode(au.ref);

        // process command
        if (the_ReplacedNode != null)
        {
            GraphObject the_parent = the_ReplacedNode.getParent();
            if (the_parent != null && ((the_parent instanceof Group) || (the_parent instanceof Svg)))
            {
                GraphObject trans = (GraphObject)the_parent;
                int N = trans.getChildrenCount();
                Object a_child;
                for (int ii = 0; ii < N; ii++)
                {
                    a_child = trans.getChild(ii);
                    if (a_child != null)
                    {
                        if (a_child instanceof GraphObject)
                        {
                            if (((GraphObject)a_child).getID() == au.ref)
                            {
                                // remove child and its children, unregister them all
                                trans.removeChild(ii);
                                // add and register the new node
                                if(au.objectToAdd != null)
                                	trans.insertChild(ii, au.objectToAdd);
                                break;
                            }
                        }
                        else
                        {
                            // remove child and its children, unregister them all
                            trans.removeChild(ii);
                            // add and register the new node
                            trans.insertChild(ii, au.objectToAdd);
                        }
                    }
                }
            }
        }
    }

    /**
     * process the lsru:ReplaceIndexed command
     *
     * @param targetNode the parent of the node to replace
     * @param index the index of the node to replace
     * @param newNode the node to replace target by
     */
    public void processReplaceIndexed(GraphObject targetNode,int index, Object newNode)
    {
        if (targetNode != null)
        {
            // check if target to replace is there
            if (index < targetNode.getChildrenCount())
            {
                targetNode.removeChild(index);
                if (newNode != null)
                {
                    targetNode.insertChild(index, (GraphObject)newNode);
                }
            }
        }
    }

    /**
     * process the lsru:DeleteIndexed command
     *
     * @param targetID the parent of the node to replace
     * @param index the index of the node to replace
     */
    public void processDeleteIndexed(int targetID,int index)
    {
        GraphObject node = m_graphObjectManager.getNode(targetID);
        if (node != null)
        {
            node.removeChild(index);
        }
    }

    /**
     * process the lsru:Delete command
     *
     */
    public void processDelete(LaserUpdate delete)
    {
        m_graphObjectManager.deleteNode(delete.ref);
    }

    /**
     * process the lsru:ReplaceIndexed command
     *
     */
    public void processReplaceAttribute(LaserUpdate update)
    {
    	GraphObject targetNode = m_graphObjectManager.getNode(update.ref);
        // Check replace child case
        if (targetNode != null)
        {
                targetNode.setDOMAttribute(update.attributeTag, update.updateValue, update.index);
        }
    }

    /**
     * process the lsru:Add command
     *
     * @param targetID the targe node
     * @param event the event
     */
    public void processSendEvent(int targetID, Event event)
    {
        // Get the targeted node
        GraphObject targetNode = m_graphObjectManager.getNode(targetID);
        if (targetNode == null)
        {
            return;
        }
        // Send the event
        if(targetNode instanceof SVGset)
        	((SVGset)targetNode).onEvent((byte)event.event, new Integer(m_renderingTime));
    }

    /**
     * process the lsru:Add command
     *
     */
    public void processAdd(LaserUpdate update)
    {
        // Get the target node
        GraphObject targetNode = m_graphObjectManager.getNode(update.ref);

    	if (targetNode != null)
        {
            if (update.attributeTag != NodeProperties.CHILDREN)
            {
                // Add
                targetNode.addDOMAttribute(update.attributeTag, update.updateValue);
            }
        }
    }

    /**
     * process the lsru:InsertIndexed command
     *
     */
    public void processInsertIndexed(LaserUpdate update) {
        GraphObject parentNode = m_graphObjectManager.getNode(update.ref); 
    	// process the attribute : children case
        if ((update.attributeTag == Value.UPDATE_CHILDREN) && (null != update.objectToAdd))
        {
            // insert under svg node by default
            if (update.ref < 0)
            {
                parentNode = (GraphObject)rootGroup.m_children.elementAt(0);
            }
            else
            {
                parentNode = m_graphObjectManager.getNode(update.ref);
            }

            // insert the new children
            if ((parentNode != null))
            {
                GraphObject node = (GraphObject)update.objectToAdd;
                parentNode.insertChild(update.index, node);
            }
        }
        else
        {
           //TODO process the other attribute cases
        }
    }

    /**
     * process the lsru:InsertIndexed command
     *
     */
    public void processInsert(LaserUpdate insert)
    {
        // process the attribute : children case
        if ( (insert.attributeTag == Value.UPDATE_CHILDREN) && ( insert.objectToAdd != null) )
        {
            // get parent
            GraphObject the_parent = null;

            // insert under svg node by default
            if (insert.ref < 0)
            {
                the_parent = (GraphObject)rootGroup.m_children.elementAt(0);
            }
            else
            {
                the_parent = m_graphObjectManager.getNode(insert.ref);
            }

            // insert the new children
            if ((the_parent != null))
            {
                the_parent.appendChild(insert.objectToAdd);
            }
        }
        else
        {
           //TODO Insert attributes as well
        }
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// Bit reader operations ///////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////

    // low level bit parsing attributes
    private DataInputStream dataIn;
    private int buffer;
    private int valid = 0;
    public int byteCount = 0;
    private boolean dataIsUsed = false;
    private boolean isSceneLoaderThread = false;

    private final static int maskBit[] = { 0x00, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };

    private final static int maskRight[] = { 0x00000000, 0x00000001, 0x00000003, 0x00000007, 0x0000000F, 0x0000001F,
                                             0x0000003F, 0x0000007F, 0x000000FF };

    public void setDataForBitReader(byte array[], boolean a_boolean)
    {
        if (Debug.DEBUG_BITREADER_TOKEN)
        {
            if (a_boolean)
            {
                System.out.println("SCENE LOADER THREAD");
            }
            else
            {
                System.out.println("RENDERER THREAD");
            }
        }
        while (dataIsUsed)
        {
            if (Debug.DEBUG_BITREADER_TOKEN)
            {
                System.out.println("Wait the Hand to use BitReader functions");
                if (a_boolean)
                {
                    System.out.println("SCENE LOADER THREAD");
                }
                else
                {
                    System.out.println("RENDERER THREAD");
                }
            }
        }
        dataIsUsed = true;
        isSceneLoaderThread = a_boolean;
        valid = 0;
        byteCount = 0;
        dataIn = new java.io.DataInputStream(new java.io.ByteArrayInputStream(array));
    }

    public void releaseDataFromBitReader()
    {
        if (Debug.DEBUG_BITREADER_TOKEN)
        {
            System.out.print("#---# - Release BitReader token by : ");
            if (isSceneLoaderThread)
            {
                System.out.println("SCENE LOADER THREAD");
            }
            else
            {
                System.out.println("RENDERER THREAD");
            }
        }
        dataIsUsed = false;
    }

    public int readBits(int n) throws java.io.IOException
    {
        if (n > 32)
        {
            throw new RuntimeException("Cannot read more than 32 bits as an int: " + n);
        }
        if (n <= 0)
        {
            return 0;
        }
        read();
        if (n <= valid)
        {
            // either we take everything from buffer
            return (buffer >>> (valid -= n)) & maskRight[n];
        }
        else
        {
            // or we will have to do more
            n -= valid;
            int tmp = (buffer & maskRight[valid]) << n;
            valid = 0;
            while (n >= 8)
            {
                int byt = 0xFF & dataIn.readByte();
                byteCount++;
                n -= 8;
                tmp |= byt << n;
            }
            if (n > 0)
            {
                buffer = 0xFF & dataIn.readByte();
                byteCount++;
                valid = 8 - n;
                tmp |= buffer >>> valid;
            }
            return tmp;
        }
    }
    
    public void flush()
    {
        if (valid > 0) {
            /*trace("flushing " + (8 - valid) + " bits: ");*/
	    	if (valid == 8)
	        {
	            return;
	        }
	        valid = 0;
	        buffer = 0;
        }
    }

    private void readFullyData(byte array[], int offset, int len) throws Exception
    {
        if (m_byteIS != null)
        {
            int read;
            while (len > 0)
            {
                read = m_byteIS.read(array, offset, len);
                offset += read;
                len -= read;
            }
        }
        else
        {
            if (Debug.DEBUG_ERROR)
            {
                System.out.println("m_byteIS is null ; because someone else close the stream");
            }
        }
    }

    void read() throws java.io.IOException
    {
        if (valid == 0)
        {
            buffer = 0xFF & dataIn.readByte();
            byteCount++;
            valid = 8;
        }
    }

    public int readInt() throws java.io.IOException
    {
        byteCount += 4;
        if (valid == 0)
        {
            return dataIn.readInt();
        }
        else
        {
            int tmp = (buffer & maskRight[valid]) << (32 - valid), t = dataIn.readInt();
            buffer = 0xFF & t;
            tmp |= (t >>> valid);
            return tmp;
        }
    }

    public boolean readBit() throws java.io.IOException
    {
        read();
        return ((buffer & maskBit[valid--]) != 0);
    }

    public int readByte() throws java.io.IOException
    {
        read();
        if (valid == 8)
        {
            valid = 0;
            return buffer;
        }
        else
        {
            int tmp = (buffer & maskRight[valid]) << (8 - valid);
            buffer = 0xFF & dataIn.readByte();
            byteCount++;
            tmp |= (buffer >>> valid);
            return tmp;
        }
    }

    public byte[] readByteArray(int l) throws java.io.IOException
    {
        byte a[] = new byte[l];
        for (int i = 0; i < l; i++)
        {
            a[i] = (byte)(0xFF & readByte());
        }
        return a;
    }

    public int readAttr_custom_fixed_16_8() throws java.io.IOException
    {
		int i = readBits(24) << 4;
        if ((i & 0x08000000) != 0) i |= 0xF0000000;
        return i;
    }
    
    /*public int readFixedPoint16_8() throws java.io.IOException {
    	int i = readFixedPoint16_8();
    	traceBits(24, i, msg);
    	return i;
    }*/
    
    public int readAttr_custom_fixed_16_8i() throws IOException {
    	boolean b = readBit();
        if (b) return Integer.MAX_VALUE;
        return readAttr_custom_fixed_16_8();
    }
    
    /*public int readFixedPoint24_8() throws IOException {
    	return readBits(32);
    }*/

    /*public String readString() throws java.io.IOException
    {
        int bitsUrlLen = readBits(5);
        int l = readBits(bitsUrlLen);
        byte a[] = new byte[l + 2];
        a[0] = (byte)((l >>> 8) & 0xFF);
        a[1] = (byte)(l & 0xFF);
        for (int i = 2; i < l + 2; i++)
        {
            a[i] = (byte)(0xFF & readByte());
        }
        java.io.ByteArrayInputStream bais = new java.io.ByteArrayInputStream(a);
        return new java.io.DataInputStream(bais).readUTF();
    }*/
    
    public String readString(int l) throws java.io.IOException {
        byte a[] = new byte[l + 2];
        a[0] = (byte)((l >>> 8) & 0xFF);
        a[1] = (byte)(l & 0xFF);
        for (int i = 2; i < l + 2; i++) {
            a[i] = (byte)(0xFF & readByte());
        }
        java.io.ByteArrayInputStream bais = new java.io.ByteArrayInputStream(a);
        return new java.io.DataInputStream(bais).readUTF();
    }
    
    public String readVlcString() throws java.io.IOException {
        int len = readVlcInt(8);
        String b = readString(len);
        /*if (Debug.DEBUG_PARSER) {
	        traceFile.print(s);
	        fill(20 - s.length());
	        traceFile.print(" n=");
	        traceFile.print(len * 8);
	        traceFile.print("\tvalue=\"");
	        traceFile.print(b);
	        traceFile.println("\"");
        }*/
        return b;
    }
    
    public String readAttr_custom_byteAlignedString() throws IOException {
    	flush();
        return readVlcString();
     }
 
///////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// SAF parsing operations ///////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////
    public abstract void parseLASeRHeader() throws Exception;

    private final static int SceneDescriptionStreamType = 0x03;
    //private final static int VisualStreamType = 0x04;
    //private final static int AudioStreamType = 0x05;
    //private final static int FontStreamType = 0x08;
    //private final static int VisualMpeg4OTI = 0x20;
    //private final static int H263OTI = 0xC2;
    //private final static int AMROTI = 0xC5;
    //private final static int AudioMpeg4OTI = 0x40;
    private final static int LASeROTI = 0x09;
    private final static int JPEGOTI = 0x6C;
    private final static int PNGOTI = 0x6D;
    //private final static int FontDataOTI = 0x06;

    private byte LASeRPacketHeader[] = new byte[11];
    private int m_SAFPacketSeqNum= -1;
    private int m_SAFPacketlength = -1;
    private int m_SAFAUStreamID = -1;
    private int m_SAFAUType = -1;
    private int m_SAFPacketTime = UNDEFINED_TIMESTAMP;
    //newly added attributes
    private boolean randomAccessPointFlag = false;
    private int flags = -1;

    /**
     * parses the SAF packet header.
     *
     * @throws Exception
     *
     */
    public void parseSAFPacketHeader() throws Exception
    {
    	randomAccessPointFlag = readBit();
        m_SAFPacketSeqNum = readBits(15);
        flags = readBits(2);
        m_SAFPacketTime = readBits(30);
        m_SAFPacketlength = readBits(16);
    }

    /**
     * parses the SAF AU header.
     *
     * @throws Exception
     *
     */
    public void parseSAFAUHeader() throws Exception
        {
        // 2 bytes of AU allready read with SAF packet header
        m_SAFPacketlength -= 2;
        m_SAFAUType = readBits(4);
        m_SAFAUStreamID = readBits(12);
        if (m_SAFAUStreamID == 0xFFF)
        {
            m_SAFAUStreamID = -1;
        }
    }

    /**
     * read and partially decode a LASeR packet
     * <p/>
     * This operation reads a LASeR packet from file/network, parses SAF Header and SAF AU Header. The operation returns
     * a boolean that tells if the read packet is the last of the content (SAF endOfStream)
     *
     * @return boolean
     * @throws java.lang.Exception
     */
    public boolean decodeLASeRPacket() throws Exception
    {
        boolean returnValue = true;
        boolean permanent;
        StreamHeader header = null;
        
        /*trace(STR_BEGIN + "decode laser packet");*/
        
        if (Debug.DEBUG_PARSER2)
        {
            System.out.println("T2 : decodeLASeRPacket begin");
        }

        // read the SAF packet header and SAF AU header
        try {
            readFullyData(LASeRPacketHeader, 0, 10);
        } catch(IndexOutOfBoundsException e) {
            return false;
        }
        
        if (Debug.DEBUG_PARSER2)
        {
            System.out.println("T2 : decodeLASeRPacket: 10 bytes header has been read");
        }

        // prepare data for parser
        setDataForBitReader(LASeRPacketHeader, true);

        // parse the SAF packet header
        parseSAFPacketHeader();

        // parse the SAF AU part
        parseSAFAUHeader() ;

        // release data from parser
        releaseDataFromBitReader();

        if (Debug.DEBUG_PARSER2)
        {
            System.out.println("T2 : decodeLASeRPacket: \t m_packetSeqNum=" + m_SAFPacketSeqNum);
        }

        // read the data part of the SAF AU
        byte data[] = null;
        if (0 != m_SAFPacketlength)
        {
            data = new byte[m_SAFPacketlength];

            // actually read on network
            readFullyData(data, 0, m_SAFPacketlength);

            // prepare data for parser
            setDataForBitReader(data, true);
        }

        // handle the different AU type
        switch(m_SAFAUType)
        {
            case 0x02: // SimpleDecoderConfigDescriptor permanent
                permanent = true;
            case 0x01: // SimpleDecoderConfigDescriptor
                permanent = false;
                if (Debug.DEBUG_PARSER2)
                {
                    System.out.println("T2 : decodeLASeRPacket: \t Stream Header "+ m_SAFPacketlength);
                }

                // create stream header object
                header = new StreamHeader();
                header.m_isPermanent = permanent;
                header.m_isRemote = false;
                header.m_objectTypeIndication = readBits(8);
                header.m_streamType = readBits(8);
                header.timeStampResolution = readBits(24);
                header.m_bufferSize = readBits(16);
                header.m_streamID = m_SAFAUStreamID;
                header.m_time = m_SAFPacketTime;
                {
	                int lengthWithoutDecoderSpecificInfo = 7;
	                if (header.m_streamType == 0xFF && header.m_objectTypeIndication == 0xFF)
	                {
	                	int mimeTypeLength = readBits(16);
	                	header.mimeType = readString(mimeTypeLength);
	                	lengthWithoutDecoderSpecificInfo += 2 + mimeTypeLength;
	                }
	                // the data left is the decoder specific info
	                if (0 < (m_SAFPacketlength - lengthWithoutDecoderSpecificInfo))
	                {
	                    header.m_decoderSpecificInfo = new byte[m_SAFPacketlength - lengthWithoutDecoderSpecificInfo];
	                    header.m_decoderSpecificInfo = readByteArray(m_SAFPacketlength - lengthWithoutDecoderSpecificInfo);
	                }
                }
                // all data has been read from primary buffer
                releaseDataFromBitReader();

                // if stream header is LASeR scene header, parse the decoder specific info
                if (header.m_streamType == SceneDescriptionStreamType && header.m_objectTypeIndication == LASeROTI)
                {
                    setDataForBitReader(header.m_decoderSpecificInfo, true);
                    parseLASeRHeader();
                    releaseDataFromBitReader();
                    executeLASeRHeader(header);
                }
                // for the other cases, create a media buffer for the media data
                else
                {
                    header.m_data = new MediaBuffer();
                }

                // anyway store header
                addStreamHeader(header);
                break;

            case 0x03: // EndOfStream
                // get stream from id
                header = getStreamHeaderFromID(m_SAFAUStreamID);
                if (header == null)
                {
                    throw new RuntimeException("No header for streamID " + m_SAFAUStreamID);
                }

                // if stream header is LASeR scene closure...
                if (header.m_streamType == SceneDescriptionStreamType && header.m_objectTypeIndication == LASeROTI)
                {
                    // indicate endOfScene
                    returnValue = false;
                }
                // remote resource
                else if (header.m_isRemote)
                {
                    // Nothing to do
                }
                // in other cases, set the media as fully received
                else
                {
                    MediaBuffer mediaBuffer = (MediaBuffer)header.m_data;
                    mediaBuffer.setMediaReceived();

                    // for still image media, copy image in image collection (other media types are not managed)
                    if (header.m_objectTypeIndication == JPEGOTI || header.m_objectTypeIndication == PNGOTI)
                    {
                        // add the media to the object table
                        m_graphObjectManager.setSingleObject(header.m_streamID, mediaBuffer.getRawData(), header.m_isPermanent);
                    }

                    // removes header associated data
                    header.m_data = null;
                }
                break;

            case 0x04: // Access Unit
                // adjust time if append scene
                if (m_decodingContext.m_newSceneIndicator) {
                    // check time special values
                    if (UNDEFINED_TIMESTAMP != m_SAFPacketTime)
                    {
                        m_SAFPacketTime = m_SAFPacketTime + m_decodingContext.m_offsetAppendSceneTime;
                    }
                }

                // get stream by id
                header = getStreamHeaderFromID(m_SAFAUStreamID);
                if (header == null)
                {
                    throw new RuntimeException("No header for streamID " + m_SAFAUStreamID);
                }

                // if stream header is LASeR scene access unit
                if (header.m_streamType == SceneDescriptionStreamType && header.m_objectTypeIndication == LASeROTI)
                {
                    //store AU to execute after when it is appropriate
                    addUnit(m_SAFPacketTime, data);
                }
                // in other cases, add received data to the media stream
                else
                {
                    // get the MediaBuffer
                    MediaBuffer mediaBuffer = (MediaBuffer)header.m_data;

                    // write the received data to the media buffer
                    mediaBuffer.write(data);
                }
                break;

            case 0x05: // EndOfSAFSession
                // indicate the end of scene reading
                returnValue = false;
                break;

            case 0x06: // Cache unit
                parseCacheUnit();
                break;
            case 0x07: // Remote Stream Header
                // create stream header object
                header = new StreamHeader();
                // Read
                header.m_isPermanent = false;
                header.m_isRemote = true;
                header.m_objectTypeIndication = readBits(8);
                header.m_streamType = readBits(8);
                header.timeStampResolution = readBits(24);
                header.m_bufferSize = readBits(16);
                header.m_streamID = m_SAFAUStreamID;
                header.m_time = m_SAFPacketTime;
                header.url = null;
                {
	                int lengthWithoutDecoderSpecificInfo = 7;
	                if (header.m_streamType == 0xFF && header.m_objectTypeIndication == 0xFF)
	                {
	                	int mimeTypeLength = readBits(16);
	                	header.mimeType = readString(mimeTypeLength);
	                	lengthWithoutDecoderSpecificInfo += 2 + mimeTypeLength;
	                }
	                header.url = readString(readBits(16));
	                lengthWithoutDecoderSpecificInfo += 2 + header.url.length();
	                // the data left is the decoder specific info
	                if (0 < (m_SAFPacketlength - lengthWithoutDecoderSpecificInfo))
	                {
	                    header.m_decoderSpecificInfo = new byte[m_SAFPacketlength - lengthWithoutDecoderSpecificInfo];
	                    header.m_decoderSpecificInfo = readByteArray(m_SAFPacketlength - lengthWithoutDecoderSpecificInfo);
	                }
                }
                // Store stream header
                addStreamHeader(header);
                // Add the url to the object table
                m_graphObjectManager.setSingleObject(header.m_streamID, header.url, header.m_isPermanent);
                break;
            default :
                if (Debug.DEBUG_PARSER2)
                {
                    System.out.println("unhandled access unit type");
                }
        }

        // release data from bit reader
        if (0 != m_SAFPacketlength)
        {
            releaseDataFromBitReader();
        }

        /*trace(STR_END + "decode laser packet");*/
        return returnValue;
    }

    /**
     * parses a cache unit. Execution is not handled yet
     *
     * @throws Exception
     *
     */
    public void parseCacheUnit() throws Exception
    {
        if (Debug.DEBUG_PARSER2)
        {
            System.out.println("T2 : cache unit is beeing parsed:");
        }

        // parse the cache unit
        readBit();                                                   // replace
        readBit();                                                   // permanent
        readBits(6);                                                 // reserved
        int len = readBits(16);                                      // url length
        /*byte url[] =*/ readByteArray(len);                             // url
        /*byte payload[] =*/ readByteArray(m_SAFPacketlength - len - 3); // payload that is associated to url

        // should now store cache object
    }

///////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// Tool operations ///////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////
    protected final int tab1[] = new int[]{ 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80, 0x100, 0x200, 0x400, 0x800,
                                            0x1000, 0x2000, 0x4000, 0x8000, 0x10000, 0x20000, 0x40000, 0x80000, 0x100000,
                                            0x200000, 0x400000, 0x800000, 0x1000000, 0x2000000, 0x4000000, 0x8000000,
                                            0x10000000, 0x20000000, 0x40000000};
    protected final int tab[] = new int[]{  0, 0, 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767,
                                            65535, 131071, 262143, 524287, 1048575, 2097151, 4194303, 8388607, 16777215,
                                            33554431, 67108863, 134417727, 268435455, 536870911};

    /**
     * Translates the integer N read from the stream to its float value, when the type of the number is not known, ie
     * in a replace field update.
     * <p/>
     * The stream contains a 4-bits number (B) followed by a 9 to 24 bits, 2's complement integer (N).
     * B is the number of bits used for N, minus 9. This means the minimum number to represent N is 9.
     * When N is read to an int, the first operation is to restore
     * the sign. If N is over 2^(B+8) - 1, it is translated to 2^(B+9)-N.
     * Then the correction for the resolution is applied. For a generic number, the resolution is constant and equal
     * to 1/256.
     *
     * @param n the 12-bit integer number picked up from the stream
     * @return the float equivalent of the 12-bit integer number decoded from the stream
     */
   public int translateGeneric(int n, int nbBits)
    {
        // read 9 to 24 bits depending on nbBits
        int res_r = Real.convertIntToRealInt(negp(n, nbBits));
        res_r = Real.divByPowerOf2(res_r, 8);

        return res_r;
    }

    public int negp(int n, int nbBits) {
        if (n > tab[nbBits])
        {
            n -= tab1[nbBits];
        }
        return n;
    }

    // methods for logging each call to read*
    
	private void fill(int n) {
	    for (int i = 0; i < n; i++) {
	        traceFile.print(" ");
	    }
	}

	/*public boolean readBit() throws java.io.IOException {
	    boolean b = readBit();
	    if (Debug.DEBUG_PARSER) {
		    traceFile.print(s);
		    fill(20 - s.length());
		    traceFile.print(" n=1\tvalue=");
		    traceFile.println(b);
	    }
	    return b;
	}*/

	/*public int readBits(int n) throws java.io.IOException {
	    int i = readBits(n);
	    traceBits(n, i, s);
	    return i;
	}*/
	
	/*public void traceBits(int n, int i, String s) {
		if (Debug.DEBUG_PARSER) {
			traceFile.print(s);
		    fill(20 - s.length());
		    traceFile.print(" n=");
		    traceFile.print(n);
		    traceFile.print("\tvalue=");
		    traceFile.println(Integer.toHexString(i));
		}
	}
	
	public void trace(String s) {
		if (Debug.DEBUG_PARSER)
			traceFile.println(s);
	}*/

	/*public int readInt() throws java.io.IOException {
	    int b = readInt();
	    if (Debug.DEBUG_PARSER) {
		    traceFile.print(s);
		    fill(20 - s.length());
		    traceFile.print(" n=32\tvalue=");
		    traceFile.println(b);
	    }
	    return b;
	}*/
	
    public int readVlcInt(int n) throws java.io.IOException {
        int k = 1;
        while (readBit()) k++;
        if (k == 0) {
            /*traceBits(1, 0, msg);*/
            return 0;
        }
        int i = readBits(k * (n - 1));
        /*traceBits(k * n, i, "vl "+msg);*/
        return i;
    }
    
    // TML
    public int readAttr_custom_0to1float() throws IOException {
        //if (readBit()) return Integer.MAX_VALUE; // inherit
        return readBits(8) << 4;
    }
    
    public int[] read_codec_valueWithUnits() throws IOException {
        int u[] = new int[2];
        u[0] = (readInt() >> 8) << 12 ; // float represented as fixed point with 8 bits mantissa
        u[1] = readBits(3); // // 0 no unit or px, 1 in, 2 cm, 3 mm, 4 pt, 5 pc, 6 %
        return u;
    }
    
    public int translateTimeUponReading(int i, boolean dur) {
        // read time as ticks
        // return time in ms
        return ((i * 1000) / m_decodingContext.m_timeResolution) + (dur ? 0 : m_SAFPacketTime);
    }
    
    protected void readAttr_custom_extension() throws IOException {
        int i = readVlcInt(5);
        readByteArray(i);
    }
}
