/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */
package iso.mpeg.streamezzo.player.util;


import iso.mpeg.streamezzo.player.node.RenderedGraphObject;
import iso.mpeg.streamezzo.player.tool.Debug;

import java.util.Vector;


/**
 * The Navigator class handles key event and propagate them to the key listeners and
 * do the navigation between objects
 */

public class Navigator
{
    static public final byte KEY_FIRE = 4;
    static public final byte KEY_UP = 0;
    static public final byte KEY_DOWN = 1;
    static public final byte KEY_LEFT = 2;
    static public final byte KEY_RIGHT = 3;

    // flag indicating if since the last time, a key was pressed
    private boolean m_isKeyPressed = false;
    // hold the pressed key code
    private int m_pressedKeyCode;
    // list of key listeners
    private EventManager m_eventManager = null;

    /** Indicates if navigation is enabled or disabled */
    private boolean m_navigationEnabled = true;

    /** Sorted list of navigatable graphobject. Sorted by index first and at end sorted by DOM order*/
    private Vector m_navigatables = null;

    /** Current focused node : null means no focus. */
    private RenderedGraphObject m_focusedElement = null;

    /** Current focused node id : -1 means no focus */
    private int m_focusedElementID = -1;

    /** Constants for navigation request */
    static private final byte NAV_NONE = -1;
    static private final byte NAV_ACTIVATE = 0;
    static private final byte NAV_UP = 1;
    static private final byte NAV_DOWN = 2;
    static private final byte NAV_RIGHT = 3;
    static private final byte NAV_LEFT = 4;

    /** The navigation current request */
    private byte m_navRequest = NAV_NONE;

    public Navigator()
    {
        m_navigatables = new Vector();
    }

    /**
     * Sets the pressed key code and the flag which tells if a key was pressed
     * since the last time
     * @param keyCode
     *
     * @see iso.mpeg.streamezzo.player.util.Navigator#processKeyEvents
     */
    public void setPressedKey(int keyCode)
    {
        m_pressedKeyCode = keyCode;
        m_isKeyPressed = true;
    }

    /**
     * Resets the flag telling whether a key was pressed
     * This method must be called at each frame of the rendering loop
     */
    public void resetKeyPressed()
    {
        m_isKeyPressed = false;
    }

    /**
     * Gets the last pressed key code
     * @return the key code
     */
    public int getPressedKeyCode()
    {
        return m_pressedKeyCode;
    }

    /**
     * Tells whether a key was pressed or not since the last rendering frame
     * @return true if a key was pressed else returns false
     */
    public boolean isKeyPressed()
    {
        return m_isKeyPressed;
    }

////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////// Key listeners ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Registers a listener
     * @param listener
     */
    public void attachEventListener(EventListener listener)
    {
        if (m_eventManager == null)
        {
            m_eventManager = new EventManager();
        }
        m_eventManager.attach(listener);
    }

    /**
     * Removes a listener from the list
     * @param listener
     */
    public void detachEventListener(EventListener listener)
    {
        if (m_eventManager != null)
        {
            m_eventManager.detach(listener);
        }
    }

    /**
     * Removes all the listeners
     */
    public void cleanEventListeners()
    {
        if (m_eventManager != null)
        {
            m_eventManager.clean();
        }
    }

    /**
     * Notifies all the regitered listeners
     * @param eventType the event type
     * @param eventData the associated event data
     * @return If the event has been handled
     */
    private boolean notifyEvent(byte eventType, Object eventData)
    {
        if (m_eventManager != null)
        {
            return m_eventManager.notify(null, eventType, eventData);
        }
        return false;
    }

////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////// Process key events ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Processes propagation of key events
     * @param sceneTimeMillis the scene time
     */
    public void processKeyEvents(int sceneTimeMillis)
    {
        boolean hasBeenHandled = false;
        if (m_isKeyPressed)
        {
            int[] params = { sceneTimeMillis, m_pressedKeyCode };
            hasBeenHandled = notifyEvent(EventListener.KEY_PRESS, params);
            if (Debug.DEBUG_EVENT) System.out.println("navigator processes key event ("+m_pressedKeyCode+") - has been handled = "+hasBeenHandled);
            if (hasBeenHandled)
            {
                m_isKeyPressed = false;
            }

            // Update navigation status if key event has not been handled
            if ( (! hasBeenHandled) && (m_navigationEnabled) )
            {
                updateNavigationRequest(m_pressedKeyCode);
            }
        }
    }


////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////// Navigation ///////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Register the passed-in node into the ordered list of navigatables
     * @param node
     */
    public void registerNodeAsNavigatable(RenderedGraphObject node)
    {
        // sanity check
        if (node == null)
        {
            return;
        }

        // Check if the node have a nav-index to insert ordered
        /*        boolean inserted = false;
        if (node.m_actualProperties.m_navIndex != -1)
        {
            // insert orderer
            int position = 0;
            while ( (! inserted) && (position < m_navigatables.size()) )
            {
                // check position
                RenderedGraphObject object = (RenderedGraphObject)m_navigatables.elementAt(position);
                if (object.m_actualProperties.m_navIndex > node.m_actualProperties.m_navIndex)
                {
                    // insert before
                    m_navigatables.insertElementAt(node, position);
                    inserted = true;
                    if (Debug.DEBUG_NAVIGATOR) System.out.println("Navigator registers a navigatable node at position "+position+" :" + node +"- nav-index="+node.m_actualProperties.m_navIndex);
                }

                // next position
                position++;
            }
        }

        if (! inserted)
        {
            // insert at the end
             m_navigatables.addElement(node);
            //if (Debug.DEBUG_NAVIGATOR) System.out.println("Navigator registers a navigatable node at position "+(m_navigatables.size() - 1)+" :" + node +"- nav-index="+node.m_actualProperties.m_navIndex);
        }*/
    }

    /**
     * Enable or disable navigation feature
     * @param isEnable true to enable navigation
     */
    public void enableNavigation(boolean isEnable)
    {
        m_navigationEnabled = isEnable;
    }

    /**
     * Update the navigation request
     * @param keyCode The keycode of the pressed key
     */
    public void updateNavigationRequest(int keyCode)
    {
        // Navigation must be active
        if (!m_navigationEnabled)
        {
            return;
        }

        m_navRequest = NAV_NONE;
        switch (keyCode)
        {
            case KEY_FIRE :
                m_navRequest = NAV_ACTIVATE;
                if (Debug.DEBUG_EVENT) System.out.println("navigator store request ACTIVATE");
                break;
            case KEY_UP :
                m_navRequest = NAV_UP;
                if (Debug.DEBUG_EVENT) System.out.println("navigator store request UP");
                break;
            case KEY_DOWN :
                m_navRequest = NAV_DOWN;
                if (Debug.DEBUG_EVENT) System.out.println("navigator store request DOWN");
                break;
            case KEY_LEFT :
                m_navRequest = NAV_LEFT;
                if (Debug.DEBUG_EVENT) System.out.println("navigator store request LEFT");
                break;
            case KEY_RIGHT :
                m_navRequest = NAV_RIGHT;
                if (Debug.DEBUG_EVENT) System.out.println("navigator store request RIGHT");
                break;
        }
    }

    /**
     * Provides the passed-in element position into the list of navigatables.
     * @param elementToFind
     * @return The position found or -1 if not found;
     */
    private int getNavigatablePosition(RenderedGraphObject elementToFind)
    {
        // Search into list
        boolean found = false;
        int position = 0;
        while ( (!found) && (position < m_navigatables.size()) )
        {
            // check current
            if (elementToFind == m_navigatables.elementAt(position))
            {
                found = true;
            }

            // next
            position++;
        }

        // return position
        if (found)
        {
            return position-1;
        }
        return -1;
    }

    /**
     * Provides the next focusable element according to nav-index order or DOM order.
     * It ignores nav-up, nav-down, nav-right, nav-left.
     * @param currentElement
     * @return RenderedGraphObject
     */
    private RenderedGraphObject getPreviousFocusable(RenderedGraphObject currentElement)
    {
        // If no current then return the last navigatable
        if (currentElement == null)
        {
            return (RenderedGraphObject)m_navigatables.lastElement();
        }

        // Get current position into the list of navigatables
        int currPos = getNavigatablePosition(currentElement);

        // Return previous
        if (currPos < 1)
        {
            return (RenderedGraphObject)m_navigatables.lastElement();
        }
        else
        {
            return (RenderedGraphObject)m_navigatables.elementAt(currPos - 1);
        }
    }

    /**
     * Provides the previous focusable element according to nav-index order or DOM order.
     * It ignores nav-up, nav-down, nav-right, nav-left.
     * @param currentElement
     * @return RenderedGraphObject
     */
    private RenderedGraphObject getNextFocusable(RenderedGraphObject currentElement)
    {
        // If no current then return the last navigatable
        if (currentElement == null)
        {
            return (RenderedGraphObject)m_navigatables.firstElement();
        }

        // Get current position into the list of navigatables
        int currPos = getNavigatablePosition(currentElement);

        // Return next
        if ( (currPos == -1) || (currPos == (m_navigatables.size() - 1)) )
        {
            return (RenderedGraphObject)m_navigatables.firstElement();
        }
        else
        {
            return (RenderedGraphObject)m_navigatables.elementAt(currPos + 1);
        }
    }

    /**
     * Provide the next focusable element according to the current focusable nav directions.
     * @param currentFocused
     * @param navDirection NAV_UP, NAV_DOWN, NAV_RIGHT, NAV_LEFT
     * @return null if cannot be found
     */
    private RenderedGraphObject getNextFocusableByNavDir(RenderedGraphObject currentFocused, int navDirection)
    {
        // Sanity check
        if (currentFocused == null)
        {
            return null;
        }

        RenderedGraphObject nextFocusable = null;
        int nextID = -1;
        switch(navDirection)
        {
            case NAV_UP :
                //nextID = currentFocused.m_actualProperties.m_navUp;
                break;
            case NAV_LEFT :
                //nextID = currentFocused.m_actualProperties.m_navLeft;
                break;
            case NAV_DOWN :
                //nextID = currentFocused.m_actualProperties.m_navDown;
                break;
            case NAV_RIGHT :
                //nextID = currentFocused.m_actualProperties.m_navRight;
                break;
        }

        // find the element by node ID
        if (nextID != -1)
        {
            boolean found = false;
            int position = 0;
            while ( (!found) && (position < m_navigatables.size()))
            {
                // Check element
                RenderedGraphObject oneFocusable = (RenderedGraphObject)m_navigatables.elementAt(position);
                if (oneFocusable != null)
                {
                    if (oneFocusable.getID() == nextID)
                    {
                        found = true;
                        nextFocusable = oneFocusable;
                    }
                }

                // next
                position++;
            }
        }

        return nextFocusable;
    }

    /**
     * Processes the navigation. Focusin, focusout and activate events will be sent.
     *
     * @param sceneTimeMs The current rendering scene time
     */
    public void processNavigation(int sceneTimeMs)
    {
        // Navigation must be active and Some navigatables must exists
        if ( (!m_navigationEnabled) || (m_navigatables.size() == 0) )
        {
            m_navRequest = NAV_NONE;
            m_navigatables.removeAllElements();
            return;
        }

         // Store previously focused
        RenderedGraphObject previousFocused = m_focusedElement;
        int previousFocusedID = m_focusedElementID;

        // Check if current focused still exists
        if (previousFocused != null)
        {
            if (! m_navigatables.contains(previousFocused) )
            {
                // Send focusout event
                Integer eventData = new Integer(sceneTimeMs);
                previousFocused.onEvent(EventListener._UI_FOCUS_OUT, eventData);
                previousFocused.notifyEvent(EventListener.UI_FOCUS_OUT, eventData);
                m_focusedElement = null;
                m_focusedElementID = -1;
             }
        }

        // Check if there is a navigation request
        if (m_navRequest == NAV_NONE)
        {
            m_navigatables.removeAllElements();
            return;
        }

        // Attempt to navigate by nav-up, nav-down, nav-right, nav-left
        boolean canBePerformedByNavIndexDir = false;
        if (previousFocused != null)
        {
            switch(m_navRequest)
            {
                case NAV_UP :
                case NAV_LEFT :
                case NAV_DOWN :
                case NAV_RIGHT :
                    m_focusedElement = getNextFocusableByNavDir(previousFocused, m_navRequest);
                    if (m_focusedElement != null)
                    {
                        canBePerformedByNavIndexDir = true;
                    }
                    break;
            }
        }

        // If navigation according to nav-index direction cannot be perform
        if (! canBePerformedByNavIndexDir)
        {
            switch(m_navRequest)
            {
                case NAV_UP :
                case NAV_LEFT :
                    m_focusedElement = getPreviousFocusable(previousFocused);
                    break;

                case NAV_DOWN :
                case NAV_RIGHT :
                    m_focusedElement = getNextFocusable(previousFocused);
                    break;
            }
        }

        // Check if focus has been set somewhere
        if (m_focusedElement == null)
        {
            m_navRequest = NAV_NONE;
            m_navigatables.removeAllElements();
            return;
        }

        // Store new focused node ID
        m_focusedElementID = m_focusedElement.getID();

        // Send suitable events
        Integer eventData = new Integer(sceneTimeMs);
        if (previousFocused != m_focusedElement)
        {
            // Focused has changed
            if (Debug.DEBUG_EVENT) System.out.println("navigator set focus on "+m_focusedElement);

            // Focusout
            if (previousFocused != null)
            {
                previousFocused.onEvent(EventListener._UI_FOCUS_OUT, eventData);
                previousFocused.notifyEvent(EventListener.UI_FOCUS_OUT, eventData);
            }

            // Focusin
            m_focusedElement.onEvent(EventListener._UI_FOCUS_IN, eventData);
            m_focusedElement.notifyEvent(EventListener.UI_FOCUS_IN, eventData);
        }
        else
        {
            // Check activation
            if (m_navRequest == NAV_ACTIVATE)
            {
                // Activate
                m_focusedElement.onEvent(EventListener._UI_ACTIVATE, eventData);
                m_focusedElement.notifyEvent(EventListener.UI_ACTIVATE, eventData);
            }
        }

        // Reset
        m_navRequest = NAV_NONE;
        m_navigatables.removeAllElements();
    }



}