/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.util;

import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.library.Rect2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.GraphObjectManager;

/**
 * This class manages a sensitive zone. It provides a set of method to check mouse
 * events whitin the zone. A sensitive zone is a normalized rectangle.
 */
public class SensitiveZone
{
    /** The sensitive normalized rectangular zone */
    public Rect2D m_rect = new Rect2D(-1, -1, -1, -1);

    /** A flag telling whether a mouse is inside the zone */
    public boolean m_bMouseIsInside = false;

////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////// Setter/Getter, construction ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
    /**
     * Constructor
     */
    public SensitiveZone()
    {
    }

    /**
     * Defines the zone
     * @param x parameter of the normalized rectangular zone to set
     * @param y parameter of the normalized rectangular zone to set
     * @param width parameter of the normalized rectangular zone to set
     * @param height parameter of the normalized rectangular zone to set
     */
    public void setZoneRect(int x, int y, int width, int height)
    {
        m_rect.x_r = x;
        m_rect.y_r = y;
        m_rect.width_r = width;
        m_rect.height_r = height;
    }

    /**
     * Defines the zone as a null zone
     */
    public void setNullZone()
    {
        m_rect.width_r = -1;
        m_rect.height_r = -1;
    }

    /**
     * Returns a reference to the rectangular zone
     * @return
     */
    public Rect2D getZone()
    {
        return m_rect;
    }

////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// Detection ////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Checks whether the mouse is enter, exit or click whithin the zone
     * @param sceneManager the scenemanager that can provide mouse position and offers mouse event management
     */
    public void checkZone(GraphObjectManager sceneManager)
    {
        // Check if the zone have to be checked
        if (! sceneManager.isItWorthToCheckSensitiveZone())
        {
            return;
        }

        // Get mouse position
        Point2D mousePosition = sceneManager.getMousePosition();
        if (null == mousePosition)
        {
            return;
        }

        // Is it inside the zone ?
        boolean isInside = m_rect.isInside(mousePosition);
        if ( (Debug.DEBUG_CURSOR_FULL) && (isInside) ) System.out.println("CURSOR : is inside the zone " +  this);

        // Check previous state
        if ( (m_bMouseIsInside) && (! isInside) )
        {
            // Mouse exit
            if (Debug.DEBUG_CURSOR) System.out.println("CURSOR : exit " +  this);
            sceneManager.onMouseExit(this);
        }
        else if ((! m_bMouseIsInside) && (isInside))
        {
            // mouse enter
            if (Debug.DEBUG_CURSOR) System.out.println("CURSOR : enter " +  this);
            sceneManager.onMouseEnter(this);
        }
        else if (isInside && sceneManager.isMouseClick())
        {
            // click TODO doesn't work
            if (Debug.DEBUG_CURSOR) System.out.println("CURSOR : click " +  this);
            sceneManager.onMouseClick(this);
        }
        else if (isInside)
        {
            // nothing
            sceneManager.onMouseMoveOver(this);
        }

        // Set new state
        m_bMouseIsInside = isInside;
    }

}
