/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.util;

import iso.mpeg.streamezzo.player.library.Point2D;
import iso.mpeg.streamezzo.player.node.Rect;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.Rendering;

import java.util.Vector;

/**
 * TextBox class represents a rectangle containing severals TextLines.
 */

public class TextBox
{
    /** Properties of the rectangle */
    private NodeProperties m_boxProperties = null;

    /** Text lines in the text box */
    public Vector m_textLines = null;

    /** The normalized top-left anchor */
    public Point2D m_topLeft = new Point2D(0, 0);

    /** The box normalized width and height */
    public int m_boxWidth, m_boxHeight = 0;

    /** Total text height in pixels */
    public int m_totalTextHeight = 0;

    /**
     * Constructor of instances of TextBox class
     */
    public TextBox(Rendering renderer, Rect linkedRect)
    {
        // Allocate vectors
        m_textLines = new Vector();

        // Get properties
        m_boxProperties = linkedRect.m_actualProperties.getCopy();

        // Initialize region properties from the linked rect
        m_topLeft.x_r = renderer.normalizeX(linkedRect.m_actualTopLeft.x_r);
        m_topLeft.y_r = renderer.normalizeY(linkedRect.m_actualTopLeft.y_r);
        m_boxWidth = renderer.normalizeW(linkedRect.m_rActualWidth);
        m_boxHeight = renderer.normalizeH(linkedRect.m_rActualHeight);
    }

    /**
     * Add a text line to the text box. The text line must be filled because total text height is computed here.
     * @param line
     * @return false if the textbox is already full
     */
    public boolean addTextLine(TextLine line)
    {
        if ( (m_totalTextHeight + line.m_lineHeight) > m_boxHeight )
        {
            if (Debug.DEBUG_TEXTFLOW) System.out.println("TextBox::addTextLine - Already full");
            return false;
        }
        m_textLines.addElement(line);
        m_totalTextHeight += line.m_lineHeight;
        if (Debug.DEBUG_TEXTFLOW) System.out.println("TextBox::addTextLine");
        return true;
    }

    /**
     * Draw the text box content
     */
    public void drawText(Rendering renderer)
    {
        // Declare local var
        int usedX = 0; // top-left position for drawString operation
        int usedY = 0; // top-left position for drawString operation
        int upLimit = 0; // top limit of the bounding rect
        int downLimit = 0;  // down limit of the bounding rect

        // Get position
        int posX = m_topLeft.x_r;
        int posY = m_topLeft.y_r;

        // Get number of lines
        int nbLines = m_textLines.size();

        // Compute limits and start point of drawing
        upLimit = posY;
        downLimit = posY + m_boxHeight;
        if (m_boxProperties.m_displayAlign == NodeProperties.DISPLAYALIGN_BEFORE)
        {
            usedY = posY;
        }
        else if (m_boxProperties.m_displayAlign == NodeProperties.DISPLAYALIGN_AFTER)
        {
            usedY = downLimit - m_totalTextHeight;
        }
        else // m_actualProperties.m_displayAlign == CENTER
        {
            usedY = posY + (m_boxHeight - m_totalTextHeight) / 2;
        }

        // Set the clipping for text
        Point2D previousOrigin = null, previousDimension = null;
        if (m_boxWidth != -1)
        {
            // save previous clip rect (intersection of all parent clip rects)
            previousOrigin = renderer.getClipRectOrigin();
            previousDimension = renderer.getClipRectDimension();
            renderer.setClipRect(new Point2D(posX, posY), new Point2D(m_boxWidth, m_boxHeight));
        }

        // Draw all lines
        for (int ii = 0; ii < nbLines; ii++)
        {
            // Get a line
            TextLine line = (TextLine) m_textLines.elementAt(ii);

            // Compute X
            if (m_boxProperties.m_text_anchor == NodeProperties.TEXTALIGN_START)
            {
                usedX = posX;
            }
            else if (m_boxProperties.m_text_anchor == NodeProperties.TEXTALIGN_END)
            {
                usedX = (posX + m_boxWidth) - line.m_lineWidth;
            }
            else // m_actualProperties.m_text_anchor == NodeProperties.ALIGN_CENTER
            {
                usedX = posX + (m_boxWidth - line.m_lineWidth) / 2;
            }

            // Check if line can be drawn and draw it
            if (((usedY) <= (downLimit)) && ((usedY + line.m_lineHeight) >= upLimit))
            {
                line.drawLine(renderer, usedX, usedY);
            }

            // Compute next Y
            usedY = usedY + line.m_lineHeight;
        }

        // Reset the clipping to previous
        if (m_boxWidth != -1)
        {
            // set previous clip rect (intersection of all parent clip rects)
            renderer.setClip(previousOrigin.x_r, previousOrigin.y_r, previousDimension.x_r, previousDimension.y_r);
        }
    }
}
