/*
 * This software was originally developed by Arnaud Caigniet, Gregory Donias, Jean-Claude Dufourd,
 * Cdric Ggout, Erwann Gouesbet, Elouan LeCoq, Gaetan Lefevre, Nicolas Pierre, Cristian Spiescu
 * and Julien Tranchant (Streamezzo), in the course of development of the ISO/IEC 14496-20
 * (Lightweight Application Scene Representation -LASeR- and Simple Aggregation Format
 * -SAF-) standard for reference purposes and its performance may not have been optimized.
 * This software module is an implementation of one or more tools as specified by the
 * LASeR and SAF standard.
 * ISO/IEC gives users free license to this software module or modifications thereof
 * for use in products claiming conformance to audiovisual and image-coding related
 * ITU Recommendations and/or ISO/IEC International Standards.
 * ISO/IEC gives users the same free license to this software module or modifications
 * thereof for research purposes and further ISO/IEC standardisation.
 * Those intending to use this software module in products are advised that its use may
 * infringe existing patents. ISO/IEC have no liability for use of this software module
 * or modifications thereof.
 * Copyright is not released for products that do not conform to audiovisual and
 * image-coding related ITU Recommendations and/or ISO/IEC International Standards.
 * Streamezzo retains full right to modify and use the code for its own purpose,
 * assign or donate the code to a third party and to inhibit third parties from using
 * the code for products that do not conform to audiovisual and image-coding related ITU
 * Recommendations and/or ISO/IEC International Standards.
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2004-2007.
 */

package iso.mpeg.streamezzo.player.util;

import iso.mpeg.streamezzo.player.library.Rect2D;
import iso.mpeg.streamezzo.player.tool.Debug;
import iso.mpeg.streamezzo.player.tool.Rendering;

import java.util.Vector;


/**
 * This class manages a text line that can be split into severals parts. Each part have their own properties
 */
public class TextLine
{
    /** The list of string in a line **/
    public Vector m_strings = null;
    /** The list of string properties. There is a link between m_strings and m_stringProperties **/
    public Vector m_stringProperties = null;
    /** The width in pixels of the line **/
    public int m_lineWidth = 0;
    /** The height in pixels of the line **/
    public int m_lineHeight = 0;
    /** The distance between top and baseline of the line **/
    public int m_lineBaselineHeight = 0;
    /** The segment index of the mark */
    private int m_markSegmentIndex = 0;
    /** The zero-based position in a the segment that defined the mark */
    private int m_markCharPosInSegment = 0;

    /**
     * Constructor of instances of TextLine class
     */
    public TextLine()
    {
        // Allocate vectors
        m_strings = new Vector();
        m_stringProperties = new Vector();

        // Init members
        m_lineWidth = 0;
        m_lineHeight = 0;
        m_lineBaselineHeight = 0;
    }

    /**
     * Reset a text line
     */
    public void clean()
    {
        m_strings.removeAllElements();
        m_stringProperties.removeAllElements();
        m_lineWidth = 0;
        m_lineHeight = 0;
        m_lineBaselineHeight = 0;
    }

    /**
     * Set a mark into the text line
     * @param segmentIndex The segment in which the mark will be set
     * @param charPos The zero-based position into the segment
     */
    public void setMark(int segmentIndex, int charPos)
    {
        m_markSegmentIndex = segmentIndex;
        m_markCharPosInSegment = charPos;
        if (Debug.DEBUG_TEXTFLOW) System.out.println("TextLine::setMark : m_markSegmentIndex = " + m_markSegmentIndex + " - m_markCharPosInSegment = " + m_markCharPosInSegment);
    }

    /**
     * Provide the segment index of the current mark
     * @return
     */
    public int getMarkSegment()
    {
        return m_markSegmentIndex;
    }

    /**
     * Provide the zero-based position into the segment of the current mark
     * @return
     */
    public int getMarkCharPos()
    {
        return m_markCharPosInSegment;
    }

    /**
     * Add a string to the end of the line.
     * It updates the line total width and the line height.
     *
     * @param renderer The renderer used to compute string width
     * @param stringToAdd The string to add at the end of the line
     * @param properties The properties of the string to add
     */
    public void addString(Rendering renderer, String stringToAdd, NodeProperties properties)
    {
        // Check string
        if (stringToAdd.length() == 0)
        {
            return;
        }

        // Add in vectors
        m_strings.addElement(stringToAdd);
        m_stringProperties.addElement(properties);

        // Update line width, height and baseline
        Object font = renderer.getFont(properties.m_fontFamily, properties.m_fontStyle, properties.m_fontSize);
        int width = renderer.stringWidth(font, stringToAdd);
        m_lineWidth += width;
        int heigth = renderer.getFontHeight(font);
        if (heigth > m_lineHeight) m_lineHeight = heigth;
        int baseline = renderer.getFontBaseline(font);
        if (baseline > m_lineBaselineHeight) m_lineBaselineHeight = baseline;
    }

    /**
     * Draw the line at the specified position
     * @param renderer The renderer
     * @param x Top-Left corner where to start drawing the line
     * @param y Top-Left corner where to start drawing the line
     */
    public void drawLine(Rendering renderer, int x, int y)
    {
        // Sanity check
        if (m_strings.size() != m_stringProperties.size())
        {
            return;
        }

        // Init position to draw strings
        int posX = x;
        int posY = y + m_lineBaselineHeight; // because anchor if at baseline-left

        // Iterate through line segments
        String lineSegmentString = null;
        NodeProperties lineSegmentProperties = null;
        Rect2D bbox = new Rect2D(-1, -1, -1, -1);
        for (int i = 0; i < m_strings.size(); i++)
        {
            // Get a string and its owner
            lineSegmentString = (String) m_strings.elementAt(i);
            lineSegmentProperties = (NodeProperties) m_stringProperties.elementAt(i);


            // Check visibility and draw string
            if (lineSegmentProperties.m_visibility != NodeProperties.VISIBILITY_HIDDEN)
            {
                // Set font and color
                Object font = renderer.getFont(lineSegmentProperties.m_fontFamily, lineSegmentProperties.m_fontStyle, lineSegmentProperties.m_fontSize);
                renderer.setFontForGraphics(font);
                renderer.setColor(lineSegmentProperties.m_fill, lineSegmentProperties.m_fillOpacity&0xFF);

                // Draw the string with baseline-left anchor in order to have a convenient
                // display when several font are used on the same line
                renderer.drawStringAtBaselineLeft(lineSegmentString, posX, posY, bbox);
            }

            // Get string segment width and update x position
            Object font = renderer.getFont(lineSegmentProperties.m_fontFamily, lineSegmentProperties.m_fontStyle, lineSegmentProperties.m_fontSize);
            posX += renderer.stringWidth(font, lineSegmentString);
        }
    }
}
